/* $Id: dexter-channel-object.c 393 2012-03-06 23:36:44Z pkovacs $ */
/*
 * libdexter - a plugin-based, distributed sampling library.
 * Copyright (C) 2006-2012 Philip Kovacs <pkovacs@users.sourceforge.net>.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 2.1
 * of the License, or (at your option) any later version.

 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.

 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
 * 02110-1301  USA
 *
*/

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <dexter-channel-object.h>
#include <dexter-channel-internal.h>
#include <dexter-messages.h>
#include <dexter-private.h>
#include <dexter-alias.h>

static void   dexter_channel_object_class_init           (DexterChannelObjectClass  *klass);
static void   dexter_channel_object_init                 (DexterChannelObject       *channel_object);
static void   dexter_channel_object_finalize             (GObject                   *object);

static void   dexter_channel_object_process_message_real (DexterChannelObject       *channel_object,
                                                          DexterChannelMessage      *message,
                                                          GError                   **error);

static void   dexter_channel_object_channel_event_real   (DexterChannelObject       *channel_object,
                                                          gint                       event);


G_DEFINE_ABSTRACT_TYPE (DexterChannelObject, dexter_channel_object, G_TYPE_OBJECT);
  

static void
dexter_channel_object_class_init (DexterChannelObjectClass *klass)
{
  GObjectClass *gobject_class;

#ifdef G_ENABLE_DEBUG
  g_message (G_STRFUNC);
#endif

  gobject_class = G_OBJECT_CLASS (klass);
  gobject_class->finalize = dexter_channel_object_finalize;

  klass->process_message = dexter_channel_object_process_message_real;
  klass->channel_event = dexter_channel_object_channel_event_real;
}


static void
dexter_channel_object_init (DexterChannelObject *channel_object)
{
#ifdef G_ENABLE_DEBUG
  g_message (G_STRFUNC);
#endif

  channel_object->channel = NULL;
  channel_object->sc = NULL;
  channel_object->id = 0;
  channel_object->mbox = NULL;  
  channel_object->wait_types = 0;
  channel_object->wait_timeout = 0;
}


static void
dexter_channel_object_finalize (GObject *object)
{
  DexterChannelObject *channel_object = DEXTER_CHANNEL_OBJECT (object);

#ifdef G_ENABLE_DEBUG
  g_message (G_STRFUNC);
#endif

  if (channel_object->channel)
    g_object_unref (channel_object->channel);
  if (channel_object->mbox)
    g_async_queue_unref (channel_object->mbox);

  /* chain up */
  G_OBJECT_CLASS (dexter_channel_object_parent_class)->finalize (object);
}


static void
dexter_channel_object_process_message_real (DexterChannelObject   *channel_object,
                                            DexterChannelMessage  *message,
                                            GError               **error)
{
#ifdef G_ENABLE_DEBUG
  g_message (G_STRFUNC);
#endif

  g_return_if_fail (DEXTER_IS_CHANNEL_OBJECT (channel_object));
  g_return_if_fail (message != NULL);

  g_object_ref (channel_object);

  /* just free the message */
  dexter_channel_message_free (message);

  g_object_unref (channel_object);
}


void
dexter_channel_object_process_message (DexterChannelObject   *channel_object,
                                       DexterChannelMessage  *message,
                                       GError               **error)
{
  DEXTER_CHANNEL_OBJECT_GET_CLASS (channel_object) -> process_message (channel_object, message, error);
}


static void
dexter_channel_object_channel_event_real (DexterChannelObject *channel_object,
                                          gint                 event)
{
  g_return_if_fail (DEXTER_IS_CHANNEL_OBJECT (channel_object));
}


void
dexter_channel_object_channel_event (DexterChannelObject *channel_object,
                                     gint                 event)
{
  DEXTER_CHANNEL_OBJECT_GET_CLASS (channel_object) -> channel_event (channel_object, event);
}


void
dexter_channel_object_send_command (DexterChannelObject  *channel_object,
                                    gint                  command,
                                    gint                  parameter,
                                    GString              *payload,
                                    GError              **error)
{
  GError *local_error = NULL;

#ifdef G_ENABLE_DEBUG
  g_message (G_STRFUNC);
#endif
    
  g_return_if_fail (DEXTER_IS_CHANNEL_OBJECT (channel_object));
  g_return_if_fail (command > 0);
  g_return_if_fail (error == NULL || *error == NULL);
    
  g_object_ref (channel_object);

  if ( (!channel_object->channel) || (channel_object->id <= 0) )
  {
    /* programming error */
    g_error (G_STRLOC " channel object not registered on a channel");
  }

  /* push command onto channel cmd queue */
  dexter_channel_command_push (channel_object->channel, 
                               channel_object, command, parameter, 
                               payload, &local_error);
  if (local_error)
  {
    g_propagate_error (error, local_error);
    g_object_unref (channel_object);
    return;
  }

  g_object_unref (channel_object);
}


DexterChannelMessage *
dexter_channel_object_wait_message (DexterChannelObject *channel_object,
                                    glong                timeout_usecs)
{
  DexterChannelMessage *msg = NULL;
 
#ifdef G_ENABLE_DEBUG
  g_message (G_STRFUNC);
#endif

  g_return_val_if_fail (DEXTER_IS_CHANNEL_OBJECT (channel_object), NULL);
  g_return_val_if_fail (timeout_usecs >= 0, NULL);

  g_object_ref (channel_object);

  if (timeout_usecs > 0)
  {
    /* wait for reply until timeout elapses */
    GTimeVal end_time;
  
    g_get_current_time (&end_time);
    g_time_val_add (&end_time, timeout_usecs);

    msg = g_async_queue_timed_pop (channel_object->mbox, &end_time);  
  }
  else
  {
    /* wait forever for msg */
    msg= g_async_queue_pop (channel_object->mbox);
  }

  g_object_unref (channel_object);
  return msg;
}


void
dexter_channel_message_free (DexterChannelMessage *message)
{
#ifdef G_ENABLE_DEBUG
  g_message (G_STRFUNC);
#endif

  if (G_UNLIKELY (!message))
    return;

  if (message->payload)
    g_string_free (message->payload, TRUE);
  g_slice_free (DexterChannelMessage, message);
}


void 
dexter_channel_object_send_wait_process (DexterChannelObject  *channel_object,
                                         gint                  command,
                                         gint                  parameter,
                                         GString              *payload,
                                         GError              **error)

{ 
  g_return_if_fail (DEXTER_IS_CHANNEL_OBJECT (channel_object));
  g_return_if_fail (command > 0);
  g_return_if_fail (error == NULL || *error == NULL);

  g_object_ref (channel_object);

  GError *local_error = NULL;
 
  /* send command */
  dexter_channel_object_send_command (channel_object, command, parameter, payload, &local_error);
  if (local_error)
  {
    g_propagate_error (error, local_error);
    g_object_unref (channel_object);
    return;
  }

  /* wait for reply */
  {
    DexterChannelMessage *reply;
 
    reply = dexter_channel_object_wait_message (channel_object, channel_object->wait_timeout);
    if (!reply)
    {
      g_set_error (error, DEXTER_CHANNEL_ERROR, DEXTER_CHANNEL_ERROR_TIMEOUT,
                   "%s: timeout waiting for message from server", G_STRFUNC);
      g_object_unref (channel_object);
      return;
    }
 
    /* process reply */
    dexter_channel_object_process_message (channel_object, reply, &local_error);
    if (local_error)
    {
      g_propagate_error (error, local_error);
      g_object_unref (channel_object);
      return;
    }
  }

  g_object_unref (channel_object);
}

#define __DEXTER_CHANNEL_OBJECT_C__
#include <dexter-aliasdef.c>
