/* $Id: dexter-channel.c 393 2012-03-06 23:36:44Z pkovacs $ */
/*
 * libdexter - a plugin-based, distributed sampling library.
 * Copyright (C) 2006-2012 Philip Kovacs <pkovacs@users.sourceforge.net>.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 2.1
 * of the License, or (at your option) any later version.

 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.

 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
 * 02110-1301  USA
 *
*/

/**
 * SECTION:dexter-channel
 * @title: Channels
 * @short_description: Using channels for network-based operation of Libdexter
 * @see_also:
 *
 * Channels are Libdexter objects used to interact with hosts on the network 
 * running Libdexter servers.  To use a channel, you must first create it with 
 * dexter_channel_new().  Next, open the channel with dexter_channel_open() to 
 * establish a network connection with the server.  You must then greet the server 
 * with dexter_channel_greet() to perform the protocol handshake that readies 
 * the channel for actual use.  You may secure your channel communications 
 * anytime after the greeting is exchanged with dexter_channel_starttls().
 * Note that the server may require that you issue dexter_channel_starttls()
 * in order to perform any meaningful data operation.  The server may also
 * require that the channel present valid X.509 certificate credentials.
 *
 * Once the greeting is exchanged, you may then create remote Libdexter objects
 * on the server by specifying the channel when creating your objects, e.g. with
 * dexter_sampler_new().  When you are done with the channel and have destroyed all 
 * objects you created on it, issue dexter_channel_part() to tell the server that
 * the channel will no longer be used.  At that point, the server will disconnect, 
 * you can then issue dexter_channel_close() and, finally, dexter_channel_free() 
 * to destroy the channel.
 *
 * When using channels to create objects on other hosts, <emphasis>two</emphasis>
 * objects are actually created: one on the remote host and one local to your
 * application.  The local object acts as a transparent proxy to the the remote 
 * object, issuing network messages to the remote object as you manipulate it.
 * There are no api differences, however, from your application's point of view,
 * i.e. working with a local object is identical to working with a remote object.
 *
 * Each channel is implemented with a single tcp/ip connection to the specified
 * host.  You may create many Libdexter objects, e.g. samplers or timed samplers, 
 * over a single channel.  There is no need to create a separate channel for each 
 * object (although you can).  The Libdexter configuration file is used to set
 * default channel settings, such as socket options, etc.
 **/

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <stdio.h>
#include <string.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <dexter.h>
#include <dexter-channel-internal.h>
#include <dexter-hash.h>
#include <dexter-messages.h>
#include <dexter-msgframer.h>
#include <dexter-network.h>
#include <dexter-private.h>
#include <dexter-alias.h>

typedef struct _channel_object_hash_entry ChannelObjectHashEntry;
typedef struct _channel_event_callback_container ChannelEventCallbackContainer;


static void     dexter_channel_class_init            (DexterChannelClass      *klass);
static void     dexter_channel_init                  (DexterChannel           *channel);
static void     dexter_channel_finalize              (GObject                 *object);
static gint     dexter_channel_set_state             (DexterChannel           *channel,
                                                      gint                     state);
static gboolean dexter_channel_is_runnable           (DexterChannel           *channel);
static void     dexter_channel_set_runnable          (DexterChannel           *channel,
                                                      gboolean                 runnable);
static void     dexter_channel_event                 (DexterChannel           *channel,
                                                      gint                     event);
static void     dexter_channel_server_frame          (DexterChannel           *channel,
                                                      DexterMsgFrame          *frame);
static void     dexter_channel_send_command          (DexterChannel           *channel,
                                                      DexterChannelMessage    *message,
                                                      GError                 **error);
static void     dexter_channel_process_message       (DexterChannelObject     *channel_object,
                                                      DexterChannelMessage    *message,
                                                      GError                 **error);
static void     dexter_channel_channel_event         (DexterChannelObject     *channel_object,
                                                      gint                     event);
static GError  *dexter_channel_convert_network_error (GError                 **network_error);

static DexterChannelSettings
               *dexter_channel_settings_dup          (const DexterChannelSettings *settings);
static void     dexter_channel_settings_free         (DexterChannelSettings       *settings);
static void     dexter_channel_settings_prepare      (DexterChannelSettings       *settings);

#ifdef HAVE_TLS
static DexterChannelTLSSettings
               *dexter_channel_tls_settings_new      (const gchar                 *cert_file,
                                                      const gchar                 *key_file,
                                                      const gchar                 *trust_file,
                                                      const gchar                 *crl_file,
                                                      gboolean                     verify_hostname);
static void     dexter_channel_tls_settings_free     (DexterChannelTLSSettings    *tls_settings);
#endif /* HAVE_TLS */

                /* GDestroyNotify to destroy channel hash entries */
static void     _destroy_object_hash_entry           (ChannelObjectHashEntry  *hash_entry);

                /* channel communication thread */
static void     _channel_thread_func                 (DexterChannel           *channel);

                /* thread function for (user) channel event callback */
static void     _event_callback_func                 (ChannelEventCallbackContainer *cb);

                /* GHFunc for notifying channel object of channel event */
static void     _event_channel_object_func           (gint                     *id,
                                                      ChannelObjectHashEntry   *hash_entry,
                                                      gint                     *event);

struct _DexterChannel
{
  DexterChannelObject       parent;

  DexterChannelSettings    *settings;     /* all user/config-supplied server settings */

  GThread                  *thread;       /* joinable thread for _channel_thread_func */

  volatile gint             runnable;     /* non-zero if channel thread allowed to continue running */
  volatile gint             state;        /* DexterChannelState enum */
  volatile gint             next_id;      /* registered channel objects get a unique (peer) id */

  int                       connfd;       /* connection fd */

  dexter_network_session_t *session;      /* network session */

  GAsyncQueue              *cmd_queue;    /* channel commands are queued up here */
  DexterHashTable          *object_hash;  /* channel object hash by (peer) id */
};

struct _DexterChannelClass 
{
  DexterChannelObjectClass parent_class;
};

struct _channel_object_hash_entry 
{
  gint                 id;
  DexterChannelObject *channel_object;
};

struct _channel_event_callback_container
{
  DexterChannel *channel;
  gint           event;
};


/* each entry corresponds to a flag in DexterChannelSettingFlags such that: flag == 1<<index,
 * i.e. the order of these elements is significant */
static const DexterConfigSetting dexter_channel_config_settings[] =
{
  { "Main", "ChannelHost",
    DEXTER_CONFIG_SETTING_TYPE_STRING, (DexterConfigSettingDefault) DEFAULT_CHANNEL_HOST },

  { "Main", "ChannelService",
    DEXTER_CONFIG_SETTING_TYPE_STRING, (DexterConfigSettingDefault) DEFAULT_CHANNEL_SERVICE },

  { "Main", "ChannelSocketOptions",
    DEXTER_CONFIG_SETTING_TYPE_STRING, (DexterConfigSettingDefault) DEFAULT_CHANNEL_SOCKET_OPTIONS },

  { "Main", "ChannelRecvBufferSize",
    DEXTER_CONFIG_SETTING_TYPE_INT, (DexterConfigSettingDefault) DEFAULT_CHANNEL_RECV_BUFFER_SIZE },

  { "Main", "ChannelTimeoutMsecs",
    DEXTER_CONFIG_SETTING_TYPE_INT, (DexterConfigSettingDefault) DEFAULT_CHANNEL_TIMEOUT_MSECS },

  /* event_callback has no config-based default since it's a function address */
  { "----", NULL, DEXTER_CONFIG_SETTING_TYPE_INT, (DexterConfigSettingDefault) 0 },

  /* tls_start_callback has no config-based default since it's a function address */
  { "----", NULL, DEXTER_CONFIG_SETTING_TYPE_INT, (DexterConfigSettingDefault) 0 },

  { "Main", "ChannelTLSCertFile",
    DEXTER_CONFIG_SETTING_TYPE_STRING, (DexterConfigSettingDefault) DEFAULT_CHANNEL_TLS_CERT_FILE },

  { "Main", "ChannelTLSKeyFile",
    DEXTER_CONFIG_SETTING_TYPE_STRING, (DexterConfigSettingDefault) DEFAULT_CHANNEL_TLS_KEY_FILE },

  { "Main", "ChannelTLSTrustFile",
    DEXTER_CONFIG_SETTING_TYPE_STRING, (DexterConfigSettingDefault) DEFAULT_CHANNEL_TLS_TRUST_FILE },

  { "Main", "ChannelTLSCrlFile",
    DEXTER_CONFIG_SETTING_TYPE_STRING, (DexterConfigSettingDefault) DEFAULT_CHANNEL_TLS_CRL_FILE },

  { "Main", "ChannelTLSVerifyHostname",
    DEXTER_CONFIG_SETTING_TYPE_BOOLEAN, (DexterConfigSettingDefault) DEFAULT_CHANNEL_TLS_VERIFY_HOSTNAME },

  /* end of array sentinel */
  { NULL, NULL, DEXTER_CONFIG_SETTING_TYPE_INT, (DexterConfigSettingDefault) 0 }
};


GQuark
dexter_channel_error_quark (void)
{
  static GQuark error_quark = 0;

  if (error_quark == 0)
    error_quark = g_quark_from_static_string ("dexter-channel-error-quark");

  return error_quark;
}


G_DEFINE_TYPE (DexterChannel, dexter_channel, DEXTER_TYPE_CHANNEL_OBJECT);


static void
dexter_channel_class_init (DexterChannelClass *klass)
{
  GObjectClass             *gobject_class;
  DexterChannelObjectClass *channel_object_class;

#ifdef G_ENABLE_DEBUG
  g_message (G_STRFUNC);
#endif

  gobject_class = G_OBJECT_CLASS (klass);
  gobject_class->finalize = dexter_channel_finalize;

  channel_object_class = DEXTER_CHANNEL_OBJECT_CLASS (klass);
  channel_object_class->process_message = dexter_channel_process_message;
  channel_object_class->channel_event = dexter_channel_channel_event;
}


static void
dexter_channel_init (DexterChannel *channel)
{
#ifdef G_ENABLE_DEBUG
  g_message (G_STRFUNC);
#endif

  channel->settings = NULL;
  channel->thread = NULL;
  g_atomic_int_set (&channel->runnable, 0);
  g_atomic_int_set (&channel->state, DEXTER_CHANNEL_STATE_CLOSED);
  g_atomic_int_set (&channel->next_id, 1);
  channel->connfd = (-1);
  channel->session = dexter_network_session_new ();
  channel->cmd_queue = g_async_queue_new ();
  channel->object_hash = dexter_hash_table_new_full (g_int_hash, g_int_equal,
                                                     NULL, (GDestroyNotify) _destroy_object_hash_entry);
}


static void
dexter_channel_finalize (GObject *object)
{
  DexterChannel *channel = DEXTER_CHANNEL (object);

#ifdef G_ENABLE_DEBUG
  g_message (G_STRFUNC);
#endif

  dexter_channel_settings_free (channel->settings);
  dexter_network_session_free (channel->session);
    
  /* purge any pending commands */
  g_async_queue_lock (channel->cmd_queue);
  while (g_async_queue_length_unlocked (channel->cmd_queue) > 0)
  {
    DexterChannelMessage *msg = g_async_queue_pop_unlocked (channel->cmd_queue);
    dexter_channel_message_free (msg);
  }
  g_async_queue_unlock (channel->cmd_queue);
  /* release async queue */
  g_async_queue_unref (channel->cmd_queue);

  dexter_hash_table_destroy (channel->object_hash);

  /* chain up */
  G_OBJECT_CLASS (dexter_channel_parent_class)->finalize (object);
}


/**
 * dexter_channel_get_state:
 * @channel: a #DexterChannel.
 *
 * Returns the current state of the channel.
 *
 * Returns: #gint from #DexterChannelState indicating current state of the channel. 
 **/
gint
dexter_channel_get_state (DexterChannel *channel)
{
  gint state;
  g_return_val_if_fail (DEXTER_IS_CHANNEL (channel), -1);

  g_object_ref (channel);

  state = g_atomic_int_get (&channel->state);

  g_object_unref (channel);

  return state;
}


static gint
dexter_channel_set_state (DexterChannel *channel,
                          gint           state)
{
  gint old_state;
  g_return_val_if_fail (DEXTER_IS_CHANNEL (channel), -1);
  g_return_val_if_fail (state == DEXTER_CHANNEL_STATE_OPEN ||
                        state == DEXTER_CHANNEL_STATE_CLOSED ||
                        state == DEXTER_CHANNEL_STATE_ABORT ||
                        state == DEXTER_CHANNEL_STATE_NOCONN, -1);

  g_object_ref (channel);

  old_state = g_atomic_int_get (&channel->state);
  g_atomic_int_set (&channel->state, state);

  g_object_unref (channel);

  return old_state;
}


static gboolean 
dexter_channel_is_runnable (DexterChannel *channel)
{
  g_return_val_if_fail (DEXTER_IS_CHANNEL (channel), FALSE);

  gboolean runnable;

  g_object_ref (channel);

  runnable = g_atomic_int_get (&channel->runnable);

  g_object_unref (channel);

  return runnable;
}


static void
dexter_channel_set_runnable (DexterChannel *channel,
                             gboolean       runnable)
{
  g_return_if_fail (DEXTER_IS_CHANNEL (channel));

  g_object_ref (channel);

  g_atomic_int_set (&channel->runnable, runnable ? 1 : 0);

  g_object_unref (channel);
}


static void
dexter_channel_event (DexterChannel *channel,
                      gint           event)
{
  ChannelEventCallbackContainer *cb;

  g_return_if_fail (DEXTER_IS_CHANNEL (channel));

  g_object_ref (channel);

  /* for each channel object registered on this channel, call its channel_event handler. */
  dexter_hash_table_foreach (channel->object_hash, (GHFunc)_event_channel_object_func, &event);

  if (channel->settings->event_callback)
  {
    /* call the user's channel event handler on the threadpool */
    cb = g_slice_new0 (ChannelEventCallbackContainer);
    cb->channel = channel; /* channel is ref'd in the thread itself */
    cb->event = event;
    dexter_threadpool_push_task (__ITHREADPOOL,
                                 (GThreadFunc) _event_callback_func,
                                 (gpointer) cb, NULL);
  }

  g_object_unref (channel);
}


static void
dexter_channel_server_frame (DexterChannel  *channel,
                             DexterMsgFrame *frame)
{
  DexterChannelMessage *msg = NULL;
    
#ifdef G_ENABLE_DEBUG
  g_message ("%s: %s", G_STRFUNC, frame->message);
#endif
    
  g_return_if_fail (DEXTER_IS_CHANNEL (channel));

  g_object_ref (channel);

  /* frame could be either an RPY or a DAT frame. */

  /* try to parse for RPY message */
  msg = g_slice_new0 (DexterChannelMessage);
  if (sscanf (frame->message, RPY_FORMAT_STRING, 
              &msg->protocol_major, &msg->protocol_minor,
              &msg->rpy.command, &msg->peer_id, &msg->rpy.status) == 5)
  {
    msg->msg_type = DEXTER_MESSAGE_REPLY;

    /* find the channel object in the hash that this reply is for */
    ChannelObjectHashEntry *hash_entry = dexter_hash_table_lookup (channel->object_hash, &msg->peer_id);
    if (!hash_entry)
    {
      g_warning ("%s: received RPY message for unregistered channel object", G_STRFUNC);
      dexter_channel_message_free (msg);
      g_object_unref (channel);
      return;
    }

    /* move payload data from frame onto msg */
    msg->payload = frame->payload;
    frame->payload = NULL;

    if (hash_entry->channel_object->wait_types & DEXTER_MESSAGE_REPLY)
    {
      /* channel object is waiting for the reply message.  push it onto channel object's mailbox. */
      g_async_queue_push (hash_entry->channel_object->mbox, msg);
    }
    else
    {
      /* channel object is not waiting for a reply message, so process it. */ 
      GError *local_error = NULL;

      dexter_channel_object_process_message (hash_entry->channel_object, msg, &local_error);
      if (local_error)
      {
        g_warning ("%s", local_error->message);
        g_clear_error (&local_error);
      }
    }
  }
  else /* try to parse for DAT message */ 
  {
    memset (msg, 0, sizeof (DexterChannelMessage));
    if (sscanf (frame->message, DAT_FORMAT_STRING, 
                &msg->protocol_major, &msg->protocol_minor,
                &msg->peer_id, &msg->dat.flags, &msg->dat.status) == 5)
    {
      msg->msg_type = DEXTER_MESSAGE_DATA;

      /* find the channel object that will handle this data message */
      ChannelObjectHashEntry *hash_entry = dexter_hash_table_lookup (channel->object_hash, 
                                                                     &msg->peer_id);
      if (!hash_entry)
      {
        g_warning ("%s: received DAT message for unregistered channel object", G_STRFUNC);
        dexter_channel_message_free (msg);
        g_object_unref (channel);
        return;
      }
     
      /* move payload data from frame onto msg */
      msg->payload = frame->payload;
      frame->payload = NULL;

      if (hash_entry->channel_object->wait_types & DEXTER_MESSAGE_DATA)
      {
        /* channel object is waiting for data message.  push it onto channel object's mailbox. */
        g_async_queue_push (hash_entry->channel_object->mbox, msg);
      }
      else
      {
        /* channel object is not waiting for data message, so process it. */
        GError *local_error = NULL;

        dexter_channel_object_process_message (hash_entry->channel_object, msg, &local_error);
        if (local_error)
        {
          g_warning ("%s", local_error->message);
          g_clear_error (&local_error);
        }
      }
    }
    else
      g_warning ("%s: bad message format from server", G_STRFUNC);
  }

  g_object_unref (channel);
}


static void
dexter_channel_send_command (DexterChannel         *channel,
                             DexterChannelMessage  *message,
                             GError               **error)
{
  GString *strmessage = NULL;
    
#ifdef G_ENABLE_DEBUG
  g_message (G_STRFUNC);
#endif

  g_return_if_fail (DEXTER_IS_CHANNEL (channel));
  g_return_if_fail (message != NULL);
  g_return_if_fail (message->msg_type == DEXTER_MESSAGE_COMMAND);
    
  g_object_ref (channel);

  /* format a CMD message */
  strmessage = g_string_new (NULL);
  g_string_printf (strmessage, CMD_FORMAT_STRING,
  message->protocol_major, message->protocol_minor,
  message->cmd.command, message->peer_id, message->cmd.parameter);
  if (message->payload)
  {
    g_string_append_printf (strmessage, "%d\n%s\n", message->payload->len, message->payload->str);
  }
  else
  {
    g_string_append (strmessage, "0\n\n");
  }

  dexter_network_session_sendall (channel->session, strmessage->str, strmessage->len, 
#ifdef HAVE_MSG_NOSIGNAL
                                  MSG_NOSIGNAL,
#else
                                  0,
#endif
                                  error);

  g_string_free (strmessage, TRUE);

  g_object_unref (channel);
}


static void
dexter_channel_process_message (DexterChannelObject  *channel_object,
                                DexterChannelMessage *message,
                                GError              **error)
{
#ifdef G_ENABLE_DEBUG
  g_message (G_STRFUNC);
#endif

  g_return_if_fail (DEXTER_IS_CHANNEL_OBJECT (channel_object));
  g_return_if_fail (message != NULL);
  g_return_if_fail (message->msg_type == DEXTER_MESSAGE_REPLY);
  g_return_if_fail (error == NULL || *error == NULL);

  g_object_ref (channel_object);

  switch (message->rpy.command)
  {
    /* these are the commands originating from the channel itself */
    case COMMAND_HELLO:
    case COMMAND_GOODBYE:
    case COMMAND_NOOP:
    {
      if (message->rpy.status != STATUS_OK)
      {
        g_set_error (error,
                     DEXTER_CHANNEL_ERROR, DEXTER_CHANNEL_ERROR_FAILED,
                     "%s: server error (%d): %s",
                     G_STRFUNC, message->rpy.status, 
                     message->payload ? message->payload->str : 
                                        dexter_protocol_strerror (message->rpy.status)); 
      }
      break;
    }
    case COMMAND_STARTTLS:
    {
      if (message->rpy.status != STATUS_OK)
      {
        g_set_error (error,
                     DEXTER_CHANNEL_ERROR,
                     DEXTER_CHANNEL_ERROR_FAILED,
                     "%s: server error (%d): %s",
                     G_STRFUNC, message->rpy.status,
                     message->payload ? message->payload->str :
                                        dexter_protocol_strerror (message->rpy.status));
      }
      else
      {
#ifndef HAVE_TLS
        /* TLS not enabled */
        g_set_error (error,
                     DEXTER_CHANNEL_ERROR, DEXTER_CHANNEL_ERROR_TLS_NOT_ENABLED,
                     "%s: TLS not enabled", G_STRFUNC);
#else
        DexterChannel *channel = DEXTER_CHANNEL (channel_object);
  
        /* if specified, issue callback to finalize TLS parameters */
        if (channel->settings->tls_start_callback)
        {
          /* copy TLS parameters */
          DexterChannelTLSSettings *tls_settings =
            dexter_channel_tls_settings_new (channel->settings->tls_cert_file,
                                             channel->settings->tls_key_file,
                                             channel->settings->tls_trust_file,
                                             channel->settings->tls_crl_file,
                                             channel->settings->tls_verify_hostname);

          /* perform the callback */
          (*channel->settings->tls_start_callback) (channel,
                                                    dexter_network_session_peer_addr (channel->session),
                                                    dexter_network_session_peer_addrlen (channel->session),
                                                    tls_settings);

          /* start TLS handshake, with adjusted parameters */
          dexter_network_session_client_starttls (channel->session,
                                                  tls_settings->cert_file,
                                                  tls_settings->key_file,
                                                  tls_settings->trust_file,
                                                  tls_settings->crl_file,
                                                  tls_settings->verify_hostname,
                                                  error);

          dexter_channel_tls_settings_free (tls_settings);
        }
        else
        {
          /* start TLS handshake, with default parameters */
          dexter_network_session_client_starttls (channel->session,
                                                  channel->settings->tls_cert_file,
                                                  channel->settings->tls_key_file,
                                                  channel->settings->tls_trust_file,
                                                  channel->settings->tls_crl_file,
                                                  channel->settings->tls_verify_hostname,
                                                  error);
        }
#endif
      }
      break;
    }
    default:
    {
      g_set_error (error,
                   DEXTER_CHANNEL_ERROR, DEXTER_CHANNEL_ERROR_FAILED,
                   "%s: unexpected server reply", G_STRFUNC);
      break;
    }
  }
    
  /* chain up */
  DEXTER_CHANNEL_OBJECT_CLASS (dexter_channel_parent_class)->process_message (channel_object, message, error);

  g_object_unref (channel_object);
}


static void
dexter_channel_channel_event (DexterChannelObject *channel_object,
                              gint                 event)
{
  DexterChannel *channel;
#ifdef G_ENABLE_DEBUG
  g_message ("%s: event %d", G_STRFUNC, event);
#endif

  g_return_if_fail (DEXTER_IS_CHANNEL_OBJECT (channel_object));

  g_object_ref (channel_object);
  channel = DEXTER_CHANNEL (channel_object);

  switch (event)
  {
    case DEXTER_CHANNEL_EVENT_ABORT:
      dexter_channel_set_state (channel, DEXTER_CHANNEL_STATE_ABORT);
      break;
    case DEXTER_CHANNEL_EVENT_NOCONN:
      dexter_channel_set_state (channel, DEXTER_CHANNEL_STATE_NOCONN);
      break;
    case DEXTER_CHANNEL_EVENT_OPEN:
      dexter_channel_set_state (channel, DEXTER_CHANNEL_STATE_OPEN);
      break;
    case DEXTER_CHANNEL_EVENT_CLOSED:
      dexter_channel_set_state (channel, DEXTER_CHANNEL_STATE_CLOSED);
      break;
    default:
      /* programming error */
      g_error (G_STRLOC " unrecognized channel event (%d)", event);
  }

  /* chain up */
  DEXTER_CHANNEL_OBJECT_CLASS (dexter_channel_parent_class)->channel_event (channel_object, event);

  g_object_unref (channel_object);
}


static GError *
dexter_channel_convert_network_error (GError **network_error)
{
  GError *channel_error = NULL;

  g_return_val_if_fail ((network_error != NULL) && (*network_error != NULL), NULL);
  g_return_val_if_fail ((*network_error)->domain == DEXTER_NETWORK_ERROR, NULL);

  channel_error = g_error_copy (*network_error);
  channel_error->domain = DEXTER_CHANNEL_ERROR;
  switch ((*network_error)->code)
  {
    case DEXTER_NETWORK_ERROR_EADDRINUSE:
      channel_error->code = DEXTER_CHANNEL_ERROR_EADDRINUSE;
      break;
    case DEXTER_NETWORK_ERROR_EAGAIN:
      channel_error->code = DEXTER_CHANNEL_ERROR_EAGAIN;
      break;
    case DEXTER_NETWORK_ERROR_ECONNREFUSED:
      channel_error->code = DEXTER_CHANNEL_ERROR_ECONNREFUSED;
      break;
    case DEXTER_NETWORK_ERROR_EINTR:
      channel_error->code = DEXTER_CHANNEL_ERROR_EINTR;
      break;
    case DEXTER_NETWORK_ERROR_ENETUNREACH:
      channel_error->code = DEXTER_CHANNEL_ERROR_ENETUNREACH;
      break;
    case DEXTER_NETWORK_ERROR_ETIMEDOUT:
      channel_error->code = DEXTER_CHANNEL_ERROR_TIMEOUT;
      break;
    case DEXTER_NETWORK_ERROR_TLS_NOT_ENABLED:
      channel_error->code = DEXTER_CHANNEL_ERROR_TLS_NOT_ENABLED;
      break;
    case DEXTER_NETWORK_ERROR_FAILED:
    default:
      channel_error->code = DEXTER_CHANNEL_ERROR_FAILED;
      break;
  }
  g_clear_error (network_error);

  return channel_error;
}


static DexterChannelSettings *
dexter_channel_settings_dup (const DexterChannelSettings *settings)
{
  DexterChannelSettings *ret_settings;

  if (!G_UNLIKELY (settings))
    return NULL;

  ret_settings = g_slice_new0 (DexterChannelSettings);
  ret_settings->flags = settings->flags;
  ret_settings->host = g_strdup (settings->host);
  ret_settings->service = g_strdup (settings->service);
  ret_settings->socket_options = g_strdup (settings->socket_options);
  ret_settings->recv_buffer_size = settings->recv_buffer_size;
  ret_settings->timeout_msecs = settings->timeout_msecs;
  ret_settings->event_callback = settings->event_callback;
  ret_settings->tls_start_callback = settings->tls_start_callback;
  ret_settings->tls_cert_file = g_strdup (settings->tls_cert_file);
  ret_settings->tls_key_file = g_strdup (settings->tls_key_file);
  ret_settings->tls_trust_file = g_strdup (settings->tls_trust_file);
  ret_settings->tls_crl_file = g_strdup (settings->tls_crl_file);
  ret_settings->tls_verify_hostname = settings->tls_verify_hostname;

  return ret_settings;
}


static void
dexter_channel_settings_free (DexterChannelSettings *settings)
{
  if (G_UNLIKELY (!settings))
    return;

  g_free (settings->host);
  g_free (settings->service);
  g_free (settings->socket_options);
  g_free (settings->tls_cert_file);
  g_free (settings->tls_key_file);
  g_free (settings->tls_trust_file);
  g_free (settings->tls_crl_file);
  g_slice_free (DexterChannelSettings, settings);
}


static void
dexter_channel_settings_prepare (DexterChannelSettings *settings)
{
  gint iter;

  if (G_UNLIKELY (!settings))
    return;

  /* examine each possible server setting. if user did not supply a value for a setting,
     fetch a default from the config-file, if available; otherwise use an internal default. */
  
  /* array's final { NULL, ... } entry is end sentinel */
  iter=0;
  while (dexter_channel_config_settings [iter].config_group)
  {
    gint setting = 1<<iter;

    /* if item is user-supplied, no default needed */
    if (settings->flags & setting)
    {
      iter++;
      continue;
    }

    /* ok, item was not supplied by user */

    /* if item has no config-based default */
    if (!dexter_channel_config_settings [iter].config_key)
    {
      switch (setting)
      {
        case DEXTER_CHANNEL_SETTING_EVENT_CALLBACK:
          settings->event_callback = NULL;
          break;
        case DEXTER_CHANNEL_SETTING_TLS_START_CALLBACK:
          settings->tls_start_callback = NULL;
          break;
        default:
          g_error (G_STRLOC " unexpected config flag (%d)", 1<<iter);
          break;
      }

      iter++;
      continue;
    }

    /* item has a config-based default. grab explicit config value or a default config value if not found */
    DexterConfigSettingDefault return_default;
    dexter_config_get_setting (__ICONFIG,
                               dexter_channel_config_settings [iter].config_group,
                               dexter_channel_config_settings [iter].config_key,
                               dexter_channel_config_settings [iter].setting_type,
                               dexter_channel_config_settings [iter].setting_default,
                               &return_default);
    switch (setting)
    {
      case DEXTER_CHANNEL_SETTING_HOST:
        settings->host = return_default.default_string;
        break;
      case DEXTER_CHANNEL_SETTING_SERVICE:
        settings->service = return_default.default_string;
        break;
      case DEXTER_CHANNEL_SETTING_SOCKET_OPTIONS:
        settings->socket_options = return_default.default_string;
        break;
      case DEXTER_CHANNEL_SETTING_RECV_BUFFER_SIZE:
        settings->recv_buffer_size = return_default.default_int;
        break;
      case DEXTER_CHANNEL_SETTING_TIMEOUT_MSECS:
        settings->timeout_msecs = return_default.default_int;
        break;
      case DEXTER_CHANNEL_SETTING_TLS_CERT_FILE:
        settings->tls_cert_file = return_default.default_string;
        break;
      case DEXTER_CHANNEL_SETTING_TLS_KEY_FILE:
        settings->tls_key_file = return_default.default_string;
        break;
      case DEXTER_CHANNEL_SETTING_TLS_TRUST_FILE:
        settings->tls_trust_file = return_default.default_string;
        break;
      case DEXTER_CHANNEL_SETTING_TLS_CRL_FILE:
        settings->tls_crl_file = return_default.default_string;
        break;
      case DEXTER_CHANNEL_SETTING_TLS_VERIFY_HOSTNAME:
        settings->tls_verify_hostname = return_default.default_boolean;
        break;
      default:
        g_error (G_STRLOC " unexpected config flag (%d)", 1<<iter);
        break;
    } /* switch */

    iter++;

  } /* while */ 
}


#ifdef HAVE_TLS
static DexterChannelTLSSettings *
dexter_channel_tls_settings_new (const gchar *cert_file,
                                 const gchar *key_file,
                                 const gchar *trust_file,
                                 const gchar *crl_file,
                                 gboolean verify_hostname)
{
  DexterChannelTLSSettings *tls_settings = g_slice_new0 (DexterChannelTLSSettings);

  tls_settings->cert_file = g_strdup (cert_file);
  tls_settings->key_file = g_strdup (key_file);
  tls_settings->trust_file = g_strdup (trust_file);
  tls_settings->crl_file = g_strdup (crl_file);
  tls_settings->verify_hostname = verify_hostname;

  return tls_settings;
}


static void
dexter_channel_tls_settings_free (DexterChannelTLSSettings *tls_settings)
{
  if (G_UNLIKELY (!tls_settings))
    return;

  g_free (tls_settings->cert_file);
  g_free (tls_settings->key_file);
  g_free (tls_settings->trust_file);
  g_free (tls_settings->crl_file);
  g_slice_free (DexterChannelTLSSettings, tls_settings);
}
#endif


static void
_destroy_object_hash_entry (ChannelObjectHashEntry *hash_entry)
{
  g_object_unref (hash_entry->channel_object);
  g_slice_free (ChannelObjectHashEntry, hash_entry);
}


static void
_channel_thread_func (DexterChannel *channel)
{
  DexterMsgFramer *framer          = NULL;         /* message framer */
  GError          *local_error     = NULL;
  gchar           *recvbuf         = NULL;
  ssize_t          msgbytes;
    
  g_return_if_fail (DEXTER_IS_CHANNEL (channel));

#ifdef G_ENABLE_DEBUG
  g_message ("channel running in thread 0x%08X", (unsigned)g_thread_self());
#endif

  /* ref channel object so it wont finalize while we run */
  g_object_ref (channel);

  recvbuf = (gchar *) g_slice_alloc0 (channel->settings->recv_buffer_size);

  /* Initialize */
  framer = g_object_new (DEXTER_TYPE_MSGFRAMER, NULL);

  do
  {
    /* check if a command message is pending */
    DexterChannelMessage *msg = g_async_queue_try_pop (channel->cmd_queue);
    if ((msg) && (msg->msg_type==DEXTER_MESSAGE_COMMAND))
    {
      /* send command message */
      dexter_channel_send_command (channel, msg, &local_error);
      if (local_error)
      {
        g_warning ("%s: error sending command to server for connfd (%d): %s",
                   G_STRFUNC, channel->connfd, local_error->message);
        dexter_channel_message_free (msg);
        g_clear_error (&local_error);
        break;
      }
      dexter_channel_message_free (msg);
    }

    /* check for bytes from server */
    if (dexter_network_session_pending (channel->session))
    {
      /* read bytes from server and frame them into dexter messages */
      msgbytes=dexter_network_session_recv (channel->session, recvbuf, channel->settings->recv_buffer_size, 
                                            0, &local_error);
      if (local_error && 
          ((local_error->code==DEXTER_NETWORK_ERROR_EINTR) ||
           (local_error->code==DEXTER_NETWORK_ERROR_EAGAIN)))
      {
        /* non-fatal EINTR/EAGAIN */
        g_clear_error (&local_error);
        continue;
      }
      else if (local_error)
      {
        /* fatal */
        g_warning ("%s: error during network session (0x%08X) operation: %s",
                   G_STRFUNC, (unsigned int) channel->session, local_error->message);
        g_clear_error (&local_error);
        break;
      }

      if (msgbytes>0)
      {
        if (dexter_msgframer_push_bytes (framer, recvbuf, msgbytes) < 0)
        {
          /* bytes from the server are somehow problematic. abort session. */
          dexter_channel_event (channel, DEXTER_CHANNEL_EVENT_ABORT);
          break;
        }

        while (dexter_msgframer_frames_queued (framer) > 0)
        {
          /* process completed message frames from server */
          DexterMsgFrame *frame = dexter_msgframer_pop_frame (framer);
          dexter_channel_server_frame (channel, frame);
          dexter_msgframer_free_frame (framer, frame);
        }
      }
      else if (msgbytes==0)
      {
        /* server has closed connection */
        dexter_channel_event (channel, DEXTER_CHANNEL_EVENT_NOCONN);
        break;
      }
      else
      {
        /* error already handled above */
        break;
      }
    }
  }
  while (dexter_channel_is_runnable (channel) && (dexter_get_state () == DEXTER_STATE_UP));

  g_object_unref (framer);
  g_slice_free1 (channel->settings->recv_buffer_size, recvbuf);
  g_object_unref (channel);
 
#ifdef G_ENABLE_DEBUG
  g_message ("channel done in thread 0x%08X", (unsigned)g_thread_self());
#endif
}


static void
_event_callback_func (ChannelEventCallbackContainer *cb)
{
  g_return_if_fail (cb != NULL);

  /* prevent channel from finalizing while we run */
  g_object_ref (cb->channel);

  /* perform callback */
  (cb->channel->settings->event_callback) (cb->channel, cb->event);

  /* release channel ref */
  g_object_unref (cb->channel);

  /* free the callback container */
  g_slice_free (ChannelEventCallbackContainer, cb);
}


static void
_event_channel_object_func (gint                     *id,
                            ChannelObjectHashEntry   *hash_entry,
                            gint                     *event)
{
  g_return_if_fail (id != NULL);
  g_return_if_fail (hash_entry != NULL);
  g_return_if_fail (event != NULL);

  /* call the channel object's channel event handler */
  dexter_channel_object_channel_event (hash_entry->channel_object, *event);
}


/**
 * dexter_channel_new:
 * @settings: address of #DexterServerSettings or %NULL.
 *
 * Creates a channel using the supplied settings. 
 *
 * %NULL may be indicated for @settings, in which case the configuration file is used for all
 * channel settings.  Internal defaults are used for items not found in the configuration file.
 *
 * No network calls are performed until the channel is opened with dexter_channel_open().
 *
 * The channel must be freed when done with dexter_channel_free().
 *
 * Returns: newly allocated #DexterChannel if successful. %NULL otherwise.
 **/
DexterChannel *
dexter_channel_new (const DexterChannelSettings *settings)
{
  DexterChannel *channel;
  DexterChannelObject *channel_object;
 
#ifdef G_ENABLE_DEBUG
  g_message (G_STRFUNC);
#endif

  channel = g_object_new (DEXTER_TYPE_CHANNEL, NULL);
  channel_object = DEXTER_CHANNEL_OBJECT (channel);

  if (settings)
  {
    channel->settings = dexter_channel_settings_dup (settings);
  }
  else
  {
    channel->settings = g_slice_new0 (DexterChannelSettings);
  }
    
  dexter_channel_settings_prepare (channel->settings);

  /* register the channel as an object capable of receiving server messages */
  channel_object->wait_timeout = (glong)(1000 * channel->settings->timeout_msecs);
  dexter_channel_register_channel_object (channel, channel_object, DEXTER_MESSAGE_REPLY);

  return channel;
}


/**
 * dexter_channel_open:
 * @channel: #DexterChannel to open.
 * @error: return location for optional #GError.
 *
 * Opens the channel created with dexter_channel_new().  A new tcp/ip
 * connection is established to the server for the channel.  It remains 
 * open until dexter_channel_close() is called, the server disconnects,
 * or a network error occurs.
 *
 * In the event an open channel is broken due to a network or server problem,
 * the channel will notify all objects running on channel and then the channel 
 * will transition to @DEXTER_CHANNEL_STATE_NOCONN.
 *
 * Returns: #gboolean indicating success or failure.
 **/
gboolean
dexter_channel_open (DexterChannel *channel,
                     GError       **error)
{
  GError *local_error = NULL;
  struct sockaddr_storage addr;
  socklen_t addrlen;

#ifdef G_ENABLE_DEBUG
  g_message (G_STRFUNC);
#endif
    
  g_return_val_if_fail (DEXTER_IS_CHANNEL (channel), FALSE);
  g_return_val_if_fail (error == NULL || *error == NULL, FALSE);

  g_object_ref (channel);

  /* error if already open */
  if (dexter_channel_get_state (channel) == DEXTER_CHANNEL_STATE_OPEN)
  {
    g_set_error (error,
                 DEXTER_CHANNEL_ERROR, DEXTER_CHANNEL_ERROR_STATE,
                 "%s: channel already open", G_STRFUNC);
    g_object_unref (channel);
    return FALSE;
  }
  /* if channel connection was lost, the channel must first be closed before an open attempt */
  else if (dexter_channel_get_state (channel) == DEXTER_CHANNEL_STATE_NOCONN)
  {
    g_set_error (error,
                 DEXTER_CHANNEL_ERROR, DEXTER_CHANNEL_ERROR_STATE,
                 "%s: channel in NOCONN state must be closed", G_STRFUNC);
    g_object_unref (channel);
    return FALSE;
  }

  /* try to connect to server */
  channel->connfd = dexter_network_connect (channel->settings->host,
                                            channel->settings->service,
                                            channel->settings->socket_options, 
                                            &addr,
                                            &addrlen,
                                            &local_error);
  if (local_error)
  {
    g_propagate_error (error, dexter_channel_convert_network_error (&local_error));
    g_object_unref (channel);
    return FALSE;
  }

#ifdef G_ENABLE_DEBUG
  dexter_network_print_socket_options (channel->connfd);
#endif

  dexter_network_session_open (channel->session, channel->connfd, &addr, addrlen);
  dexter_channel_set_runnable (channel, TRUE);

  /* start threaded channel function.  we join this thread when we close the channel. */
  channel->thread = g_thread_create ((GThreadFunc) _channel_thread_func,
                                     channel,
                                     TRUE, /* joinable */
                                     &local_error);
  if (local_error)
  {
    g_propagate_error (error, local_error);
    g_object_unref (channel);
    return FALSE;
  }

  dexter_channel_event (channel, DEXTER_CHANNEL_EVENT_OPEN);

  g_object_unref (channel);
  
  return TRUE;
}


/**
 * dexter_channel_close:
 * @channel: #DexterChannel to close.
 * @error: return location for optional #GError.
 *
 * Closes a #DexterChannel previously opened with dexter_channel_open().
 * If the channel is already closed, no error occurs and %TRUE is returned.
 *
 * Returns: #gboolean indicating success or failure.
 **/
gboolean
dexter_channel_close (DexterChannel *channel,
                      GError       **error)
{
#ifdef G_ENABLE_DEBUG
  g_message (G_STRFUNC);
#endif
    
  g_return_val_if_fail (DEXTER_IS_CHANNEL (channel), FALSE);
  g_return_val_if_fail (error == NULL || *error == NULL, FALSE);

  g_object_ref (channel);

  /* ok if channel already closed */ 
  if (dexter_channel_get_state (channel) == DEXTER_CHANNEL_STATE_CLOSED)
  {
    g_object_unref (channel);
    return TRUE;
  }
    
  /* tell channel thread to stop */
  dexter_channel_set_runnable (channel, FALSE);

  /* wait for channel thread */
  if (channel->thread)
  {
    g_thread_join (channel->thread);
    channel->thread = NULL;
  }

  dexter_network_session_close (channel->session);
  dexter_channel_event (channel, DEXTER_CHANNEL_EVENT_CLOSED);

  g_object_unref (channel);

  return TRUE;
}


/**
 * dexter_channel_greet:
 * @channel: #DexterChannel to send greeting.
 * @error: return location for optional #GError.
 *
 * Sends the protocol greeting on a #DexterChannel previously opened with dexter_channel_open().
 * The greeting must be sent before using the channel, e.g. to create samplers.
 *
 * Returns: #gboolean indicating success or failure.
 **/
gboolean
dexter_channel_greet (DexterChannel *channel,
                      GError       **error)
{
  DexterChannelObject  *channel_object;
  GError               *local_error = NULL;

#ifdef G_ENABLE_DEBUG
  g_message (G_STRFUNC);
#endif

  g_return_val_if_fail (DEXTER_IS_CHANNEL (channel), FALSE);
  g_return_val_if_fail (error == NULL || *error == NULL, FALSE);

  g_object_ref (channel);

  channel_object = DEXTER_CHANNEL_OBJECT (channel);
  
  /* check if channel not open */
  if (G_UNLIKELY (dexter_channel_get_state (channel) != DEXTER_CHANNEL_STATE_OPEN))
  {
    g_set_error (error,
                 DEXTER_CHANNEL_ERROR, DEXTER_CHANNEL_ERROR_STATE,
                 "%s: channel not open", G_STRFUNC);
    g_object_unref (channel);
    return FALSE;
  }

  /* send hello command to open conversation */
  dexter_channel_object_send_wait_process (channel_object, COMMAND_HELLO, 0, NULL, &local_error);
  if (local_error)
  {
    g_propagate_error (error, local_error);
    g_object_unref (channel);
    return FALSE;
  }

  g_object_unref (channel);

  return TRUE;
}


/**
 * dexter_channel_noop:
 * @channel: #DexterChannel to send noop.
 * @error: return location for optional #GError.
 *
 * Sends a "no operation" command over the #DexterChannel to the connected server.
 *
 * Returns: #gboolean indicating success or failure.
 *
 * Since: 0.2.0
 **/
gboolean
dexter_channel_noop (DexterChannel *channel,
                     GError       **error)
{
  DexterChannelObject  *channel_object;
  GError               *local_error = NULL;

#ifdef G_ENABLE_DEBUG
  g_message (G_STRFUNC);
#endif

  g_return_val_if_fail (DEXTER_IS_CHANNEL (channel), FALSE);
  g_return_val_if_fail (error == NULL || *error == NULL, FALSE);

  g_object_ref (channel);

  channel_object = DEXTER_CHANNEL_OBJECT (channel);

  /* check if channel not open */
  if (G_UNLIKELY (dexter_channel_get_state (channel) != DEXTER_CHANNEL_STATE_OPEN))
  {
    g_set_error (error,
                 DEXTER_CHANNEL_ERROR, DEXTER_CHANNEL_ERROR_STATE,
                 "%s: channel not open", G_STRFUNC);
    g_object_unref (channel);
    return FALSE;
  }

  /* send noop command */
  dexter_channel_object_send_wait_process (channel_object, COMMAND_NOOP, 0, NULL, &local_error);
  if (local_error)
  {
    g_propagate_error (error, local_error);
    g_object_unref (channel);
    return FALSE;
  }

  g_object_unref (channel);

  return TRUE;
}


/**
 * dexter_channel_starttls:
 * @channel: #DexterChannel to send start tls request.
 * @error: return location for optional #GError.
 *
 * Initializes TLS over the currently opened channel.
 *
 * Returns: #gboolean indicating success or failure.
 * 
 * Since: 0.2.0
 **/
gboolean
dexter_channel_starttls (DexterChannel *channel,
                         GError       **error)
{
#ifdef HAVE_TLS
  DexterChannelObject *channel_object;
  GError              *local_error = NULL;
#endif

#ifdef G_ENABLE_DEBUG
  g_message (G_STRFUNC);
#endif

  g_return_val_if_fail (DEXTER_IS_CHANNEL (channel), FALSE);
  g_return_val_if_fail (error == NULL || *error == NULL, FALSE);

#ifndef HAVE_TLS
  /* TLS not enabled */
  g_set_error (error,
               DEXTER_CHANNEL_ERROR, DEXTER_CHANNEL_ERROR_TLS_NOT_ENABLED,
               "%s: TLS not enabled", G_STRFUNC);
  return FALSE;
#else
  /* TLS enabled, continue */
  
  g_object_ref (channel);

  channel_object = DEXTER_CHANNEL_OBJECT (channel);

  /* check if channel not open */
  if (G_UNLIKELY (dexter_channel_get_state (channel) != DEXTER_CHANNEL_STATE_OPEN))
  {
    g_set_error (error,
                 DEXTER_CHANNEL_ERROR, DEXTER_CHANNEL_ERROR_STATE,
                 "%s: channel not open", G_STRFUNC);
    g_object_unref (channel);
    return FALSE;
  }

  /* check if TLS already started */
  if (G_UNLIKELY (dexter_network_session_tls_running (channel->session)))
  {
    g_set_error (error,
                 DEXTER_CHANNEL_ERROR, DEXTER_CHANNEL_ERROR_FAILED,
                 "%s: TLS already running", G_STRFUNC);
    g_object_unref (channel);
    return FALSE;
  }

  /* send starttls command to server */
  dexter_channel_object_send_wait_process (channel_object, COMMAND_STARTTLS, 0, NULL, &local_error);
  if (local_error)
  {
    g_propagate_error (error, local_error);
    g_object_unref (channel);
    return FALSE;
  }

  g_object_unref (channel);

  return TRUE;
#endif /* HAVE_TLS */
}


/**
 * dexter_channel_part:
 * @channel: #DexterChannel to send parting message.
 * @error: return location for optional #GError.
 *
 * Sends protocol parting message on a #DexterChannel over which dexter_channel_greet()
 * has already been called.  Should be called prior to dexter_channel_close().
 *
 * Returns: #gboolean indicating success or failure.
 **/
gboolean
dexter_channel_part (DexterChannel  *channel,
                     GError        **error)
{
  DexterChannelObject  *channel_object;
  GError               *local_error = NULL;

#ifdef G_ENABLE_DEBUG
  g_message (G_STRFUNC);
#endif

  g_return_val_if_fail (DEXTER_IS_CHANNEL (channel), FALSE);
  g_return_val_if_fail (error == NULL || *error == NULL, FALSE);

  g_object_ref (channel);

  channel_object = DEXTER_CHANNEL_OBJECT (channel);
  
  /* check if channel not open */
  if (G_UNLIKELY (dexter_channel_get_state (channel) != DEXTER_CHANNEL_STATE_OPEN))
  {
    g_set_error (error,
                 DEXTER_CHANNEL_ERROR, DEXTER_CHANNEL_ERROR_STATE,
                 "%s: channel not open", G_STRFUNC);
    g_object_unref (channel);
    return FALSE;
  }

  /* send goodbye command to close conversation */
  dexter_channel_object_send_wait_process (channel_object, COMMAND_GOODBYE, 0, NULL, &local_error);
  if (local_error)
  {
    g_propagate_error (error, local_error);
    g_object_unref (channel);
    return FALSE;
  }

  g_object_unref (channel);

  return TRUE;
}


/**
 * dexter_channel_free:
 * @channel: #DexterChannel to free.
 *
 * Frees a #DexterChannel created with dexter_channel_new().  If the channel is not closed,
 * it will be closed for you, although it is better form to use dexter_channel_close() 
 * before freeing the channel.
 **/
void
dexter_channel_free (DexterChannel *channel)
{
#ifdef G_ENABLE_DEBUG
  g_message (G_STRFUNC);
#endif

  g_return_if_fail (channel == NULL || DEXTER_IS_CHANNEL (channel));

  if (G_UNLIKELY (!channel))
    return;

  g_object_ref (channel);

  if (dexter_channel_get_state (channel) != DEXTER_CHANNEL_STATE_CLOSED)
    dexter_channel_close (channel, NULL);

  /* deregister the channel as an object capable of receiving server messages */
  dexter_channel_deregister_channel_object (channel, DEXTER_CHANNEL_OBJECT (channel));

  /* release this api's ref */
  g_object_unref (channel);

  /* release object -- channel will finalize only after its channel_objects finalize 
   * since they hold channel references */
  g_object_unref (channel);
}


const DexterChannelSettings *
dexter_channel_get_settings (DexterChannel *channel)
{
  const DexterChannelSettings *settings;

  g_return_val_if_fail (DEXTER_IS_CHANNEL (channel), FALSE);

  g_object_ref (channel);

  settings = channel->settings;
  
  g_object_unref (channel);

  return settings;
}


gboolean
dexter_channel_register_channel_object (DexterChannel       *channel,
                                        DexterChannelObject *channel_object,
                                        gint                 wait_types)
{
  ChannelObjectHashEntry *hash_entry = NULL;

#ifdef G_ENABLE_DEBUG
  g_message (G_STRFUNC);
#endif

  g_return_val_if_fail (DEXTER_IS_CHANNEL (channel), FALSE);
  g_return_val_if_fail (DEXTER_IS_CHANNEL_OBJECT (channel_object), FALSE);

  g_object_ref (channel);
  g_object_ref (channel_object);

  /* create hash entry for channel object */
  hash_entry = g_slice_new0 (ChannelObjectHashEntry);
  hash_entry->id = g_atomic_int_add (&channel->next_id, 1);
  hash_entry->channel_object = g_object_ref (channel_object);
  dexter_hash_table_insert (channel->object_hash, &hash_entry->id, hash_entry);

  /* assign data for channel object */
  channel_object->channel = g_object_ref (channel);
  channel_object->id = hash_entry->id;
  channel_object->mbox = g_async_queue_new ();
  channel_object->wait_types = wait_types;
    
  g_object_unref (channel_object);
  g_object_unref (channel);

  return TRUE;
}


gboolean
dexter_channel_deregister_channel_object (DexterChannel       *channel,
                                          DexterChannelObject *channel_object)
{
#ifdef G_ENABLE_DEBUG
  g_message (G_STRFUNC);
#endif
    
  g_return_val_if_fail (DEXTER_IS_CHANNEL (channel), FALSE);
  g_return_val_if_fail (DEXTER_IS_CHANNEL_OBJECT (channel_object), FALSE);

  g_object_ref (channel);
  g_object_ref (channel_object);

  /* contained channel ref better be this channel */
  if (channel_object->channel != channel)
  {
    g_warning ("%s: channel object not registered on this channel", G_STRFUNC);
    g_object_unref (channel_object);
    g_object_unref (channel);
    return FALSE;
  }

  /* remove channel object from channel hash */
  if ( (channel_object->id <= 0) ||
       (!dexter_hash_table_remove (channel->object_hash, &channel_object->id)) )
  {
    g_warning ("%s: channel object not registered.  cannot deregister.", G_STRFUNC);
    g_object_unref (channel_object);
    g_object_unref (channel);
    return FALSE;
  }

  /* release channel ref to channel object */
  g_object_unref (channel_object->channel);
  channel_object->channel = NULL;
  channel_object->id = 0;

  /* remove any pending messages from channel object mbox */
  g_async_queue_lock (channel_object->mbox);
  while (g_async_queue_length_unlocked (channel_object->mbox) > 0)
  {
    DexterChannelMessage *msg = g_async_queue_pop_unlocked (channel_object->mbox);
    dexter_channel_message_free (msg);
  }
  g_async_queue_unlock (channel_object->mbox);

  /* release channel object mbox */
  g_async_queue_unref (channel_object->mbox);
  channel_object->mbox = NULL;
    
  g_object_unref (channel_object);
  g_object_unref (channel);

  return TRUE;
}


void
dexter_channel_command_push (DexterChannel       *channel,
                             DexterChannelObject *channel_object,
                             gint                 command,
                             gint                 parameter,
                             GString             *payload,
                             GError             **error)
{
  DexterChannelMessage *msg = NULL;
    
#ifdef G_ENABLE_DEBUG
  g_message (G_STRFUNC);
#endif

  g_return_if_fail (DEXTER_IS_CHANNEL (channel));
  g_return_if_fail (DEXTER_IS_CHANNEL_OBJECT (channel_object));
  g_return_if_fail (command > 0);
  g_return_if_fail (channel == channel_object->channel);
  g_return_if_fail (error == NULL || *error == NULL);

  g_object_ref (channel);
  g_object_ref (channel_object);

  if ((command != COMMAND_HELLO) && (dexter_channel_get_state (channel) != DEXTER_CHANNEL_STATE_OPEN))
  {
    g_set_error (error,
                 DEXTER_CHANNEL_ERROR, DEXTER_CHANNEL_ERROR_STATE,
                 "%s: cannot send command: channel is not open", G_STRFUNC);
    g_object_unref (channel_object);
    g_object_unref (channel);
    return;
  }

  msg = g_slice_new0 (DexterChannelMessage);
  msg->msg_type = DEXTER_MESSAGE_COMMAND;
  msg->protocol_major = DEXTER_PROTOCOL_MAJOR;
  msg->protocol_minor = DEXTER_PROTOCOL_MINOR;
  msg->peer_id = channel_object->id;
  msg->payload = payload;
  msg->cmd.command = command;
  msg->cmd.parameter = parameter;

  g_async_queue_push (channel->cmd_queue, msg);

  g_object_unref (channel_object);
  g_object_unref (channel);
}

#define __DEXTER_CHANNEL_C__
#include <dexter-aliasdef.c>
