/* $Id: dexter-channel.h 393 2012-03-06 23:36:44Z pkovacs $ */
/*
 * libdexter - a plugin-based, distributed sampling library.
 * Copyright (C) 2006-2012 Philip Kovacs <pkovacs@users.sourceforge.net>.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 2.1
 * of the License, or (at your option) any later version.

 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.

 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
 * 02110-1301  USA
 *
*/

#ifndef __DEXTER_CHANNEL_H__
#define __DEXTER_CHANNEL_H__

#include <sys/types.h>
#include <sys/socket.h>
#include <glib.h>

G_BEGIN_DECLS

/**
 * DEXTER_CHANNEL_ERROR:
 *
 * The error domain of the libdexter channel subsystem.
 **/
#define DEXTER_CHANNEL_ERROR dexter_channel_error_quark()
GQuark dexter_channel_error_quark (void);

/**
 * DexterChannelError:
 * @DEXTER_CHANNEL_ERROR_EADDRINUSE: local address in use.
 * @DEXTER_CHANNEL_ERROR_EAGAIN: no more free ports or insufficient routing cache entries.
 * @DEXTER_CHANNEL_ERROR_ECONNREFUSED: no one listening on the remote address.
 * @DEXTER_CHANNEL_ERROR_EINTR: system call interrupted by signal that was caught.
 * @DEXTER_CHANNEL_ERROR_ENETUNREACH: network is unreachable.
 * @DEXTER_CHANNEL_ERROR_STATE: operation not allowed in current channel state.
 * @DEXTER_CHANNEL_ERROR_TIMEOUT: operation timed out.
 * @DEXTER_CHANNEL_ERROR_TLS_NOT_ENABLED: TLS not enabled on this installation.
 * @DEXTER_CHANNEL_ERROR_FAILED: other failure.
 *
 * Error codes for #DexterChannel operations.
 **/
typedef enum
{
  DEXTER_CHANNEL_ERROR_EADDRINUSE,
  DEXTER_CHANNEL_ERROR_EAGAIN,
  DEXTER_CHANNEL_ERROR_ECONNREFUSED,
  DEXTER_CHANNEL_ERROR_EINTR,
  DEXTER_CHANNEL_ERROR_ENETUNREACH,
  DEXTER_CHANNEL_ERROR_STATE,
  DEXTER_CHANNEL_ERROR_TIMEOUT,
  DEXTER_CHANNEL_ERROR_TLS_NOT_ENABLED,
  DEXTER_CHANNEL_ERROR_FAILED
} DexterChannelError;


/**
 * DexterChannel:
 *
 * Opaque type with no public members.
 **/
typedef struct _DexterChannel DexterChannel;


/**
 * DexterChannelState:
 * @DEXTER_CHANNEL_STATE_OPEN: channel is open.
 * @DEXTER_CHANNEL_STATE_ABORT: channel aborted connection.
 * @DEXTER_CHANNEL_STATE_NOCONN: server disconnected.
 * @DEXTER_CHANNEL_STATE_CLOSED: channel is closed.
 *
 * All possible channel states.
 **/
typedef enum {
  DEXTER_CHANNEL_STATE_OPEN,
  DEXTER_CHANNEL_STATE_ABORT,
  DEXTER_CHANNEL_STATE_NOCONN,
  DEXTER_CHANNEL_STATE_CLOSED,
} DexterChannelState;

/**
 * DexterChannelEvent:
 * @DEXTER_CHANNEL_EVENT_OPEN: channel opened successfully.
 * @DEXTER_CHANNEL_EVENT_ABORT: channel aborted connection.
 * @DEXTER_CHANNEL_EVENT_NOCONN: server disconnected.
 * @DEXTER_CHANNEL_EVENT_CLOSED: channel closed successfully.
 *
 * All possible channel events.
 **/
typedef enum {
  DEXTER_CHANNEL_EVENT_OPEN,
  DEXTER_CHANNEL_EVENT_ABORT,
  DEXTER_CHANNEL_EVENT_NOCONN,
  DEXTER_CHANNEL_EVENT_CLOSED
} DexterChannelEvent;


/**
 * DexterChannelEventCallback:
 * @channel: #DexterChannel firing the event.
 * @event: #DexterChannelEvent fired.
 *
 * Signature of callback function used to receive channel event notifications.
 **/
typedef void (*DexterChannelEventCallback)  (DexterChannel *channel,
                                             gint           event);


typedef struct _DexterChannelTLSSettings DexterChannelTLSSettings;
/**
 * DexterChannelTLSSettings:
 * @cert_file: full path to X.509 certificate file in PEM format.
 * @key_file: full path to X.509 certificate private key file in PEM format.
 * @trust_file: full path to X.509 trust file in PEM format.
 * @crl_file: full path to X.509 certificate revocation list file in PEM format.
 * @verify_hostname: verify that the hostname of server matches its X.509 certificate.
 *
 * TLS settings that the channel will use when starting TLS.  They are copied from the 
 * #DexterChannelSettings structure provided to dexter_channel_new ().
 *
 * When this structure is seen within the context of a #DexterChannelTLSStartCallback,
 * you may change the elements as needed.  For any non-%NULL %gchar elements that you want
 * to change, you must first g_free() the element and then allocate a replacement string
 * using a GLib string allocation api, e.g. g_strdup_printf().  The Library will free
 * everything for you when the callback exits.
 *
 * Since: 0.2.0
 **/
struct _DexterChannelTLSSettings
{
  /*< public >*/
  gchar    *cert_file;        /* ChannelTLSCertFile */ 
  gchar    *key_file;         /* ChannelTLSKeyFile */
  gchar    *trust_file;       /* ChannelTLSTrustFile */
  gchar    *crl_file;         /* ChannelTLSCrlFile  */
  gboolean  verify_hostname;  /* ChannelTLSVerifyHostname */
};


/**
 * DexterChannelTLSStartCallback:
 * @channel: #DexterChannel on which TLS is starting.
 * @server_sa: server socket address.
 * @server_salen: length of server socket address.
 * @tls_settings: #DexterChannelTLSSettings of TLS settings.
 *
 * Signature of callback function used to adjust TLS settings, as needed for this server address, 
 * as TLS starts.
 *
 * If specificed in dexter_channel_starttls(), this callback allows you to inspect the server
 * socket address and adjust your channel TLS parameters as needed.  In this manner you can decide 
 * which certificate, trust and revocation files to use based on the server you are connected to.
 *
 * Since: 0.2.0
 **/
typedef void (*DexterChannelTLSStartCallback) (DexterChannel *channel,
                                               const struct sockaddr *server_sa,
                                               socklen_t server_salen,
                                               DexterChannelTLSSettings *tls_settings);

/**
 * DexterChannelSettingFlags:
 * @DEXTER_CHANNEL_SETTING_HOST: flag for host element of #DexterServerSettings.
 * @DEXTER_CHANNEL_SETTING_SERVICE: flag for service element of #DexterServerSettings.
 * @DEXTER_CHANNEL_SETTING_SOCKET_OPTIONS: flag for socket_options element of #DexterServerSettings.
 * @DEXTER_CHANNEL_SETTING_RECV_BUFFER_SIZE: flag for recv_buffer_size element of #DexterServerSettings.
 * @DEXTER_CHANNEL_SETTING_TIMEOUT_MSECS: flag for timeout_msecs element of #DexterServerSettings.
 * @DEXTER_CHANNEL_SETTING_EVENT_CALLBACK: flag for event_callback element of #DexterServerSettings.
 * @DEXTER_CHANNEL_SETTING_TLS_START_CALLBACK: flag for tls_start_callback element of #DexterServerSettings.
 * @DEXTER_CHANNEL_SETTING_TLS_CERT_FILE: flag for tls_cert_file element of #DexterServerSettings.
 * @DEXTER_CHANNEL_SETTING_TLS_KEY_FILE: flag for tls_key_file element of #DexterServerSettings.
 * @DEXTER_CHANNEL_SETTING_TLS_TRUST_FILE: flag for tls_trust_file element of #DexterServerSettings.
 * @DEXTER_CHANNEL_SETTING_TLS_CRL_FILE: flag for tls_cert_file element of #DexterServerSettings.
 * @DEXTER_CHANNEL_SETTING_TLS_VERIFY_HOSTNAME: flag for tls_verify_hostname element of #DexterServerSettings.
 *
 * Flags to indicate which elements of #DexterChannelSettings have data on them.
 *
 * Since 0.2.0
 **/
typedef enum
{
  DEXTER_CHANNEL_SETTING_HOST                 = 1 << 0,
  DEXTER_CHANNEL_SETTING_SERVICE              = 1 << 1,
  DEXTER_CHANNEL_SETTING_SOCKET_OPTIONS       = 1 << 2,
  DEXTER_CHANNEL_SETTING_RECV_BUFFER_SIZE     = 1 << 3,
  DEXTER_CHANNEL_SETTING_TIMEOUT_MSECS        = 1 << 4,
  DEXTER_CHANNEL_SETTING_EVENT_CALLBACK       = 1 << 5,
  DEXTER_CHANNEL_SETTING_TLS_START_CALLBACK   = 1 << 6,
  DEXTER_CHANNEL_SETTING_TLS_CERT_FILE        = 1 << 7,
  DEXTER_CHANNEL_SETTING_TLS_KEY_FILE         = 1 << 8,
  DEXTER_CHANNEL_SETTING_TLS_TRUST_FILE       = 1 << 9,
  DEXTER_CHANNEL_SETTING_TLS_CRL_FILE         = 1 << 10,
  DEXTER_CHANNEL_SETTING_TLS_VERIFY_HOSTNAME  = 1 << 11
} DexterChannelSettingFlags;


typedef struct _DexterChannelSettings DexterChannelSettings;
/**
 * DexterChannelSettings:
 * @flags: Bitwise OR of flags from #DexterChannelSettingFlags.
 * @host: host name or host ip.
 * @service: service name or port number.
 * @socket_options: tab or space-delimited list of socket options.
 * @recv_buffer_size: size of channel socket receive buffer.
 * @timeout_msecs: milliseconds to wait for server response before timing out.
 * @event_callback: callback to receive channel event notifications.
 * @tls_start_callback: callback to override TLS settings as TLS starts.
 * @tls_cert_file: full path to X.509 certificate file in PEM format.
 * @tls_key_file: full path to X.509 certificate private key file in PEM format.
 * @tls_trust_file: full path to X.509 trust file in PEM format.
 * @tls_crl_file: full path to X.509 certificate revocation list file in PEM format.
 * @tls_verify_hostname: whether or not client requires that the server's DNS hostname match
 *                       its X.509 certificate.
 *
 * Channel settings.
 *
 * The @flags element indicates which elements have user-supplied data.
 *
 * Elements for which no flag is set, i.e. have no user-supplied data, will be set to a default:
 * either from the configuration file, if available, or an internal default will be used.
 *
 * Since: 0.2.0
 **/
struct _DexterChannelSettings
{
  /*< public >*/
  DexterChannelSettingFlags     flags;
  gchar   *host;                /* ChannelHost */
  gchar   *service;             /* ChannelService */
  gchar   *socket_options;      /* ChannelSocketOptions */
  gint     recv_buffer_size;    /* ChannelRecvBufferSize */
  gint     timeout_msecs;       /* ChannelTimeoutMsecs */
  DexterChannelEventCallback    event_callback;         /* -- */
  DexterChannelTLSStartCallback tls_start_callback;     /* -- */
  gchar   *tls_cert_file;       /* ChannelTLSCertFile */
  gchar   *tls_key_file;        /* ChannelTLSKeyFile */
  gchar   *tls_trust_file;      /* ChannelTLSTrustFile */
  gchar   *tls_crl_file;        /* ChannelTLSCrlFile */
  gboolean tls_verify_hostname; /* ChannelTLSVerifyHostname */
};


DexterChannel   *dexter_channel_new         (const DexterChannelSettings *settings) G_GNUC_MALLOC;

gboolean         dexter_channel_open        (DexterChannel *channel,
                                             GError       **error);

gboolean         dexter_channel_close       (DexterChannel *channel,
                                             GError       **error);

gboolean         dexter_channel_greet       (DexterChannel *channel,
                                             GError       **error);

gboolean         dexter_channel_noop        (DexterChannel *channel,
                                             GError       **error);

gboolean         dexter_channel_starttls    (DexterChannel *channel,
                                             GError       **error);

gboolean         dexter_channel_part        (DexterChannel *channel,
                                             GError       **error);

void             dexter_channel_free        (DexterChannel *channel);

gint             dexter_channel_get_state   (DexterChannel *channel);

G_END_DECLS

#endif /* __DEXTER_CHANNEL_H__ */
