/* $Id: dexter-config.c 393 2012-03-06 23:36:44Z pkovacs $ */
/*
 * libdexter - a plugin-based, distributed sampling library.
 * Copyright (C) 2006-2012 Philip Kovacs <pkovacs@users.sourceforge.net>.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 2.1
 * of the License, or (at your option) any later version.

 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.

 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
 * 02110-1301  USA
 *
*/

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <dexter-config.h>
#include <dexter-alias.h>

/* GObject implementing iconfig interface */
static void  dexter_config_class_init      (DexterConfigClass              *klass);
static void  dexter_config_init            (DexterConfig                   *config);
static void  dexter_config_finalize        (GObject                        *object);
static void  dexter_config_iconfig_init    (DexterPluginIConfigInterface   *iconfig);

struct _DexterConfig 
{
  GObject   parent;
  GKeyFile *keyfile;
};

struct _DexterConfigClass 
{
  GObjectClass parent_class;
};


G_DEFINE_TYPE_WITH_CODE (DexterConfig, dexter_config, G_TYPE_OBJECT,
                         G_IMPLEMENT_INTERFACE (DEXTERPLUGIN_TYPE_ICONFIG,
                                                dexter_config_iconfig_init));

static void
dexter_config_class_init (DexterConfigClass *klass)
{
  GObjectClass *gobject_class;

#ifdef G_ENABLE_DEBUG
  g_message (G_STRFUNC);
#endif

  gobject_class = G_OBJECT_CLASS (klass);
  gobject_class->finalize = dexter_config_finalize;
}


static void
dexter_config_init (DexterConfig *config)
{
#ifdef G_ENABLE_DEBUG
  g_message (G_STRFUNC);
#endif

  config->keyfile = g_key_file_new ();
}


static void
dexter_config_finalize (GObject *object)
{
  DexterConfig *config = DEXTER_CONFIG (object);

#ifdef G_ENABLE_DEBUG
  g_message (G_STRFUNC);
#endif

  g_key_file_free (config->keyfile);

  /* chain up */
  G_OBJECT_CLASS (dexter_config_parent_class)->finalize (object);
}


static void
dexter_config_iconfig_init (DexterPluginIConfigInterface *iconfig)
{
#ifdef G_ENABLE_DEBUG
  g_message (G_STRFUNC);
#endif
    
  iconfig->has_group = dexter_config_has_group;
  iconfig->get_groups = dexter_config_get_groups;
  iconfig->get_keys = dexter_config_get_keys;
  iconfig->get_string = dexter_config_get_string;
  iconfig->get_string_list = dexter_config_get_string_list;
  iconfig->get_boolean = dexter_config_get_boolean;
  iconfig->get_boolean_list = dexter_config_get_boolean_list;
  iconfig->get_integer = dexter_config_get_integer;
  iconfig->get_integer_list = dexter_config_get_integer_list;
  iconfig->key_error = dexter_config_key_error;
}


gboolean
dexter_config_load  (DexterConfig *config,
                     const gchar  *filename,
                     GError      **error)
{
  GError *local_error = NULL;

  g_return_val_if_fail (DEXTER_IS_CONFIG (config), FALSE);
  g_return_val_if_fail (filename != NULL, FALSE);
  g_return_val_if_fail (error == NULL || *error == NULL, FALSE);

  g_key_file_load_from_file (config->keyfile, filename, G_KEY_FILE_NONE, &local_error);
  if (local_error)
  {
    g_propagate_error (error, local_error);
    return FALSE;
  }

  return TRUE;
}


gboolean 
dexter_config_has_group (DexterPluginIConfig *iconfig,
                         const gchar         *group)
{
  DexterConfig *config = NULL;

  g_return_val_if_fail (DEXTERPLUGIN_IS_ICONFIG (iconfig), FALSE);
  g_return_val_if_fail (group != NULL, FALSE);

  config = DEXTER_CONFIG (iconfig);
  return g_key_file_has_group (config->keyfile, group);
}


gboolean
dexter_config_get_groups (DexterPluginIConfig   *iconfig,
                          gchar               ***return_groups,
                          gsize                 *length)
{
  DexterConfig *config = NULL;

  g_return_val_if_fail (DEXTERPLUGIN_IS_ICONFIG (iconfig), FALSE);
  g_return_val_if_fail (return_groups != NULL, FALSE);

  config = DEXTER_CONFIG (iconfig);
  *return_groups = g_key_file_get_groups (config->keyfile, length);
  return TRUE;
}


gboolean
dexter_config_get_keys (DexterPluginIConfig   *iconfig,
                        const gchar           *group,
                        gchar               ***return_keys,
                        gsize                  *length,
                        GError                **error)
{
  GError        *local_error = NULL;
  gchar        **values;
  gsize          len;
  DexterConfig  *config;

  g_return_val_if_fail (DEXTERPLUGIN_IS_ICONFIG (iconfig), FALSE);
  g_return_val_if_fail (group != NULL, FALSE);
  g_return_val_if_fail (return_keys != NULL, FALSE);
  g_return_val_if_fail (error == NULL || *error == NULL, FALSE);

  config = DEXTER_CONFIG (iconfig);
  values = g_key_file_get_keys (config->keyfile, group, &len, &local_error);
  if (local_error)
  {
    g_propagate_error (error, local_error);
    return FALSE;
  }

  *return_keys = values;
  *length = len;
  return TRUE;
}


gboolean
dexter_config_get_string (DexterPluginIConfig  *iconfig,
                          const gchar          *group,
                          const gchar          *key,
                          gchar               **return_string,
                          GError              **error)
{
  GError       *local_error = NULL;
  gchar        *value;
  DexterConfig *config;

  g_return_val_if_fail (DEXTERPLUGIN_IS_ICONFIG (iconfig), FALSE);
  g_return_val_if_fail (group != NULL, FALSE);
  g_return_val_if_fail (key != NULL, FALSE);
  g_return_val_if_fail (return_string != NULL, FALSE);
  g_return_val_if_fail (error == NULL || *error == NULL, FALSE);
    
  config = DEXTER_CONFIG (iconfig);
  value = g_key_file_get_string (config->keyfile, group, key, &local_error);
  if (local_error)
  {
    g_propagate_error (error, local_error);
    return FALSE;
  }
    
  *return_string = value;
  return TRUE;
}


gboolean
dexter_config_get_string_list (DexterPluginIConfig   *iconfig,
                               const gchar           *group,
                               const gchar           *key,
                               gchar               ***return_strings,
                               gsize                 *length,
                               GError               **error)
{
  GError        *local_error = NULL;
  gchar        **values;
  gsize          len;   
  DexterConfig  *config;

  g_return_val_if_fail (DEXTERPLUGIN_IS_ICONFIG (iconfig), FALSE);
  g_return_val_if_fail (group != NULL, FALSE);
  g_return_val_if_fail (key != NULL, FALSE);
  g_return_val_if_fail (return_strings != NULL, FALSE);
  g_return_val_if_fail (error == NULL || *error == NULL, FALSE);

  config = DEXTER_CONFIG (iconfig);
  values = g_key_file_get_string_list (config->keyfile, group, key, &len, &local_error);
  if (local_error)
  {
    g_propagate_error (error, local_error);
    return FALSE;
  }

  *return_strings = values;
  *length = len;
  return TRUE;
}


gboolean
dexter_config_get_boolean (DexterPluginIConfig  *iconfig,
                           const gchar          *group,
                           const gchar          *key,
                           gboolean             *return_boolean,
                           GError              **error)
{
  GError       *local_error = NULL;
  gboolean      value;
  DexterConfig *config;

  g_return_val_if_fail (DEXTERPLUGIN_IS_ICONFIG (iconfig), FALSE);
  g_return_val_if_fail (group != NULL, FALSE);
  g_return_val_if_fail (key != NULL, FALSE);
  g_return_val_if_fail (return_boolean != NULL, FALSE);
  g_return_val_if_fail (error == NULL || *error == NULL, FALSE);

  config = DEXTER_CONFIG (iconfig);
  value = g_key_file_get_boolean (config->keyfile, group, key, &local_error);
  if (local_error)
  {
    g_propagate_error (error, local_error);
    return FALSE;
  }

  *return_boolean = value;
  return TRUE;
}

gboolean
dexter_config_get_boolean_list (DexterPluginIConfig   *iconfig,
                                const gchar           *group,
                                const gchar           *key,
                                gboolean             **return_booleans,
                                gsize                 *length,
                                GError              **error)
{
  GError       *local_error = NULL;
  gboolean     *values;
  gsize         len;
  DexterConfig *config;

  g_return_val_if_fail (DEXTERPLUGIN_IS_ICONFIG (iconfig), FALSE);
  g_return_val_if_fail (group != NULL, FALSE);
  g_return_val_if_fail (key != NULL, FALSE);
  g_return_val_if_fail (return_booleans != NULL, FALSE);
  g_return_val_if_fail (error == NULL || *error == NULL, FALSE);

  config = DEXTER_CONFIG (iconfig);
  values = g_key_file_get_boolean_list (config->keyfile, group, key, &len, &local_error);
  if (local_error)
  {
    g_propagate_error (error, local_error);
    return FALSE;
  }

  *return_booleans = values;
  *length = len;
  return TRUE;
}


gboolean
dexter_config_get_integer (DexterPluginIConfig  *iconfig,
                           const gchar          *group,
                           const gchar          *key,
                           gint                 *return_int,
                           GError              **error)
{
  GError       *local_error = NULL;
  gint          value;
  DexterConfig *config;

  g_return_val_if_fail (DEXTERPLUGIN_IS_ICONFIG (iconfig), FALSE);
  g_return_val_if_fail (group != NULL, FALSE);
  g_return_val_if_fail (key != NULL, FALSE);
  g_return_val_if_fail (return_int != NULL, FALSE);
  g_return_val_if_fail (error == NULL || *error == NULL, FALSE);

  config = DEXTER_CONFIG (iconfig);
  value = g_key_file_get_integer (config->keyfile, group, key, &local_error);
  if (local_error)
  {
    g_propagate_error (error, local_error);
     return FALSE;
  }

  *return_int = value;
  return TRUE;
}


gboolean
dexter_config_get_integer_list (DexterPluginIConfig   *iconfig,
                                const gchar           *group,
                                const gchar           *key,
                                gint                **return_ints,
                                gsize                 *length,
                                GError               **error)
{
  GError       *local_error = NULL;
  gint         *values;
  gsize         len;
  DexterConfig *config;

  g_return_val_if_fail (DEXTERPLUGIN_IS_ICONFIG (iconfig), FALSE);
  g_return_val_if_fail (group != NULL, FALSE);
  g_return_val_if_fail (key != NULL, FALSE);
  g_return_val_if_fail (return_ints != NULL, FALSE);
  g_return_val_if_fail (error == NULL || *error == NULL, FALSE);

  config = DEXTER_CONFIG (iconfig);
  values = g_key_file_get_integer_list (config->keyfile, group, key, &len, &local_error);
  if (local_error)
  {
    g_propagate_error (error, local_error);
    return FALSE;
  }

  *return_ints = values;
  *length = len;
  return TRUE;
}


gboolean
dexter_config_key_error (DexterPluginIConfig *iconfig,
                         GError              *error)
{
  g_return_val_if_fail (DEXTERPLUGIN_IS_ICONFIG (iconfig), FALSE);
  g_return_val_if_fail (error != NULL, FALSE);

  return (error->code == G_KEY_FILE_ERROR_KEY_NOT_FOUND ||
          error->code == G_KEY_FILE_ERROR_GROUP_NOT_FOUND);
}


gboolean
dexter_config_get_setting (DexterPluginIConfig        *iconfig,
                           const gchar                *group,
                           const gchar                *key,
                           DexterConfigSettingType     setting_type,
                           DexterConfigSettingDefault  setting_default,
                           DexterConfigSettingDefault *return_setting)
{
  g_return_val_if_fail (DEXTERPLUGIN_IS_ICONFIG (iconfig), FALSE);
  g_return_val_if_fail (group != NULL, FALSE);
  g_return_val_if_fail (key != NULL, FALSE);
  g_return_val_if_fail (return_setting != NULL, FALSE);

  switch (setting_type)
  {
    case DEXTER_CONFIG_SETTING_TYPE_BOOLEAN:
      {
        if (!dexter_config_get_boolean (iconfig,
                                        group,
                                        key,
                                        &return_setting->default_boolean,
                                        NULL))
        {
          g_warning ("error reading config item (%s). using default (%d).",
                     key, setting_default.default_boolean);
          return_setting->default_boolean = setting_default.default_boolean;
        };
        break;
      }
    case DEXTER_CONFIG_SETTING_TYPE_INT:
      {
        if (!dexter_config_get_integer (iconfig,
                                        group,
                                        key,
                                        &return_setting->default_int,
                                        NULL))
        {
          g_warning ("error reading config item (%s). using default (%d).",
                     key, setting_default.default_int);
          return_setting->default_int = setting_default.default_int;
        }
        break;
      }
    case DEXTER_CONFIG_SETTING_TYPE_STRING:
      {
        if (!dexter_config_get_string (iconfig,
                                       group,
                                       key,
                                       &return_setting->default_string,
                                       NULL))
        {
          g_warning ("error reading config item (%s). using default (%s).",
                     key, setting_default.default_string);
         return_setting->default_string = g_strdup (setting_default.default_string);
        }
        break;
      }
    default:
      /* programming error */
      g_error (G_STRLOC " unexpected config setting type (%d)", setting_type);
      return FALSE;
  };

  return TRUE;
}

#define __DEXTER_CONFIG_C__
#include <dexter-aliasdef.c>
