/* $Id: dexter-event-loop.c 393 2012-03-06 23:36:44Z pkovacs $ */
/*
 * libdexter - a plugin-based, distributed sampling library.
 * Copyright (C) 2006-2012 Philip Kovacs <pkovacs@users.sourceforge.net>.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 2.1
 * of the License, or (at your option) any later version.

 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.

 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
 * 02110-1301  USA
 *
*/

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <dexter-event-loop.h>
#include <dexter-private.h>
#include <dexter-alias.h>

static void     dexter_event_loop_class_init      (DexterEventLoopClass *klass);
static void     dexter_event_loop_init            (DexterEventLoop      *event_loop);
static void     dexter_event_loop_finalize        (GObject              *object);

static void     _event_loop_thread_func           (DexterEventLoop      *event_loop);


G_DEFINE_TYPE (DexterEventLoop, dexter_event_loop, G_TYPE_OBJECT);


static void
dexter_event_loop_class_init (DexterEventLoopClass *klass)
{
  GObjectClass *gobject_class;

#ifdef G_ENABLE_DEBUG
  g_message (G_STRFUNC);
#endif

  gobject_class = G_OBJECT_CLASS (klass);
  gobject_class->finalize = dexter_event_loop_finalize;
}


static void
dexter_event_loop_init (DexterEventLoop *event_loop)
{
#ifdef G_ENABLE_DEBUG
  g_message (G_STRFUNC);
#endif

  event_loop->context = g_main_context_new ();
  event_loop->main_loop = g_main_loop_new (event_loop->context, FALSE);
}


static void
dexter_event_loop_finalize (GObject *object)
{
  DexterEventLoop *event_loop = DEXTER_EVENT_LOOP (object);

#ifdef G_ENABLE_DEBUG
  g_message (G_STRFUNC);
#endif

  if (event_loop->main_loop)
    g_main_loop_unref (event_loop->main_loop);

  if (event_loop->context)
    g_main_context_unref (event_loop->context);

  /* chain up */
  G_OBJECT_CLASS (dexter_event_loop_parent_class)->finalize (object);
}


static void
_event_loop_thread_func (DexterEventLoop *event_loop)
{
  g_return_if_fail (DEXTER_IS_EVENT_LOOP (event_loop));

  /* ref event loop object so it won't finalize while we run */
  g_object_ref (event_loop);

  g_main_context_acquire (event_loop->context);
  g_main_loop_run (event_loop->main_loop);

  /* 
   * Won't go below here until dexter_event_loop_quit () called.
   */

  g_main_context_release (event_loop->context);
  g_object_unref (event_loop);
}


void
dexter_event_loop_run (DexterEventLoop *event_loop)
{
  GError *local_error = NULL;

#ifdef G_ENABLE_DEBUG
  g_message (G_STRFUNC);
#endif

  g_return_if_fail (DEXTER_IS_EVENT_LOOP (event_loop));

  if (g_main_loop_is_running (event_loop->main_loop))
  {
    g_warning ("%s: event loop is already running", G_STRFUNC);
    return;
  }

  dexter_threadpool_push_task (__ITHREADPOOL,
                              (GThreadFunc) _event_loop_thread_func,
                              event_loop,
                              &local_error);
  if (local_error)
  {
    g_warning ("%s: %s", G_STRFUNC, local_error->message);
    g_clear_error (&local_error);
  }
}


void
dexter_event_loop_quit (DexterEventLoop *event_loop)
{
#ifdef G_ENABLE_DEBUG
  g_message (G_STRFUNC);
#endif

  g_return_if_fail (DEXTER_IS_EVENT_LOOP (event_loop));

  if (!g_main_loop_is_running (event_loop->main_loop))
  {
    g_warning ("%s: event loop not running", G_STRFUNC);
    return;
  }

  g_main_loop_quit (event_loop->main_loop);
}


gboolean
dexter_event_loop_is_running (DexterEventLoop *event_loop)
{
#ifdef G_ENABLE_DEBUG
  g_message (G_STRFUNC);
#endif

  g_return_val_if_fail (DEXTER_IS_EVENT_LOOP (event_loop), FALSE);

  return g_main_loop_is_running (event_loop->main_loop);
}


void
dexter_event_loop_source_attach (DexterEventLoop *event_loop,
                                 GSource         *source)
{
#ifdef G_ENABLE_DEBUG
  g_message (G_STRFUNC);
#endif

  g_return_if_fail (DEXTER_IS_EVENT_LOOP (event_loop));
  g_return_if_fail (source != NULL);

  g_source_attach (source, event_loop->context);
}

#define __DEXTER_EVENT_LOOP_C__
#include <dexter-aliasdef.c>
