/* $Id: dexter-factory.c 393 2012-03-06 23:36:44Z pkovacs $ */
/*
 * libdexter - a plugin-based, distributed sampling library.
 * Copyright (C) 2006-2012 Philip Kovacs <pkovacs@users.sourceforge.net>.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 2.1
 * of the License, or (at your option) any later version.

 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.

 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
 * 02110-1301  USA
 *
*/

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <dexter-factory.h>
#include <dexter-module.h>
#include <dexter-private.h>
#include <dexter-alias.h>

/* used as a list accumulator as we iterate over the service hash */
typedef struct 
{
  GList *service_list;
  GType  iface_type;
} HashAccumulator;


static void  dexter_factory_class_init        (DexterFactoryClass  *klass);
static void  dexter_factory_init              (DexterFactory       *factory);
static void  dexter_factory_finalize          (GObject             *object);

             /* helper func */
static void  _inspect_service_for_iface_type  (gpointer             key,
                                               DexterPluginService *service,
                                               HashAccumulator     *accum);

struct _DexterFactory 
{
  GObject     parent;

  GHashTable *authorization_hash; /* hash of uuid->uuid (authorized services) */
  GHashTable *service_hash;       /* hash of uuid->DexterPluginService */
};

struct _DexterFactoryClass 
{
  GObjectClass parent_class;
};

static GHashTable *module_hash = NULL;   /* never freed as per GTypeModule docs */
    

G_DEFINE_TYPE (DexterFactory, dexter_factory, G_TYPE_OBJECT);


GQuark
dexter_factory_error_quark (void)
{
  static GQuark error_quark = 0;

  if (error_quark == 0)
    error_quark = g_quark_from_static_string ("dexter-factory-error-quark");

  return error_quark;
}


static void
dexter_factory_class_init (DexterFactoryClass *klass)
{
  GObjectClass *gobject_class;

#ifdef G_ENABLE_DEBUG
  g_message (G_STRFUNC);
#endif

  gobject_class = G_OBJECT_CLASS (klass);
  gobject_class->finalize = dexter_factory_finalize;
}


static void
dexter_factory_init (DexterFactory *factory)
{
#ifdef G_ENABLE_DEBUG
  g_message (G_STRFUNC);
#endif

  factory->authorization_hash = 
    g_hash_table_new_full (g_str_hash, g_str_equal, NULL, (GDestroyNotify) g_free);
    
  factory->service_hash = 
    g_hash_table_new_full (g_str_hash, g_str_equal, NULL, (GDestroyNotify) dexterplugin_service_free);
}


static void
dexter_factory_finalize (GObject *object)
{
  DexterFactory *factory = DEXTER_FACTORY (object);

#ifdef G_ENABLE_DEBUG
  g_message (G_STRFUNC);
#endif

  g_hash_table_destroy (factory->service_hash);

  /* chain up */
  G_OBJECT_CLASS (dexter_factory_parent_class)->finalize (object);
}


void
dexter_factory_load_plugins (DexterFactory  *factory,
                             GError        **error)
{
  GError        *local_error  = NULL;
  DexterModule  *module = NULL;
  gchar        **taglist = NULL;
  gchar         *plugin_path = NULL;
  gchar         *uuid = NULL;
  gchar         *tmp_uuid = NULL;
  gsize          length;
  gint           iter;
    
#ifdef G_ENABLE_DEBUG
  g_message (G_STRFUNC);
#endif

  g_return_if_fail (DEXTER_IS_FACTORY (factory));
  g_return_if_fail (error == NULL || *error == NULL);

  /* ******************************
   *  is module loading supported? 
   * ******************************
   */
    
  if (!g_module_supported())
  {
      g_set_error (error,
                   DEXTER_FACTORY_ERROR,
                   DEXTER_FACTORY_ERROR_MODULES_NOT_SUPPORTED,
                   "%s: modules not supported", G_STRFUNC);
      return;
  }

  /* ***********************************
   *  build hash of authorized services
   * ***********************************
   */
    
  /* Fetch taglist of authorized services, e.g. AuthorizedServices=1;2;3;4 */
  dexter_config_get_string_list (__ICONFIG, 
                                 "Main", "AuthorizedServices",
                                 &taglist, &length, &local_error);
  if (local_error)
  {
    g_propagate_error (error, local_error);
    g_strfreev (taglist);
    return;
  }
                     
  /* for each authorized service tag */
  for (iter=0; iter<length; iter++)
  {
    /* look up uuid of authorized service in [Services] group */
    dexter_config_get_string (__ICONFIG, 
                              "Services", taglist[iter],
                              &uuid, &local_error);
    if (local_error)
    {
      g_propagate_error (error, local_error);
      g_free (uuid);
      g_strfreev (taglist);
      return;
    }
        
    /* is uuid in authorization hash? */  
    tmp_uuid = g_hash_table_lookup (factory->authorization_hash, uuid);
    if (!tmp_uuid)
    {
      /* add uuid to hash */
      tmp_uuid = g_strdup (uuid);
      g_hash_table_insert (factory->authorization_hash, tmp_uuid, tmp_uuid);
    }
  
    g_free (uuid);
    uuid=NULL;
  }

  g_strfreev (taglist);
  taglist=NULL;
    
  /* ******************************************************************
   *  build module hash, inspecting each service the module registers.  
   *  only authorized services are served up by the library.  
   * ******************************************************************
   */
    
  /* Fetch taglist of plugin tags, e.g. LoadPlugins=1;2;3;4 */
  dexter_config_get_string_list (__ICONFIG, 
                                 "Main", "LoadPlugins",
                                 &taglist, &length, &local_error);
  if (local_error)
  {
    g_propagate_error (error, local_error);
    g_strfreev (taglist);
    return;
  }

  /* for each plugin tag */
  for (iter=0; iter<length; iter++)
  {
    /* look up plugin full path name in [Plugins] group */
    dexter_config_get_string (__ICONFIG, 
                              "Plugins",  taglist[iter],
                              &plugin_path, &local_error);
    if (local_error)
    {
      g_propagate_error (error, local_error);
      g_strfreev (taglist);
      g_free (plugin_path);
      return;
    }

    /* disallow relative paths */
    if (!g_path_is_absolute (plugin_path))
    {
      g_set_error (error,
                   DEXTER_FACTORY_ERROR,
                   DEXTER_FACTORY_ERROR_NOT_ABSOLUTE_PATH,
                   "%s: %s is not an absolute path",
                   G_STRFUNC, plugin_path);
      g_strfreev (taglist);
      g_free (plugin_path);
      return;
    }
      
    /* add module suffix to plugin path name, if not indicated */
    if (!g_str_has_suffix (plugin_path, "." G_MODULE_SUFFIX))
    {
      gchar *filename = g_strdup_printf ("%s.%s", plugin_path, G_MODULE_SUFFIX);
      g_free (plugin_path);
      plugin_path = filename;
    }
        
    /* create module hash if needed */
    if (!module_hash)
      module_hash = g_hash_table_new (g_str_hash, g_str_equal);

    /* is the module already loaded? */
    module = g_hash_table_lookup (module_hash, plugin_path);
    if (!module)
    {
      /* ok, so create the module */
      DexterPluginServiceGroup *service_group   = NULL;
      DexterPluginService *service = NULL;
      gint iter;

      module = dexter_module_new (plugin_path);
      if (!module)
      {
          g_set_error (error,
                       DEXTER_FACTORY_ERROR,
                       DEXTER_FACTORY_ERROR_FAILED,
                       "%s: %s module not created",
                       G_STRFUNC, plugin_path);
          g_strfreev (taglist);
          g_free (plugin_path);
          return;
      }
      
      /* load the module for the first time so it can register its types */
      if (!g_type_module_use (G_TYPE_MODULE (module)))
      {
        g_set_error (error,
                     DEXTER_FACTORY_ERROR,
                     DEXTER_FACTORY_ERROR_FAILED,
                     "%s: %s module didn't load",
                     G_STRFUNC, plugin_path);
        g_strfreev (taglist);
        g_free (plugin_path);
        return;     
      }

      /* store new module in permanent module hash */
      g_hash_table_insert (module_hash, (G_TYPE_MODULE (module))->name, module);

      /* fetch the service group from the plugin module. */
      service_group = dexter_module_get_services (module);

      /* check config for each service to see if it's authorized */
      for (iter=0; iter<service_group->n_services; iter++)
      {
        service = service_group->services + iter;

        tmp_uuid = g_hash_table_lookup (factory->authorization_hash, service->uuid);
        if (tmp_uuid)
        {
          /* add it to the factory's hash of (authorized) services */
          DexterPluginService *service_dup = dexterplugin_service_dup (service);
          g_hash_table_insert (factory->service_hash, service_dup->uuid, service_dup);
        }
        else
        {
          g_warning ("unauthorized service %s", service->uuid);
        }
      }
      /* free the (processed) service group that the module created for us */
      dexterplugin_service_group_free (service_group);
      /* release the module. it will unload itself and reload automatically as needed. */
      g_type_module_unuse (G_TYPE_MODULE (module));
    }
    else
    {
      g_warning ("plugin [%s] already loaded", plugin_path);
    }

    g_free (plugin_path);
    plugin_path=NULL;
  }
    
  g_strfreev (taglist);
}

static void
_inspect_service_for_iface_type (gpointer             key,
                                 DexterPluginService *service,
                                 HashAccumulator     *accum)
{
  /* ascertain if the service supports the requested interface: accum->iface_type.
   * if yes, prepend it to the list of services we are maintaining: accum->service_list,
   * with each pass through this helper function. */

  GType *ifaces_implemented;
  guint  n_ifaces, iter;
    
  g_return_if_fail (service != NULL);
  g_return_if_fail (accum != NULL);
  g_return_if_fail (accum->iface_type != G_TYPE_INVALID);
    
  ifaces_implemented = g_type_interfaces (service->type, &n_ifaces);

  for (iter=0; iter<n_ifaces; iter++)
  {
    if (ifaces_implemented[iter] == accum->iface_type)  /* service implements iface of interest */
    {
      accum->service_list = g_list_prepend (accum->service_list, dexterplugin_service_dup (service));
      g_free (ifaces_implemented);
      return;
    }
  }

  g_free (ifaces_implemented);
}

DexterPluginServiceGroup *
dexter_factory_get_services (DexterFactory *factory,
                                   GType                iface_type)
{ 
  DexterPluginServiceGroup *service_group = NULL;
  HashAccumulator           accum;
  gint                      service_list_length;

#ifdef G_ENABLE_DEBUG
  g_message (G_STRFUNC);
#endif

  g_return_val_if_fail (DEXTER_IS_FACTORY (factory), NULL);
  g_return_val_if_fail (iface_type != G_TYPE_INVALID, NULL);

  accum.service_list = NULL;
  accum.iface_type = iface_type;

  /* inspect each loaded service for object types that implement the requested interface. */
  g_hash_table_foreach (factory->service_hash, (GHFunc)_inspect_service_for_iface_type, &accum);
        
  /* GList accum.service_list now contains matching services. */
    
  /* prepare a service group. */
  service_list_length = g_list_length (accum.service_list);
  if (service_list_length > 0)
  {
    GList *element=NULL;
    gint iter;
  
    service_group = dexterplugin_service_group_new (service_list_length);
  
    for (element=g_list_first (accum.service_list), iter=0; 
         element; 
         element=g_list_next (element), iter++)
    {
      DexterPluginService *service = (DexterPluginService *) element->data;
      dexterplugin_service_copy (service_group->services + iter, service);
      dexterplugin_service_free (service);
    }
  }

  /* free working GList of services */
  if (accum.service_list)
    g_list_free (accum.service_list);

  return service_group;
}

#define __DEXTER_FACTORY_C__
#include <dexter-aliasdef.c>
