/* $Id: dexter-hash.c 393 2012-03-06 23:36:44Z pkovacs $ */
/*
 * libdexter - a plugin-based, distributed sampling library.
 * Copyright (C) 2006-2012 Philip Kovacs <pkovacs@users.sourceforge.net>.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 2.1
 * of the License, or (at your option) any later version.

 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.

 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
 * 02110-1301  USA
 *
*/

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <dexter-hash.h>
#include <dexter-alias.h>

struct _dexter_hash_table
{
  GHashTable *hash;
  GMutex     *mutex;
};

DexterHashTable *
dexter_hash_table_new (GHashFunc  hash_func,
                       GEqualFunc key_equal_func)
{
  return dexter_hash_table_new_full (hash_func, key_equal_func, NULL, NULL);
}


DexterHashTable *
dexter_hash_table_new_full (GHashFunc      hash_func,
                            GEqualFunc     key_equal_func,
                            GDestroyNotify key_destroy_func,
                            GDestroyNotify value_destroy_func)
{
  DexterHashTable * table = g_slice_new0 (DexterHashTable);

  table->hash = g_hash_table_new_full (hash_func, key_equal_func, key_destroy_func, value_destroy_func);
  table->mutex = g_mutex_new ();
    
  return table;
}


void
dexter_hash_table_insert (DexterHashTable *hash_table,
                          gpointer         key,
                          gpointer         value)
{
  g_return_if_fail (hash_table != NULL);

  g_mutex_lock (hash_table->mutex);
  g_hash_table_insert (hash_table->hash, key, value);
  g_mutex_unlock (hash_table->mutex);
}


void
dexter_hash_table_replace (DexterHashTable *hash_table,
                           gpointer         key,
                           gpointer         value)
{
  g_return_if_fail (hash_table != NULL);

  g_mutex_lock (hash_table->mutex);
  g_hash_table_replace (hash_table->hash, key, value);
  g_mutex_unlock (hash_table->mutex);
}


gint
dexter_hash_table_size (DexterHashTable *hash_table)
{
  gint size;

  g_return_val_if_fail (hash_table != NULL, -1);

  g_mutex_lock (hash_table->mutex);
  size = g_hash_table_size (hash_table->hash);
  g_mutex_unlock (hash_table->mutex);
  return size;
}


gpointer
dexter_hash_table_lookup (DexterHashTable *hash_table,
                          gconstpointer    key)
{
    gpointer data;

    g_return_val_if_fail (hash_table != NULL, NULL);

    g_mutex_lock (hash_table->mutex);
    data = g_hash_table_lookup (hash_table->hash, key);
    g_mutex_unlock (hash_table->mutex);
    return data;
}


gboolean
dexter_hash_table_lookup_extended (DexterHashTable *hash_table,
                                   gconstpointer    lookup_key,
                                   gpointer        *orig_key,
                                   gpointer        *value)
{
  gboolean key_found;

  g_return_val_if_fail (hash_table != NULL, FALSE);

  g_mutex_lock (hash_table->mutex);
  key_found = g_hash_table_lookup_extended (hash_table->hash, lookup_key, orig_key, value);
  g_mutex_unlock (hash_table->mutex);
  return key_found;
}


void
dexter_hash_table_foreach (DexterHashTable *hash_table,
                           GHFunc           func,
                           gpointer         user_data)
{
  g_return_if_fail (hash_table != NULL);

  g_mutex_lock (hash_table->mutex);
  g_hash_table_foreach (hash_table->hash, func, user_data);
  g_mutex_unlock (hash_table->mutex);
}


gpointer
dexter_hash_table_find (DexterHashTable *hash_table,
                        GHRFunc          predicate,
                        gpointer         user_data)
{
  gpointer value;
    
  g_return_val_if_fail (hash_table != NULL, NULL);

  g_mutex_lock (hash_table->mutex);
  value = g_hash_table_find (hash_table->hash, predicate, user_data);
  g_mutex_unlock (hash_table->mutex);
  return value;
}


gboolean
dexter_hash_table_remove (DexterHashTable *hash_table,
                          gconstpointer    key)
{
  gboolean key_found;

  g_return_val_if_fail (hash_table != NULL, FALSE);

  g_mutex_lock (hash_table->mutex);
  key_found = g_hash_table_remove (hash_table->hash, key);
  g_mutex_unlock (hash_table->mutex);
  return key_found;
}


gboolean
dexter_hash_table_steal (DexterHashTable *hash_table,
                         gconstpointer    key)
{
  gboolean key_found;

  g_return_val_if_fail (hash_table != NULL, FALSE);

  g_mutex_lock (hash_table->mutex);
  key_found = g_hash_table_steal (hash_table->hash, key);
  g_mutex_unlock (hash_table->mutex);
  return key_found;
}


gint
dexter_hash_table_foreach_remove (DexterHashTable *hash_table,
                                  GHRFunc          func,
                                  gpointer         user_data)
{
  gint keys_removed;

  g_return_val_if_fail (hash_table != NULL, -1);

  g_mutex_lock (hash_table->mutex);
  keys_removed = g_hash_table_foreach_remove (hash_table->hash, func, user_data);
  g_mutex_unlock (hash_table->mutex);
  return keys_removed;
}


gint
dexter_hash_table_foreach_steal (DexterHashTable *hash_table,
                                 GHRFunc          func,
                                 gpointer         user_data)
{
  gint keys_removed;

  g_return_val_if_fail (hash_table != NULL, -1);

  g_mutex_lock (hash_table->mutex);
  keys_removed = g_hash_table_foreach_steal (hash_table->hash, func, user_data);
  g_mutex_unlock (hash_table->mutex);
  return keys_removed;
}


void
dexter_hash_table_destroy (DexterHashTable *hash_table)
{
  g_return_if_fail (hash_table != NULL);

  g_mutex_lock (hash_table->mutex);
  g_hash_table_destroy (hash_table->hash);
  g_mutex_unlock (hash_table->mutex);
  g_mutex_free (hash_table->mutex);
  g_slice_free (DexterHashTable, hash_table);
}

#define __DEXTER_HASH_C__
#include <dexter-aliasdef.c>
