/* $Id: dexter-log.c 393 2012-03-06 23:36:44Z pkovacs $ */
/*
 * libdexter - a plugin-based, distributed sampling library.
 * Copyright (C) 2006-2012 Philip Kovacs <pkovacs@users.sourceforge.net>.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 2.1
 * of the License, or (at your option) any later version.

 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.

 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
 * 02110-1301  USA
 *
*/

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <stdarg.h>
#include <syslog.h>
#include <dexter-log.h>
#include <dexter-alias.h>

static int _dexter_log_level_to_syslog_priority (gint level);

#ifdef HAVE_STRUCT_SYSLOG_DATA
static struct syslog_data sdata = SYSLOG_DATA_INIT;
#endif

static int 
_dexter_log_level_to_syslog_priority (gint level)
{
  int priority;

  switch (level)
  {
    case DEXTER_LOG_CRIT:
      priority=LOG_CRIT;
      break;
    case DEXTER_LOG_ERR:
      priority=LOG_ERR;
      break;
    case DEXTER_LOG_WARNING:
      priority=LOG_WARNING;
      break;
    case DEXTER_LOG_NOTICE:
      priority=LOG_NOTICE;
      break;
    case DEXTER_LOG_INFO:
      priority=LOG_INFO;
      break;
    case DEXTER_LOG_DEBUG:
      priority=LOG_DEBUG;
      break;
    default:
      /* programming error */
      priority= -1;
      g_error (G_STRLOC " error unexpected log level code: %d", level);
      break;
  }

  return priority;
}


void
dexter_log_init (void)
{
#ifdef G_ENABLE_DEBUG
  g_message (G_STRFUNC);
#endif

#ifdef HAVE_OPENLOG_R
  openlog_r (g_get_prgname (), LOG_NDELAY, LOG_USER, &sdata);
#else
  openlog (g_get_prgname (), LOG_NDELAY, LOG_USER);
#endif
}


void dexter_log_exit (void)
{
#ifdef G_ENABLE_DEBUG
  g_message (G_STRFUNC);
#endif

#ifdef HAVE_CLOSELOG_R
  closelog_r (&logdata);
#else
  closelog ();
#endif
}

void 
dexter_log (gint level, const gchar *format, ...)
{
  va_list ap;
  int priority = _dexter_log_level_to_syslog_priority (level);

  va_start (ap, format);

  /* Highest priority goes to using the _r reentrant variants.  Most modern distros 
   * already have thread-safe syslog, but if the _r variant is available, then we know
   * that the base version is not thread-safe.
   *
   * Next priority is to use the vsyslog/vsyslog_r variants as we can just pass it our 
   * va_list without having to produce the finished string via a memory allocation.
   */

#if defined (HAVE_VSYSLOG_R)
  vsyslog_r (priority, &sdata, format, ap);
#elif defined (HAVE_SYSLOG_R)
  {
    gchar *log_string = g_strdup_vprintf (format, ap);
    syslog_r (priority, &sdata, "%s", log_string);
    g_free (log_string);
  }
#elif defined (HAVE_VSYSLOG)
  vsyslog (priority, format, ap);
#else
  {
    gchar *log_string = g_strdup_vprintf (format, ap);
    syslog (priority, "%s", log_string);
    g_free (log_string);
  }
#endif

  va_end (ap);
}

#define __DEXTER_LOG_C__
#include <dexter-aliasdef.c>
