/* $Id: dexter-messages.h 393 2012-03-06 23:36:44Z pkovacs $ */
/*
 * libdexter - a plugin-based, distributed sampling library.
 * Copyright (C) 2006-2012 Philip Kovacs <pkovacs@users.sourceforge.net>.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 2.1
 * of the License, or (at your option) any later version.

 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.

 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
 * 02110-1301  USA
 *
*/

#ifndef __DEXTER_MESSAGES_H__
#define __DEXTER_MESSAGES_H__

#include <glib.h>

G_BEGIN_DECLS

/*
 * Dexter (Network) Messages are encoded as UTF-8 and have the following forms:
 *
 * -- client to server messages --
 * DEXTER/<major>.<minor> CMD <cmd> <peer> <param>\n
 * <num payload bytes>\n
 * <payload bytes>\n
 *
 * -- server to client replies --
 * DEXTER/<major>.<minor> RPY <cmd> <peer> <status>\n
 * <num payload bytes>\n
 * <payload bytes>\n
 *
 * -- server to client stream data --
 * DEXTER/<major>.<minor> DAT <peer> <flags> <status>\n
 * <num payload bytes>\n
 * <payload bytes>\n
 *
 * <major>.<minor> is the protocol revision, e.g. 1.0;
 * <cmd> is an integer command (see enum below);
 * <peer> is the sampler's peer id, assigned at creation time
 * <param> is a parameter value for the command, if needed
 * <flags> indicates if the data starts, continues and/or ends a complete data message.
 * <status> is the error status of the reply or data message (see enum below)
 */

#define DEXTER_PROTOCOL_MAJOR 1
#define DEXTER_PROTOCOL_MINOR 1

#define CMD_FORMAT_STRING     "DEXTER/%d.%d CMD %d %d %d\n"
#define RPY_FORMAT_STRING     "DEXTER/%d.%d RPY %d %d %d\n"
#define DAT_FORMAT_STRING     "DEXTER/%d.%d DAT %d %d %d\n"

enum {
    COMMAND_HELLO                     = 100,
    COMMAND_NOOP                      = 110,  /* DEXTER/1.1 */
    COMMAND_STARTTLS                  = 120,  /* DEXTER/1.1 */
    COMMAND_GET_SERVICES              = 200,
    COMMAND_SAMPLER_NEW               = 300,
    COMMAND_SAMPLER_INITIALIZE        = 310,
    COMMAND_SAMPLER_SAMPLE            = 320,
    COMMAND_SAMPLER_FREE              = 390,
    COMMAND_TIMEDSAMPLER_NEW          = 400,
    COMMAND_TIMEDSAMPLER_INITIALIZE   = 410,
    COMMAND_TIMEDSAMPLER_START        = 420,
    COMMAND_TIMEDSAMPLER_STOP         = 430,
    COMMAND_TIMEDSAMPLER_FREE         = 490,
    COMMAND_GOODBYE                   = 900,
    COMMAND_UNKNOWN                   = 999
} DexterProtocolCommands;

enum {
    STATUS_OK                         = 0,
    ERROR_MESSAGE_BAD_FORMAT          = 9010,
    ERROR_INCOMPATIBLE_PROTOCOL       = 9020,
    ERROR_GREETING_NOT_EXCHANGED      = 9030,
    ERROR_GREETING_ALREADY_EXCHANGED  = 9040,
    ERROR_NO_PAYLOAD_DATA             = 9050,
    ERROR_PAYLOAD_DATA_INVALID        = 9060,
    ERROR_SERVICE_UNAVAILABLE         = 9070,
    ERROR_COMMAND_UNKNOWN             = 9080,
    ERROR_COMMAND_FAILED              = 9090,
    ERROR_COMMAND_INVALID_PARAMETER   = 9100,
    ERROR_SAMPLER_PEER_NOT_FOUND      = 9110,
    ERROR_SAMPLER_DATA                = 9120,
    ERROR_TLS_NOT_STARTED             = 9130, /* DEXTER/1.1 */
    ERROR_TLS_ALREADY_STARTED         = 9140, /* DEXTER/1.1 */ 
    ERROR_TLS_NOT_AVAILABLE           = 9150  /* DEXTER/1.1 */
} DexterProtocolStatusCodes;

enum {
    DAT_CHUNK_FIRST                   = 1<<0,
    DAT_CHUNK_CONTINUATION            = 1<<1,
    DAT_CHUNK_LAST                    = 1<<2
} DexterApiDatFlags;

const gchar *dexter_protocol_strerror     (gint) G_GNUC_INTERNAL;

G_END_DECLS

#endif /* __DEXTER_MESSAGES_H__ */
