/* $Id: dexter-module.c 393 2012-03-06 23:36:44Z pkovacs $ */
/*
 * libdexter - a plugin-based, distributed sampling library.
 * Copyright (C) 2006-2012 Philip Kovacs <pkovacs@users.sourceforge.net>.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 2.1
 * of the License, or (at your option) any later version.

 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.

 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
 * 02110-1301  USA
 *
*/

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <dexter-module.h>
#include <dexter-alias.h>

static void      dexter_module_class_init   (DexterModuleClass *klass);
static void      dexter_module_init         (DexterModule      *plugin_module);
                 /* no finalize for type modules */
static gboolean  dexter_module_load         (GTypeModule       *type_module);
static void      dexter_module_unload       (GTypeModule       *type_module);


G_DEFINE_TYPE (DexterModule, dexter_module, G_TYPE_TYPE_MODULE);


static void
dexter_module_class_init (DexterModuleClass *klass)
{
  GTypeModuleClass *gtype_module_class;

#ifdef G_ENABLE_DEBUG
  g_message (G_STRFUNC);
#endif
   
  dexter_module_parent_class = g_type_class_peek_parent (klass);

  gtype_module_class = G_TYPE_MODULE_CLASS (klass);
  gtype_module_class->load = dexter_module_load;
  gtype_module_class->unload = dexter_module_unload;
}


static void
dexter_module_init (DexterModule *plugin_module)
{
#ifdef G_ENABLE_DEBUG
  g_message (G_STRFUNC);
#endif
      
  plugin_module->library = NULL;
}


static gboolean
dexter_module_load (GTypeModule *type_module)
{
  DexterModule *plugin_module = DEXTER_MODULE (type_module);

  /* load the shared library. */
#ifdef G_ENABLE_DEBUG
  g_message (G_STRFUNC);
#endif
  plugin_module->library = g_module_open (type_module->name, G_MODULE_BIND_LAZY | G_MODULE_BIND_LOCAL);

  if (!plugin_module->library)
  {
    g_warning ("Failed to load plugin `%s': %s", type_module->name, g_module_error ());
    return FALSE;
  }

  /* load the symbols we expect plugin to export */
  if (!g_module_symbol (plugin_module->library, 
                        "dexter_plugin_init", (gpointer) &plugin_module->dexter_plugin_init) ||
      !g_module_symbol (plugin_module->library, 
                        "dexter_plugin_exit", (gpointer) &plugin_module->dexter_plugin_exit) ||
      !g_module_symbol (plugin_module->library, 
                      "dexter_plugin_get_services",(gpointer) &plugin_module->dexter_plugin_get_services))
  {
    g_warning ("Plugin %s lacks required symbols", type_module->name);
    g_module_close (plugin_module->library);
    return FALSE;
  }

  /* initialize the plugin */
  (*plugin_module->dexter_plugin_init) (plugin_module);

  return TRUE;
}


static void
dexter_module_unload (GTypeModule *type_module)
{
  DexterModule *plugin_module = DEXTER_MODULE (type_module);
 
#ifdef G_ENABLE_DEBUG
  g_message (G_STRFUNC);
#endif
  
  /* de-initialize the plugin */
  (*plugin_module->dexter_plugin_exit) ();

  /* unload plugin */
  g_module_close (plugin_module->library);

  plugin_module->library = NULL;
  plugin_module->dexter_plugin_init = NULL;
  plugin_module->dexter_plugin_exit = NULL;
  plugin_module->dexter_plugin_get_services = NULL;
}


DexterModule*
dexter_module_new (const gchar *filename)
{
  DexterModule *plugin_module;

  g_return_val_if_fail (filename != NULL, NULL);
  g_return_val_if_fail (*filename > 0, NULL);

  plugin_module = g_object_new (DEXTER_TYPE_MODULE, NULL);
  if (plugin_module)
    g_type_module_set_name (G_TYPE_MODULE (plugin_module), filename);

  return plugin_module;
}


DexterPluginServiceGroup *
dexter_module_get_services (const DexterModule *plugin_module)
{
  g_return_val_if_fail (DEXTER_IS_MODULE (plugin_module), NULL);

  return (plugin_module->dexter_plugin_get_services) ();
}

#define __DEXTER_MODULE_C__
#include <dexter-aliasdef.c>
