/* $Id: dexter-msgframer.c 393 2012-03-06 23:36:44Z pkovacs $ */
/*
 * libdexter - a plugin-based, distributed sampling library.
 * Copyright (C) 2006-2012 Philip Kovacs <pkovacs@users.sourceforge.net>.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 2.1
 * of the License, or (at your option) any later version.

 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.

 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
 * 02110-1301  USA
 *
*/

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <errno.h>
#include <string.h>
#include <stdlib.h>
#include <dexter-msgframer.h>
#include <dexter-alias.h>

#define MAX_PAYLOAD_SIZE  16777216  /* 16 MiB */

/* 
 * Dexter messages take the form:
 *
 * <message>\n
 * <payload size>\n
 * <payload>\n
 * 
 */


enum   /* message framer state */
{
  FRAME_START = 0,
  FRAME_SCANNING_MESSAGE,
  FRAME_SCANNING_PAYLOAD_SIZE,
  FRAME_SCANNING_PAYLOAD,
  FRAME_FINALIZE
};

static void   dexter_msgframer_class_init  (DexterMsgFramerClass *klass);
static void   dexter_msgframer_init        (DexterMsgFramer      *framer);
static void   dexter_msgframer_finalize    (GObject              *object);

              /* GFunc for g_queue_foreach */ 
static void   _delete_frame                (DexterMsgFrame       *frame, 
                                            gpointer              unused);

struct _DexterMsgFramer 
{
  GObject         parent;
  GQueue         *frame_queue;        /* queue of frames as they are assembled */
  DexterMsgFrame *pending_frame;      /* the frame we are assembling */
  guint           state;              /* the state of the framer */
  guint           position;           /* relative position from start of message, payload size or payload */
  gchar           payload_csize[32];  /* bytes on the payload size line */
  gint            payload_size;       /* payload size converted from payload_csize after those bytes arrive */
};

struct _DexterMsgFramerClass 
{
  GObjectClass    parent_class;
};


G_DEFINE_TYPE (DexterMsgFramer, dexter_msgframer, G_TYPE_OBJECT);
  

static void
dexter_msgframer_class_init (DexterMsgFramerClass *klass)
{
  GObjectClass *gobject_class;

#ifdef G_ENABLE_DEBUG
  g_message (G_STRFUNC);
#endif

  gobject_class = G_OBJECT_CLASS (klass);
  gobject_class->finalize = dexter_msgframer_finalize;
}


static void
dexter_msgframer_init (DexterMsgFramer *framer)
{
#ifdef G_ENABLE_DEBUG
  g_message (G_STRFUNC);
#endif

  framer->frame_queue = g_queue_new();
  framer->pending_frame = NULL;
  framer->state = FRAME_START;
}


static void
dexter_msgframer_finalize (GObject *object)
{
  DexterMsgFramer *framer = DEXTER_MSGFRAMER (object);

#ifdef G_ENABLE_DEBUG
  g_message (G_STRFUNC);
#endif

  g_queue_foreach (framer->frame_queue, (GFunc) _delete_frame, NULL);
  g_queue_free (framer->frame_queue);
  if (framer->pending_frame)
    _delete_frame (framer->pending_frame, NULL);
    
  /* chain up */
  G_OBJECT_CLASS (dexter_msgframer_parent_class)->finalize (object);
}


static void
_delete_frame (DexterMsgFrame *frame,
               gpointer        unused)
{
  if (G_UNLIKELY (!frame))
    return;

  if (frame->payload)
    g_string_free (frame->payload, TRUE);
  g_slice_free (DexterMsgFrame, frame);
}


gint
dexter_msgframer_push_bytes (DexterMsgFramer *framer,
                             gchar           *buffer,
                             guint            buffersize)
{
  gchar *pbuffer = buffer;
  guint  pbuffersize = buffersize;

#ifdef G_ENABLE_DEBUG
  g_message (G_STRFUNC);
#endif

  g_return_val_if_fail (DEXTER_IS_MSGFRAMER (framer), DEXTER_MSGFRAMER_ERROR_FAILED);
  g_return_val_if_fail (buffer != NULL, DEXTER_MSGFRAMER_ERROR_FAILED);

start:

  if (pbuffersize == 0)
    return 0;

  /* -debug-
  {
    char format[64];
    g_print ("*******************************************\n");
    sprintf (format,"pushing %d bytes (state %d)\n", buffersize, framer->state);
    g_print (format, buffer);
    g_print ("*******************************************\n");
  }
  */

  /* 
   * ***********
   * START STATE
   * ***********
   */
  if (framer->state == FRAME_START)
  {
    /* sanity check on first few bytes of message */
    if (strncmp ("DEXTER", pbuffer, MIN (buffersize,6)) != 0)
    {
      /* badly formed message */
      return DEXTER_MSGFRAMER_ERROR_MESSAGE;
    }
  
    /* state transition */
    framer->pending_frame = g_slice_new0 (DexterMsgFrame);
    framer->state = FRAME_SCANNING_MESSAGE;
    framer->position=0;
    memset (framer->payload_csize, 0, sizeof (framer->payload_csize));

    /*g_print ("state changed to FRAME_SCANNING_MESSAGE (pbuffersize = %d)\n", pbuffersize);*/
  }

  /* 
   * **********************
   * SCANNING MESSAGE STATE
   * **********************
   */
  while ((framer->state == FRAME_SCANNING_MESSAGE) && (pbuffersize > 0))
  {
    if (*pbuffer == '\n')
    {
      /* DEXTER message read -- do state transition. */ 
      framer->state = FRAME_SCANNING_PAYLOAD_SIZE;
      framer->position=0;
      pbuffer++;
      pbuffersize--;

      /*g_print ("state changed to FRAME_SCANNING_PAYLOAD_SIZE (pbuffersize = %d)\n", pbuffersize);*/
    }
    else if (framer->position >= sizeof (framer->pending_frame->message)-1)
    {
      /* message line too long */
      return DEXTER_MSGFRAMER_ERROR_BUFFER_OVERRUN;
    }
    else 
    {
      /* keep reading and saving message bytes */
      framer->pending_frame->message[framer->position++] = *pbuffer++;
      pbuffersize--;
    }
  }

  /*
   * ***************************
   * SCANNING PAYLOAD SIZE STATE
   * ***************************
   */
  while ((framer->state == FRAME_SCANNING_PAYLOAD_SIZE) && (pbuffersize > 0))
  {
    if (*pbuffer == '\n')
    {
      /* payload csize is read -- convert it to an integer. */
      errno=0;
      framer->payload_size = strtol (framer->payload_csize, (char **)NULL, 10);
      if (errno)
      {
        /* error converting payload_csize */
        return DEXTER_MSGFRAMER_ERROR_CONVERSION;
      }

      /* range sanity check */
      if ((framer->payload_size < 0) || (framer->payload_size > MAX_PAYLOAD_SIZE))
      {
        /* payload_size out of range */
        return DEXTER_MSGFRAMER_ERROR_RANGE;
      }

      /* state transition */
      framer->state = FRAME_SCANNING_PAYLOAD;
      framer->position=0;
      pbuffer++;
      pbuffersize--;
      if (framer->payload_size>0)
      {
        framer->pending_frame->payload = g_string_sized_new (framer->payload_size);
        framer->state = FRAME_SCANNING_PAYLOAD;
      }
      else
      {
        framer->state = FRAME_FINALIZE;
      }

      /*g_print ("state changed to FRAME_SCANNING_PAYLOAD (pbuffersize = %d)\n", pbuffersize);*/
    }
    else if (framer->position >= sizeof (framer->payload_csize)-1)
    {
      /* payload_csize buffer overrun */
      return DEXTER_MSGFRAMER_ERROR_BUFFER_OVERRUN;
    }
    else
    {
      /* keep reading and saving payload csize bytes */
      framer->payload_csize[framer->position++] = *pbuffer++;
      pbuffersize--;
    }
  }

  /*
   * **********************
   * SCANNING PAYLOAD STATE
   * **********************
   */
  while ((framer->state == FRAME_SCANNING_PAYLOAD) && (pbuffersize > 0))
  {
    if ((framer->position == framer->payload_size) && (*pbuffer=='\n'))
    {
      /* entire payload has been read and the frame is now complete */
      framer->state = FRAME_FINALIZE;
      framer->position=0;
      /* don't move past the last newline so finalize can acknowledge it 
      pbuffer++;
      buffersize--;
      */

      /*g_print ("state changed to FRAME_FINALIZE (pbuffersize = %d)\n", pbuffersize);*/
    }
    else if (framer->position > framer->payload_size)
    {
      /* message rambles on beyond expected size */
      return DEXTER_MSGFRAMER_ERROR_MESSAGE;
    }
    else
    {
      /* keep reading and saving payload bytes */
      g_string_append_c (framer->pending_frame->payload, *pbuffer++);
      framer->position++;
      pbuffersize--;
    }
  }

  /*
   * **************
   * FINALIZE STATE
   * **************
   */
  while ((framer->state == FRAME_FINALIZE) && (pbuffersize > 0))
  {
    if (*pbuffer == '\n')
    {
      /* frame is complete */
      g_queue_push_tail (framer->frame_queue, framer->pending_frame);
      framer->pending_frame=NULL;
      framer->state = FRAME_START;
      pbuffer++;
      pbuffersize--;
      /*g_print ("state changed to FRAME_START (pbuffersize= %d)\n", pbuffersize);*/
      goto start;
    }
    else
    {
      /* badly formed message or frame */
      return DEXTER_MSGFRAMER_ERROR_MESSAGE;
    }
  }

  return 0;
}


gint
dexter_msgframer_frames_queued (DexterMsgFramer *framer)
{
  gint len;

#ifdef G_ENABLE_DEBUG
  g_message (G_STRFUNC);
#endif

  g_return_val_if_fail (DEXTER_IS_MSGFRAMER (framer), -1);

  len = g_queue_get_length (framer->frame_queue);
  return len;
}


DexterMsgFrame *
dexter_msgframer_pop_frame (DexterMsgFramer *framer)
{
  DexterMsgFrame *frame;

#ifdef G_ENABLE_DEBUG
  g_message (G_STRFUNC);
#endif
  
  g_return_val_if_fail (DEXTER_IS_MSGFRAMER (framer), NULL);

  frame = g_queue_pop_head (framer->frame_queue);
  return frame;
}


void
dexter_msgframer_free_frame  (DexterMsgFramer *framer,
                              DexterMsgFrame  *frame)
{
#ifdef G_ENABLE_DEBUG
  g_message (G_STRFUNC);
#endif
  
  g_return_if_fail (DEXTER_IS_MSGFRAMER (framer));

  if (G_UNLIKELY (!frame))
    return;

  _delete_frame (frame, NULL);
}

#define __DEXTER_MSGFRAMER_C__
#include <dexter-aliasdef.c>
