/*
 * libdexter - a plugin-based, distributed sampling library.
 * Copyright (C) 2006-2012 Philip Kovacs <pkovacs@users.sourceforge.net>.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 2.1
 * of the License, or (at your option) any later version.

 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.

 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
 * 02110-1301  USA
 *
*/

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

/*
 * On some systems, <netinet/ip.h> doesn't include its required headers,
 * <sys/types.h>, <netinet/in_systm.h> and <netinet/in.h>, 
 * so we include them here.
 */

#include <errno.h>
#include <stdlib.h>
#include <string.h>
#include <sys/select.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <sys/types.h>
#include <netdb.h>
#include <netinet/tcp.h>
#include <netinet/in_systm.h>
#include <netinet/in.h>
#include <netinet/ip.h>
#include <unistd.h>
#ifdef HAVE_TLS
#include <dexter-tls.h>
#endif
#include <dexter-network.h>
#include <dexter-alias.h>

#define DEFAULT_BACKLOG 10

static int      _dexter_network_connect            (int                     sockfd,
                                                   const struct sockaddr  *serv_addr,
                                                   socklen_t               addrlen,
                                                   GError                **error);

static void     _dexter_network_set_socket_options (int                     sockfd,
                                                    const char             *options,
                                                    GError                **error);

#ifdef HAVE_TLS
static gboolean _dexter_network_session_starttls   (dexter_network_session_t  *session,
                                                    gboolean                   role,
                                                    const gchar               *cert_file,
                                                    const gchar               *key_file,
                                                    const gchar               *trust_file,
                                                    const gchar               *crl_file,
                                                    guint                      flags,
                                                    GError                   **error);
#endif

struct _dexter_network_session_t
{
  int sockfd;                   /* socket file descriptor */
  struct sockaddr_storage addr; /* peer socket address info */
  socklen_t addrlen;            /* peer socket address length (family-dependent) */
#ifdef HAVE_TLS
  dexter_tls_t *tls;            /* active tls session, if non-null */
#endif
};

enum socket_option_type {OPT_BOOL, OPT_INT, OPT_ON};

typedef struct _dexter_socket_option
{
    const char *name;
    int level;
    int option;
    int value;
    int opttype;
} dexter_socket_option;

static const dexter_socket_option socket_options[] = 
{
#ifdef SO_KEEPALIVE
  {"SO_KEEPALIVE",      SOL_SOCKET,    SO_KEEPALIVE,    0,                 OPT_BOOL},
#endif
#ifdef SO_REUSEADDR
  {"SO_REUSEADDR",      SOL_SOCKET,    SO_REUSEADDR,    0,                 OPT_BOOL},
#endif
#ifdef TCP_NODELAY
  {"TCP_NODELAY",       IPPROTO_TCP,   TCP_NODELAY,     0,                 OPT_BOOL},
#endif
#ifdef TCP_KEEPCNT
  {"TCP_KEEPCNT",       IPPROTO_TCP,   TCP_KEEPCNT,     0,                 OPT_INT},
#endif
#ifdef TCP_KEEPIDLE
  {"TCP_KEEPIDLE",      IPPROTO_TCP,   TCP_KEEPIDLE,    0,                 OPT_INT},
#endif
#ifdef TCP_KEEPINTVL
  {"TCP_KEEPINTVL",     IPPROTO_TCP,   TCP_KEEPINTVL,   0,                 OPT_INT},
#endif
#ifdef IPTOS_LOWDELAY
  {"IPTOS_LOWDELAY",    IPPROTO_IP,    IP_TOS,          IPTOS_LOWDELAY,    OPT_ON},
#endif
#ifdef IPTOS_THROUGHPUT
  {"IPTOS_THROUGHPUT",  IPPROTO_IP,    IP_TOS,          IPTOS_THROUGHPUT,  OPT_ON},
#endif
#ifdef SO_REUSEPORT
  {"SO_REUSEPORT",      SOL_SOCKET,    SO_REUSEPORT,    0,                 OPT_BOOL},
#endif
#ifdef SO_SNDBUF
  {"SO_SNDBUF",         SOL_SOCKET,    SO_SNDBUF,       0,                 OPT_INT},
#endif
#ifdef SO_RCVBUF
  {"SO_RCVBUF",         SOL_SOCKET,    SO_RCVBUF,       0,                 OPT_INT},
#endif
#if defined(SO_SNDLOWAT) && !defined(DEXTER_PLATFORM_WIN32)
  {"SO_SNDLOWAT",       SOL_SOCKET,    SO_SNDLOWAT,     0,                 OPT_INT},
#endif
#if defined(SO_RCVLOWAT) && !defined(DEXTER_PLATFORM_WIN32)
  {"SO_RCVLOWAT",       SOL_SOCKET,    SO_RCVLOWAT,     0,                 OPT_INT},
#endif
#ifdef SO_SNDTIMEO
  {"SO_SNDTIMEO",       SOL_SOCKET,    SO_SNDTIMEO,     0,                 OPT_INT},
#endif
#ifdef SO_RCVTIMEO
  {"SO_RCVTIMEO",       SOL_SOCKET,    SO_RCVTIMEO,     0,                 OPT_INT},
#endif
#ifdef TCP_FASTACK
  {"TCP_FASTACK",       IPPROTO_TCP,   TCP_FASTACK,     0,                 OPT_BOOL},
#endif
  {NULL,0,0,0,0}
};


GQuark
dexter_network_error_quark (void)
{
  static GQuark error_quark = 0;

  if (error_quark == 0)
    error_quark = g_quark_from_static_string ("dexter-network-error-quark");

  return error_quark;
}


static int
_dexter_network_connect (int                    sockfd,
                         const struct sockaddr *addr,
                         socklen_t              addrlen,
                         GError               **error)
{
  int n;

  g_return_val_if_fail (sockfd > 0, -1);
  g_return_val_if_fail (addr != NULL, -1);
  g_return_val_if_fail (error == NULL || *error == NULL, -1);     

  if ((n = connect (sockfd, addr, addrlen)) == -1)
  {
    gint dxt_errno;
    switch (errno)
    {
      case EADDRINUSE:
        dxt_errno = DEXTER_NETWORK_ERROR_EADDRINUSE;
        break;
      case EAGAIN:
        dxt_errno = DEXTER_NETWORK_ERROR_EAGAIN;
        break;
      case ECONNREFUSED:
        dxt_errno = DEXTER_NETWORK_ERROR_ECONNREFUSED;
        break;
      case EINTR:
        dxt_errno = DEXTER_NETWORK_ERROR_EINTR;
        break;
      case ENETUNREACH:
        dxt_errno = DEXTER_NETWORK_ERROR_ENETUNREACH;
        break;
      case ETIMEDOUT:
        dxt_errno = DEXTER_NETWORK_ERROR_ETIMEDOUT;
        break;
      default:
        dxt_errno = DEXTER_NETWORK_ERROR_FAILED;
        break;
    }
    g_set_error (error,
                 DEXTER_NETWORK_ERROR,
                 dxt_errno,
                 "%s: %s",
                 G_STRFUNC, g_strerror (errno));
  }
  return n;
}


static void
_dexter_network_set_socket_options (int          sockfd,
                                    const char  *options,
                                    GError     **error)
{
  gchar  **optarray = NULL;
  gchar  **optsplit   = NULL;
  gchar  **optcurrent = NULL;

  g_return_if_fail (sockfd > 0);
  g_return_if_fail (error == NULL || *error == NULL);

  /* options is a string of space or tab delimited socket options, 
   * e.g. "SO_RCVBUF=8096 SO_SNDBUF=8095".  split it 
   * up into an array of name[=value] strings.
   */
  optarray = g_strsplit_set (options, " \t", 0);
  for (optcurrent=optarray; optcurrent && *optcurrent; optcurrent++)
  {
    if (**optcurrent)
    {
      gchar   *optname  = NULL;
      gchar   *optvalue = NULL;
      int      value = 1;
      int      ret = 0;
      int      i;
      gboolean got_value = FALSE;

      /* parse the name[=value] current option */
      optsplit = g_strsplit (*optcurrent, "=", 0);
      optname = *optsplit;
      if (optsplit+1)
      {
        optvalue = *(optsplit+1);
        if (optvalue)
        {
          errno=0;
          value = strtol (optvalue, (char **)NULL, 10);
          if (errno)
          {
            g_set_error (error,
                         DEXTER_NETWORK_ERROR,
                         DEXTER_NETWORK_ERROR_FAILED,
                         "%s: Error converting (%s) socket option value (%s)",
                         G_STRFUNC, optname, optvalue);
            goto done;
          }
          else
          {
            got_value = TRUE;
          }
        }
      }

      /* now find the socket option by name from our static list */
      for (i=0;socket_options[i].name;i++)
      if (strcmp(socket_options[i].name,optname)==0)
        break;

      /* did we find the option? */
      if (!socket_options[i].name) 
      {
        /* not found */
        g_warning ("%s: ignoring unknown socket option %s", G_STRFUNC, optname);
        continue; 
      }

      switch (socket_options[i].opttype) 
      {
          case OPT_BOOL:
          case OPT_INT:
            {
              ret = setsockopt(sockfd,socket_options[i].level,
                               socket_options[i].option,(char *)&value,sizeof(int));
              break;
            }
          case OPT_ON:
            {
              if (got_value)
              {
                g_warning ("%s: option (%s) does not take a value",G_STRFUNC, optname);
              }
              {
                int on = socket_options[i].value;
                ret = setsockopt(sockfd,socket_options[i].level,
                socket_options[i].option,(char *)&on,sizeof(int));
              }
              break;
            }
          default:
            {
              /* programming error */
              g_error (G_STRLOC " invalid socket option type (%d)", socket_options[i].opttype);
            }
      }
      
      if (ret)
      {
        g_set_error (error,
                     DEXTER_NETWORK_ERROR,
                     DEXTER_NETWORK_ERROR_FAILED,
                     "%s: Error (%d) failed to set socket option %s: (%s)",
                     G_STRFUNC, errno, optname, g_strerror(errno));
        goto done;
      }
      
      g_strfreev (optsplit);
      optsplit = NULL;
    }
  }

done:
  g_strfreev (optsplit);
  g_strfreev (optarray);
}


#ifdef HAVE_TLS
static gboolean
_dexter_network_session_starttls (dexter_network_session_t  *session,
                                  gboolean                   role,
                                  const gchar               *cert_file,
                                  const gchar               *key_file,
                                  const gchar               *trust_file,
                                  const gchar               *crl_file,
                                  guint                      flags,
                                  GError                   **error)
{
  GError *local_error = NULL;

  g_return_val_if_fail (session != NULL, FALSE);
  g_return_val_if_fail (error == NULL || *error == NULL, FALSE);

#ifdef G_ENABLE_DEBUG
  g_message (G_STRFUNC);
  g_message ("TLS cert file = %s", cert_file);
  g_message ("TLS key file = %s", key_file);
  g_message ("TLS trust file = %s", trust_file);
  g_message ("TLS crl file = %s", crl_file);
  g_message ("TLS flags = %d", flags);
#endif

  /* check if TLS already started */
  if (session->tls)
  {
    g_set_error (error,
                 DEXTER_NETWORK_ERROR,
                 DEXTER_NETWORK_ERROR_FAILED,
                 "%s: TLS alreayd started", G_STRFUNC);
    return FALSE;
  }

  /* create session TLS data */
  session->tls = dexter_tls_new (role,
                                 cert_file,
                                 key_file,
                                 trust_file,
                                 crl_file,
                                 flags,
                                 &local_error);
  if (local_error)
  {
    g_set_error (error,
                 DEXTER_NETWORK_ERROR,
                 DEXTER_NETWORK_ERROR_FAILED,
                 "%s: error creating TLS session: %s", G_STRFUNC, local_error->message);
    g_clear_error (&local_error);
    return FALSE;
  }

  /* TLS handshake */
  dexter_tls_handshake (session->tls,
                        session->sockfd,
                        &local_error);
  if (local_error)
  {
    g_set_error (error,
                 DEXTER_NETWORK_ERROR,
                 DEXTER_NETWORK_ERROR_FAILED,
                 "%s: error during TLS handshake: %s", G_STRFUNC, local_error->message);
    g_clear_error (&local_error);
    return FALSE;
  }

  /* verify peer */
  if (flags & DEXTER_TLS_VERIFY_HOSTNAME)
  {
    /* resolve peer's hostname */
    char hbuf[NI_MAXHOST];
    if (getnameinfo ((const struct sockaddr *)&session->addr, 
                      session->addrlen, hbuf, sizeof(hbuf), NULL, 0, NI_NAMEREQD))
    {
      g_set_error (error,
                   DEXTER_NETWORK_ERROR,
                   DEXTER_NETWORK_ERROR_FAILED,
                   "%s: could not resolve peer hostname", G_STRFUNC);
      return FALSE;
    }
    /* verify peer with hostname check */
    dexter_tls_verify_peer (session->tls,
                            hbuf,
                            &local_error);
  }
  else
  {
    /* verify peer without hostname check */
    dexter_tls_verify_peer (session->tls,
                            NULL,
                            &local_error);
  }
  if (local_error)
  {
    g_set_error (error,
                 DEXTER_NETWORK_ERROR,
                 DEXTER_NETWORK_ERROR_FAILED,
                 "%s: error verifying TLS peer: %s", G_STRFUNC, local_error->message);
    g_clear_error (&local_error);
    return FALSE;
  }

  return TRUE;
}
#endif


int
dexter_network_connect (const char                *hostname,
                        const char                *service,
                        const char                *socket_options,
                        struct sockaddr_storage   *addr,
                        socklen_t                 *addrlen,
                        GError                   **error)
{
  int              sockfd, n;
  const int        on = 1;
  struct addrinfo  hints, *res, *res0;
  GError          *local_error = NULL;
    
  g_return_val_if_fail (hostname != NULL, -1);
  g_return_val_if_fail (service != NULL, -1);
  g_return_val_if_fail (error == NULL || *error == NULL, -1);

  memset(&hints, 0, sizeof(struct addrinfo));
  hints.ai_family   = AF_UNSPEC;
  hints.ai_socktype = SOCK_STREAM;

  if ((n = getaddrinfo (hostname, service, &hints, &res)))
  {
    gint dxt_errno;
    switch (n)
    {
      case EAI_AGAIN:
        dxt_errno = DEXTER_NETWORK_ERROR_EAGAIN;
        break;
      default:
        dxt_errno = DEXTER_NETWORK_ERROR_FAILED;
        break;
    }
    g_set_error (error,
                 DEXTER_NETWORK_ERROR,
                 dxt_errno,
                 "%s: %s",        
                 G_STRFUNC, gai_strerror (errno));
                 return -1;
  }

  res0 = res;

  do {
    g_clear_error (&local_error);
    if ((sockfd = socket (res->ai_family, res->ai_socktype, res->ai_protocol)) == -1)
      continue;                   /* ignore */

    if (socket_options)
    {
      _dexter_network_set_socket_options (sockfd, socket_options, &local_error);
      if (local_error)
      {
        g_propagate_error (error, local_error);
        freeaddrinfo (res0);
        return -1;
      }
    }
    /* always use TCP_NODELAY */
    if (setsockopt (sockfd, IPPROTO_TCP, TCP_NODELAY, &on, sizeof(on)) == -1)
    {
      g_set_error (error,
                   DEXTER_NETWORK_ERROR,
                   DEXTER_NETWORK_ERROR_FAILED,
                   "%s: Error (%d) failed to set socket option: (%s)",
                   G_STRFUNC, errno, g_strerror(errno));
      freeaddrinfo (res0);
      return -1;
    }

    if (_dexter_network_connect (sockfd, res->ai_addr, res->ai_addrlen, &local_error) == 0)
      break;                      /* success */

    if (close (sockfd) < 0)       /* ignore */
      continue;

  } while ((res = res->ai_next));

  if (!res)                       /* all attempts failed, last error is the one we pass up */
  {
    if (local_error)
    {
      g_propagate_error (error, local_error);
    }
    else
    {
      g_set_error (error,
                   DEXTER_NETWORK_ERROR,
                   DEXTER_NETWORK_ERROR_FAILED,
                   "%s: unable to connect to host %s, service %s: %s",
                   G_STRFUNC, hostname, service, g_strerror (errno));

    }
    freeaddrinfo (res0);
    return -1;
  }

  /* if addr was provided, copy the res->ai_addr for the caller */
  if (addr)
    memcpy (addr, res->ai_addr, res->ai_addrlen);

  /* if addrlen was provided, copy the res->ai_addrlen for the caller */
  if (addrlen)
    *addrlen = res->ai_addrlen;

  freeaddrinfo (res0);
  return sockfd;
}


int
dexter_network_listen (const char                *hostname,
                       const char                *service,
                       const char                *socket_options,
                       struct sockaddr_storage   *addr,
                       socklen_t                 *addrlen,
                       GError                   **error)
{
  int              listenfd, n;
  const int        on = 1;
  struct addrinfo  hints, *res, *res0;
  GError          *local_error = NULL;

  g_return_val_if_fail (hostname != NULL, -1);
  g_return_val_if_fail (service != NULL, -1);
  g_return_val_if_fail (error == NULL || *error == NULL, -1);

  memset(&hints, 0, sizeof (struct addrinfo));
  hints.ai_flags  = AI_PASSIVE;
  hints.ai_family   = AF_UNSPEC;
  hints.ai_socktype   = SOCK_STREAM;

  if ((n = getaddrinfo (hostname, service, &hints, &res)))
  {
    gint dxt_errno;
    switch (n)
    {
      case EAI_AGAIN:
        dxt_errno = DEXTER_NETWORK_ERROR_EAGAIN;
        break;
      default:
        dxt_errno = DEXTER_NETWORK_ERROR_FAILED;
        break;
    }
    g_set_error (error,
                 DEXTER_NETWORK_ERROR,
                 dxt_errno,
                 "%s: %s",
                 G_STRFUNC, gai_strerror (errno));
    return -1;
  }

  res0 = res;

  do {
    g_clear_error (&local_error);
    if ((listenfd = socket (res->ai_family, res->ai_socktype, res->ai_protocol)) == -1)
      continue;           /* try next one */

    if (socket_options)
    {
      _dexter_network_set_socket_options (listenfd, socket_options, &local_error);
      if (local_error)
      {
        g_propagate_error (error, local_error);
        freeaddrinfo (res0);
        return -1;
      }
    }

    /* always use SO_REUSEADDR and TCP_NODELAY */
    if ( (setsockopt (listenfd, SOL_SOCKET, SO_REUSEADDR, &on, sizeof(on)) == -1) ||
         (setsockopt (listenfd, IPPROTO_TCP, TCP_NODELAY, &on, sizeof(on)) == -1) )
    {
      g_set_error (error,
                   DEXTER_NETWORK_ERROR,
                   DEXTER_NETWORK_ERROR_FAILED,
                   "%s: Error (%d) failed to set socket option: (%s)",
                   G_STRFUNC, errno, g_strerror(errno));
      freeaddrinfo (res0);
      return -1;
    }

    if (bind (listenfd, res->ai_addr, res->ai_addrlen) == 0)
      break;              /* success */

    if (close (listenfd) < 0)       /* bind error, close and try next one */
      return -1;

  } while ((res = res->ai_next));

  if (!res)                         /* all attempts failed, last error is the one we pass up */
  {
    if (local_error)
    {
      g_propagate_error (error, local_error);
    }
    else
    {
      g_set_error (error,
                   DEXTER_NETWORK_ERROR,
                   DEXTER_NETWORK_ERROR_FAILED,
                   "%s: Error unable to listen on host %s, service %s: %s",
                   G_STRFUNC, hostname, service, g_strerror (errno));
    }
    freeaddrinfo (res0);
    return -1;
  }

  if (listen (listenfd, DEFAULT_BACKLOG) == -1)
  {
    g_set_error (error,
                 DEXTER_NETWORK_ERROR,
                 DEXTER_NETWORK_ERROR_FAILED,
                 "%s: Error unable to listen on host %s, service %s: %s",
                 G_STRFUNC, hostname, service, g_strerror (errno));

    freeaddrinfo (res0);
    return -1;
  }

  /* if addr was provided, copy the res->ai_addr for the caller */
  if (addr)
    memcpy (addr, res->ai_addr, res->ai_addrlen);

   /* if addrlen was provided, copy the res->ai_addrlen for the caller */
  if (addrlen)
    *addrlen = res->ai_addrlen;

  freeaddrinfo (res0);
  return listenfd;
}


void
dexter_network_print_socket_options (int sockfd)
{
  int       value;
  socklen_t vlen = sizeof (int);
  const dexter_socket_option *p = socket_options;

  g_return_if_fail (sockfd > 0);

  for (; p->name; p++)
  {
    if (getsockopt (sockfd, p->level, p->option, (void *)&value, &vlen) == -1)
    {
      g_warning ("%s: could not test socket option %s", G_STRFUNC, p->name);
    }
    else
    {
      g_message ("socket option %s = %d", p->name,value);
    }
  }
}


dexter_network_session_t *
dexter_network_session_new (void)
{
  dexter_network_session_t *session;
  
  session = g_slice_new0 (dexter_network_session_t);

  return (session);
}


gboolean
dexter_network_session_open (dexter_network_session_t  *session,
                             int                        sockfd,
                             struct sockaddr_storage   *addr,
                             socklen_t                  addrlen)
{
  g_return_val_if_fail (session != NULL, FALSE);
  g_return_val_if_fail (sockfd > 0, FALSE);
  g_return_val_if_fail (addr != NULL, FALSE);
  g_return_val_if_fail (addrlen > 0, FALSE);

  session->sockfd = sockfd;
  memcpy (&session->addr, addr, sizeof (struct sockaddr_storage));
  session->addrlen = addrlen;

  return TRUE;
}


gboolean
dexter_network_session_client_starttls (dexter_network_session_t  *session,
                                        const gchar               *cert_file,
                                        const gchar               *key_file,
                                        const gchar               *trust_file,
                                        const gchar               *crl_file,
                                        gboolean                   verify_hostname,
                                        GError                   **error)
{
#ifndef HAVE_TLS
  g_set_error (error,
               DEXTER_NETWORK_ERROR,
               DEXTER_NETWORK_ERROR_TLS_NOT_ENABLED,
               "%s: TLS not enabled", G_STRFUNC);
  return FALSE;
#else
  return _dexter_network_session_starttls (session,
                                           DEXTER_TLS_ROLE_CLIENT,
                                           cert_file,
                                           key_file,
                                           trust_file,
                                           crl_file,
                                           verify_hostname ? DEXTER_TLS_VERIFY_HOSTNAME : 0,
                                           error);
#endif
}

gboolean
dexter_network_session_server_starttls (dexter_network_session_t  *session,
                                        const gchar               *cert_file,
                                        const gchar               *key_file,
                                        const gchar               *trust_file,
                                        const gchar               *crl_file,
                                        guint                      client_cert_flag,
                                        gboolean                   verify_hostname,
                                        GError                   **error)
{
#ifndef HAVE_TLS
  g_set_error (error,
               DEXTER_NETWORK_ERROR,
               DEXTER_NETWORK_ERROR_TLS_NOT_ENABLED,
               "%s: TLS not enabled", G_STRFUNC);
  return FALSE;
#else
  return _dexter_network_session_starttls (session,
                                           DEXTER_TLS_ROLE_SERVER,
                                           cert_file,
                                           key_file,
                                           trust_file,
                                           crl_file,
                                           (1<<client_cert_flag) |
                                              (verify_hostname ? DEXTER_TLS_VERIFY_HOSTNAME : 0),
                                           error);
#endif
}


gboolean 
dexter_network_session_tls_running (dexter_network_session_t *session)
{
  g_return_val_if_fail (session != NULL, FALSE);

#ifndef HAVE_TLS
  return FALSE;
#else
  return (session->tls != NULL);
#endif
}


const struct sockaddr *
dexter_network_session_peer_addr (dexter_network_session_t *session)
{
  g_return_val_if_fail (session != NULL, NULL);

  return (const struct sockaddr *)&session->addr;
}


socklen_t
dexter_network_session_peer_addrlen (dexter_network_session_t *session)
{
  g_return_val_if_fail (session != NULL, 0);

  return session->addrlen;
}


/* returns TRUE if bytes pending to read */
gboolean
dexter_network_session_pending (dexter_network_session_t *session)
{
  struct timeval timeout;
  fd_set         rset;
  
  g_return_val_if_fail (session != NULL, FALSE);
  g_return_val_if_fail (session->sockfd > 0, FALSE);

  timeout.tv_sec= 0;
  timeout.tv_usec = 20;
  FD_ZERO (&rset);
  FD_SET (session->sockfd, &rset);

  if ((select (session->sockfd+1, &rset, NULL, NULL, &timeout) < 0) && (errno == EINTR))
    return FALSE;

  return FD_ISSET (session->sockfd, &rset);
}


ssize_t
dexter_network_session_recv (dexter_network_session_t  *session,
                             void                      *buffer,
                             size_t                     bufsize,
                             int                        flags,
                             GError                   **error)
{
  ssize_t n;

  g_return_val_if_fail (session != NULL, -1);
  g_return_val_if_fail (session->sockfd > 0, -1);
  g_return_val_if_fail (buffer != NULL, -1);
  g_return_val_if_fail (bufsize > 0, -1);
  g_return_val_if_fail (error == NULL || *error == NULL, -1);

#ifdef HAVE_TLS
  /* use TLS api if active */
  if (session->tls)
  {
    GError *local_error = NULL;

    n = dexter_tls_recv (session->tls, buffer, bufsize, &local_error);
    if (local_error)
    {
      n = -1;
      switch (local_error->code)
      {
        case DEXTER_TLS_ERROR_EINTR:
          {
            g_set_error (error,
                         DEXTER_NETWORK_ERROR,
                         DEXTER_NETWORK_ERROR_EINTR,
                         "%s: a signal occurred before any data was delivered",
                         G_STRFUNC);
            g_clear_error (&local_error);
            break;
          }
        case DEXTER_TLS_ERROR_EAGAIN:
          {
            g_set_error (error,
                         DEXTER_NETWORK_ERROR,
                         DEXTER_NETWORK_ERROR_EINTR,
                         "%s: operation would block",
                         G_STRFUNC);
            g_clear_error (&local_error);
            break;
          }
        default:
          {
            g_set_error (error,
                         DEXTER_NETWORK_ERROR,
                         DEXTER_NETWORK_ERROR_FAILED,
                         "%s: %s",
                         G_STRFUNC, local_error->message);
            g_clear_error (&local_error);
            break;
          }
      }
    }
  } /* if (session->tls) */

  else 
#endif /* HAVE_TLS */
  
  /* use ordinary network api */
  if ((n = recv (session->sockfd, buffer, bufsize, flags)) == -1)
  {
    if (errno == EINTR)
    {
      g_set_error (error,
                   DEXTER_NETWORK_ERROR,
                   DEXTER_NETWORK_ERROR_EINTR,
                   "%s: a signal occurred before any data was delivered",
                   G_STRFUNC);
    }
    else if (errno == EAGAIN)
    {
      g_set_error (error,
                   DEXTER_NETWORK_ERROR,
                   DEXTER_NETWORK_ERROR_EAGAIN,
                   "%s: operation would block",
                   G_STRFUNC);
    }
    else
    { 
      g_set_error (error,
                   DEXTER_NETWORK_ERROR,
                   DEXTER_NETWORK_ERROR_FAILED,
                   "%s: %s",
                   G_STRFUNC, g_strerror (errno));
    }
  }

  return n;
}


ssize_t
dexter_network_session_sendall (dexter_network_session_t  *session,
                                const void                *buffer,
                                size_t                     bufsize,
                                int                        flags,
                                GError                   **error)
{
  ssize_t total=0, n;
  size_t  remaining=bufsize;

  g_return_val_if_fail (session != NULL, -1);
  g_return_val_if_fail (session->sockfd > 0, -1);
  g_return_val_if_fail (buffer != NULL, -1);
  g_return_val_if_fail (bufsize > 0, -1);
  g_return_val_if_fail (error == NULL || *error == NULL, -1);

  while (total < remaining)
  {
#ifdef HAVE_TLS
    /* use TLS api if active */
    if (session->tls)
    {
      GError *local_error = NULL;
      n = dexter_tls_send (session->tls, buffer+total, bufsize, &local_error);
      if (local_error)
      {
        if ((local_error->code == DEXTER_TLS_ERROR_EINTR) ||
            (local_error->code == DEXTER_TLS_ERROR_EAGAIN))
        {
          g_clear_error (&local_error);
          continue;
        }
        else
        {
          g_set_error (error,
                       DEXTER_NETWORK_ERROR,
                       DEXTER_NETWORK_ERROR_FAILED,
                       "%s: %s", 
                       G_STRFUNC, local_error->message);
          g_clear_error (&local_error);
          return -1;
        }
      }
    } /* if (session->tls) */

    else 
#endif
    /* use ordinary network api */
    if ((n = send (session->sockfd, buffer+total, bufsize, flags)) == -1)
    {
      if ((errno == EINTR) || (errno == EAGAIN))
        continue;
      else
      {
        g_set_error (error,
                     DEXTER_NETWORK_ERROR,
                     DEXTER_NETWORK_ERROR_FAILED,
                     "%s: %s",
                     G_STRFUNC, g_strerror (errno));
        return -1;
      }
    }

    total += n;
    remaining -= n;
  }

  return total;
}


void
dexter_network_session_close (dexter_network_session_t *session)
{
  g_return_if_fail (session != NULL);
  g_return_if_fail (session->sockfd > 0);

#ifdef HAVE_TLS
  if (session->tls)
    dexter_tls_close (session->tls);
#endif

  shutdown (session->sockfd, SHUT_RDWR);
  close (session->sockfd);
}


void
dexter_network_session_free (dexter_network_session_t *session)
{
  if (!session)
    return;

#ifdef HAVE_TLS
  if (session->tls)
    dexter_tls_free (session->tls);
#endif

  g_slice_free (dexter_network_session_t, session);
}

#define __DEXTER_NETWORK_C__
#include <dexter-aliasdef.c>
