/* $Id: dexter-private.c 393 2012-03-06 23:36:44Z pkovacs $ */
/*
 * libdexter - a plugin-based, distributed sampling library.
 * Copyright (C) 2006-2012 Philip Kovacs <pkovacs@users.sourceforge.net>.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 2.1
 * of the License, or (at your option) any later version.

 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.

 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
 * 02110-1301  USA
 *
*/

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <dexter.h>
#include <dexter-private.h>
#include <dexterplugin-main.h>
#include <dexter-alias.h>

static DexterPrivate      _dexter_private        = { (DexterConfig *)         0,
                                                     (DexterEventLoop *)      0,
                                                     (DexterFactory *)        0,
                                                     (DexterThreadPool *)     0,
                                                     (DexterVersion *)        0
                                                   };

static DexterPluginIMain  _dexter_private_imain  = { (DexterPluginIConfig  *) 0,
                                                     (DexterPluginIVersion *) 0,
                                                     (void *)                 0,
                                                     (void *)                 0,
                                                     (void *)                 0,
                                                     (void *)                 0,
                                                     (void *)                 0,
                                                     (void *)                 0,
                                                     (void *)                 0,
                                                     (void *)                 0,
                                                     (void *)                 0,
                                                     (void *)                 0,
                                                     (void *)                 0,
                                                     (void *)                 0,
                                                     (void *)                 0,
                                                     (void *)                 0
                                                   };

static GError             *dexter_convert_factory_error (GError **factory_error);

gboolean
dexter_private_init (const gchar  *config_filename,
                     GError      **error)
{
  GError *local_error = NULL;

  g_return_val_if_fail (config_filename != NULL, FALSE);
  g_return_val_if_fail (error == NULL || *error == NULL, FALSE);

  g_assert (_dexter_private.config == NULL);

  /* Init config object */
  _dexter_private.config = g_object_new (DEXTER_TYPE_CONFIG, NULL);
  dexter_config_load (_dexter_private.config, config_filename, &local_error);
  if (local_error)
  {
    g_propagate_error (error, local_error);
    return FALSE;
  }
  _dexter_private_imain.iconfig = DEXTERPLUGIN_ICONFIG (_dexter_private.config);

  /* Init version object */
  _dexter_private.version = g_object_new (DEXTER_TYPE_VERSION, NULL);
  _dexter_private_imain.iversion = DEXTERPLUGIN_IVERSION (_dexter_private.version);

  /* imain interface is ready.  Pass it to libdexterplugin (before loading plugins). */
  dexterplugin_main_imain_set (&_dexter_private_imain);

  /* Init threadpool object */
  _dexter_private.threadpool = g_object_new (DEXTER_TYPE_THREADPOOL, NULL);
  dexter_threadpool_tp_init (_dexter_private.threadpool, &local_error);
  if (local_error)
  {
    g_propagate_error (error, local_error);
    return FALSE;
  }

  /* Init event loop object (after threadpool init) */
  _dexter_private.eventloop = g_object_new (DEXTER_TYPE_EVENT_LOOP, NULL);
  dexter_event_loop_run (_dexter_private.eventloop);

  /* Init factory object */
  _dexter_private.factory = g_object_new (DEXTER_TYPE_FACTORY, NULL);
  dexter_factory_load_plugins (_dexter_private.factory, &local_error);
  if (local_error)
  {
    g_propagate_error (error, dexter_convert_factory_error (&local_error));
    return FALSE;
  }

  return TRUE;
}


void
dexter_private_exit (void)
{
  g_assert (_dexter_private.config != NULL);

  /* de-init event loop before stopping threadpool */
  dexter_event_loop_quit (_dexter_private.eventloop);
  g_object_unref (_dexter_private.eventloop);
  g_object_unref (_dexter_private.version);
  g_object_unref (_dexter_private.config);
  dexter_threadpool_tp_exit (_dexter_private.threadpool);
  g_object_unref (_dexter_private.threadpool);
  g_object_unref (_dexter_private.factory);
}


const DexterPrivate *
dexter_private_get_priv (void)
{
  return &_dexter_private;
}


static GError *
dexter_convert_factory_error (GError **factory_error)
{
  GError *dexter_error = NULL;

  g_return_val_if_fail ((factory_error != NULL) && (*factory_error != NULL), NULL);
  g_return_val_if_fail ((*factory_error)->domain == DEXTER_FACTORY_ERROR, NULL);

  dexter_error = g_error_copy (*factory_error);
  dexter_error->domain = DEXTER_ERROR;
  switch ((*factory_error)->code)
  {
    case DEXTER_FACTORY_ERROR_MODULES_NOT_SUPPORTED:
      dexter_error->code = DEXTER_ERROR_MODULES_NOT_SUPPORTED;
      break;
    case DEXTER_FACTORY_ERROR_NOT_ABSOLUTE_PATH:
      dexter_error->code = DEXTER_ERROR_NOT_ABSOLUTE_PATH;
      break;
    case DEXTER_FACTORY_ERROR_FAILED:
    default:
      dexter_error->code = DEXTER_ERROR_FAILED;
      break;
  }
  g_clear_error (factory_error);

  return dexter_error;
}

#define __DEXTER_PRIVATE_C__
#include <dexter-aliasdef.c>
