/* $Id: dexter-sampler.c 393 2012-03-06 23:36:44Z pkovacs $ */
/*
 * libdexter - a plugin-based, distributed sampling library.
 * Copyright (C) 2006-2012 Philip Kovacs <pkovacs@users.sourceforge.net>.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 2.1
 * of the License, or (at your option) any later version.

 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.

 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
 * 02110-1301  USA
 *
*/

/**
 * SECTION:dexter-sampler
 * @title: Samplers
 * @short_description: Using samplers for data snapshots
 * @see_also:
 *
 * Samplers are Libdexter objects that are capable of sampling data in some meaningful
 * domain, storing the data internally and returning the data to the caller.  Each sampler
 * "service" is implemented in a Libdexter plug-in and has a <emphasis>universally unique 
 * identifier </emphasis> (uuid) to distinguish it from all other Libdexter services.  
 * Libdexter provides the framework for implementing these plug-in services, instantiating 
 * them within your application and running them remotely over the network.
 *
 * In order to instantiate sampler objects, a service broker must first be used to obtain 
 * a list of available services from your local process or from a remote host.  You then 
 * select a desired service from the list and specify it in dexter_sampler_new() to create 
 * the sampler.  Once created, you may then take data "snapshots" with dexter_sampler_sample(),
 * and copy and/or duplicate the sampler data, e.g. with dexter_sampler_data_copy(), etc.
 *
 * Callbacks can be used to receive asynchronous notifications of data changes in the sampler
 * after dexter_sampler_sample() is called. To use callbacks, create a null-terminated array 
 * of #DexterSamplerDataCallback's and specify that array to dexter_sampler_new(). You can free 
 * the callback array immediately since the sampler creates a private copy.  Callbacks are
 * each invoked with a private copy of the sampler's data, rather than chaining the same memory 
 * to each callback. The data copy is freed by Libdexter automatically when the callback returns.  
 *
 * If a channel is specified in dexter_sampler_new(), the returned sampler is a transparent 
 * proxy to a partner object that is created on the channel's remote host. The remote object 
 * does the actual sampling while Libdexter handles the transmission of the data back to the proxy 
 * sampler running locally in your application.  There is no difference in the way local and 
 * remote samplers are used from the application's perspective.
 **/

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <string.h>
#include <dexter-channel-internal.h>
#include <dexter-messages.h>
#include <dexter-sampler-internal.h>
#include <dexter-private.h>
#include <dexterplugin-isampler.h>
#include <dexterplugin-sampler.h>
#include <dexter-alias.h>

static void dexter_sampler_class_init      (DexterSamplerClass    *klass);
static void dexter_sampler_init            (DexterSampler         *sampler);
static void dexter_sampler_dispose         (GObject               *object);
static void dexter_sampler_finalize        (GObject               *object);
static void dexter_sampler_process_message (DexterChannelObject   *channel_object,
                                            DexterChannelMessage  *message,
                                            GError               **error);
static void dexter_sampler_channel_event   (DexterChannelObject   *channel_object,
                                            gint                   event);

GQuark
dexter_sampler_error_quark (void)
{
  static GQuark error_quark = 0;

  if (error_quark == 0)
    error_quark = g_quark_from_static_string ("dexter-sampler-error-quark");

  return error_quark;
}

G_DEFINE_TYPE (DexterSampler, dexter_sampler, DEXTER_TYPE_CHANNEL_OBJECT);


static void
dexter_sampler_class_init (DexterSamplerClass *klass)
{
  GObjectClass             *gobject_class;
  DexterChannelObjectClass *channel_object_class;

#ifdef G_ENABLE_DEBUG
  g_message (G_STRFUNC);
#endif

  gobject_class = G_OBJECT_CLASS (klass);
  gobject_class->dispose = dexter_sampler_dispose;
  gobject_class->finalize = dexter_sampler_finalize;

  channel_object_class = DEXTER_CHANNEL_OBJECT_CLASS (klass);
  channel_object_class->process_message = dexter_sampler_process_message;
  channel_object_class->channel_event = dexter_sampler_channel_event;
}


static void
dexter_sampler_init (DexterSampler *sampler)
{
#ifdef G_ENABLE_DEBUG
  g_message (G_STRFUNC);
#endif

  sampler->dispose_has_run = FALSE;
  sampler->initialized = 0;
  sampler->plugin_sampler = NULL;
  sampler->mutex = g_mutex_new ();
}


static void
dexter_sampler_dispose (GObject *object)
{
  DexterSampler *sampler = DEXTER_SAMPLER (object);
    
#ifdef G_ENABLE_DEBUG
  g_message (G_STRFUNC);
#endif

  if (sampler->dispose_has_run)
    return;

  if (G_IS_OBJECT (sampler->plugin_sampler))
    g_object_unref (sampler->plugin_sampler);
    
  sampler->dispose_has_run = TRUE;

  /* chain up */
  G_OBJECT_CLASS (dexter_sampler_parent_class)->dispose (object);
}


static void
dexter_sampler_finalize (GObject *object)
{
  DexterSampler *sampler = DEXTER_SAMPLER (object); 
    
#ifdef G_ENABLE_DEBUG
  g_message (G_STRFUNC);
#endif

  g_mutex_free (sampler->mutex);

  /* chain up */
  G_OBJECT_CLASS (dexter_sampler_parent_class)->finalize (object);
}


static void
dexter_sampler_process_message (DexterChannelObject   *channel_object,
                                DexterChannelMessage  *message,
                                GError               **error)
{
#ifdef G_ENABLE_DEBUG
  g_message (G_STRFUNC);
#endif

  g_return_if_fail (DEXTER_IS_CHANNEL_OBJECT (channel_object));
  g_return_if_fail (message != NULL);
  g_return_if_fail (error == NULL || *error == NULL);

  g_object_ref (channel_object);

  if (message->msg_type == DEXTER_MESSAGE_REPLY)
  {
    if (message->rpy.status != STATUS_OK)
    {
      g_set_error (error,
                   DEXTER_SAMPLER_ERROR,
                   DEXTER_SAMPLER_ERROR_FAILED,
                   "%s: server error (%d): %s",
                   G_STRFUNC, message->rpy.status,
                   message->payload ? message->payload->str : NULL);
    }
    else
    {
      switch (message->rpy.command)
      {
        case COMMAND_SAMPLER_NEW:
        case COMMAND_SAMPLER_INITIALIZE:
        case COMMAND_SAMPLER_SAMPLE:
        case COMMAND_SAMPLER_FREE:
          break;
        default:
        {
          /* programming error */
          g_error (G_STRLOC " reply received from server for unimplemented command (%d)", message->rpy.command);
        }
      }
    }
  }
  else if (message->msg_type == DEXTER_MESSAGE_DATA)
  {
    DexterSampler *sampler = DEXTER_SAMPLER (channel_object);

    if (G_LIKELY (message->dat.status == STATUS_OK))
    {
      gpointer  sampler_data;
      GError   *local_error   = NULL;

      sampler_data = dexter_sampler_data_deserialize (sampler, message->payload, &local_error);
      if (local_error)
      {
        g_propagate_error (error, local_error);
      }
      else
      {
        dexter_sampler_data_xfer (sampler, NULL, sampler_data);
      }
      dexter_sampler_data_free (sampler, sampler_data);
      /* queue the callbacks onto threadpool */
      dexterplugin_sampler_callbacks (DEXTERPLUGIN_SAMPLER (sampler->plugin_sampler));
    }
  }
  else
  {
    /* programming error */
    g_error (G_STRLOC " Error invalid message type (%d) received from server", message->msg_type);
  }

  /* chain up */
  DEXTER_CHANNEL_OBJECT_CLASS (dexter_sampler_parent_class) -> process_message (channel_object, message, error);

  g_object_unref (channel_object);
}


static void
dexter_sampler_channel_event (DexterChannelObject *channel_object,
                              gint                 event)
{
  DexterSampler *sampler;
#ifdef G_ENABLE_DEBUG
  g_message (G_STRFUNC);
#endif

  g_return_if_fail (DEXTER_IS_CHANNEL_OBJECT (channel_object));

  g_object_ref (channel_object);
  sampler = DEXTER_SAMPLER (channel_object);

  switch (event)
  {
    case DEXTER_CHANNEL_EVENT_NOCONN:
    case DEXTER_CHANNEL_EVENT_CLOSED:
    case DEXTER_CHANNEL_EVENT_OPEN:
      break;
    default:
      /* programming error */
      g_error (G_STRLOC " unrecognized channel event (%d)", event);
    }

  /* chain up */
  DEXTER_CHANNEL_OBJECT_CLASS (dexter_sampler_parent_class) -> channel_event (channel_object, event);

  g_object_unref (channel_object);
}


gboolean
dexter_sampler_is_initialized (DexterSampler *sampler)
{
  gboolean ret;

  g_return_val_if_fail (DEXTER_IS_SAMPLER (sampler), FALSE);

  g_object_ref (sampler);
  ret = g_atomic_int_get (&sampler->initialized) > 0;
  g_object_unref (sampler);

  return ret;

}


DexterSampler *
dexter_sampler_new_actual (const DexterPluginService       *service,
                           const DexterSamplerDataCallback *callbacks,
                           DexterChannel                   *channel,
                           DexterPluginIThreadPool         *ithreadpool)
{
  DexterSamplerDataCallback *callbacks_copy;
  DexterPluginSampler       *plugin_sampler;
  DexterSampler             *sampler;
  guint                      n_callbacks;
  gsize                      callbacks_size;

  g_return_val_if_fail (service != NULL, NULL);
  g_return_val_if_fail (service->type != G_TYPE_INVALID, NULL);
  g_return_val_if_fail (channel == NULL || DEXTER_IS_CHANNEL (channel), NULL);
  g_return_val_if_fail (ithreadpool == NULL || DEXTERPLUGIN_IS_ITHREADPOOL (ithreadpool), NULL);

  if (channel)
    g_object_ref (channel);

  /* create a copy of the callback array */
  callbacks_copy = (DexterSamplerDataCallback *) callbacks;
  callbacks_size = 0;
  if (callbacks)
  {
    for (n_callbacks = 0; callbacks_copy && *callbacks_copy++; n_callbacks++);
    callbacks_size = (n_callbacks+1) * sizeof (DexterSamplerDataCallback);
    callbacks_copy = (DexterSamplerDataCallback *) g_slice_alloc0 (callbacks_size);
    memcpy (callbacks_copy, callbacks, callbacks_size);
  }

  /* instantiate the sampler container */
  sampler = g_object_new (DEXTER_TYPE_SAMPLER, NULL);

  /* instantiate the plugin sampler */
  plugin_sampler = g_object_new (service->type, 
                                 "private1", G_OBJECT (sampler), 
                                 "private2", callbacks_copy,
                                 "private3", ithreadpool,
                                 "private4", callbacks_size,
                                  NULL);
   
  /* set the sampler container public data */
  sampler->plugin_sampler = G_OBJECT (plugin_sampler);

  if (channel)
  {
    DexterChannelObject *channel_object = DEXTER_CHANNEL_OBJECT (sampler);

    /* register sampler as an object capable of receiving server messages */
    channel_object->wait_timeout = (glong)(1000 * dexter_channel_get_settings (channel) -> timeout_msecs);
    dexter_channel_register_channel_object (channel, channel_object, DEXTER_MESSAGE_REPLY|DEXTER_MESSAGE_DATA);
    g_object_unref (channel);
  }
  return sampler;
}


/**
 * dexter_sampler_new:
 * @service: a #DexterPluginService.
 * @callbacks: optional, null-terminated array of #DexterSamplerDataCallback functions, or %NULL.
 * @channel: optional #DexterChannel over which the sampler operates, or %NULL.
 * @error: return location for optional #GError.
 *
 * Creates a #DexterSampler implementing the indicated #DexterPluginService.  
 *
 * If a callback array is specified, it is duplicated and stored internally (so you may free
 * the @callbacks immediately after this function).  Each callback function will be called 
 * asynchronously in a thread after dexter_sampler_sample() updates the sampler.
 *
 * If a channel is indicated, the sampler operates transparently as a proxy to a sampler created 
 * on the remote machine designated during channel creation.  The channel must be opened with 
 * dexter_channel_open() prior to calling this function.
 *
 * If no channel is indicated, the sampler operates locally (in-process).
 *
 * The sampler must be freed when done with dexter_sampler_free().
 *
 * Returns: #DexterSampler if successful or %NULL otherwise.
 **/
DexterSampler *
dexter_sampler_new (const DexterPluginService        *service,
                    const DexterSamplerDataCallback  *callbacks,
                    DexterChannel                    *channel,
                    GError                          **error)
{
  DexterSampler *sampler;

#ifdef G_ENABLE_DEBUG
  g_message (G_STRFUNC);
#endif
  
  g_return_val_if_fail (service != NULL, NULL);
  g_return_val_if_fail (service->type != G_TYPE_INVALID, NULL);
  g_return_val_if_fail (channel == NULL || DEXTER_IS_CHANNEL (channel), NULL);
  g_return_val_if_fail (error == NULL || *error == NULL, NULL);

  if (channel && (dexter_channel_get_state (channel) == DEXTER_CHANNEL_STATE_OPEN)) 
  {
    DexterPluginService *local_service = NULL;
    GString             *payload;
    GError              *local_error = NULL;

    /* if a channel is indicated, we assume that the indicated service is part of
     * of a remote service group.  the gtype of that service may be different from
     * the local version of the same service, so we need to find the local version
     * of this service. */
    if (!dexter_sampler_find_local_service (&local_service, service))
    {
      g_warning ("%s: requested service is not available locally", G_STRFUNC);
      return NULL;
    }

    /* create the sampler */
    sampler = dexter_sampler_new_actual (local_service, callbacks, channel, __ITHREADPOOL);
    dexterplugin_service_free (local_service);
  
    /* send command to create network peer */
    payload = g_string_new (service->uuid);
    dexter_channel_object_send_wait_process (DEXTER_CHANNEL_OBJECT (sampler), 
                                             COMMAND_SAMPLER_NEW, 0, payload, &local_error);
    if (local_error)
    {
      g_propagate_error (error, local_error);
      /* release local object due to error */
      g_object_unref (sampler);
      return NULL;
    }
  }
  else if (channel)
  {
    g_set_error (error,
                 DEXTER_CHANNEL_ERROR,
                 DEXTER_CHANNEL_ERROR_STATE,
                 "%s: operation requires an open channel", G_STRFUNC);
    return NULL;
  }
  else
  {
    /* create a standalone local sampler */
    sampler = dexter_sampler_new_actual (service, callbacks, channel, __ITHREADPOOL);
  }

  return sampler;
}


/**
 * dexter_sampler_free:
 * @sampler: a #DexterSampler.
 * @error: return location of optional #GError.
 *
 * Frees a #DexterSampler created with dexter_sampler_new().
 *
 * If the sampler was created over a channel, the channel should be open when calling this function.
 * A message will be sent to the server to free the peer.  You will be notified if an error occurs
 * during that message transmission, but the local object is always freed in any event.
 **/
void
dexter_sampler_free (DexterSampler  *sampler,
                     GError        **error)
{
  DexterChannelObject *channel_object;
    
#ifdef G_ENABLE_DEBUG
  g_message (G_STRFUNC);
#endif
      
  g_return_if_fail (DEXTER_IS_SAMPLER (sampler));
  g_return_if_fail (error == NULL || *error == NULL);

  g_object_ref (sampler);
  channel_object = DEXTER_CHANNEL_OBJECT (sampler);
    
  /* if channel is open, send message to free peer */
  if ((channel_object->channel) && 
      (dexter_channel_get_state (channel_object->channel) == DEXTER_CHANNEL_STATE_OPEN))
  {
    GError *local_error = NULL;
    
    /* send command to free network peer */
    dexter_channel_object_send_wait_process (channel_object, COMMAND_SAMPLER_FREE, 0, NULL, &local_error);
    if (local_error)
    {
      g_propagate_error (error, local_error);
      goto done;
    }
    /* deregister sampler as an object capable of receiving server messages */
    dexter_channel_deregister_channel_object (channel_object->channel, channel_object);
  }

  /* if the channel is not open, just deregister sampler */
  else if (channel_object->channel)
  {
    /* deregister sampler as an object capable of receiving server messages */
    dexter_channel_deregister_channel_object (channel_object->channel, channel_object);
  }

done:
  /* release this api ref */
  g_object_unref (sampler);

  /* unref the local sampler */ 
  g_object_unref (sampler);
}


/**
 * dexter_sampler_initialize:
 * @sampler: a #DexterSampler.
 * @param: address of an optional #gint parameter, or %NULL.
 * @input: optional #gconstpointer to sampler-specific initialization data, or %NULL.
 * @error: return location of optional #GError.
 *
 * Performs one-time initialization of the sampler created with dexter_sampler_new(). If the sampler
 * requires simple integer data for initialization, indicate the address to that integer, or %NULL,
 * if no integer is required. If the sampler requires other, complex or structured input data for 
 * initialization, indicate the address to that input structure, or %NULL, if no input is required.
 *
 * The sampler should be initialized once after creation with dexter_sampler_new() and before 
 * sampling with dexter_sampler_sample().
 *
 * If the sampler was created over a channel, the channel should be open when calling this function.
 *
 * Returns: #gboolean indicating success or failure;
 **/
gboolean
dexter_sampler_initialize (DexterSampler  *sampler,
                           const gint     *param,
                           gconstpointer   input,
                           GError        **error)
{
  DexterChannelObject *channel_object;
    
#ifdef G_ENABLE_DEBUG
  g_message (G_STRFUNC);
#endif
      
  g_return_val_if_fail (DEXTER_IS_SAMPLER (sampler), FALSE);
  g_return_val_if_fail (error == NULL || *error == NULL, FALSE);

  g_object_ref (sampler);
  channel_object = DEXTER_CHANNEL_OBJECT (sampler);

  /* error if already initialized */
  if (dexter_sampler_is_initialized (sampler))
  {
    g_set_error (error,
                 DEXTER_SAMPLER_ERROR,
                 DEXTER_SAMPLER_ERROR_FAILED,
                 "%s: sampler already initialized", G_STRFUNC);
    g_object_unref (sampler);
    return FALSE;
  }

  /* if channel is open, send message for peer to initialize */
  if ((channel_object->channel) &&
      (dexter_channel_get_state (channel_object->channel) == DEXTER_CHANNEL_STATE_OPEN))
  {
    GString *payload = NULL;
    GError  *local_error = NULL;

    if (input)
    {
      /* serialize the input to the payload */
      payload = g_string_new (NULL);
      dexter_sampler_input_serialize (sampler, payload, input, &local_error);
      if (local_error)
      {
        g_string_free (payload, TRUE);
        g_propagate_error (error, local_error);
        g_object_unref (sampler);
        return FALSE;
      }
    }

    /* send the initialize command to the network peer */
    dexter_channel_object_send_wait_process (channel_object, COMMAND_SAMPLER_INITIALIZE,
                                             param ? *param : 0,
                                             payload,
                                             &local_error);
    if (local_error)
    {
      g_propagate_error (error, local_error);
      if (payload)
        g_string_free (payload, TRUE);
      g_object_unref (sampler);
      return FALSE;
    }
    if (payload)
      g_string_free (payload, TRUE);
  }
  else if (channel_object->channel)
  {
    g_set_error (error,
                 DEXTER_CHANNEL_ERROR,
                 DEXTER_CHANNEL_ERROR_STATE,
                 "%s: operation requires an open channel", G_STRFUNC);
    g_object_unref (sampler);
    return FALSE;
  }
  else
  {
    gboolean ret;

    /* initialize local sampler with input */
    g_mutex_lock (sampler->mutex);
    ret = 
      dexterplugin_isampler_initialize (DEXTERPLUGIN_ISAMPLER (sampler->plugin_sampler), param, input, error);
    g_mutex_unlock (sampler->mutex);
    if (!ret)
    {
      g_object_unref (sampler);
      return ret;
    }
  }

  g_atomic_int_set (&sampler->initialized, 1);

  g_object_unref (sampler);
  return TRUE;
}


/**
 * dexter_sampler_sample:
 * @sampler: a #DexterSampler.
 * @error: return location for an optional #GError.
 *
 * Have the sampler refresh its data sample.
 *
 * If the sampler was created over a channel, the channel should be open when calling this function.
 *
 * Returns: #gboolean indicating success or failure.
 **/
gboolean
dexter_sampler_sample (DexterSampler  *sampler,
                       GError        **error)
{
  DexterChannelObject  *channel_object;
  GError               *local_error = NULL;

#ifdef G_ENABLE_DEBUG
  g_message (G_STRFUNC);
#endif
  
  g_return_val_if_fail (DEXTER_IS_SAMPLER (sampler), FALSE);
  g_return_val_if_fail (error == NULL || *error == NULL, FALSE);

  g_object_ref (sampler);
  channel_object = DEXTER_CHANNEL_OBJECT (sampler);

  /* error if not initialized */
  if (!dexter_sampler_is_initialized (sampler))
  {
    g_set_error (error,
                 DEXTER_SAMPLER_ERROR,
                 DEXTER_SAMPLER_ERROR_FAILED,
                 "%s: sampler not initialized", G_STRFUNC);
    g_object_unref (sampler);
    return FALSE;
  }

  /* if an open channel send message for peer to perform sample */
  if ((channel_object->channel) &&
      (dexter_channel_get_state (channel_object->channel) == DEXTER_CHANNEL_STATE_OPEN))
  {
    gboolean done=FALSE;

    /* send command to have network peer perform a sample */
    dexter_channel_object_send_wait_process (channel_object, COMMAND_SAMPLER_SAMPLE, 0, NULL, &local_error);
    if (local_error)
    {
      g_propagate_error (error, local_error);
      g_object_unref (sampler);
      return FALSE;
    }

    do /* now wait for a data message --
          [future]: data messages may be divided up by the appl. protocol into several data chunks */
    {
      DexterChannelMessage *msg = 
        dexter_channel_object_wait_message (channel_object, channel_object->wait_timeout);
      if (!msg)
      {
          g_set_error (error,
                       DEXTER_CHANNEL_ERROR,
                       DEXTER_CHANNEL_ERROR_TIMEOUT,
                       "%s: timeout waiting for message from server", G_STRFUNC);
          g_object_unref (sampler);
          return FALSE;
      }
      else if (msg->msg_type==DEXTER_MESSAGE_DATA)
      { /* process message */
        gboolean last_data_chunk = msg->dat.flags & DAT_CHUNK_LAST;
        dexter_sampler_process_message (channel_object, msg, error);
        if (last_data_chunk)
          done=TRUE;
      }
      else
      {
        /* programming error */
        g_error (G_STRLOC " unexpected message type delivered to channel object");
      }
    }
    while (!done);
  }
  else if (channel_object->channel)
  {
    g_set_error (error,
                 DEXTER_CHANNEL_ERROR,
                 DEXTER_CHANNEL_ERROR_STATE,
                 "%s: operation requires an open channel", G_STRFUNC);
    g_object_unref (sampler);
    return FALSE;
  }
  else
  {
    gboolean ret;
    /* do local sample */
    g_mutex_lock (sampler->mutex);
    ret=dexterplugin_isampler_sample (DEXTERPLUGIN_ISAMPLER (sampler->plugin_sampler), &local_error);
    g_mutex_unlock (sampler->mutex);
    if (local_error)
    {
      g_propagate_error (error, local_error);
      g_object_unref (sampler);
      return ret;
    }

    /* queue the callbacks onto threadpool */
    dexterplugin_sampler_callbacks (DEXTERPLUGIN_SAMPLER (sampler->plugin_sampler));
  }

  g_object_unref (sampler);
  return TRUE;
}


/**
 * dexter_sampler_data_new:
 * @sampler: a #DexterSampler. 
 * @populate: #gboolean to determine if returned memory is populated with current sampler data or zero-filled.
 *
 * Allocates and returns a new sampler data structure.  The plugin implementing the sampler defines this 
 * structure -- it is completely opaque to the library.  
 *
 * If @populate is %TRUE, the returned memory is populated with current sampler data.  If %FALSE,
 * it is zero-filled.
 *
 * The memory must freed with dexter_sampler_data_free() when done, unless within a 
 * #DexterSamplerDataCallback, in which case the library will free the memory for you after 
 * the callback returns.
 *
 * Returns: #gpointer to newly allocated sample memory.
 **/
gpointer
dexter_sampler_data_new (DexterSampler *sampler,
                         gboolean       populate) 
{
  gpointer data;

  g_return_val_if_fail (DEXTER_IS_SAMPLER (sampler), NULL);
  g_return_val_if_fail (DEXTERPLUGIN_IS_ISAMPLER (sampler->plugin_sampler), NULL);

  /* need thread-safety locking if populating with existing data */
  g_object_ref (sampler);

  if (populate && (!dexter_sampler_is_initialized (sampler)))
    g_warning ("%s: sampler not initialized", G_STRFUNC);

  if (populate)
    g_mutex_lock (sampler->mutex);
  data = dexterplugin_isampler_data_new (DEXTERPLUGIN_ISAMPLER (sampler->plugin_sampler), populate);
  if (populate)
    g_mutex_unlock (sampler->mutex);
  
  g_object_unref (sampler);
  return data;
}


/**
 * dexter_sampler_data_dup:
 * @sampler: a #DexterSampler.
 * @source_data: #gconstpointer to sample data to duplicate.
 *
 * Allocates sampler data memory and copies the contents of @source_data onto the new memory.
 *
 * The user must free this memory with dexter_sampler_data_free().
 *
 * Returns: #gpointer to newly allocated memory containing the duplicated data.
 **/
gpointer
dexter_sampler_data_dup (DexterSampler *sampler,
                         gconstpointer  source_data)
{
  gpointer data;

  g_return_val_if_fail (DEXTER_IS_SAMPLER (sampler), NULL);
  g_return_val_if_fail (DEXTERPLUGIN_IS_ISAMPLER (sampler->plugin_sampler), NULL);
  g_return_val_if_fail (source_data != NULL, NULL);

  g_object_ref (sampler);

  data = dexterplugin_isampler_data_dup (DEXTERPLUGIN_ISAMPLER (sampler->plugin_sampler), source_data);
  
  g_object_unref (sampler);
  return data;
}


/**
 * dexter_sampler_data_xfer:
 * @sampler: a #DexterSampler.
 * @old_data: optional #gpointer to memory to receive a copy of the sampler's current (old) data, or %NULL.
 * @new_data: optional #gconstpointer to memory to replace the sampler's current data, or %NULL.
 *
 * Transfers (copies) data into and/or out of the sampler.  If @old_data is indicated, the sampler data
 * is copied into that memory, before being replaced with @new_data.  If @new_data is indicated, that data 
 * will be copied over existing internal sampler data, after existing data is copied out to @old_data.  
 *
 * Note that use of the @new_data parameter by users is probably never appropriate.  The proper way for users 
 * to replace internal sampler data is with dexter_sampler_sample(), i.e. simply take a new sample.
 *
 * This api is used internally by the library to transfer network-serialized sampler data into and out of 
 * samplers.  Users may also find it useful to copy out existing sampler data.
 *
 * <informalexample><programlisting>
 * /<!-- -->* One way to get the sampler's current data. *<!-- -->/
 * gpointer current_data = dexter_sampler_data_new (sampler, TRUE);
 * <!-- --> 
 * /<!-- -->* Another way *<!-- -->/
 * gpointer current_data = dexter_sampler_data_new (sampler, FALSE);
 * dexter_sampler_data_xfer (sampler, current_data, NULL);
 * </programlisting></informalexample>
 **/
void
dexter_sampler_data_xfer (DexterSampler *sampler,
                          gpointer       old_data,
                          gconstpointer  new_data)
{
  g_return_if_fail (DEXTER_IS_SAMPLER (sampler));
  g_return_if_fail (DEXTERPLUGIN_IS_ISAMPLER (sampler->plugin_sampler));

  g_object_ref (sampler);

  if (!dexter_sampler_is_initialized (sampler))
    g_warning ("%s: sampler not initialized", G_STRFUNC);

  /* need thread-safety locking */
  g_mutex_lock (sampler->mutex);
  dexterplugin_isampler_data_xfer (DEXTERPLUGIN_ISAMPLER (sampler->plugin_sampler), old_data, new_data);
  g_mutex_unlock (sampler->mutex);

  g_object_unref (sampler);
}


/**
 * dexter_sampler_data_copy:
 * @sampler: a #DexterSampler.
 * @dest_data: #gpointer to destination memory to receive the copy.
 * @source_data: #gconstpointer to source memory for the copy.
 *
 * Copies the sampler data indicated by @source_data to the memory indicated by @dest_data.
 **/
void
dexter_sampler_data_copy (DexterSampler *sampler,
                          gpointer       dest_data,
                          gconstpointer  source_data)
{
  g_return_if_fail (DEXTER_IS_SAMPLER (sampler));
  g_return_if_fail (DEXTERPLUGIN_IS_ISAMPLER (sampler->plugin_sampler));
  g_return_if_fail (dest_data != NULL);
  g_return_if_fail (source_data != NULL);

  g_object_ref (sampler);
  
  dexterplugin_isampler_data_copy (DEXTERPLUGIN_ISAMPLER (sampler->plugin_sampler), dest_data, source_data);
  
  g_object_unref (sampler);
}


/**
 * dexter_sampler_data_free:
 * @sampler: a #DexterSampler.
 * @data: #gpointer to sampler data to free.
 *
 * Frees sampler data previously allocated with dexter_sampler_data_new() or dexter_sampler_data_dup().
 **/
void
dexter_sampler_data_free (DexterSampler *sampler,
                          gpointer       data)
{
  g_return_if_fail (DEXTER_IS_SAMPLER (sampler));
  g_return_if_fail (DEXTERPLUGIN_IS_ISAMPLER (sampler->plugin_sampler));

  g_object_ref (sampler);
  
  dexterplugin_isampler_data_free (DEXTERPLUGIN_ISAMPLER (sampler->plugin_sampler), data);
  
  g_object_unref (sampler);
}


gboolean
dexter_sampler_data_serialize (DexterSampler *sampler,
                               GString       *dest_buffer,
                               gconstpointer  source_data,
                               GError        **error)
{
  gboolean ret;

  g_return_val_if_fail (DEXTER_IS_SAMPLER (sampler), FALSE);
  g_return_val_if_fail (DEXTERPLUGIN_IS_ISAMPLER (sampler->plugin_sampler), FALSE);
  g_return_val_if_fail (dest_buffer != NULL, FALSE);
  g_return_val_if_fail (source_data != NULL, FALSE);
  g_return_val_if_fail (error == NULL || *error == NULL, FALSE);

  g_object_ref (sampler);
  
  ret = dexterplugin_isampler_data_serialize (DEXTERPLUGIN_ISAMPLER (sampler->plugin_sampler), 
                                              dest_buffer, source_data, error);
  
  g_object_unref (sampler);
  return ret;
}


gpointer
dexter_sampler_data_deserialize (DexterSampler  *sampler,
                                 const GString  *source_buffer,
                                 GError        **error)
{
  gpointer data;

  g_return_val_if_fail (DEXTER_IS_SAMPLER (sampler), NULL);
  g_return_val_if_fail (DEXTERPLUGIN_IS_ISAMPLER (sampler->plugin_sampler), NULL);
  g_return_val_if_fail (source_buffer != NULL, NULL);
  g_return_val_if_fail (error == NULL || *error == NULL, NULL);

  g_object_ref (sampler);
  
  data = dexterplugin_isampler_data_deserialize (DEXTERPLUGIN_ISAMPLER (sampler->plugin_sampler),
                                                 source_buffer, error);
  
  g_object_unref (sampler);
  return data;
}


/**
 * dexter_sampler_input_new:
 * @sampler: a #DexterSampler.
 *
 * Allocates and returns a new sampler input structure.  The plugin implementing the sampler
 * defines this structure -- it is completely opaque to the library.
 *
 * The memory must be freed with dexter_sampler_input_free() when done.
 *
 * Returns: #gpointer to newly allocated input structure.
 **/
gpointer
dexter_sampler_input_new (DexterSampler *sampler)
{
  gpointer data;

  g_return_val_if_fail (DEXTER_IS_SAMPLER (sampler), NULL);
  g_return_val_if_fail (DEXTERPLUGIN_IS_ISAMPLER (sampler->plugin_sampler), NULL);

  g_object_ref (sampler);
  
  data = dexterplugin_isampler_input_new (DEXTERPLUGIN_ISAMPLER (sampler->plugin_sampler));
  
  g_object_unref (sampler);
  return data;
}


/**
 * dexter_sampler_input_free:
 * @sampler: a #DexterSampler.
 * @input: #gpointer to sampler input to free.
 *
 * Frees an input structure previously allocated with dexter_sampler_input_new().
 **/
void
dexter_sampler_input_free (DexterSampler *sampler,
                           gpointer       input)
{
  g_return_if_fail (DEXTER_IS_SAMPLER (sampler));
  g_return_if_fail (DEXTERPLUGIN_IS_ISAMPLER (sampler->plugin_sampler));

  g_object_ref (sampler);
  
  dexterplugin_isampler_input_free (DEXTERPLUGIN_ISAMPLER (sampler->plugin_sampler), input);
  
  g_object_unref (sampler);
}


gboolean
dexter_sampler_input_serialize (DexterSampler  *sampler,
                                GString        *dest_buffer,
                                gconstpointer   source_input,
                                GError        **error)
{
  gboolean ret;

  g_return_val_if_fail (DEXTER_IS_SAMPLER (sampler), FALSE);
  g_return_val_if_fail (DEXTERPLUGIN_IS_ISAMPLER (sampler->plugin_sampler), FALSE);
  g_return_val_if_fail (dest_buffer != NULL, FALSE);
  g_return_val_if_fail (source_input != NULL, FALSE);
  g_return_val_if_fail (error == NULL || *error == NULL, FALSE);

  g_object_ref (sampler);
  
  ret = dexterplugin_isampler_input_serialize (DEXTERPLUGIN_ISAMPLER (sampler->plugin_sampler),
                                               dest_buffer, source_input, error);
  
  g_object_unref (sampler);
  return ret;
}

    
gpointer
dexter_sampler_input_deserialize (DexterSampler  *sampler,
                                  const GString  *source_buffer,
                                  GError        **error)
{
  gpointer data;

  g_return_val_if_fail (DEXTER_IS_SAMPLER (sampler), NULL);
  g_return_val_if_fail (DEXTERPLUGIN_IS_ISAMPLER (sampler->plugin_sampler), NULL);
  g_return_val_if_fail (source_buffer != NULL, NULL);
  g_return_val_if_fail (error == NULL || *error == NULL, NULL);

  g_object_ref (sampler);
  
  data = dexterplugin_isampler_input_deserialize (DEXTERPLUGIN_ISAMPLER (sampler->plugin_sampler),
                                                  source_buffer, error);
  
  g_object_unref (sampler);
  return data;
}


gboolean
dexter_sampler_find_local_service (DexterPluginService       **local_service,
                                   const DexterPluginService  *remote_service)
{
  DexterPluginServiceGroup *lservice_group;
  DexterPluginService      *lservice = NULL;

  g_return_val_if_fail (local_service != NULL, FALSE);
  g_return_val_if_fail (*local_service == NULL, FALSE);
  g_return_val_if_fail (remote_service != NULL, FALSE);

  lservice_group = dexter_factory_get_services (__FACTORY, DEXTERPLUGIN_TYPE_ISAMPLER);
  if (!lservice_group)
    return FALSE;

  /* find the local service matching the uuid of the remote service */
  if (!dexterplugin_service_group_find_uuid (&lservice, lservice_group, remote_service->uuid))
    return FALSE;

  *local_service = dexterplugin_service_dup (lservice);
  dexterplugin_service_group_free (lservice_group);

  return TRUE;
}

#define __DEXTER_SAMPLER_C__
#include <dexter-aliasdef.c>
