/* $Id: dexter-server.c 393 2012-03-06 23:36:44Z pkovacs $ */
/*
 * libdexter - a plugin-based, distributed sampling library.
 * Copyright (C) 2006-2012 Philip Kovacs <pkovacs@users.sourceforge.net>.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 2.1
 * of the License, or (at your option) any later version.

 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.

 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
 * 02110-1301  USA
 *
*/

/**
 * SECTION:dexter-server
 * @title: Servers
 * @short_description: Using servers to host Libdexter network operations
 * @see_also:
 *
 * Servers are Libdexter objects that allow applications to host Libdexter
 * services on the network.   You can write a dedicated server application 
 * with just a few lines of code, or you can augment your application to 
 * allow it to host Libdexter operations as add-on functionality.
 * The configuration file determines which plugins are loaded and which
 * services are authorized for hosting.
 *
 * After creating the server with dexter_server_new() and starting it with
 * dexter_server_start(), a listening thread will begin accepting incoming 
 * connections, initiated from other hosts using dexter_channel_open().  
 * The listening thread will continue accepting connections until dexter_server_stop() 
 * is called (or the server process exits). 
 *
 * Several settings in the configuration file control the default behavior of 
 * Libdexter server objects, including default host name/ip, listening service/port, 
 * socket options, TLS settings, etc.  
 *
 * Applications should stop any running servers when done with dexter_server_stop().  
 * You should then free the server objects as a last step with dexter_server_free().
 **/

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <errno.h>
#include <fcntl.h>
#include <netdb.h>
#include <string.h>
#include <sys/select.h>
#include <sys/socket.h>
#include <sys/types.h>
#include <syslog.h>
#ifdef HAVE_LIBWRAP
#include <tcpd.h>
#endif
#include <unistd.h>
#include <dexter.h>
#include <dexter-log.h>
#include <dexter-server-internal.h>
#include <dexter-network.h>
#include <dexter-msgframer.h>
#include <dexter-servercomm.h>
#ifdef HAVE_TLS
#include <dexter-tls.h>
#endif
#include <dexter-private.h>
#include <dexter-alias.h>

typedef struct _dexter_service_data DexterServiceData;

static void     dexter_server_class_init            (DexterServerClass *klass);
static void     dexter_server_init                  (DexterServer      *server);
static void     dexter_server_finalize              (GObject           *object);
static gint     dexter_server_set_state             (DexterServer      *server,
                                                     gint               state);
static gboolean dexter_server_is_runnable           (DexterServer      *server);
static void     dexter_server_set_runnable          (DexterServer      *server,
                                                     gboolean           runnable);
static GError  *dexter_server_convert_network_error (GError           **network_error);

static DexterServerSettings
               *dexter_server_settings_dup          (const DexterServerSettings *settings);
static void     dexter_server_settings_free         (DexterServerSettings       *settings);
static void     dexter_server_settings_prepare      (DexterServerSettings       *settings);

#ifdef HAVE_TLS
                /* tls seed refresh runs in event loop */
static gboolean _tls_seed_refresh_func              (gpointer           unused);
#endif
                /* accept loop runs in a thread */
static void     _server_thread_func                 (DexterServer      *server);

                /* each client serviced in a thread */
static void     _service_thread_func                (DexterServiceData *data);


struct _DexterServer
{
  GObject               parent;

  DexterServerSettings *settings;         /* all user/config-supplied server settings */
  
  GThread              *thread;           /* joinable thread for _server_thread_func */

  volatile gint         runnable;         /* non-zero if server allowed to continue running */
  volatile gint         state;            /* server state from DexterServerState enum */
  volatile gint         client_count;     /* total number of active clients */
  
  int                   listenfd;         /* listen fd */
  
  struct
  sockaddr_storage      addr;             /* socket address */
  socklen_t             addrlen;          /* socket address length (family-dependent) */
  gchar                 host_numeric[NI_MAXHOST];       /* numeric hostname */
  gchar                 serv_numeric[NI_MAXSERV];;      /* port number */
};

struct _DexterServerClass
{
  GObjectClass parent_class;
};

struct _dexter_service_data
{
  DexterServer *server;                   /* server */
  int connfd;                             /* connection fd */
  struct sockaddr_storage addr;           /* socket address struct */
  socklen_t addrlen;                      /* socket address length (family-dependent) */
};


/* each entry corresponds to a flag in DexterServerSettingFlags such that: flag == 1<<index,
 * i.e. the order of these elements is significant */
static const DexterConfigSetting dexter_server_config_settings[] =
{
  { "Main", "ServerHost",
    DEXTER_CONFIG_SETTING_TYPE_STRING, (DexterConfigSettingDefault) DEFAULT_SERVER_HOST },

  { "Main", "ServerService", 
    DEXTER_CONFIG_SETTING_TYPE_STRING, (DexterConfigSettingDefault) DEFAULT_SERVER_SERVICE }, 

  { "Main", "ServerSocketOptions",
    DEXTER_CONFIG_SETTING_TYPE_STRING, (DexterConfigSettingDefault) DEFAULT_SERVER_SOCKET_OPTIONS },

  { "Main", "ServerMaxBacklog",
    DEXTER_CONFIG_SETTING_TYPE_INT, (DexterConfigSettingDefault) DEFAULT_SERVER_MAX_BACKLOG },

  { "Main", "ServerMaxClients",
    DEXTER_CONFIG_SETTING_TYPE_INT, (DexterConfigSettingDefault) DEFAULT_SERVER_MAX_CLIENTS },

  { "Main", "ServerRecvBufferSize",
    DEXTER_CONFIG_SETTING_TYPE_INT, (DexterConfigSettingDefault) DEFAULT_SERVER_RECV_BUFFER_SIZE },

  { "Main", "ServerTLSSupportFlag",
    DEXTER_CONFIG_SETTING_TYPE_INT, (DexterConfigSettingDefault) DEFAULT_SERVER_TLS_SUPPORT_FLAG },

  { "Main", "ServerTLSClientCertFlag",
    DEXTER_CONFIG_SETTING_TYPE_INT, (DexterConfigSettingDefault) DEFAULT_SERVER_TLS_CLIENT_CERT_FLAG },

  /* tls_start_callback has no config-based default since it's a function address */
  { "----", NULL, DEXTER_CONFIG_SETTING_TYPE_INT, (DexterConfigSettingDefault) 0 },

  { "Main", "ServerTLSCertFile",
    DEXTER_CONFIG_SETTING_TYPE_STRING, (DexterConfigSettingDefault) DEFAULT_SERVER_TLS_CERT_FILE },

  { "Main", "ServerTLSKeyFile",
    DEXTER_CONFIG_SETTING_TYPE_STRING, (DexterConfigSettingDefault) DEFAULT_SERVER_TLS_KEY_FILE },

  { "Main", "ServerTLSTrustFile",
    DEXTER_CONFIG_SETTING_TYPE_STRING, (DexterConfigSettingDefault) DEFAULT_SERVER_TLS_TRUST_FILE },

  { "Main", "ServerTLSCrlFile",
    DEXTER_CONFIG_SETTING_TYPE_STRING, (DexterConfigSettingDefault) DEFAULT_SERVER_TLS_CRL_FILE },

  { "Main", "ServerTLSVerifyHostname",
    DEXTER_CONFIG_SETTING_TYPE_BOOLEAN, (DexterConfigSettingDefault) DEFAULT_SERVER_TLS_VERIFY_HOSTNAME },
    
  /* end of array sentinel */
  { NULL, NULL, DEXTER_CONFIG_SETTING_TYPE_INT, (DexterConfigSettingDefault) 0 }
};


#ifdef HAVE_LIBWRAP
static struct request_info _request;                /* client request for libwrap */
G_LOCK_DEFINE_STATIC (_request);                    /* assume libwrap is not thread-safe */
static gboolean _request_initialized = FALSE;
#endif

#ifdef HAVE_TLS
static GSource *_tls_seed_refresh_source = NULL;    /* event source to refresh tls seed periodically */
G_LOCK_DEFINE_STATIC (_tls_seed_refresh_source);    /* mutex for tls seed refresh operations */
static volatile int _tls_server_count = 0;          /* tls seed refresh source destroyed when count 0 */
#endif


GQuark
dexter_server_error_quark (void)
{
  static GQuark error_quark = 0;

  if (error_quark == 0)
    error_quark = g_quark_from_static_string ("dexter-server-error-quark");

  return error_quark;
}


G_DEFINE_TYPE (DexterServer, dexter_server, G_TYPE_OBJECT);


static void
dexter_server_class_init (DexterServerClass *klass)
{
  GObjectClass *gobject_class;

#ifdef G_ENABLE_DEBUG
  g_message (G_STRFUNC);
#endif

  gobject_class = G_OBJECT_CLASS (klass);
  gobject_class->finalize = dexter_server_finalize;
}


static void
dexter_server_init (DexterServer *server)
{
#ifdef G_ENABLE_DEBUG
  g_message (G_STRFUNC);
#endif

  server->settings = NULL;
  server->thread = NULL;
  g_atomic_int_set (&server->runnable, 0);
  g_atomic_int_set (&server->state, DEXTER_SERVER_STATE_STOPPED);
  g_atomic_int_set (&server->client_count, 0);
  server->listenfd = (-1);
  memset (&server->addr, 0, sizeof (struct sockaddr_storage));
  server->addrlen = (socklen_t)0;
  memset (server->host_numeric, 0, sizeof (server->host_numeric));
  memset (server->serv_numeric, 0, sizeof (server->serv_numeric));
}


static void
dexter_server_finalize (GObject *object)
{
  DexterServer *server = DEXTER_SERVER (object);

#ifdef G_ENABLE_DEBUG
  g_message (G_STRFUNC);
#endif

  /* each service thread holds a ref to the server, so we cannot finalize until 
   * they are all done.  there should therefore only be multiple, concurrent
   * _readers_ of server->settings. */
  dexter_server_settings_free (server->settings);

  /* chain up */
  G_OBJECT_CLASS(dexter_server_parent_class)->finalize (object);
}


#ifdef HAVE_LIBWRAP
gboolean
dexter_server_hosts_access (int sockfd)
{
  gboolean allow;

  G_LOCK (_request);

  if (!_request_initialized)
  {
    request_init (&_request, RQ_DAEMON, g_get_prgname (), 0);
    _request_initialized=TRUE;
  }

  /* Check whether connections are denied from this host. */
  request_set (&_request, RQ_FILE, sockfd, 0);
  fromhost (&_request);
  allow = (hosts_access (&_request) != 0);
  
  G_UNLOCK (_request);

  return allow;
}
#endif


/**
 * dexter_server_get_state:
 * @server: a #DexterServer.
 *
 * Returns the current state of the server.
 *
 * Returns: #gint from #DexterServerState indicating current state of the server.
 **/
gint
dexter_server_get_state (DexterServer *server)
{
    gint state;
    g_return_val_if_fail (DEXTER_IS_SERVER (server), -1);

    g_object_ref (server);

    state = g_atomic_int_get (&server->state);

    g_object_unref (server);

    return state;
}

static gint
dexter_server_set_state (DexterServer *server,
                         gint          state)
{
  gint old_state;
  g_return_val_if_fail (DEXTER_IS_SERVER(server), -1);
  g_return_val_if_fail (state == DEXTER_SERVER_STATE_STARTED || 
                        state == DEXTER_SERVER_STATE_STOPPED, -1);

  g_object_ref (server);

  old_state = g_atomic_int_get (&server->state);
  g_atomic_int_set (&server->state, state);

  g_object_unref (server);

  return old_state;
}


static gboolean 
dexter_server_is_runnable (DexterServer *server)
{
  gboolean runnable;

  g_return_val_if_fail (DEXTER_IS_SERVER(server), FALSE);

  g_object_ref (server);

  runnable = g_atomic_int_get (&server->runnable) > 0;
  
  g_object_unref (server);

  return runnable;
}


static void
dexter_server_set_runnable (DexterServer *server,
                            gboolean      runnable)
{
  g_return_if_fail (DEXTER_IS_SERVER(server));

  g_object_ref (server);
  g_atomic_int_set (&server->runnable, runnable ? 1 : 0);
  g_object_unref (server);
}


static GError *
dexter_server_convert_network_error (GError **network_error)
{
  GError *server_error = NULL;

  g_return_val_if_fail ((network_error != NULL) && (*network_error != NULL), NULL);
  g_return_val_if_fail ((*network_error)->domain == DEXTER_NETWORK_ERROR, NULL);

  server_error = g_error_copy (*network_error);
  server_error->domain = DEXTER_SERVER_ERROR;
  switch ((*network_error)->code)
  {
    case DEXTER_NETWORK_ERROR_EACCES:
      server_error->code = DEXTER_SERVER_ERROR_EACCES;
      break;
    case DEXTER_NETWORK_ERROR_EADDRINUSE:
      server_error->code = DEXTER_SERVER_ERROR_EADDRINUSE;
      break;
    case DEXTER_NETWORK_ERROR_EAGAIN:
      server_error->code = DEXTER_SERVER_ERROR_EAGAIN;
      break;
    case DEXTER_NETWORK_ERROR_FAILED:
    default:
      server_error->code = DEXTER_SERVER_ERROR_FAILED;
      break;
  }
  g_clear_error (network_error);
  return server_error;
}


static DexterServerSettings *
dexter_server_settings_dup (const DexterServerSettings *settings)
{
  DexterServerSettings *ret_settings;

  if (!G_UNLIKELY (settings))
    return NULL;

  ret_settings = g_slice_new0 (DexterServerSettings);
  ret_settings->flags = settings->flags;
  ret_settings->host = g_strdup (settings->host);
  ret_settings->service = g_strdup (settings->service);
  ret_settings->socket_options = g_strdup (settings->socket_options);
  ret_settings->max_backlog = settings->max_backlog;
  ret_settings->max_clients = settings->max_clients;
  ret_settings->recv_buffer_size = settings->recv_buffer_size;
  ret_settings->tls_support_flag = settings->tls_support_flag;
  ret_settings->tls_client_cert_flag = settings->tls_client_cert_flag;
  ret_settings->tls_start_callback = settings->tls_start_callback;
  ret_settings->tls_cert_file = g_strdup (settings->tls_cert_file);
  ret_settings->tls_key_file = g_strdup (settings->tls_key_file);
  ret_settings->tls_trust_file = g_strdup (settings->tls_trust_file);
  ret_settings->tls_crl_file = g_strdup (settings->tls_crl_file);
  ret_settings->tls_verify_hostname = settings->tls_verify_hostname;

  return ret_settings;
}


static void
dexter_server_settings_free (DexterServerSettings *settings)
{
  if (G_UNLIKELY (!settings))
    return;

  g_free (settings->host);
  g_free (settings->service);
  g_free (settings->socket_options);
  g_free (settings->tls_cert_file);
  g_free (settings->tls_key_file);
  g_free (settings->tls_trust_file);
  g_free (settings->tls_crl_file);
  g_slice_free (DexterServerSettings, settings);
}


static void
dexter_server_settings_prepare (DexterServerSettings *settings)
{
  gint iter;

  if (G_UNLIKELY (!settings))
    return;

  /* examine each possible server setting. if user did not supply a value for a setting,
     fetch a default from the config-file, if available; otherwise use an internal default. */

  /* setting array's final { NULL, ... } entry is end sentinel */
  iter=0;
  while (dexter_server_config_settings [iter].config_group) 
  {
    gint setting = 1<<iter;

    /* if item is user-supplied, no default needed */
    if (settings->flags & setting)
    {
      iter++;
      continue;
    }

    /* item not user-supplied */

    /* if item has no config-based default */
    if (!dexter_server_config_settings [iter].config_key)
    {
      switch (setting)
      {
        case DEXTER_SERVER_SETTING_TLS_START_CALLBACK:
          settings->tls_start_callback = NULL;
          break;
        default:
          g_error (G_STRLOC " unexpected config flag (%d)", 1<<iter);
          break;
      }

      iter++;
      continue;
    }

    /* item has a config-based default. grab explicit config value or a default config value if not found */
    DexterConfigSettingDefault return_default;
    dexter_config_get_setting (__ICONFIG,
                               dexter_server_config_settings [iter].config_group,
                               dexter_server_config_settings [iter].config_key,
                               dexter_server_config_settings [iter].setting_type,
                               dexter_server_config_settings [iter].setting_default,
                               &return_default);
    switch (setting)
    {
      case DEXTER_SERVER_SETTING_HOST:
        settings->host = return_default.default_string;
        break;
      case DEXTER_SERVER_SETTING_SERVICE:
        settings->service = return_default.default_string;
        break;
      case DEXTER_SERVER_SETTING_SOCKET_OPTIONS:
        settings->socket_options = return_default.default_string;
        break;
      case DEXTER_SERVER_SETTING_MAX_BACKLOG:
        settings->max_backlog = return_default.default_int;
        break;
      case DEXTER_SERVER_SETTING_MAX_CLIENTS:
        settings->max_clients = return_default.default_int;
        break;
      case DEXTER_SERVER_SETTING_RECV_BUFFER_SIZE:
        settings->recv_buffer_size = return_default.default_int;
        break;
      case DEXTER_SERVER_SETTING_TLS_SUPPORT_FLAG:
        settings->tls_support_flag = return_default.default_int;
        break;
      case DEXTER_SERVER_SETTING_TLS_CLIENT_CERT_FLAG:
        settings->tls_client_cert_flag = return_default.default_int;
        break;
      case DEXTER_SERVER_SETTING_TLS_CERT_FILE:
        settings->tls_cert_file = return_default.default_string;
        break;
      case DEXTER_SERVER_SETTING_TLS_KEY_FILE:
        settings->tls_key_file = return_default.default_string;
        break;
      case DEXTER_SERVER_SETTING_TLS_TRUST_FILE:
        settings->tls_trust_file = return_default.default_string;
        break;
      case DEXTER_SERVER_SETTING_TLS_CRL_FILE:
        settings->tls_crl_file = return_default.default_string;
        break;
      case DEXTER_SERVER_SETTING_TLS_VERIFY_HOSTNAME:
        settings->tls_verify_hostname = return_default.default_boolean;
        break;
      default:
        g_error (G_STRLOC " unexpected config flag (%d)", 1<<iter);
        break;
    } /* switch */

    iter++;

  } /* while */
}


#ifdef HAVE_TLS
static gboolean 
_tls_seed_refresh_func (gpointer unused)
{
#ifdef G_ENABLE_DEBUG
  g_message (G_STRFUNC);
#endif
  /* seed the TLS parameters for certificate servers */
  dexter_tls_server_seed ();

  return TRUE;
}
#endif


static void
_server_thread_func (DexterServer *server)
{
  DexterServiceData *service_data = NULL;
  GError        *local_error = NULL;
  int            sockflags;
  struct timeval timeout;
  fd_set         rset;
  char           client_host[NI_MAXHOST];

  g_object_ref (server);

#ifdef G_ENABLE_DEBUG
  g_message ("server running in thread 0x%08X", (unsigned)g_thread_self());
#endif

  /* get server's numeric hostname/port and log message that server has started */
  if (getnameinfo ((const struct sockaddr *)&server->addr, server->addrlen, 
                    server->host_numeric, sizeof (server->host_numeric),
                    server->serv_numeric, sizeof (server->serv_numeric), 
                    NI_NUMERICHOST) != 0)
  {
    g_warning ("%s: unable to get numeric hostname/port of server: %s",
                G_STRFUNC, gai_strerror (errno));
    goto server_thread_done;
  }

  dexter_log (DEXTER_LOG_INFO, "server started on host %s, port %s", 
              server->host_numeric, server->serv_numeric);

#ifdef G_ENABLE_DEBUG
  dexter_network_print_socket_options (server->listenfd);
#endif
  /* set the listen socket to non-blocking mode.  we will poll for connections
   * instead of blocking on accept().  this is done so we can allow the server
   * to be stopped with dexter_server_stop (). */
  if ( (sockflags = fcntl (server->listenfd, F_GETFL, 0) == -1) ||
       (fcntl (server->listenfd, F_SETFL, sockflags | O_NONBLOCK) == -1) )
  {
    g_warning ("%s: fcntl error on listenfd %d: %s", G_STRFUNC, server->listenfd, g_strerror (errno));
    goto server_thread_done;
  }

  /* each service thread will be passed an allocated DexterServiceData structure */
  service_data = g_slice_new0 (DexterServiceData);
  service_data->server = server;
  
  do
  {
    /* addrlen is a value-result parameter for accept(), so reset it each loop */
    service_data->addrlen = sizeof (struct sockaddr_storage);
    timeout.tv_sec= 0;
    timeout.tv_usec = 20000;
    FD_ZERO (&rset);
    FD_SET (server->listenfd, &rset);
    if ((select (server->listenfd+1, &rset, NULL, NULL, &timeout) < 0) && (errno == EINTR))
      continue;
    if (FD_ISSET (server->listenfd, &rset))
    {
      /* non-blocking accept attempt */
      if ((service_data->connfd = accept (server->listenfd,
                                          (struct sockaddr *)&service_data->addr, 
                                          &service_data->addrlen)) < 0)
      {
        if ( (errno == EWOULDBLOCK) || (errno == ECONNABORTED) ||
#ifdef EPROTO
             (errno == EPROTO) ||
#endif
             (errno == EINTR) )
          /* non-fatal errors, just loop */
          continue;

        else
        {
          /* fatal error, die */
          g_warning ("%s: error (%d) during accept: %s", G_STRFUNC, errno, g_strerror (errno));
          break;
        } 
      }

      /* OK, we have a connection. */

      /* If max clients currently being serviced, close connection. */
      if (g_atomic_int_get (&server->client_count) >= server->settings->max_clients)
      {
        close (service_data->connfd);
        continue;
      }

      /* Grab numeric name of client host */
      if (getnameinfo ((const struct sockaddr *)&service_data->addr, service_data->addrlen, 
                       client_host, sizeof (client_host),
                       NULL, (size_t)0, NI_NUMERICHOST) != 0)
      {
        g_warning ("%s: unable to get numeric hostname of client connection: %s", 
                   G_STRFUNC, gai_strerror (errno));
        break;
      }

#ifdef HAVE_LIBWRAP
      /* If using libwrap, check access permission for the connecting host */
      if (!dexter_server_hosts_access (service_data->connfd))
      {
        close (service_data->connfd);
        dexter_log (DEXTER_LOG_INFO, "server %s:%s refused connection from host %s", 
                    server->host_numeric, server->serv_numeric, client_host);
        continue;
      }
#endif

      dexter_log (DEXTER_LOG_INFO, "server %s:%s accepted connection from host %s", 
                  server->host_numeric, server->serv_numeric, client_host);

      /* Use a blocking socket in the service thread, so explicitly unset O_NONBLOCK 
       * on connfd.  This is not necessary on Linux, as the fd returned from accept() 
       * does not inherit the listening socket's O_NONBLOCK status flag.  We do it for 
       * portability since some platforms do inherit O_NONBLOCK. */
      if ( (sockflags = fcntl (service_data->connfd, F_GETFL, 0) == -1) ||
           (fcntl (service_data->connfd, F_SETFL, sockflags & ~O_NONBLOCK) == -1) )
      {
        g_warning ("%s: fcntl error on connfd %d: %s",
                   G_STRFUNC, service_data->connfd, g_strerror (errno));
        break;
      }
    
      /* Service runs in threadpool. */
      dexter_threadpool_push_task (__ITHREADPOOL,
                                   (GThreadFunc) _service_thread_func,
                                   service_data,
                                   &local_error);
      if (local_error)
      {
        /* GLib's g_thread_pool_push only returns an error for non-exclusive pools when 
         * no new thread could be created.  GLib still enqueues the task, however. */
        g_warning ("%s: %s", G_STRFUNC, local_error->message);
        g_clear_error (&local_error);
        break;
      }

      /* new DexterServiceData needed for next connection */
      service_data = g_slice_new0 (DexterServiceData);
      service_data->server = server;
    }
  }
  while (dexter_server_is_runnable (server) && (dexter_get_state () == DEXTER_STATE_UP));
    
server_thread_done:
  dexter_server_set_state (server, DEXTER_SERVER_STATE_STOPPED);
  close (server->listenfd);

  dexter_log (DEXTER_LOG_INFO, "server stopped on host %s, port %s", 
              server->host_numeric, server->serv_numeric);

  if (service_data)
    g_slice_free (DexterServiceData, service_data);

  g_object_unref (server);

#ifdef G_ENABLE_DEBUG
  g_message ("server done in thread 0x%08X", (unsigned)g_thread_self());
#endif
}


static void 
_service_thread_func (DexterServiceData *sd)
{
  DexterMsgFramer  *framer;
  DexterServerComm *servercomm;
  dexter_network_session_t *session;
  GError           *local_error = NULL;
  gchar            *recvbuf = NULL;
  ssize_t           msgbytes;
  gboolean          done = FALSE;

  g_return_if_fail (sd != NULL);
  g_return_if_fail (DEXTER_IS_SERVER (sd->server));

  /* Increment server ref count so it won't finalize while we run */
  g_object_ref (sd->server);

  /* Increment server's client count */
  g_atomic_int_inc (&sd->server->client_count);

#ifdef G_ENABLE_DEBUG
  g_message ("service (count %d) running in thread 0x%08X", 
             g_atomic_int_get (&sd->server->client_count), (unsigned) g_thread_self());
#endif

  /* Initialize */
  recvbuf = (gchar *) g_slice_alloc0 (sd->server->settings->recv_buffer_size);
  framer = g_object_new (DEXTER_TYPE_MSGFRAMER, NULL);
  servercomm = g_object_new (DEXTER_TYPE_SERVERCOMM, NULL);
  session = dexter_network_session_new ();
  dexter_network_session_open (session, sd->connfd, &sd->addr, sd->addrlen);
  servercomm->server = sd->server;
  servercomm->session = session;
  servercomm->settings = sd->server->settings;

  do
  {
    /* check for bytes from client */
    if (dexter_network_session_pending (session))
    {
      /* read bytes from client and frame them into dexter messages */
      msgbytes=dexter_network_session_recv (session, recvbuf, 
                                            sd->server->settings->recv_buffer_size, 0, &local_error);
      if (local_error && 
          ((local_error->code==DEXTER_NETWORK_ERROR_EINTR) || 
           (local_error->code==DEXTER_NETWORK_ERROR_EAGAIN)))
      {
         /* non-fatal EINTR/EAGAIN */
         g_clear_error (&local_error);
         continue;
      }
      else if (local_error)
      {
        /* fatal */
        g_warning ("%s: error during network session (0x%08X) operation: %s",
                    G_STRFUNC, (unsigned int) session, local_error->message);
        g_clear_error (&local_error);
        break;
      }

      if (msgbytes>0)
      {
        if (dexter_msgframer_push_bytes (framer, recvbuf, msgbytes) < 0)
        {
          /* bytes from the channel are somehow problematic.  close connection. */
          break;
        }

        if (dexter_msgframer_frames_queued (framer) > 0)
        {
          /* process completed message frame from client */
          DexterMsgFrame *frame = dexter_msgframer_pop_frame (framer);
          dexter_servercomm_client_frame (servercomm, frame, &done, &local_error);
          if (local_error)
          {
            g_warning ("%s: error processing client frame in network session (0x%08X): %s", 
                       G_STRFUNC, (unsigned int) session, local_error->message);
            g_clear_error (&local_error);
            break;
          }
          dexter_msgframer_free_frame (framer, frame);
        }
      }
      else if (msgbytes==0)
      {
        /* client has closed connection */
        break;
      }
      else
      {
        /* error already handled above */
        break;
      }
    }
    else /* nothing to read from client */
    {
      /* send one data frame, if any are pending */
      dexter_servercomm_send_data_frame (servercomm, &local_error);
      if (local_error)
      {
        g_clear_error (&local_error);
        break;
      }
    }
  }
  while ((!done) && (dexter_server_is_runnable (sd->server)) && (dexter_get_state () == DEXTER_STATE_UP));
    
  dexter_network_session_close (session);
  dexter_network_session_free (session);
    
  /* decrement client count. */
  g_atomic_int_add (&sd->server->client_count, -1);

  g_object_unref (framer);
  g_object_unref (servercomm);
  g_object_unref (sd->server);
  g_slice_free1 (sd->server->settings->recv_buffer_size, recvbuf);
  g_slice_free (DexterServiceData, sd);

#ifdef G_ENABLE_DEBUG
  g_message ("service done in thread 0x%08X", (unsigned)g_thread_self ());
#endif
}


/**
 * dexter_server_new:
 * @settings: address of #DexterServerSettings or %NULL.
 *  If %NULL, the configuration file is used for all server settings.
 *
 * Creates a server using the supplied settings.
 *
 * %NULL may be indicated for @settings, in which case the configuration file is used for all 
 * server settings.  Internal defaults are used for items not found in the configuration file.
 *
 * No network calls are performed until the server is started with dexter_server_start().
 *
 * The server must be freed when done with dexter_server_free().
 *
 * Returns: newly allocated #DexterServer if successful. %NULL otherwise.
 **/
DexterServer *
dexter_server_new (const DexterServerSettings *settings)
{
  DexterServer *server;
    
#ifdef G_ENABLE_DEBUG
  g_message (G_STRFUNC);
#endif

  server = g_object_new (DEXTER_TYPE_SERVER, NULL);

  if (settings)
  {
    server->settings = dexter_server_settings_dup (settings);
  }
  else
  {
    server->settings = g_slice_new0 (DexterServerSettings);
  }

  dexter_server_settings_prepare (server->settings);

  /* we must not modify the server->settings after we start the server since 
   * multiple, concurrent readers will be inspecting them across threads. */

  return server;
}


/**
 * dexter_server_start:
 * @server: #DexterServer to start.
 * @error: return location for optional #GError.
 *
 * Starts the server created with dexter_server_new().
 *
 * The server is threaded, so control will pass immediately to the next statement in the 
 * program.  The server will accept connections immediately and continue to accept connections 
 * until dexter_server_stop() is called.
 *
 * Returns: #gboolean indicating success or failure.
 **/
gboolean
dexter_server_start (DexterServer  *server,
                     GError        **error)
{
  GError *local_error = NULL;
     
#ifdef G_ENABLE_DEBUG
  g_message (G_STRFUNC);
#endif
   
  g_return_val_if_fail (DEXTER_IS_SERVER (server), FALSE);
  g_return_val_if_fail (error == NULL || *error == NULL, FALSE);

#ifdef HAVE_TLS
  /* Possibly seed TLS parameters for first time */
  if ((server->settings->tls_support_flag == DEXTER_SERVER_TLS_SUPPORT_AVAILABLE) ||
      (server->settings->tls_support_flag == DEXTER_SERVER_TLS_SUPPORT_REQUIRED))
  {
    G_LOCK (_tls_seed_refresh_source);

    if (!_tls_seed_refresh_source)
    {
      /* no refresh source is running in the event loop, so we know the the TLS parameters
       * have never yet been seeded, i.e. this is the first TLS-enabled server to be started.
       * seed the parameters for the first time now. */
      dexter_tls_server_seed ();
    }

    G_UNLOCK (_tls_seed_refresh_source);
  }        
#else
  if (server->settings->tls_support_flag != DEXTER_SERVER_TLS_SUPPORT_NONE)
  {
    /* specifying tls_settings parameter on installations without TLS is not permitted */
    g_set_error (error,
                 DEXTER_SERVER_ERROR,
                 DEXTER_SERVER_ERROR_TLS_NOT_ENABLED,
                 "%s: TLS not enabled", G_STRFUNC);
    return FALSE;
  }
#endif /* HAVE_TLS */

  g_object_ref (server);

  /* check if server started */
  if (dexter_server_get_state (server) == DEXTER_SERVER_STATE_STARTED)
  {
    g_set_error (error,
                 DEXTER_SERVER_ERROR,
                 DEXTER_SERVER_ERROR_STATE,
                 "%s: server already started", G_STRFUNC);
    g_object_unref (server);
    return FALSE;
  }

  /* try to listen on this address */
  memset (&server->addr, 0, sizeof (struct sockaddr_storage));
  server->addrlen = 0;
  server->listenfd = dexter_network_listen (server->settings->host,
                                            server->settings->service,
                                            server->settings->socket_options,
                                            &server->addr,
                                            &server->addrlen,
                                            &local_error);
  if (local_error)
  {
    g_propagate_error (error, dexter_server_convert_network_error (&local_error));
    g_object_unref (server);
    return FALSE;
  }
     
  dexter_server_set_runnable (server, TRUE);

  /* start threaded server function.  we join this thread when we stop the server. */
  server->thread = g_thread_create ((GThreadFunc) _server_thread_func,
                                    server,
                                    TRUE, /* joinable */
                                    &local_error);
  if (local_error)
  {
    g_propagate_error (error, local_error);
    g_object_unref (server);
    return FALSE;
  }

  dexter_server_set_state (server, DEXTER_SERVER_STATE_STARTED);

#ifdef HAVE_TLS
  if ((server->settings->tls_support_flag == DEXTER_SERVER_TLS_SUPPORT_AVAILABLE) ||
      (server->settings->tls_support_flag == DEXTER_SERVER_TLS_SUPPORT_REQUIRED))
  {
    G_LOCK (_tls_seed_refresh_source);

    /* if not yet started, start a refresh source func to reseed TLS parameters daily */
    if (!_tls_seed_refresh_source)
    {
      _tls_seed_refresh_source = g_timeout_source_new (1000 /* millisec/sec */ *
                                                       60   /* sec/min */ *
                                                       60   /* min/hour */ *
                                                       24   /* hour/day */);  /* millisec/day */
      g_source_set_callback (_tls_seed_refresh_source,
                             (GSourceFunc) _tls_seed_refresh_func,
                             NULL,
                             NULL);
      g_source_set_can_recurse (_tls_seed_refresh_source, FALSE);
      dexter_event_loop_source_attach (__EVENTLOOP, _tls_seed_refresh_source);
    }

    G_UNLOCK (_tls_seed_refresh_source);

    /* increment count of TLS-enabled servers */
    g_atomic_int_add (&_tls_server_count, 1);
  }
#endif /* HAVE_TLS */

  g_object_unref (server);

  return TRUE;
}


/**
 * dexter_server_stop:
 * @server: #DexterServer to stop.
 * @error: return location for optional #GError.
 *
 * Stops a #DexterServer previously opened with dexter_channel_open().
 *
 * No new connections will be accepted and existing connections are shutdown.
 *
 * Returns: #gboolean indicating success or failure.
 **/
gboolean
dexter_server_stop (DexterServer  *server,
                    GError       **error)
{
#ifdef G_ENABLE_DEBUG
  g_message (G_STRFUNC);
#endif
      
  g_return_val_if_fail (DEXTER_IS_SERVER (server), FALSE);
  g_return_val_if_fail (error == NULL || *error == NULL, FALSE);
  
  g_object_ref (server);

  /* check if server not started */
  if (dexter_server_get_state (server) != DEXTER_SERVER_STATE_STARTED)
  {
    g_set_error (error,
                 DEXTER_SERVER_ERROR,
                 DEXTER_SERVER_ERROR_STATE,
                 "%s: server not started", G_STRFUNC);
    g_object_unref (server);
    return FALSE;
  }

  /* tell server thread to stop */
  dexter_server_set_runnable (server, FALSE);

  /* wait for server thread */
  if (server->thread)
  {
    g_thread_join (server->thread);
    server->thread = NULL;
  }
  dexter_server_set_state (server, DEXTER_SERVER_STATE_STOPPED);

#ifdef HAVE_TLS
  if (g_atomic_int_dec_and_test (&_tls_server_count))
  {
    /* no more TLS servers running, so stop the TLS parameter refresh source */
    G_LOCK (_tls_seed_refresh_source);
    g_source_destroy (_tls_seed_refresh_source);
    g_source_unref (_tls_seed_refresh_source);
    _tls_seed_refresh_source=NULL;
    G_UNLOCK (_tls_seed_refresh_source);
  }
#endif

  g_object_unref (server);

  return TRUE;
}


/**
 * dexter_server_free:
 * @server: #DexterServer to free.
 *
 * Frees a #DexterServer created with dexter_server_new().
 *
 * If the server is started, it will be stopped for you, although it is better form to use dexter_server_stop()
 * before freeing the server.   Future versions may be stricter on this point.
 **/
void
dexter_server_free (DexterServer *server)
{
#ifdef G_ENABLE_DEBUG
  g_message (G_STRFUNC);
#endif

  g_return_if_fail (server == NULL || DEXTER_IS_SERVER (server));

  if (G_UNLIKELY (!server))
    return;

  g_object_ref (server);

  if (dexter_server_get_state (server) == DEXTER_SERVER_STATE_STARTED)
    dexter_server_stop (server, NULL);

  g_object_unref (server);

  g_object_unref (server);
}

#define __DEXTER_SERVER_C__
#include <dexter-aliasdef.c>
