/* $Id: dexter-server.h 393 2012-03-06 23:36:44Z pkovacs $ */
/*
 * libdexter - a plugin-based, distributed sampling library.
 * Copyright (C) 2006-2012 Philip Kovacs <pkovacs@users.sourceforge.net>.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 2.1
 * of the License, or (at your option) any later version.

 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.

 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
 * 02110-1301  USA
 *
*/

#ifndef __DEXTER_SERVER_H__
#define __DEXTER_SERVER_H__

#include <glib.h>

G_BEGIN_DECLS

/**
 * DEXTER_SERVER_ERROR:
 *
 * The error domain of the libdexter server subsystem.
 **/
#define DEXTER_SERVER_ERROR   dexter_server_error_quark()
GQuark dexter_server_error_quark (void);

/**
 * DexterServerError:
 * @DEXTER_SERVER_ERROR_EACCES: address is protected, and the user is not the superuser.
 * @DEXTER_SERVER_ERROR_EADDRINUSE: local address in use.
 * @DEXTER_SERVER_ERROR_EAGAIN: temporary name server failure.
 * @DEXTER_SERVER_ERROR_STATE: operation not allowed in current server state.
 * @DEXTER_SERVER_ERROR_TLS_NOT_ENABLED: TLS not enabled on this installation.
 * @DEXTER_SERVER_ERROR_FAILED: other failure.
 *
 * Error codes for dexter_server operations.
 **/
typedef enum
{
  DEXTER_SERVER_ERROR_EACCES,
  DEXTER_SERVER_ERROR_EADDRINUSE,
  DEXTER_SERVER_ERROR_EAGAIN,
  DEXTER_SERVER_ERROR_STATE,
  DEXTER_SERVER_ERROR_TLS_NOT_ENABLED,
  DEXTER_SERVER_ERROR_FAILED
} DexterServerError;


/**
 * DexterServerState:
 * @DEXTER_SERVER_STATE_STARTED: server is started and available for use.
 * @DEXTER_SERVER_STATE_STOPPED: server is stopped.
 *
 * All possible server states.
 **/
typedef enum 
{
  DEXTER_SERVER_STATE_STARTED,
  DEXTER_SERVER_STATE_STOPPED,
} DexterServerState;


/**
 * DexterServer:
 *
 * Opaque type with no public members.
 **/
typedef struct _DexterServer DexterServer;


/**
 * DexterServerTLSSupportFlag:
 * @DEXTER_SERVER_TLS_SUPPORT_NONE: server does not support TLS.
 * @DEXTER_SERVER_TLS_SUPPORT_AVAILABLE: server supports TLS.
 * @DEXTER_SERVER_TLS_SUPPORT_REQUIRED: server requires TLS.
 *
 * Flag to set the level of TLS support offered by the server.
 *
 * For @DEXTER_SERVER_TLS_SUPPORT_AVAILABLE, the client may issue dexter_channel_starttls(),
 * but is not required to do so.
 *
 * For @DEXTER_SERVER_TLS_SUPPORT_REQUIRED, the client must issue dexter_channel_starttls()
 * after dexter_channel_greet().  The server will respond negatively to any channel operation,
 * except dexter_channel_part(), until TLS is started by the client.
 * 
 * Since: 0.2.0
 **/
typedef enum
{
  DEXTER_SERVER_TLS_SUPPORT_NONE      = 0,
  DEXTER_SERVER_TLS_SUPPORT_AVAILABLE = 1,
  DEXTER_SERVER_TLS_SUPPORT_REQUIRED  = 2,
} DexterServerTLSSupportFlag;


/**
 * DexterServerTLSClientCertFlag:
 * @DEXTER_SERVER_TLS_CLIENT_CERT_NONE: server will not request client certificate.
 * @DEXTER_SERVER_TLS_CLIENT_CERT_REQUESTED: server requests client certificate.
 * @DEXTER_SERVER_TLS_CLIENT_CERT_REQUIRED: server requires client certificate.
 *
 * Flag to set server requirements for client X.509 certificates.
 *
 * Since: 0.2.0
 **/
typedef enum
{
  DEXTER_SERVER_TLS_CLIENT_CERT_NONE      = 0,
  DEXTER_SERVER_TLS_CLIENT_CERT_REQUESTED = 1,
  DEXTER_SERVER_TLS_CLIENT_CERT_REQUIRED  = 2
} DexterServerTLSClientCertFlag;


typedef struct _DexterServerTLSSettings DexterServerTLSSettings;

/**
 * DexterServerTLSSettings:
 * @cert_file: full path to X.509 certificate file in PEM format.
 * @key_file: full path to X.509 certificate private key file in PEM format.
 * @trust_file: full path to X.509 trust file in PEM format.
 * @crl_file: full path to X.509 certificate revocation list file in PEM format.
 * @client_cert_flag: server requirements for client certificates.
 * @verify_hostname: verify that the hostname of client matches its X.509 certificate.
 *
 * TLS settings that the server will use when starting TLS for a client. They are copied 
 * from the #DexterServerSettings structure provided to dexter_server_new ().
 *
 * When this structure is seen within the context of a #DexterServerTLSStartCallback, 
 * you may change the elements as needed.  For any non-%NULL %gchar elements that you want
 * to change, you must first g_free() the element and then allocate a replacement string 
 * using a GLib string allocation api, e.g. g_strdup_printf().  The Library will free 
 * everything for you when the callback exits.
 *
 * Since: 0.2.0
 **/
struct _DexterServerTLSSettings
{
  /*< public >*/
  gchar   *cert_file;       /* ServerTLSCertFile */
  gchar   *key_file;        /* ServerTLSKeyFile */
  gchar   *trust_file;      /* ServerTLSTrustFile */
  gchar   *crl_file;        /* ServerTLSCrlFile */
  DexterServerTLSClientCertFlag client_cert_flag;  /* ServerTLSClientCertFlag */
  gboolean verify_hostname; /* ServerTLSVerifyHostname */
};


/**
 * DexterServerTLSStartCallback:
 * @server: #DexterServer on which TLS is starting.
 * @client_sa: socket address of client.
 * @client_salen: length of client socket address.
 * @tls_settings: #DexterServerTLSSettings of TLS settings.
 *
 * Signature of callback function used to adjust TLS settings, as needed for this client address, 
 * as TLS starts.
 *
 * If specificed in #DexterServerTLSSettings, this callback allows you to inspect the client 
 * socket address and adjust your server TLS settings as needed for that client connection.  
 * In this manner you can decide which certificate, trust and revocation files to use based 
 * on the client, and whether or not that particular client must present X.509 certificates.  
 *
 * Since: 0.2.0
 **/
typedef void (*DexterServerTLSStartCallback) (DexterServer *server,
                                              const struct sockaddr *client_sa,
                                              socklen_t client_salen,
                                              DexterServerTLSSettings *tls_settings);


/**
 * DexterServerSettingFlags:
 * @DEXTER_SERVER_SETTING_HOST: flag for host element of #DexterServerSettings.
 * @DEXTER_SERVER_SETTING_SERVICE: flag for service element of #DexterServerSettings.
 * @DEXTER_SERVER_SETTING_SOCKET_OPTIONS: flag for socket_options element of #DexterServerSettings.
 * @DEXTER_SERVER_SETTING_MAX_BACKLOG: flag for max_backlog element of #DexterServerSettings.
 * @DEXTER_SERVER_SETTING_MAX_CLIENTS: flag for max_clients element of #DexterServerSettings.
 * @DEXTER_SERVER_SETTING_RECV_BUFFER_SIZE: flag for recv_buffer_size element of #DexterServerSettings.
 * @DEXTER_SERVER_SETTING_TLS_SUPPORT_FLAG: flag for tls_support_flag element of #DexterServerSettings.
 * @DEXTER_SERVER_SETTING_TLS_CLIENT_CERT_FLAG: flag for tls_client_cert_flag element of #DexterServerSettings.
 * @DEXTER_SERVER_SETTING_TLS_START_CALLBACK: flag for tls_start_callback element of #DexterServerSettings.
 * @DEXTER_SERVER_SETTING_TLS_CERT_FILE: flag for tls_cert_file element of #DexterServerSettings.
 * @DEXTER_SERVER_SETTING_TLS_KEY_FILE: flag for tls_key_file element of #DexterServerSettings.
 * @DEXTER_SERVER_SETTING_TLS_TRUST_FILE: flag for tls_trust_file element of #DexterServerSettings.
 * @DEXTER_SERVER_SETTING_TLS_CRL_FILE: flag for tls_cert_file element of #DexterServerSettings.
 * @DEXTER_SERVER_SETTING_TLS_VERIFY_HOSTNAME: flag for tls_verify_hostname element of #DexterServerSettings.
 *
 * Flags to indicate which elements of #DexterServerSettings have data on them.
 *
 * Since 0.2.0
 **/
typedef enum
{
  DEXTER_SERVER_SETTING_HOST                 = 1 << 0,
  DEXTER_SERVER_SETTING_SERVICE              = 1 << 1,
  DEXTER_SERVER_SETTING_SOCKET_OPTIONS       = 1 << 2,
  DEXTER_SERVER_SETTING_MAX_BACKLOG          = 1 << 3,
  DEXTER_SERVER_SETTING_MAX_CLIENTS          = 1 << 4,
  DEXTER_SERVER_SETTING_RECV_BUFFER_SIZE     = 1 << 5,
  DEXTER_SERVER_SETTING_TLS_SUPPORT_FLAG     = 1 << 6,
  DEXTER_SERVER_SETTING_TLS_CLIENT_CERT_FLAG = 1 << 7,
  DEXTER_SERVER_SETTING_TLS_START_CALLBACK   = 1 << 8,
  DEXTER_SERVER_SETTING_TLS_CERT_FILE        = 1 << 9,
  DEXTER_SERVER_SETTING_TLS_KEY_FILE         = 1 << 10,
  DEXTER_SERVER_SETTING_TLS_TRUST_FILE       = 1 << 11,
  DEXTER_SERVER_SETTING_TLS_CRL_FILE         = 1 << 12,
  DEXTER_SERVER_SETTING_TLS_VERIFY_HOSTNAME  = 1 << 13
} DexterServerSettingFlags;


typedef struct _DexterServerSettings DexterServerSettings;
/**
 * DexterServerSettings:
 * @flags: Bitwise OR of flags from #DexterServerSettingFlags.
 * @host: host name or host ip.
 * @service: service name or port number.
 * @socket_options: tab or space-delimited list of socket options.
 * @max_backlog: maximum number of data messages in per-channel backlog queue.
 * @max_clients: maximum number of simulateous clients.
 * @recv_buffer_size: size of service socket receive buffer.
 * @tls_support_flag: indicates server TLS support.
 * @tls_client_cert_flag: indicates server requirement for client certificates.
 * @tls_start_callback: callback to override TLS settings as TLS starts.
 * @tls_cert_file: full path to X.509 certificate file in PEM format.
 * @tls_key_file: full path to X.509 certificate private key file in PEM format.
 * @tls_trust_file: full path to X.509 trust file in PEM format.
 * @tls_crl_file: full path to X.509 certificate revocation list file in PEM format.
 * @tls_verify_hostname: whether or not server requires that the client's DNS hostname match 
 *                       its X.509 certificate.
 *
 * Server settings.
 *
 * The @flags element indicates which elements have user-supplied data.
 *
 * Elements for which no flag is set, i.e. have no user-supplied data, will be set to a default: 
 * either from the configuration file, if available, or an internal default will be used.
 *
 * Since: 0.2.0
 **/
struct _DexterServerSettings
{
  /*< public >*/
  DexterServerSettingFlags      flags;
  gchar   *host;                /* ServerHost */
  gchar   *service;             /* ServerService */
  gchar   *socket_options;      /* ServerSocketOptions */
  gint     max_backlog;         /* ServerMaxBacklog */
  gint     max_clients;         /* ServerMaxClients */
  gint     recv_buffer_size;    /* ServerRecvBufferSize */
  DexterServerTLSSupportFlag    tls_support_flag;       /* ServerTLSSupportFlag */
  DexterServerTLSClientCertFlag tls_client_cert_flag;   /* ServerTLSClientCertFlag */
  DexterServerTLSStartCallback  tls_start_callback;     /* -- */
  gchar   *tls_cert_file;       /* ServerTLSCertFile */
  gchar   *tls_key_file;        /* ServerTLSKeyFile */
  gchar   *tls_trust_file;      /* ServerTLSTrustFile */
  gchar   *tls_crl_file;        /* ServerTLSCrlFile */
  gboolean tls_verify_hostname; /* ServerTLSVerifyHostname */
};


DexterServer   *dexter_server_new       (const DexterServerSettings *settings) G_GNUC_MALLOC;

gboolean        dexter_server_start     (DexterServer  *server,
                                         GError       **error);

gboolean        dexter_server_stop      (DexterServer  *server,
                                         GError       **error);
    
void            dexter_server_free      (DexterServer  *server);

gint            dexter_server_get_state (DexterServer  *server);

G_END_DECLS

#endif /* __DEXTER_SERVER_H__ */
