/* $Id: dexter-servercomm.c 393 2012-03-06 23:36:44Z pkovacs $ */
/*
 * libdexter - a plugin-based, distributed sampling library.
 * Copyright (C) 2006-2012 Philip Kovacs <pkovacs@users.sourceforge.net>.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 2.1
 * of the License, or (at your option) any later version.

 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.

 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
 * 02110-1301  USA
 *
*/

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <stdio.h>
#include <string.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <dexter-channel-object.h>
#include <dexter-messages.h>
#include <dexter-network.h>
#include <dexter-private.h>
#include <dexter-servercomm.h>
#include <dexter-service-broker.h>
#include <dexter-sampler-internal.h>
#include <dexter-timedsampler-internal.h>
#include <dexterplugin-isampler.h>
#include <dexterplugin-sampler.h>
#include <dexterplugin-serialize.h>
#include <dexter-alias.h>

#define REQUIRE_GREETING \
  if (!servercomm->greeting_exchanged) \
  { \
    dexter_servercomm_reply (servercomm, msg.cmd.command, msg.peer_id, ERROR_GREETING_NOT_EXCHANGED, NULL, \
                             error); \
    break; \
  }

#define FIND_SAMPLER \
  { \
    /* find the sampler using the peer id from the client */ \
    sampler = g_hash_table_lookup (servercomm->sampler_hash, &msg.peer_id); \
    if (!sampler) \
    { \
      dexter_servercomm_reply (servercomm, msg.cmd.command, msg.peer_id, ERROR_SAMPLER_PEER_NOT_FOUND, NULL, \
                               error); \
      break; \
    } \
  }


static void dexter_servercomm_class_init (DexterServerCommClass  *klass);
static void dexter_servercomm_init       (DexterServerComm       *servercomm);
static void dexter_servercomm_finalize   (GObject                *object);
static void dexter_servercomm_reply      (DexterServerComm       *servercomm,
                                          gint                    command,
                                          gint                    peer_id,
                                          gint                    status,
                                          GString                *payload,
                                          GError                **error);
#ifdef HAVE_TLS
static DexterServerTLSSettings
           *dexter_servercomm_tls_settings_new  (const gchar                   *cert_file,
                                                 const gchar                   *key_file,
                                                 const gchar                   *trust_file,
                                                 const gchar                   *crl_file,
                                                 DexterServerTLSClientCertFlag  client_cert_flag,
                                                 gboolean                       verify_hostname);
static void dexter_servercomm_tls_settings_free (DexterServerTLSSettings       *tls_settings);
static void dexter_servercomm_starttls          (DexterServerComm              *servercomm,
                                                 const DexterChannelMessage    *msg,
                                                 GError                       **error);
#endif

            /* function to destroy [timed]sampler when removed from sampler hash */
static void _sampler_hash_destroy_notify (GObject               *object);

            /* DexterSamplerDataCallback to create a DAT message for client */
static void _data_message_producer       (GObject               *sampler,
                                          gpointer               sampler_data);


G_DEFINE_TYPE (DexterServerComm, dexter_servercomm, G_TYPE_OBJECT);
  

static void
dexter_servercomm_class_init (DexterServerCommClass *klass)
{
  GObjectClass *gobject_class;

#ifdef G_ENABLE_DEBUG
  g_message (G_STRFUNC);
#endif

  gobject_class = G_OBJECT_CLASS (klass);
  gobject_class->finalize = dexter_servercomm_finalize;
}


static void
dexter_servercomm_init (DexterServerComm *servercomm)
{
#ifdef G_ENABLE_DEBUG
  g_message (G_STRFUNC);
#endif

  /* these members are set by service thread.
  servercomm->server = 
  servercomm->session = 
  servercomm->settings = 
  */
  servercomm->greeting_exchanged = FALSE;
  servercomm->sampler_service_group = 
    dexter_factory_get_services (__FACTORY, DEXTERPLUGIN_TYPE_ISAMPLER);
  servercomm->sampler_hash = 
    g_hash_table_new_full (g_int_hash, g_int_equal, NULL, (GDestroyNotify) _sampler_hash_destroy_notify);
  servercomm->dat_queue = g_async_queue_new ();
  servercomm->cur_backlog = 0;

#ifdef HAVE_TLS
  servercomm->tls_settings = NULL;
#endif
}


static void
dexter_servercomm_finalize (GObject *object)
{
  DexterServerComm *servercomm = DEXTER_SERVERCOMM (object);

#ifdef G_ENABLE_DEBUG
  g_message (G_STRFUNC);
#endif

  if (servercomm->sampler_hash)
    g_hash_table_destroy (servercomm->sampler_hash);

  dexterplugin_service_group_free (servercomm->sampler_service_group);

  /* purge any messages on async queue */
  g_async_queue_lock (servercomm->dat_queue);
  while (g_async_queue_length_unlocked (servercomm->dat_queue) > 0)
  {
    GString *message = g_async_queue_pop_unlocked (servercomm->dat_queue);
    g_string_free (message, TRUE);
  }
  g_async_queue_unlock (servercomm->dat_queue);
  /* release async queue */
  g_async_queue_unref (servercomm->dat_queue);

#ifdef HAVE_TLS
  dexter_servercomm_tls_settings_free (servercomm->tls_settings);
#endif

  /* chain up */
  G_OBJECT_CLASS(dexter_servercomm_parent_class)->finalize (object);
}


static void
dexter_servercomm_reply (DexterServerComm  *servercomm,
                         gint               command,
                         gint               peer_id,
                         gint               status,
                         GString           *payload,
                         GError           **error)
{
  GString *message;

#ifdef G_ENABLE_DEBUG
  g_message (G_STRFUNC);
#endif
  
  g_return_if_fail (DEXTER_IS_SERVERCOMM (servercomm));

  message = g_string_new (NULL);
  g_string_append_printf (message, RPY_FORMAT_STRING, DEXTER_PROTOCOL_MAJOR, DEXTER_PROTOCOL_MINOR,
                          command, peer_id, status);
  g_string_append_printf (message, "%d\n", payload ? payload->len : 0);
  if (payload)
  {
    g_string_append_len (message, payload->str, payload->len);
  }
  g_string_append_c (message, '\n');

  dexter_network_session_sendall (servercomm->session, message->str, message->len,
#ifdef HAVE_MSG_NOSIGNAL
                                  MSG_NOSIGNAL,
#else
                                  0,
#endif
                                  error);

  g_string_free (message, TRUE);
}


#ifdef HAVE_TLS
static DexterServerTLSSettings *
dexter_servercomm_tls_settings_new (const gchar *cert_file,
                                    const gchar *key_file,
                                    const gchar *trust_file,
                                    const gchar *crl_file,
                                    DexterServerTLSClientCertFlag client_cert_flag,
                                    gboolean verify_hostname)
{
  DexterServerTLSSettings *tls_settings;

  g_return_val_if_fail (client_cert_flag == DEXTER_SERVER_TLS_CLIENT_CERT_NONE ||
                        client_cert_flag == DEXTER_SERVER_TLS_CLIENT_CERT_REQUESTED ||
                        client_cert_flag == DEXTER_SERVER_TLS_CLIENT_CERT_REQUIRED, NULL);

  tls_settings = g_slice_new0 (DexterServerTLSSettings);
  tls_settings->cert_file = g_strdup (cert_file);
  tls_settings->key_file = g_strdup (key_file);
  tls_settings->trust_file = g_strdup (trust_file);
  tls_settings->crl_file = g_strdup (crl_file);
  tls_settings->client_cert_flag = client_cert_flag;
  tls_settings->verify_hostname = verify_hostname;

  return tls_settings;
}


static void
dexter_servercomm_tls_settings_free (DexterServerTLSSettings *tls_settings)
{
  if (G_UNLIKELY (!tls_settings))
    return;

  g_free (tls_settings->cert_file);
  g_free (tls_settings->key_file);
  g_free (tls_settings->trust_file);
  g_free (tls_settings->crl_file);
  g_slice_free (DexterServerTLSSettings, tls_settings);
}


static void 
dexter_servercomm_starttls (DexterServerComm *servercomm,
                            const DexterChannelMessage *msg,
                            GError **error)
{
  GError *local_error = NULL;

  g_return_if_fail (DEXTER_IS_SERVERCOMM (servercomm));
  g_return_if_fail (msg != NULL);
  g_return_if_fail (error == NULL || *error == NULL);

  g_object_ref (servercomm);

  /* check if server supports TLS */
  if (servercomm->settings->tls_support_flag == DEXTER_SERVER_TLS_SUPPORT_NONE)
  {
    dexter_servercomm_reply (servercomm, msg->cmd.command, msg->peer_id, ERROR_TLS_NOT_AVAILABLE, NULL, error);
    g_object_unref (servercomm);
    return;
  }

  /* check if TLS already running */
  if (G_UNLIKELY (dexter_network_session_tls_running (servercomm->session)))
  {
    dexter_servercomm_reply (servercomm, msg->cmd.command, msg->peer_id, ERROR_TLS_ALREADY_STARTED, NULL,error);
    g_object_unref (servercomm);
    return;
  }

  /* tell client to proceed with TLS handshake */
  dexter_servercomm_reply (servercomm, msg->cmd.command, msg->peer_id, STATUS_OK, NULL, error);

  /* if specified, issue callback to finalize TLS parameters */
  if (servercomm->settings->tls_start_callback)
  {
    /* copy TLS parameters */
    DexterServerTLSSettings *tls_settings = 
      dexter_servercomm_tls_settings_new (servercomm->settings->tls_cert_file, 
                                          servercomm->settings->tls_key_file, 
                                          servercomm->settings->tls_trust_file,
                                          servercomm->settings->tls_crl_file,
                                          servercomm->settings->tls_client_cert_flag,
                                          servercomm->settings->tls_verify_hostname);

    /* perform the callback */
    (*servercomm->settings->tls_start_callback) (servercomm->server,
                                                 dexter_network_session_peer_addr (servercomm->session),
                                                 dexter_network_session_peer_addrlen (servercomm->session),
                                                 tls_settings);

    /* start TLS handshake, with adjusted parameters */
    dexter_network_session_server_starttls (servercomm->session,
                                            tls_settings->cert_file,
                                            tls_settings->key_file,
                                            tls_settings->trust_file,
                                            tls_settings->crl_file,
                                            tls_settings->client_cert_flag,
                                            tls_settings->verify_hostname,
                                            &local_error);
    dexter_servercomm_tls_settings_free (tls_settings);
  }
  else
  {
    /* start TLS handshake, with default parameters */
    dexter_network_session_server_starttls (servercomm->session,
                                            servercomm->settings->tls_cert_file,
                                            servercomm->settings->tls_key_file,
                                            servercomm->settings->tls_trust_file,
                                            servercomm->settings->tls_crl_file,
                                            servercomm->settings->tls_client_cert_flag,
                                            servercomm->settings->tls_verify_hostname,
                                            &local_error);
  }

  if (local_error)
  {
    g_propagate_error (error, local_error);
    g_object_unref (servercomm);
    return;
  }
                                
  g_object_unref (servercomm);
}
#endif


static void
_sampler_hash_destroy_notify (GObject *object)
{
  if (DEXTER_IS_SAMPLER (object))
  {
    dexter_sampler_free (DEXTER_SAMPLER (object), NULL);
  }
  else if (DEXTER_IS_TIMEDSAMPLER (object))
  {
    dexter_timedsampler_free (DEXTER_TIMEDSAMPLER (object), NULL);
  }
  else
  {
    /* programming error */
    g_error (G_STRLOC " unexpected type removed from sampler hash");
  }
}


static void            
_data_message_producer (GObject  *sampler,
                        gpointer  sampler_data)
{
  DexterChannelObject *channel_object;
  DexterServerComm    *servercomm;
    
  g_return_if_fail (G_IS_OBJECT (sampler));
  g_return_if_fail (sampler_data != NULL);

  channel_object = DEXTER_CHANNEL_OBJECT (sampler);

  if (G_UNLIKELY (!DEXTER_IS_SERVERCOMM (channel_object->sc)))
    return;

  servercomm = DEXTER_SERVERCOMM (channel_object->sc);

  /* if the backlog queue is full, do not produce a new data message */
  g_assert (servercomm->settings->max_backlog > 0);
  if (g_atomic_int_get (&servercomm->cur_backlog) >= servercomm->settings->max_backlog)
    return;

  {
    GString *message;
    GString *payload;
    GError  *local_error = NULL;
    gint     status = STATUS_OK;

    message = g_string_new (NULL);
    payload = g_string_new (NULL);
    
    if (DEXTER_IS_SAMPLER (sampler))
    {
      dexter_sampler_data_serialize (DEXTER_SAMPLER (sampler), payload, sampler_data, &local_error);
    }
    else if (DEXTER_IS_TIMEDSAMPLER (sampler))
    {
      dexter_timedsampler_data_serialize (DEXTER_TIMEDSAMPLER (sampler), payload, sampler_data, &local_error);
    }
    else
    {
      /* programming error */
      g_error (G_STRLOC " unexpected object type at data serialization");
    }
    if (local_error)
    {
      payload = g_string_truncate (payload, 0);
      g_string_printf (payload, "%s", local_error->message);
      status = ERROR_SAMPLER_DATA;
      g_clear_error (&local_error);
    }

    /* TODO: break up DAT message into sequence of DAT frames if payload exceeds
     * config-specified maximum size, i.e. MaxDataChunkSize. */
    
    g_string_printf (message, DAT_FORMAT_STRING,
                     DEXTER_PROTOCOL_MAJOR, DEXTER_PROTOCOL_MINOR, channel_object->id, 
                     DAT_CHUNK_FIRST /*DAT_CHUNK_CONTINUATION*/ | DAT_CHUNK_LAST, status);
    g_string_append_printf (message, "%d\n%s\n", payload->len, payload->str);
    g_string_free (payload, TRUE);
    
    /* finally push new message onto the delivery queue */
    g_async_queue_push (channel_object->mbox, message);

    /* and increment the backlog queue length */
    g_atomic_int_add (&servercomm->cur_backlog, 1);
  }
}


void
dexter_servercomm_client_frame (DexterServerComm      *servercomm,
                                const DexterMsgFrame  *frame,
                                gboolean              *done,
                                GError               **error)
{
  DexterChannelMessage msg;

#ifdef G_ENABLE_DEBUG
  g_message ("%s: %s", G_STRFUNC, frame->message);
#endif

  g_return_if_fail (DEXTER_IS_SERVERCOMM (servercomm));
  g_return_if_fail (frame != NULL);

  /* All frames coming from the client must be type CMD.  Parse the message. */

  /* is message formatted correctly? */
  if (sscanf (frame->message, CMD_FORMAT_STRING, &msg.protocol_major, &msg.protocol_minor, 
              &msg.cmd.command, &msg.peer_id, &msg.cmd.parameter) != 5)
  {
    dexter_servercomm_reply (servercomm, COMMAND_UNKNOWN, 0, ERROR_MESSAGE_BAD_FORMAT, NULL, error);
    return;
  }

  /* is the protocol revision of the message compatible with this server? */
  if ((msg.protocol_major > DEXTER_PROTOCOL_MAJOR) || 
      ((msg.protocol_major == DEXTER_PROTOCOL_MAJOR) && (msg.protocol_minor > DEXTER_PROTOCOL_MINOR)))
  {
    dexter_servercomm_reply (servercomm, msg.cmd.command, msg.peer_id, ERROR_INCOMPATIBLE_PROTOCOL, NULL, 
                             error);
    return;
  }

  switch (msg.cmd.command)
  {
    case COMMAND_HELLO:
    {
      /* has client already greeted? */
      if (servercomm->greeting_exchanged)
      {
        dexter_servercomm_reply (servercomm, msg.cmd.command, msg.peer_id, 
                                 ERROR_GREETING_ALREADY_EXCHANGED, NULL, error);
        break;
      }
      servercomm->greeting_exchanged = TRUE;
      dexter_servercomm_reply (servercomm, msg.cmd.command, msg.peer_id, STATUS_OK, NULL, error);
      break;
    }
    case COMMAND_NOOP:
    {
      REQUIRE_GREETING;

      dexter_servercomm_reply (servercomm, msg.cmd.command, msg.peer_id, STATUS_OK, NULL, error);
      break;
    }
    case COMMAND_STARTTLS:
    {
      REQUIRE_GREETING;
#ifdef HAVE_TLS
      dexter_servercomm_starttls (servercomm, &msg, error);
#else
      dexter_servercomm_reply (servercomm, msg.cmd.command, msg.peer_id, ERROR_TLS_NOT_AVAILABLE, NULL, error);
#endif
      break;
    }
    case COMMAND_GET_SERVICES:
    {
      GType                    iface_type;
      GString                  *payload;
      DexterPluginServiceGroup *service_group;
      GError                   *local_error = NULL;

      REQUIRE_GREETING;

      /* verify that the parameter is valid */
      switch (msg.cmd.parameter)
      {
        case DEXTER_SERVICE_SAMPLER:
          iface_type = DEXTERPLUGIN_TYPE_ISAMPLER;
          break;
        default:
          iface_type = G_TYPE_INVALID;
          break;
      }
      if (iface_type == G_TYPE_INVALID)
      {
        dexter_servercomm_reply (servercomm, msg.cmd.command, msg.peer_id, 
                                 ERROR_COMMAND_INVALID_PARAMETER, NULL, error);
        break;
      }
  
      payload = g_string_new (NULL);
      service_group = dexter_factory_get_services (__FACTORY, iface_type);

      dexterplugin_serialize_service_group_serialize (service_group, payload, &local_error);
      if (local_error)
      {
        g_string_printf (payload, "%s", local_error->message);
        dexter_servercomm_reply (servercomm, msg.cmd.command, msg.peer_id, ERROR_COMMAND_FAILED, payload,
                                 error);
        g_string_free (payload, TRUE);
        g_clear_error (&local_error);
        break;
      }
      dexter_servercomm_reply (servercomm, msg.cmd.command, msg.peer_id, STATUS_OK, payload, error);
      dexterplugin_service_group_free (service_group);
      g_string_free (payload, TRUE);
      break;
    }
    case COMMAND_SAMPLER_NEW:
    {
      DexterPluginService       *service = NULL;
      DexterSampler             *sampler;
      DexterChannelObject       *channel_object;
      DexterSamplerDataCallback *callbacks;
      
      REQUIRE_GREETING;

      /* message payload must contain the uuid of the desired sampler */
      if ((!frame->payload) || (frame->payload->len <= 0))
      {
        dexter_servercomm_reply (servercomm, msg.cmd.command, msg.peer_id, ERROR_NO_PAYLOAD_DATA, NULL, error);
        break;
      }
      
      /* use the uuid to find the service */
      if (!dexterplugin_service_group_find_uuid (&service, servercomm->sampler_service_group, 
                                                 frame->payload->str))
      {
        dexter_servercomm_reply (servercomm, msg.cmd.command, msg.peer_id, ERROR_SERVICE_UNAVAILABLE, NULL,
                                 error);
        break;
      }

      /* create sampler with data message producer as callback */
      callbacks = (DexterSamplerDataCallback *) g_slice_alloc0 (2 * sizeof (DexterSamplerDataCallback));
      callbacks[0] = (DexterSamplerDataCallback) _data_message_producer;
      sampler = dexter_sampler_new_actual (service, callbacks, NULL, NULL);
     
      /* base channel_object is used to queue up outbound DAT messages for client */
      channel_object = DEXTER_CHANNEL_OBJECT (sampler);
      channel_object->id = msg.peer_id;
      channel_object->mbox = g_async_queue_ref (servercomm->dat_queue);
      channel_object->sc = G_OBJECT (servercomm);
      g_slice_free1 (2 * sizeof (DexterSamplerDataCallback), callbacks);
      
      /* insert sampler into hash */
      g_hash_table_insert (servercomm->sampler_hash, &msg.peer_id, sampler);
      
      dexter_servercomm_reply (servercomm, msg.cmd.command, msg.peer_id, STATUS_OK, NULL, error);
      break;
    }
    case COMMAND_TIMEDSAMPLER_NEW:
    {
      DexterPluginService         *service = NULL;
      DexterTimedSampler          *sampler;
      DexterChannelObject         *channel_object;
      DexterSamplerDataCallback   *callbacks;
      gchar      uuid[37];       
      glong      usecs_interval;
      int        rc;

      REQUIRE_GREETING;

      /* message payload must contain the uuid and interval time of the desired sampler */
      if ((!frame->payload) || (frame->payload->len <= 0))
      {
        dexter_servercomm_reply (servercomm, msg.cmd.command, msg.peer_id, ERROR_NO_PAYLOAD_DATA, NULL, error);
        break;
      }

      /* scan for uuid and timing interval */
      memset (uuid, 0, sizeof(uuid));
      rc = sscanf (frame->payload->str, "%36c %ld", uuid, &usecs_interval);
      if (rc != 2)
      {
        dexter_servercomm_reply (servercomm, msg.cmd.command, msg.peer_id, ERROR_PAYLOAD_DATA_INVALID, NULL,
                                 error);
        break;
      }
      /*g_print ("uuid=[%s] usecs_interval=%ld\n", uuid, usecs_interval);*/

      /* use the uuid to find the service */
      if (!dexterplugin_service_group_find_uuid (&service, servercomm->sampler_service_group, uuid))
      {
        dexter_servercomm_reply (servercomm, msg.cmd.command, msg.peer_id, ERROR_SERVICE_UNAVAILABLE, NULL,
                                 error);
        break;
      }

      /* create sampler with data message producer as callback */
      callbacks = (DexterSamplerDataCallback *) g_slice_alloc0 (2 * sizeof (DexterSamplerDataCallback));
      callbacks[0] = (DexterSamplerDataCallback) _data_message_producer;
      sampler = dexter_timedsampler_new_actual (service, usecs_interval, callbacks, NULL, NULL);

      /* base channel_object is used to queue up outbound DAT messages for client */
      channel_object = DEXTER_CHANNEL_OBJECT (sampler);
      channel_object->id = msg.peer_id;
      channel_object->mbox = g_async_queue_ref (servercomm->dat_queue);
      channel_object->sc = G_OBJECT (servercomm);
      g_slice_free1 (2 * sizeof (DexterSamplerDataCallback), callbacks);

      /* insert sampler into hash */
      g_hash_table_insert (servercomm->sampler_hash, &msg.peer_id, sampler);

      dexter_servercomm_reply (servercomm, msg.cmd.command, msg.peer_id, STATUS_OK, NULL, error);
      break;
    }
    case COMMAND_SAMPLER_INITIALIZE:
    {
      DexterSampler *sampler;
      gpointer       input;
      GError        *local_error = NULL;

      REQUIRE_GREETING;

      FIND_SAMPLER;

      /* deserialize sampler input from payload, if present */
      if (frame->payload)
      {
        input = dexter_sampler_input_deserialize (sampler, frame->payload, &local_error);
        if (local_error)
        {
          GString *payload = g_string_new (local_error->message);
          dexter_servercomm_reply (servercomm, msg.cmd.command, msg.peer_id, ERROR_COMMAND_FAILED, payload, 
                                   error);
          g_string_free (payload, TRUE);
          dexter_sampler_input_free (sampler, input);
          g_clear_error (&local_error);
          break;
        }
      }

      /* initialize the sampler */
      dexter_sampler_initialize (sampler, (const gint *)&msg.cmd.parameter, input, &local_error);
      if (local_error)
      {
        GString *payload = g_string_new (local_error->message);
        dexter_servercomm_reply (servercomm, msg.cmd.command, msg.peer_id, ERROR_COMMAND_FAILED, payload,
                                 error);
        g_string_free (payload, TRUE);
        g_clear_error (&local_error);
        break;
      }
      /* free the input */
      dexter_sampler_input_free (sampler, input);

      dexter_servercomm_reply (servercomm, msg.cmd.command, msg.peer_id, STATUS_OK, NULL, error);
      break;
    }
    case COMMAND_TIMEDSAMPLER_INITIALIZE:
    {
      DexterTimedSampler *sampler;
      gpointer            input;
      GError             *local_error = NULL;

      REQUIRE_GREETING;

      FIND_SAMPLER;

      /* deserialize sampler input from payload, if present */
      if (frame->payload)
      {
        input = dexter_timedsampler_input_deserialize (sampler, frame->payload, &local_error);
        if (local_error)
        {
          GString *payload = g_string_new (local_error->message);
          dexter_servercomm_reply (servercomm, msg.cmd.command, msg.peer_id, ERROR_COMMAND_FAILED, payload,
                                   error);
          g_string_free (payload, TRUE);
          dexter_timedsampler_input_free (sampler, input);
          g_clear_error (&local_error);
          break;
        }
      }

      /* initialize the sampler */
      dexter_timedsampler_initialize (sampler, (const gint *)&msg.cmd.parameter, input, &local_error);
      if (local_error)
      {
        GString *payload = g_string_new (local_error->message);
        dexter_servercomm_reply (servercomm, msg.cmd.command, msg.peer_id, ERROR_COMMAND_FAILED, payload,
                                 error);
        g_string_free (payload, TRUE);
        g_clear_error (&local_error);
        break;
      }
      /* free the input */
      dexter_timedsampler_input_free (sampler, input);

      dexter_servercomm_reply (servercomm, msg.cmd.command, msg.peer_id, STATUS_OK, NULL, error);
      break;
    }
    case COMMAND_SAMPLER_SAMPLE:
    {
      DexterSampler *sampler;
      GError        *local_error = NULL;
      
      REQUIRE_GREETING;

      FIND_SAMPLER;

      dexter_sampler_sample (sampler, &local_error);
      if (local_error)
      {
        GString *payload = g_string_new (local_error->message);
        dexter_servercomm_reply (servercomm, msg.cmd.command, msg.peer_id, ERROR_COMMAND_FAILED, payload,
                                 error);
        g_string_free (payload, TRUE);
        g_clear_error (&local_error);
        break;
      }

      dexter_servercomm_reply (servercomm, msg.cmd.command, msg.peer_id, STATUS_OK, NULL, error);
      break;        
    }
    case COMMAND_SAMPLER_FREE:
    case COMMAND_TIMEDSAMPLER_FREE:
    {
      DexterSampler *sampler;

      REQUIRE_GREETING;

      FIND_SAMPLER;

      /* removing sampler from hash also frees it since GDestroyNotify was specified
       * with g_hash_table_new_full() */
      g_hash_table_remove (servercomm->sampler_hash, &msg.peer_id);

      /* TODO: remove any pending DAT messages for this sampler. */

      dexter_servercomm_reply (servercomm, msg.cmd.command, msg.peer_id, STATUS_OK, NULL, error);
      break;
    }
    case COMMAND_TIMEDSAMPLER_START:
    {
      DexterSampler *sampler;
      GError        *local_error = NULL;

      REQUIRE_GREETING;

      FIND_SAMPLER;

      dexter_timedsampler_start (DEXTER_TIMEDSAMPLER (sampler), &local_error);
      if (local_error)
      {
        GString *payload = g_string_new (local_error->message);
        dexter_servercomm_reply (servercomm, msg.cmd.command, msg.peer_id, ERROR_COMMAND_FAILED, payload,
                                 error);
        g_string_free (payload, TRUE);
        g_clear_error (&local_error);
        break;
      }

      dexter_servercomm_reply (servercomm, msg.cmd.command, msg.peer_id, STATUS_OK, NULL, error);
      break;
    }
    case COMMAND_TIMEDSAMPLER_STOP:
    {
      DexterSampler *sampler;
      GError        *local_error = NULL;

      REQUIRE_GREETING;

      FIND_SAMPLER;

      dexter_timedsampler_stop (DEXTER_TIMEDSAMPLER (sampler), &local_error);
      if (local_error)
      {
        GString *payload = g_string_new (local_error->message);
        dexter_servercomm_reply (servercomm, msg.cmd.command, msg.peer_id, ERROR_COMMAND_FAILED, payload,
                                 error);
        g_string_free (payload, TRUE);
        g_clear_error (&local_error);
        break;
      }

      dexter_servercomm_reply (servercomm, msg.cmd.command, msg.peer_id, STATUS_OK, NULL, error);
      break;
    }
    case COMMAND_GOODBYE:
    {
      REQUIRE_GREETING;

      dexter_servercomm_reply (servercomm, msg.cmd.command, msg.peer_id, STATUS_OK, NULL, error);
      *done=TRUE;
      break;
    }
    default:
      dexter_servercomm_reply (servercomm, msg.cmd.command, msg.peer_id, ERROR_COMMAND_UNKNOWN, NULL, error);
      break;
    }
}


void
dexter_servercomm_send_data_frame (DexterServerComm  *servercomm,
                                   GError           **error)
{
  GString *message;
    
  g_return_if_fail (DEXTER_IS_SERVERCOMM (servercomm));
  g_return_if_fail (error == NULL || *error == NULL);

  /* grab oldest message */
  message = g_async_queue_try_pop (servercomm->dat_queue);
  if (message)
  {
    dexter_network_session_sendall (servercomm->session, message->str, message->len,
#ifdef HAVE_MSG_NOSIGNAL
                                    MSG_NOSIGNAL,
#else
                                    0,
#endif
                                    error);
    g_string_free (message, TRUE);
    message=NULL;

    /* decrement backlog queue length */
    g_atomic_int_add (&servercomm->cur_backlog, -1);
  }
}

#define __DEXTER_SERVERCOMM_C__
#include <dexter-aliasdef.c>
