/* $Id: dexter-service-broker.c 393 2012-03-06 23:36:44Z pkovacs $ */
/*
 * libdexter - a plugin-based, distributed sampling library.
 * Copyright (C) 2006-2012 Philip Kovacs <pkovacs@users.sourceforge.net>.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 2.1
 * of the License, or (at your option) any later version.

 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.

 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
 * 02110-1301  USA
 *
*/

/**
 * SECTION:dexter-service-broker
 * @title: Service Brokers
 * @short_description: Using service brokers to request available services
 * @see_also:
 *
 * Service brokers are Libdexter objects used to fetch available service 
 * information and return it in the form of a collection known as a service group.  
 * If a channel is specified during dexter_service_broker_new(), the service broker 
 * retrieves the service group from the remote host over the channel, otherwise it 
 * operates locally (in-process).  
 *
 * Service brokers must be freed with dexter_service_broker_free() when done.
 **/

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <dexter-service-broker-internal.h>
#include <dexter-channel-object.h>
#include <dexter-channel-internal.h>
#include <dexter-messages.h>
#include <dexter-private.h>
#include <dexterplugin-isampler.h>
#include <dexterplugin-serialize.h>
#include <dexter-alias.h>

static void  dexter_service_broker_class_init      (DexterServiceBrokerClass *klass);
static void  dexter_service_broker_init            (DexterServiceBroker      *broker);
static void  dexter_service_broker_finalize        (GObject                  *object);

static void  dexter_service_broker_process_message (DexterChannelObject      *channel_object,
                                                    DexterChannelMessage     *message,
                                                    GError                  **error);
static void  dexter_service_broker_channel_event   (DexterChannelObject      *channel_object,
                                                    gint                      event);

struct _DexterServiceBroker 
{ 
  DexterChannelObject       parent;

  /* temp var for process_message */
  DexterPluginServiceGroup *service_group;
};

struct _DexterServiceBrokerClass 
{
  DexterChannelObjectClass  parent_class;
};


GQuark
dexter_service_broker_error_quark (void)
{
  static GQuark error_quark = 0;

  if (error_quark == 0)
    error_quark = g_quark_from_static_string ("dexter-service-broker-error-quark");

  return error_quark;
}


G_DEFINE_TYPE (DexterServiceBroker, dexter_service_broker, DEXTER_TYPE_CHANNEL_OBJECT);


static void
dexter_service_broker_class_init (DexterServiceBrokerClass *klass)
{
  GObjectClass             *gobject_class;
  DexterChannelObjectClass *channel_object_class;
    
#ifdef G_ENABLE_DEBUG
  g_message (G_STRFUNC);
#endif

  gobject_class = G_OBJECT_CLASS (klass);
  gobject_class->finalize = dexter_service_broker_finalize;

  channel_object_class = DEXTER_CHANNEL_OBJECT_CLASS (klass);
  channel_object_class->process_message = dexter_service_broker_process_message;
  channel_object_class->channel_event = dexter_service_broker_channel_event;
}


static void
dexter_service_broker_init (DexterServiceBroker *broker)
{
#ifdef G_ENABLE_DEBUG
  g_message (G_STRFUNC);
#endif

  broker->service_group = NULL;
}


static void
dexter_service_broker_finalize (GObject *object)
{
  DexterServiceBroker *broker = DEXTER_SERVICE_BROKER (object);

#ifdef G_ENABLE_DEBUG
  g_message (G_STRFUNC);
#endif

  dexterplugin_service_group_free (broker->service_group);
      
  /* chain up */
  G_OBJECT_CLASS (dexter_service_broker_parent_class)->finalize (object);
}


static void
dexter_service_broker_process_message (DexterChannelObject   *channel_object,
                                       DexterChannelMessage  *message,
                                       GError               **error)
{
  DexterServiceBroker *broker = NULL;

#ifdef G_ENABLE_DEBUG
  g_message (G_STRFUNC);
#endif

  g_return_if_fail (DEXTER_IS_CHANNEL_OBJECT (channel_object));
  g_return_if_fail (message != NULL);
  g_return_if_fail (error == NULL || *error == NULL);

  g_object_ref (channel_object);
  broker = DEXTER_SERVICE_BROKER (channel_object);

  if (message->msg_type == DEXTER_MESSAGE_REPLY)
  {
    if (message->rpy.status != STATUS_OK)
    {
      g_set_error (error,
                   DEXTER_SERVICE_BROKER_ERROR,
                   DEXTER_SERVICE_BROKER_ERROR_FAILED,
                   "%s: server replied with error: command (%d) status (%d) details (%s)",
                   G_STRFUNC, message->rpy.command, message->rpy.status,
                    message->payload ? message->payload->str : NULL);
    }
    else
    {
      switch (message->rpy.command)
      {
        case COMMAND_GET_SERVICES:
        {
          /* deserialize service group */
          broker->service_group = dexterplugin_serialize_service_group_deserialize (message->payload, error);
          break;
        }
        default:
        {
          /* programming error */
          g_error (G_STRLOC " reply received from server for unimplemented command (%d)", message->rpy.command);
        }
      }
    }
  }
  else
  {
    /* programming error */
    g_error (G_STRLOC " invalid message type (%d) received from server", message->msg_type);
  }

  /* chain up */
  DEXTER_CHANNEL_OBJECT_CLASS (dexter_service_broker_parent_class) 
    -> process_message (channel_object, message, error);

  g_object_unref (channel_object);
}


static void
dexter_service_broker_channel_event (DexterChannelObject *channel_object,
                                     gint                 event)
{
  DexterServiceBroker *broker;
#ifdef G_ENABLE_DEBUG
  g_message (G_STRFUNC);
#endif

  g_return_if_fail (DEXTER_IS_CHANNEL_OBJECT (channel_object));

  g_object_ref (channel_object);
  broker = DEXTER_SERVICE_BROKER (channel_object);

  switch (event)
  {
    case DEXTER_CHANNEL_EVENT_NOCONN:
    case DEXTER_CHANNEL_EVENT_CLOSED:
    case DEXTER_CHANNEL_EVENT_OPEN:
      break;
    default:
    {
      /* programming error */
      g_error (G_STRLOC " unrecognized channel event (%d)", event);
    }
  }

  /* chain up */
  DEXTER_CHANNEL_OBJECT_CLASS (dexter_service_broker_parent_class) -> channel_event (channel_object, event);

  g_object_unref (channel_object);
}


/**
 * dexter_service_broker_new:
 * @channel: #DexterChannel over which to create the service broker, or %NULL for local operation.
 *
 * Creates a #DexterServiceBroker over the specified #DexterChannel, if indicated.
 *
 * If no #DexterChannel is indicated, the service broker operates locally (in-process).
 *
 * The service broker must be freed when done with dexter_service_broker_free().
 *
 * Returns: pointer to newly allocated #DexterServiceBroker.
 **/
DexterServiceBroker *
dexter_service_broker_new (DexterChannel *channel)
{
  DexterServiceBroker *broker = NULL;

#ifdef G_ENABLE_DEBUG
  g_message (G_STRFUNC);
#endif

  g_return_val_if_fail (channel == NULL || DEXTER_IS_CHANNEL (channel), NULL);
    
  /* create the object */
  broker = g_object_new (DEXTER_TYPE_SERVICE_BROKER, NULL);

  if (channel)
  {
    /* register broker as an object capable of receiving server messages */
    dexter_channel_register_channel_object (channel, DEXTER_CHANNEL_OBJECT (broker), DEXTER_MESSAGE_REPLY);
  }

  return broker;
}


/**
 * dexter_service_broker_get_services:
 * @service_broker: a #DexterServiceBroker.
 * @service_type: boolean OR of service type flags from #DexterServiceType.
 * @error: return location for optional #GError.
 *
 * Requests and delivers a service group, either for the local process, or for the remote host 
 * if created on a #DexterChannel.
 *
 * The service group must be freed with dexterplugin_service_group_free() when done.
 *
 * Returns: #DexterPluginServiceGroup service group.
 **/
DexterPluginServiceGroup *
dexter_service_broker_get_services (DexterServiceBroker  *service_broker,
                                    gint                  service_type,
                                    GError              **error)
{
  DexterChannelObject      *channel_object = NULL;
  DexterPluginServiceGroup *service_group = NULL;
  GError                   *local_error  = NULL;

#ifdef G_ENABLE_DEBUG
  g_message (G_STRFUNC);
#endif

  g_return_val_if_fail (DEXTER_IS_SERVICE_BROKER (service_broker), NULL);
  g_return_val_if_fail (service_type > 0, NULL);
  g_return_val_if_fail (error == NULL || *error == NULL, NULL);

  g_object_ref (service_broker);
  channel_object = DEXTER_CHANNEL_OBJECT (service_broker);

  if ((channel_object->channel) && /* remote operation */
      (dexter_channel_get_state (channel_object->channel) == DEXTER_CHANNEL_STATE_OPEN))
  {
    switch (service_type)
    {
      case DEXTER_SERVICE_SAMPLER:
        break;
      default:
      {
        /* programming error */
        g_error (G_STRLOC " invalid service type indicated");
      }
    }

    dexter_channel_object_send_wait_process (channel_object, 
                                             COMMAND_GET_SERVICES, service_type, NULL, &local_error);
    if (local_error)
    {
      g_propagate_error (error, local_error);
      g_object_unref (service_broker);
      return NULL;
    }
    else
    {
      /* set return data */
      service_group = service_broker->service_group; 
    }
  }
  else if (channel_object->channel)
  {
    g_set_error (error,
                 DEXTER_CHANNEL_ERROR,
                 DEXTER_CHANNEL_ERROR_STATE,
                 "%s: operation requires an open channel", G_STRFUNC);
    g_object_unref (service_broker);
    return NULL;
  }
  else /* local operation */
  {
    GType iface_type = G_TYPE_INVALID;
    switch (service_type)
    {
      case DEXTER_SERVICE_SAMPLER:
        iface_type = DEXTERPLUGIN_TYPE_ISAMPLER;
        break;
      default:
      {
        /* programming error */
        g_error (G_STRLOC " invalid service type indicated");
      }
    }
    service_group = dexter_factory_get_services (__FACTORY, iface_type);
  }
    
  /* reset temps */
  service_broker->service_group = NULL;      /* do not hold pointer to returned data */
          
  g_object_unref (service_broker);

  return service_group;
}

/**
 * dexter_service_broker_free:
 * @service_broker: a #DexterServiceBroker.
 *
 * Frees the #DexterServiceBroker previously created with dexter_service_broker_new().
 **/
void
dexter_service_broker_free (DexterServiceBroker *broker)
{
  DexterChannelObject *channel_object = NULL;

#ifdef G_ENABLE_DEBUG
  g_message (G_STRFUNC);
#endif
    
  g_return_if_fail (broker == NULL || DEXTER_IS_SERVICE_BROKER (broker));

  if (G_UNLIKELY (!broker))
    return;

  g_object_ref (broker);
  channel_object = DEXTER_CHANNEL_OBJECT (broker);
    
  if (channel_object->channel)
  {
    /* deregister broker as an object capable of receiving server messages */
    dexter_channel_deregister_channel_object (channel_object->channel, channel_object);
  }
  
  /* release this api's ref */
  g_object_unref (broker);
  /* release object */
  g_object_unref (broker);
}

#define __DEXTER_SERVICE_BROKER_C__
#include <dexter-aliasdef.c>
