/* $Id: dexter-threadpool.c 393 2012-03-06 23:36:44Z pkovacs $ */
/*
 * libdexter - a plugin-based, distributed sampling library.
 * Copyright (C) 2006-2012 Philip Kovacs <pkovacs@users.sourceforge.net>.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 2.1
 * of the License, or (at your option) any later version.

 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.

 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
 * 02110-1301  USA
 *
*/

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <dexter.h>
#include <dexter-private.h>
#include <dexter-alias.h>

typedef struct _DexterThreadPoolTask 
{
  GThreadFunc func;
  gpointer    data;

} DexterThreadPoolTask;


static void   dexter_threadpool_class_init       (DexterThreadPoolClass            *klass);
static void   dexter_threadpool_finalize         (GObject                          *object);
static void   dexter_threadpool_ithreadpool_init (DexterPluginIThreadPoolInterface *ithreadpool);

static void   _threadpool_dispatcher             (gpointer  data,
                                                  gpointer  user_data);


G_DEFINE_TYPE_WITH_CODE (DexterThreadPool, dexter_threadpool, G_TYPE_OBJECT,
                         G_IMPLEMENT_INTERFACE (DEXTERPLUGIN_TYPE_ITHREADPOOL,
                                                dexter_threadpool_ithreadpool_init));


static void
dexter_threadpool_class_init (DexterThreadPoolClass *klass)
{
  GObjectClass *gobject_class;

#ifdef G_ENABLE_DEBUG
  g_message (G_STRFUNC);
#endif

  gobject_class = G_OBJECT_CLASS (klass);
  gobject_class->finalize = dexter_threadpool_finalize;
}


static void
dexter_threadpool_init (DexterThreadPool *threadpool)
{
#ifdef G_ENABLE_DEBUG
  g_message (G_STRFUNC);
#endif

  threadpool->exclusive = DEFAULT_THREADPOOL_INIT_EXCLUSIVE;
  threadpool->exit_immediate = DEFAULT_THREADPOOL_EXIT_IMMEDIATE;
    
  threadpool->threadpool = NULL;
  threadpool->initialized = 0;

  if (!dexter_config_get_integer (__ICONFIG, "Main", "ThreadPoolNumThreads",
                                  &threadpool->max_threads, NULL))
  {
    threadpool->max_threads = DEFAULT_THREADPOOL_INIT_MAX_THREADS;
  }
  if (threadpool->max_threads <= 0)
  {
    threadpool->max_threads = DEFAULT_THREADPOOL_INIT_MAX_THREADS;
    g_warning ("%s: ThreadPoolNumThreads must be > 0. Using default %d threads.", 
               G_STRFUNC, threadpool->max_threads);
  }

  if (!dexter_config_get_boolean (__ICONFIG, "Main", "ThreadPoolExitWait",
                                  &threadpool->exit_wait, NULL))
  {
    threadpool->exit_wait = DEFAULT_THREADPOOL_EXIT_WAIT;
  }
}


static void
dexter_threadpool_finalize (GObject *object)
{
  DexterThreadPool *threadpool;

#ifdef G_ENABLE_DEBUG
  g_message (G_STRFUNC);
#endif

  threadpool = DEXTER_THREADPOOL (object);

  /* chain up */
  G_OBJECT_CLASS (dexter_threadpool_parent_class)->finalize (object);
}


static void
dexter_threadpool_ithreadpool_init (DexterPluginIThreadPoolInterface *ithreadpool)
{
#ifdef G_ENABLE_DEBUG
  g_message (G_STRFUNC);
#endif

  ithreadpool->push_task = dexter_threadpool_push_task;
}


static void
_threadpool_dispatcher (gpointer data,
                        gpointer user_data /* unused */)
{
  g_return_if_fail (data != NULL);

  DexterThreadPoolTask  *task = (DexterThreadPoolTask *) data;
  task->func (task->data);
  g_slice_free (DexterThreadPoolTask, task);
}


gboolean
dexter_threadpool_tp_init (DexterThreadPool  *threadpool,
                           GError           **error)
{
  GError *local_error = NULL;
    
  g_return_val_if_fail (threadpool != NULL, FALSE);
  g_return_val_if_fail (error == NULL || *error == NULL, FALSE);
  g_assert (dexter_threadpool_is_initialized (threadpool) == FALSE);

  /* Init glib thread system */
  if (!g_thread_supported ()) g_thread_init (NULL);

  threadpool->threadpool = g_thread_pool_new (_threadpool_dispatcher,
                                                    NULL, /* unused */
                                                    threadpool->max_threads,
                                                    threadpool->exclusive,
                                                    &local_error);
  if (local_error)
  {
    g_propagate_error (error, local_error);
    return FALSE;
  }
    
  g_atomic_int_set (&threadpool->initialized, 1);

  return TRUE;
}


gboolean
dexter_threadpool_is_initialized (DexterThreadPool *threadpool)
{
  g_return_val_if_fail (DEXTER_IS_THREADPOOL (threadpool), FALSE);

  return (g_atomic_int_get (&threadpool->initialized) > 0);
}


gboolean
dexter_threadpool_push_task (DexterPluginIThreadPool  *ithreadpool,
                             GThreadFunc               func,
                             gpointer                  data,
                             GError                  **error)
{
  GError               *local_error = NULL;
  DexterThreadPool     *threadpool = NULL; 
  DexterThreadPoolTask *task = NULL;

  g_return_val_if_fail (DEXTERPLUGIN_IS_ITHREADPOOL (ithreadpool), FALSE);
  g_return_val_if_fail (func != NULL, FALSE);
  g_return_val_if_fail (data != NULL, FALSE);
  g_return_val_if_fail (error == NULL || *error == NULL, FALSE);

  threadpool = DEXTER_THREADPOOL (ithreadpool);

  /* silently ignore new tasks if library is shutting down */
  if (dexter_get_state () == DEXTER_STATE_STOPPING)
    return TRUE;

  g_assert (dexter_threadpool_is_initialized (threadpool) == TRUE);

#ifdef G_ENABLE_DEBUG
  g_message (G_STRFUNC);
#endif

  task = g_slice_new0 (DexterThreadPoolTask);
  task->func = func;
  task->data = data;
  g_thread_pool_push (threadpool->threadpool, task, &local_error);
  if (local_error)
  {
    g_propagate_error (error, local_error);
    g_slice_free (DexterThreadPoolTask, task);
    return FALSE;
  }

  return TRUE;
}


void
dexter_threadpool_tp_exit (DexterThreadPool *threadpool)
{
  g_return_if_fail (DEXTER_IS_THREADPOOL (threadpool));
  g_assert (dexter_threadpool_is_initialized (threadpool) == TRUE);

  g_atomic_int_set (&threadpool->initialized, 0);

  g_thread_pool_free (threadpool->threadpool, threadpool->exit_immediate, threadpool->exit_wait);
  threadpool->threadpool = NULL;
}

#define __DEXTER_THREADPOOL_C__
#include <dexter-aliasdef.c>
