/* $Id: dexter-timedsampler.c 393 2012-03-06 23:36:44Z pkovacs $ */
/*
 * libdexter - a plugin-based, distributed sampling library.
 * Copyright (C) 2006-2012 Philip Kovacs <pkovacs@users.sourceforge.net>.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 2.1
 * of the License, or (at your option) any later version.

 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.

 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
 * 02110-1301  USA
 *
*/

/**
 * SECTION:dexter-timedsampler
 * @title: Timed Samplers
 * @short_description: Using timed samplers for repetitive data snapshots
 * @see_also:
 *
 * Timed samplers, like ordinary samplers, are Libdexter objects that are capable of sampling
 * and storing data. They differ from ordinary samplers in that they operate in a timing loop 
 * and take recurrent samples at regular intervals.  Any Libdexter service returned by a service 
 * broker using the #DEXTER_SERVICE_SAMPLER service type may be instantiated either as an 
 * ordinary sampler or a timed sampler.
 *
 * The desired timing interval is specified during sampler creation with dexter_timedsampler_new().  
 * You then start the timed sampler with dexter_timedsampler_start() and stop it with 
 * dexter_timedsampler_stop().  It is best to use callbacks for data change notifications when using 
 * timed samplers since they will reliably notify you of data changes after each update cycle.
 *
 * When running a timed sampler over the network, i.e. on a channel, the remote sampler becomes a 
 * streaming data source. After the timed sampler is started, the Libdexter server hosting it will 
 * stream its serialized data over the channel back to the local host.  The data continues to stream 
 * until the local host issues a stop request to the server.  As with ordinary samplers, the application 
 * manipulates local and remote timed samplers in exactly the same manner.
 **/

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <string.h>
#include <dexter-channel-internal.h>
#include <dexter-messages.h>
#include <dexter-private.h>
#include <dexter-timedsampler-internal.h>
#include <dexterplugin-isampler.h>
#include <dexterplugin-sampler.h>
#include <dexter-alias.h>

static void     dexter_timedsampler_class_init      (DexterTimedSamplerClass  *klass);
static void     dexter_timedsampler_init            (DexterTimedSampler       *tsampler);
static void     dexter_timedsampler_dispose         (GObject                  *object);
static void     dexter_timedsampler_finalize        (GObject                  *object);
static gint     dexter_timedsampler_set_state       (DexterTimedSampler       *tsampler,
                                                     gint                      state);
static void     dexter_timedsampler_process_message (DexterChannelObject      *channel_object,
                                                     DexterChannelMessage     *message,
                                                     GError                  **error);
static void     dexter_timedsampler_channel_event   (DexterChannelObject      *channel_object,
                                                     gint                      event);
                 /* timing is in main event loop */
static gboolean _timedsampler_source_func           (DexterTimedSampler   *tsampler);


GQuark
dexter_timedsampler_error_quark (void)
{
  static GQuark error_quark = 0;

  if (error_quark == 0)
    error_quark = g_quark_from_static_string ("dexter-timedsampler-error-quark");

  return error_quark;
}

G_DEFINE_TYPE (DexterTimedSampler, dexter_timedsampler, DEXTER_TYPE_CHANNEL_OBJECT);


static void
dexter_timedsampler_class_init (DexterTimedSamplerClass *klass)
{
  GObjectClass             *gobject_class;
  DexterChannelObjectClass *channel_object_class;
    
#ifdef G_ENABLE_DEBUG
  g_message (G_STRFUNC);
#endif

  gobject_class = G_OBJECT_CLASS (klass);
  gobject_class->dispose = dexter_timedsampler_dispose;
  gobject_class->finalize = dexter_timedsampler_finalize;

  channel_object_class = DEXTER_CHANNEL_OBJECT_CLASS (klass);
  channel_object_class->process_message = dexter_timedsampler_process_message;
  channel_object_class->channel_event = dexter_timedsampler_channel_event;
}


static void
dexter_timedsampler_init (DexterTimedSampler *tsampler)
{
#ifdef G_ENABLE_DEBUG
  g_message (G_STRFUNC);
#endif

  tsampler->dispose_has_run = FALSE;
  tsampler->initialized = 0;
  tsampler->state = DEXTER_TIMEDSAMPLER_STATE_STOPPED;
  tsampler->mutex = g_mutex_new ();
  tsampler->source = NULL;
}


static void
dexter_timedsampler_dispose (GObject *object)
{
  DexterTimedSampler *tsampler = DEXTER_TIMEDSAMPLER (object);

#ifdef G_ENABLE_DEBUG
  g_message (G_STRFUNC);
#endif

  if (tsampler->dispose_has_run)
    return;

  if (G_IS_OBJECT (tsampler->plugin_sampler))
    g_object_unref (tsampler->plugin_sampler);

  tsampler->dispose_has_run = TRUE;

  /* chain up */
  G_OBJECT_CLASS (dexter_timedsampler_parent_class)->dispose (object);
}


static void
dexter_timedsampler_finalize (GObject *object)
{
  DexterTimedSampler    *tsampler;
    
#ifdef G_ENABLE_DEBUG
  g_message (G_STRFUNC);
#endif
  tsampler = DEXTER_TIMEDSAMPLER (object);

  if (G_UNLIKELY (tsampler->source))
  {
    g_warning ("%s: timed server finalizing with active timeout source.", G_STRFUNC);
    g_source_destroy (tsampler->source);
  }

  if (tsampler->mutex)
    g_mutex_free (tsampler->mutex);

  if (tsampler->source)
    g_source_unref (tsampler->source);
}


/**
 * dexter_timedsampler_get_state:
 * @tsampler: a #DexterTimedSampler.
 *
 * Returns the current state of the timed sampler.
 *
 * Returns: #gint from #DexterTimedSamplerState indicating current state of the timed sampler.
 **/
gint
dexter_timedsampler_get_state (DexterTimedSampler *tsampler)
{
  gint state;

  g_return_val_if_fail (DEXTER_IS_TIMEDSAMPLER (tsampler), -1);

  g_object_ref (tsampler);

  state = g_atomic_int_get (&tsampler->state);

  g_object_unref (tsampler);

  return state;
}


static gint
dexter_timedsampler_set_state (DexterTimedSampler *tsampler,
                               gint                state)
{
  gint old_state;

  g_return_val_if_fail (tsampler != NULL, -1);
  g_return_val_if_fail (state == DEXTER_TIMEDSAMPLER_STATE_STARTED || 
                        state == DEXTER_TIMEDSAMPLER_STATE_STOPPED, -1);

  g_object_ref (tsampler);

  old_state = g_atomic_int_get (&tsampler->state);
  g_atomic_int_set (&tsampler->state, state);

  g_object_unref (tsampler);

  return old_state;
}


static void 
dexter_timedsampler_process_message (DexterChannelObject   *channel_object,
                                     DexterChannelMessage  *message,
                                     GError               **error)
{
  DexterTimedSampler *tsampler;

#ifdef G_ENABLE_DEBUG
  g_message (G_STRFUNC);
#endif

  g_return_if_fail (DEXTER_IS_CHANNEL_OBJECT (channel_object));
  g_return_if_fail (message != NULL);
  g_return_if_fail (error == NULL || *error == NULL);
  tsampler = DEXTER_TIMEDSAMPLER (channel_object);

  g_object_ref (channel_object);
  if (message->msg_type == DEXTER_MESSAGE_REPLY)
  {
    if (message->rpy.status != STATUS_OK)
    {
      g_set_error (error,
                   DEXTER_TIMEDSAMPLER_ERROR,
                   DEXTER_TIMEDSAMPLER_ERROR_FAILED,
                   "%s: server error: command (%d) status (%d) details (%s)",
                   G_STRFUNC, message->rpy.command, message->rpy.status, 
                   message->payload ? message->payload->str : NULL);    
                   dexter_channel_message_free (message);
      g_object_unref (channel_object);
      return;
    }
    else
    {
      switch (message->rpy.command)
      {
        case COMMAND_TIMEDSAMPLER_NEW:
        case COMMAND_TIMEDSAMPLER_INITIALIZE:
        case COMMAND_TIMEDSAMPLER_FREE:
        case COMMAND_TIMEDSAMPLER_START:
        case COMMAND_TIMEDSAMPLER_STOP:
          break;
        default:
        {
          /* programming error */
          g_error (G_STRLOC " reply received from server for unimplemented command (%d)", message->rpy.command);
        }
      }
    }
  }
  else if (message->msg_type == DEXTER_MESSAGE_DATA)
  {
    DexterTimedSampler *tsampler = DEXTER_TIMEDSAMPLER (channel_object);

    if (G_LIKELY (message->dat.status == STATUS_OK))
    {
      gpointer  sampler_data;
      GError   *local_error   = NULL;

      sampler_data = dexter_timedsampler_data_deserialize (tsampler, message->payload, &local_error);
      if (local_error)
      {
        g_propagate_error (error, local_error);
      }
      else
      {
        dexter_timedsampler_data_xfer (tsampler, NULL, sampler_data);
      }
      dexter_timedsampler_data_free (tsampler, sampler_data);
      /* queue the callbacks onto threadpool */
      dexterplugin_sampler_callbacks (DEXTERPLUGIN_SAMPLER (tsampler->plugin_sampler));
    }
  }
  else
  {
    /* programming error */
    g_error (G_STRLOC " invalid message type (%d) received from server", message->msg_type);
  }

  /* chain up */
  DEXTER_CHANNEL_OBJECT_CLASS (dexter_timedsampler_parent_class) 
    -> process_message (channel_object, message, error);

  g_object_unref (channel_object);
}


static void 
dexter_timedsampler_channel_event (DexterChannelObject *channel_object,
                                   gint                 event)
{
  DexterTimedSampler *tsampler;
#ifdef G_ENABLE_DEBUG
  g_message (G_STRFUNC);
#endif

  g_return_if_fail (DEXTER_IS_CHANNEL_OBJECT (channel_object));

  g_object_ref (channel_object);
  tsampler = DEXTER_TIMEDSAMPLER (channel_object);

  switch (event)
  {
    case DEXTER_CHANNEL_EVENT_NOCONN:
    case DEXTER_CHANNEL_EVENT_CLOSED:
      {
        /* if channel connection was broken or closed. set our state to STOPPED. */
        dexter_timedsampler_set_state (tsampler, DEXTER_TIMEDSAMPLER_STATE_STOPPED);
        break;
      }
    case DEXTER_CHANNEL_EVENT_OPEN:
      break;
    default:
    {
      /* programming error */
      g_error (G_STRLOC " unrecognized channel event (%d)", event);
    }
  }

  /* chain up */
  DEXTER_CHANNEL_OBJECT_CLASS (dexter_timedsampler_parent_class)
    -> channel_event (channel_object, event);

  g_object_unref (channel_object);
}


gboolean
_timedsampler_source_func (DexterTimedSampler *tsampler)
{
  GError *local_error = NULL;

  g_return_val_if_fail (DEXTER_IS_TIMEDSAMPLER (tsampler), FALSE);

  g_object_ref (tsampler);

  /* do the sample */
  g_mutex_lock (tsampler->mutex);
  dexterplugin_isampler_sample (DEXTERPLUGIN_ISAMPLER (tsampler->plugin_sampler), &local_error);
  g_mutex_unlock (tsampler->mutex);

  if (local_error)
  {
    g_warning ("%s: %s", G_STRFUNC, local_error->message);
    g_clear_error (&local_error);
    g_object_unref (tsampler);
    return TRUE;  /* keep loop going */
  }

  /* queue callbacks onto threadpool */
  dexterplugin_sampler_callbacks (DEXTERPLUGIN_SAMPLER (tsampler->plugin_sampler));

  g_object_unref (tsampler);

  return TRUE;
}


gboolean
dexter_timedsampler_is_initialized (DexterTimedSampler *tsampler)
{
  gboolean ret;

  g_return_val_if_fail (DEXTER_IS_TIMEDSAMPLER (tsampler), FALSE);

  g_object_ref (tsampler);
  ret = g_atomic_int_get (&tsampler->initialized) > 0;
  g_object_unref (tsampler);

  return ret;

}


DexterTimedSampler *
dexter_timedsampler_new_actual (const DexterPluginService       *service,
                                glong                            msecs_interval,
                                const DexterSamplerDataCallback *callbacks,
                                DexterChannel                   *channel,
                                DexterPluginIThreadPool         *ithreadpool)
{
  DexterSamplerDataCallback   *callbacks_copy;
  DexterPluginSampler         *plugin_sampler;
  DexterTimedSampler          *timed_sampler;
  guint                        n_callbacks;
  gsize                        callbacks_size;

  g_return_val_if_fail (service != NULL, NULL);
  g_return_val_if_fail (service->type != G_TYPE_INVALID, NULL);
  g_return_val_if_fail (channel == NULL || DEXTER_IS_CHANNEL (channel), NULL);
  g_return_val_if_fail (ithreadpool == NULL || DEXTERPLUGIN_IS_ITHREADPOOL (ithreadpool), NULL);

  if (channel)
    g_object_ref (channel);

  /* create a copy of the callback array */
  callbacks_copy = (DexterSamplerDataCallback *)callbacks;
  callbacks_size = 0;
  if (callbacks)
  {
    for (n_callbacks = 0; callbacks_copy && *callbacks_copy++; n_callbacks++);
    callbacks_size = (n_callbacks+1) * sizeof (DexterSamplerDataCallback);
    callbacks_copy = (DexterSamplerDataCallback *) g_slice_alloc0 (callbacks_size);
    memcpy (callbacks_copy, callbacks, callbacks_size);
  }

  /* instantiate the timed sampler container */
  timed_sampler = g_object_new (DEXTER_TYPE_TIMEDSAMPLER, NULL);

  /* instantiate the plugin sampler */
  plugin_sampler = g_object_new (service->type,
                                 "private1", G_OBJECT (timed_sampler),
                                 "private2", callbacks_copy,
                                 "private3", ithreadpool,
                                 "private4", callbacks_size,
                                  NULL);

  /* set the timed sampler container public data */
  timed_sampler->msecs_interval = msecs_interval;
  timed_sampler->plugin_sampler = G_OBJECT (plugin_sampler); 

  if (channel)
  {
    DexterChannelObject *channel_object = DEXTER_CHANNEL_OBJECT (timed_sampler);

    /* register sampler as an object capable of receiving server messages */
    channel_object->wait_timeout = (glong)(1000 * dexter_channel_get_settings (channel) -> timeout_msecs);
    dexter_channel_register_channel_object (channel, channel_object, DEXTER_MESSAGE_REPLY);
    g_object_unref (channel);
  }  

  return timed_sampler;
}


/**
 * dexter_timedsampler_new:
 * @service: a #DexterPluginService.
 * @msecs_interval: #guint refresh interval in milliseconds.
 * @callbacks: optional, null-terminated array of #DexterSamplerDataCallback functions, or %NULL.
 * @channel: optional #DexterChannel over which the sampler operates, or %NULL.
 * @error: return location for optional #GError.
 *
 * Creates a #DexterTimedSampler implementing the indicated #DexterPluginService.
 *
 * The timed sampler will begin automatic sampling at each @msecs_interval after dexter_timedsampler_start()
 * is called.  It will continue to sample cyclically until dexter_timedsampler_stop() is called.
 *
 * If a callback array is specified, it is duplicated and stored internally (so you may free
 * the @callbacks immediately after this function).  Each callback function will be called
 * asynchronously in a thread as the timed sampler is updated cyclically. 
 *
 * If a channel is indicated, the timed sampler operates transparently as a proxy to a timed sampler 
 * created on the remote machine designated during channel creation.  The channel must be opened with
 * dexter_channel_open() prior to calling this function.
 *
 * If no channel is indicated, the timed sampler operates locally (in-process).
 *
 * The timed sampler must be freed when done with dexter_timedsampler_free().
 *
 * Returns: #DexterTimedSampler if successful or %NULL otherwise.
 **/
DexterTimedSampler *
dexter_timedsampler_new (const DexterPluginService        *service,
                         guint                             msecs_interval,
                         const DexterSamplerDataCallback  *callbacks,
                         DexterChannel                    *channel,
                         GError                          **error)
{
  DexterTimedSampler *timed_sampler;

#ifdef G_ENABLE_DEBUG
  g_message (G_STRFUNC);
#endif

  g_return_val_if_fail (service != NULL, NULL);
  g_return_val_if_fail (service->type != G_TYPE_INVALID, NULL);
  g_return_val_if_fail (channel == NULL || DEXTER_IS_CHANNEL (channel), NULL);
  g_return_val_if_fail (error == NULL || *error == NULL, NULL);

  if (channel && (dexter_channel_get_state (channel) == DEXTER_CHANNEL_STATE_OPEN))
  {
    DexterPluginService *local_service = NULL;
    GString             *payload;
    GError              *local_error = NULL;

    /* if a channel is indicated, we assume that the indicated service is part of
     * of a remote service group.  the gtype of that service may be different from
     * the local version of the same service, so we need to find the local version
     * of this service. */
    if (!dexter_sampler_find_local_service (&local_service, service))
    {
      g_warning ("%s: requested service is not available locally", G_STRFUNC);
      return NULL;
    }

    /* create the timed sampler */
    timed_sampler = dexter_timedsampler_new_actual (local_service, msecs_interval, callbacks, channel,
                                                    __ITHREADPOOL);
    dexterplugin_service_free (local_service);
  
    /* send command to create network peer */
    payload = g_string_new (NULL);
    g_string_append_printf (payload, "%s %d", service->uuid, timed_sampler->msecs_interval);
    dexter_channel_object_send_wait_process (DEXTER_CHANNEL_OBJECT (timed_sampler),
                                             COMMAND_TIMEDSAMPLER_NEW, 0, payload, &local_error);
    if (local_error)
    {
      g_propagate_error (error, local_error);
      /* release local object due to error */
      g_object_unref (timed_sampler);
      return NULL;
    }
  }
  else if (channel)
  {
    g_set_error (error,
                 DEXTER_CHANNEL_ERROR,
                 DEXTER_CHANNEL_ERROR_STATE,
                 "%s: operation requires an open channel", G_STRFUNC);
    return NULL;
  }
  else
  {
    /* create a standalone local sampler */
    timed_sampler = dexter_timedsampler_new_actual (service, msecs_interval, callbacks, channel,
                                                    __ITHREADPOOL); 
  }
 
  return timed_sampler;
}


/**
 * dexter_timedsampler_free:
 * @tsampler: a #DexterTimedSampler.
 * @error: return location of optional #GError.
 *
 * Frees a #DexterTimedSampler created with dexter_timedsampler_new().  If the timed sampler is started, 
 * it will be stopped for you, although it is better form to use dexter_timedsampler_stop() before freeing 
 * the timed sampler.  
 *
 * If the timed sampler is operating on a channel, a message is sent to the server to free the peer.  
 * You will be notified if an error occurs during that message transmission, but the local object is 
 * always freed in any event.
 **/
void
dexter_timedsampler_free (DexterTimedSampler  *tsampler,
                          GError             **error)
{
  DexterChannelObject *channel_object;

#ifdef G_ENABLE_DEBUG
  g_message (G_STRFUNC);
#endif

  g_return_if_fail (DEXTER_IS_TIMEDSAMPLER (tsampler));
  g_return_if_fail (error == NULL || *error == NULL);

  g_object_ref (tsampler);
  channel_object = DEXTER_CHANNEL_OBJECT (tsampler);

  /* if sampler is started, stop it */
  if (dexter_timedsampler_get_state (tsampler) == DEXTER_TIMEDSAMPLER_STATE_STARTED)
    dexter_timedsampler_stop (tsampler, NULL);

  /* if channel is open, send message to free peer */
  if ((channel_object->channel) &&
      (dexter_channel_get_state (channel_object->channel) == DEXTER_CHANNEL_STATE_OPEN))
  {
    GError *local_error = NULL;

    /* send command to free network peer */
    dexter_channel_object_send_wait_process (channel_object,
                                             COMMAND_TIMEDSAMPLER_FREE, 0, NULL, &local_error);
    if (local_error)
    {
      g_propagate_error (error, local_error);
      goto done;
    }
        
    /* deregister sampler as an object capable of receiving server messages */
    dexter_channel_deregister_channel_object (channel_object->channel, channel_object);
  }

  /* if the channel is not open, just deregister sampler */
  else if (channel_object->channel)
  {
    /* deregister sampler as an object capable of receiving server messages */
    dexter_channel_deregister_channel_object (channel_object->channel, channel_object);
  }

done:
  /* release this api ref */
  g_object_unref (tsampler);

  /* release the object */
  g_object_unref (tsampler);
}


/**
 * dexter_timedsampler_initialize:
 * @tsampler: a #DexterTimedSampler.
 * @param: address of an optional #gint parameter, or %NULL.
 * @input: optional #gconstpointer to plugin-specific initialization data, or %NULL.
 * @error: return location of optional #GError.
 *
 * Performs one-time initialization of the timed sampler created with dexter_timedsampler_new().
 * If the sampler requires simple integer data for initialization, indicate the address to that integer, 
 * or %NULL, if no integer is required.  If the timed sampler requires other, complex or structured input 
 * data for initialization, indicate the address to that input structure, or %NULL, if no input is required.
 *
 * The timed sampler should be initialized once after creation with dexter_timedsampler_new() and 
 * before starting with dexter_timedsampler_start().
 *
 * If the timed sampler was created over a channel, the channel should be open when calling this function.
 *
 * Returns: #gboolean indicating success or failure.
 **/
gboolean
dexter_timedsampler_initialize (DexterTimedSampler  *tsampler,
                                const gint          *param,
                                gconstpointer        input,
                                GError             **error)
{
  DexterChannelObject *channel_object;

#ifdef G_ENABLE_DEBUG
  g_message (G_STRFUNC);
#endif

  g_return_val_if_fail (DEXTER_IS_TIMEDSAMPLER (tsampler), FALSE);
  g_return_val_if_fail (error == NULL || *error == NULL, FALSE);

  channel_object = DEXTER_CHANNEL_OBJECT (tsampler);
  g_object_ref (tsampler);

  /* error if already initialized */
  if (dexter_timedsampler_is_initialized (tsampler))
  {
    g_set_error (error,
                 DEXTER_TIMEDSAMPLER_ERROR,
                 DEXTER_TIMEDSAMPLER_ERROR_FAILED,
                 "%s: timed sampler already initialized", G_STRFUNC);
    g_object_unref (tsampler);
    return FALSE;
  }

  /* if channel is open, send message for peer to initialize */
  if ((channel_object->channel) &&
      (dexter_channel_get_state (channel_object->channel) == DEXTER_CHANNEL_STATE_OPEN))
  {
    GString *payload = NULL;
    GError  *local_error = NULL;

    if (input)
    {
      /* serialize the input to the payload */
      payload = g_string_new (NULL);
      dexter_timedsampler_input_serialize (tsampler, payload, input, &local_error);
      if (local_error)
      {
        g_string_free (payload, TRUE);
        g_propagate_error (error, local_error);
        g_object_unref (tsampler);
        return FALSE;
      }
    }

    /* send the initialize command to the network peer */
    dexter_channel_object_send_wait_process (channel_object, COMMAND_TIMEDSAMPLER_INITIALIZE,
                                        param ? *param : 0, 
                                        payload,
                                        &local_error);
    if (local_error)
    {
      g_propagate_error (error, local_error);
      if (payload)
        g_string_free (payload, TRUE);
      g_object_unref (tsampler);
      return FALSE;
    }
    if (payload)
      g_string_free (payload, TRUE);
  }
  else if (channel_object->channel)
  {
    g_set_error (error,
                 DEXTER_CHANNEL_ERROR,
                 DEXTER_CHANNEL_ERROR_STATE,
                 "%s: operation requires an open channel", G_STRFUNC);

    g_object_unref (tsampler);
    return FALSE;
  }
  else
  {
    gboolean ret;

    /* initialize local sampler with input */
    g_mutex_lock (tsampler->mutex);
    ret = 
      dexterplugin_isampler_initialize (DEXTERPLUGIN_ISAMPLER (tsampler->plugin_sampler), param, input, error);
    g_mutex_unlock (tsampler->mutex);
    if (!ret)
    {
      g_object_unref (tsampler);
      return ret;
    }
  }

  g_atomic_int_set (&tsampler->initialized, 1);

  g_object_unref (tsampler);
  return TRUE;
}


/**
 * dexter_timedsampler_start:
 * @tsampler: a #DexterTimedSampler.
 * @error: return location of optional #GError.
 *
 * Start the timed sampler and begin receiving updates every @msecs_interval, 
 * as specified in dexter_timedsampler_new().  The sampler updates recurrently 
 * until dexter_timedsampler_stop() is called.
 *
 * If operating over a channel and the channel closes for any reason, the sampler 
 * state will transition to @DEXTER_TIMEDSAMPLER_STATE_STOPPED.
 *
 * Returns: #gboolean %TRUE is started successfully, %FALSE otherwise.
 **/
gboolean
dexter_timedsampler_start (DexterTimedSampler  *tsampler,
                           GError             **error)
{
  DexterChannelObject *channel_object;
  GError              *local_error = NULL;

#ifdef G_ENABLE_DEBUG
  g_message (G_STRFUNC);
#endif

  g_return_val_if_fail (DEXTER_IS_TIMEDSAMPLER (tsampler), FALSE);
  g_return_val_if_fail (error == NULL || *error == NULL, FALSE);

  g_object_ref (tsampler);
  channel_object = DEXTER_CHANNEL_OBJECT (tsampler);

  /* error if not initialized */
  if (!dexter_timedsampler_is_initialized (tsampler))
  {
    g_set_error (error,
                 DEXTER_TIMEDSAMPLER_ERROR,
                 DEXTER_TIMEDSAMPLER_ERROR_FAILED,
                 "%s: timed sampler not initialized", G_STRFUNC);
    g_object_unref (tsampler);
    return FALSE;
  }
    
  /* error if already started */
  if (dexter_timedsampler_get_state (tsampler) == DEXTER_TIMEDSAMPLER_STATE_STARTED)
  {
    g_set_error (error,
                 DEXTER_TIMEDSAMPLER_ERROR,
                 DEXTER_TIMEDSAMPLER_ERROR_STARTED,
                 "%s: timed sampler already started", G_STRFUNC);
    g_object_unref (tsampler);
    return FALSE;
  }

  /* if a channel-based timed sampler, must have an open channel */
  if ((channel_object->channel) &&
      (dexter_channel_get_state (channel_object->channel) != DEXTER_CHANNEL_STATE_OPEN))
  {
    g_set_error (error,
                 DEXTER_CHANNEL_ERROR,
                 DEXTER_CHANNEL_ERROR_STATE,
                 "%s: operation requires an open channel", G_STRFUNC);
    g_object_unref (tsampler);
    return FALSE;
  }

  /* a channel-based timed sampler and channel is open  */
  else if ((channel_object->channel) &&
           (dexter_channel_get_state (channel_object->channel) == DEXTER_CHANNEL_STATE_OPEN))
  {
    /* send start command to network peer */
    dexter_channel_object_send_wait_process (channel_object, COMMAND_TIMEDSAMPLER_START, 0, NULL, &local_error);
    if (local_error)
    {
      g_propagate_error (error, local_error);
      g_object_unref (tsampler);
      return FALSE;
    }
  }
  else /* not a channel-based timed sampler -- a local sampler */
  {
    /* create timed sampler's timeout source, set its callback and attach it to the main loop */
    g_assert (tsampler->source == NULL);
    tsampler->source = g_timeout_source_new (tsampler->msecs_interval);
    g_source_set_callback (tsampler->source,
                           (GSourceFunc)_timedsampler_source_func,
                           tsampler,
                           NULL);
    g_source_set_can_recurse (tsampler->source, FALSE);
    dexter_event_loop_source_attach (__EVENTLOOP, tsampler->source);
    /* main loop increments source ref, so total is now 2. */
  }

  dexter_timedsampler_set_state (tsampler, DEXTER_TIMEDSAMPLER_STATE_STARTED);
  g_object_unref (tsampler);
  return TRUE;
}


/**
 * dexter_timedsampler_stop:
 * @tsampler: a #DexterTimedSampler.
 * @error: return location of optional #GError.
 *
 * Stops the timed sampler previously started with dexter_timedsampler_start().
 * If the timed sampler is already stopped, no error occurs and %TRUE is returned.
 *
 * Returns: #gboolean %TRUE if stopped successfully, %FALSE otherwise.
 **/
gboolean
dexter_timedsampler_stop (DexterTimedSampler  *tsampler,
                          GError             **error)
{
  DexterChannelObject       *channel_object;

#ifdef G_ENABLE_DEBUG
  g_message (G_STRFUNC);
#endif

  g_return_val_if_fail (DEXTER_IS_TIMEDSAMPLER (tsampler), FALSE);
  g_return_val_if_fail (error == NULL || *error == NULL, FALSE);

  g_object_ref (tsampler);
  channel_object = DEXTER_CHANNEL_OBJECT (tsampler);

  /* error if not initialized */
  if (!dexter_timedsampler_is_initialized (tsampler))
  {
    g_set_error (error,
                 DEXTER_TIMEDSAMPLER_ERROR,
                 DEXTER_TIMEDSAMPLER_ERROR_FAILED,
                 "%s: timed sampler not initialized", G_STRFUNC);
    g_object_unref (tsampler);
    return FALSE;
  }

  /* stopping a timed sampler that is already stopped will not be considered an error */
  if (dexter_timedsampler_get_state (tsampler) == DEXTER_TIMEDSAMPLER_STATE_STOPPED)
  {
    g_object_unref (tsampler);
    return TRUE;
  }

  /* the channel-based timed sampler is started, but the channel not open -- highly unlikely. */
  if (G_UNLIKELY ((channel_object->channel) &&
                  (dexter_channel_get_state (channel_object->channel) != DEXTER_CHANNEL_STATE_OPEN)))
  {
    g_set_error (error,
                 DEXTER_CHANNEL_ERROR,
                 DEXTER_CHANNEL_ERROR_STATE,
                 "%s: operation requires an open channel", G_STRFUNC);
    g_object_unref (tsampler);
    return FALSE;
  }

  /* if a channel-based timed sampler, the channel is open here */
  else if (channel_object->channel) 
  {
    GError *local_error = NULL;

    /* send stop command to network peer */
    dexter_channel_object_send_wait_process (channel_object, COMMAND_TIMEDSAMPLER_STOP, 0, NULL, &local_error);
    if (local_error)
    {
      g_propagate_error (error, local_error);
      g_object_unref (tsampler);
      return FALSE;
    }
  }
  else /* not a channel-based timed sampler -- a local sampler */
  {
    /* destroy timed sampler's timeout source which stops it from running in event loop */
    g_assert (tsampler->source != NULL);
    g_source_destroy (tsampler->source);
    /* release last ref to source */
    g_source_unref (tsampler->source);
    tsampler->source=NULL;
  }

  dexter_timedsampler_set_state (tsampler, DEXTER_TIMEDSAMPLER_STATE_STOPPED);
  g_object_unref (tsampler);
  return TRUE;
}


/**
 * dexter_timedsampler_data_new:
 * @tsampler: a #DexterTimedSampler.
 * @populate: #gboolean to determine if returned memory is populated with current sampler data or zero-filled.
 *
 * Allocates memory to contain timed sampler data.  The plugin implementing the timed sampler defines this 
 * structure -- it is completely opaque to the library.  
 *
 * If @populate is %TRUE, the returned memory is populated with current timed sampler data.  If %FALSE,
 * it is zero-filled.
 *
 * The memory must freed with dexter_timedsampler_data_free() when done, unless within a 
 * #DexterSamplerDataCallback, in which case the library will free the memory for you after the 
 * callback returns.
 *
 * Returns: #gpointer to newly allocated sample memory.
 */
gpointer
dexter_timedsampler_data_new (DexterTimedSampler *tsampler,
                              gboolean            populate)
{
  gpointer  data;

  g_return_val_if_fail (DEXTER_IS_TIMEDSAMPLER (tsampler), NULL);
  g_return_val_if_fail (DEXTERPLUGIN_IS_ISAMPLER (tsampler->plugin_sampler), NULL);

  g_object_ref (tsampler);

  if (populate && (!dexter_timedsampler_is_initialized (tsampler)))
    g_warning ("%s: timed sampler not initialized", G_STRFUNC);

  /* need thread-safety locking if populating with existing data */
  if (populate)
    g_mutex_lock (tsampler->mutex);
  data = dexterplugin_isampler_data_new (DEXTERPLUGIN_ISAMPLER (tsampler), populate);
  if (populate)
    g_mutex_unlock (tsampler->mutex);

  g_object_unref (tsampler);
  return data;
}


/**
 * dexter_timedsampler_data_dup:
 * @tsampler: a #DexterTimedSampler.
 * @source_data: #gconstpointer to timed sample data to duplicate.
 *
 * Allocates timed sampler data memory and copies the contents of @source_data onto the new memory.
 *
 * The user must free this memory with dexter_timedsampler_data_free().
 *
 * Returns: #gpointer to newly allocated memory containing the duplicated data.
 **/
gpointer
dexter_timedsampler_data_dup (DexterTimedSampler *tsampler,
                              gconstpointer       source_data)
{
  gpointer data;

  g_return_val_if_fail (DEXTER_IS_TIMEDSAMPLER (tsampler), NULL);
  g_return_val_if_fail (DEXTERPLUGIN_IS_ISAMPLER (tsampler->plugin_sampler), NULL);
  g_return_val_if_fail (source_data != NULL, NULL);

  g_object_ref (tsampler);

  data = dexterplugin_isampler_data_dup (DEXTERPLUGIN_ISAMPLER (tsampler->plugin_sampler), source_data);
  
  g_object_unref (tsampler);
  return data;
}


/**
 * dexter_timedsampler_data_xfer:
 * @tsampler: a #DexterTimedSampler.
 * @old_data: optional #gpointer to memory to receive a copy of the timed sampler's current (old) data, or %NULL.
 * @new_data: optional #gconstpointer to memory to replace the timed sampler's current data, or %NULL.
 *
 * Transfers (copies) data into and/or out of the timed sampler.  If @old_data is indicated, the timed sampler 
 * data is copied into that memory, before being replaced with @new_data.  If @new_data is indicated, that data
 * will be copied over existing internal timed sampler data, after existing data is copied out to @old_data.
 *
 * Note that use of the @new_data parameter by users is probably never appropriate.  The proper way for users
 * to replace internal timed sampler data is to start the sampler with dexter_timedsampler_start() and let it 
 * update cyclically on its own.
 *
 * This api is used internally by the library to transfer network-serialized sampler data into and out of
 * timed samplers.  Users may also find it useful to copy out existing sampler data.
 *
 * <informalexample><programlisting>
 * /<!-- -->* One way to get the timed sampler's current data. *<!-- -->/
 * gpointer current_data = dexter_timedsampler_data_new (tsampler, TRUE);
 * <!-- -->
 * /<!-- -->* Another way  *<!-- -->/
 * gpointer current_data = dexter_timedsampler_data_new (tsampler, FALSE);
 * dexter_timedsampler_data_xfer (tsampler, current_data, NULL);
 * </programlisting></informalexample>
 **/
void
dexter_timedsampler_data_xfer (DexterTimedSampler *tsampler,
                               gpointer            old_data,
                               gconstpointer       new_data)
{
  g_return_if_fail (DEXTER_IS_TIMEDSAMPLER (tsampler));
  g_return_if_fail (DEXTERPLUGIN_IS_ISAMPLER (tsampler->plugin_sampler));

  /* need thread-safety locking */
  g_object_ref (tsampler);

  if (!dexter_timedsampler_is_initialized (tsampler))
    g_warning ("%s: timed sampler not initialized", G_STRFUNC);
  
  g_mutex_lock (tsampler->mutex);
  dexterplugin_isampler_data_xfer (DEXTERPLUGIN_ISAMPLER (tsampler->plugin_sampler), old_data, new_data);
  g_mutex_unlock (tsampler->mutex);
  
  g_object_unref (tsampler);
}


/**
 * dexter_timedsampler_data_copy:
 * @tsampler: a #DexterTimedSampler.
 * @dest_data: #gpointer to destination memory to receive the copy.
 * @source_data: #gconstpointer to source memory for the copy.
 *
 * Copies the timed sampler data indicated by @source_data to the memory indicated by @dest_data.
 **/
void
dexter_timedsampler_data_copy (DexterTimedSampler *tsampler,
                               gpointer            dest_data,
                               gconstpointer       source_data)
{
  g_return_if_fail (DEXTER_IS_TIMEDSAMPLER (tsampler));
  g_return_if_fail (DEXTERPLUGIN_IS_ISAMPLER (tsampler->plugin_sampler));
  g_return_if_fail (dest_data != NULL);
  g_return_if_fail (source_data != NULL);

  g_object_ref (tsampler);
  
  dexterplugin_isampler_data_copy (DEXTERPLUGIN_ISAMPLER (tsampler->plugin_sampler), dest_data, source_data);
  
  g_object_unref (tsampler);
}


/**
 * dexter_timedsampler_data_free:
 * @tsampler: a #DexterTimedSampler.
 * @data: #gpointer to timed sampler data to free.
 *
 * Frees timed sampler data previously allocated with dexter_timedsampler_data_new() or 
 * dexter_timedsampler_data_dup().
 **/
void
dexter_timedsampler_data_free (DexterTimedSampler *tsampler,
                               gpointer            data)
{
  g_return_if_fail (DEXTER_IS_TIMEDSAMPLER (tsampler));
  g_return_if_fail (DEXTERPLUGIN_IS_ISAMPLER (tsampler->plugin_sampler));

  g_object_ref (tsampler);
  
  dexterplugin_isampler_data_free (DEXTERPLUGIN_ISAMPLER (tsampler->plugin_sampler), data);
  
  g_object_unref (tsampler);
}


gboolean
dexter_timedsampler_data_serialize (DexterTimedSampler  *tsampler,
                                    GString             *dest_buffer,
                                    gconstpointer        source_data,
                                    GError             **error)
{
  gboolean ret;

  g_return_val_if_fail (DEXTER_IS_TIMEDSAMPLER (tsampler), FALSE);
  g_return_val_if_fail (DEXTERPLUGIN_IS_ISAMPLER (tsampler->plugin_sampler), FALSE);
  g_return_val_if_fail (dest_buffer != NULL, FALSE);
  g_return_val_if_fail (source_data != NULL, FALSE);
  g_return_val_if_fail (error == NULL || *error == NULL, FALSE);

  g_object_ref (tsampler);
  ret = dexterplugin_isampler_data_serialize (DEXTERPLUGIN_ISAMPLER (tsampler->plugin_sampler),
                                              dest_buffer, source_data, error);
  g_object_unref (tsampler);
  return ret;
}


gpointer
dexter_timedsampler_data_deserialize (DexterTimedSampler  *tsampler,
                                      const GString       *source_buffer,
                                      GError             **error)
{
  gpointer data;

  g_return_val_if_fail (DEXTER_IS_TIMEDSAMPLER (tsampler), NULL);
  g_return_val_if_fail (DEXTERPLUGIN_IS_ISAMPLER (tsampler->plugin_sampler), NULL);
  g_return_val_if_fail (source_buffer != NULL, NULL);
  g_return_val_if_fail (error == NULL || *error == NULL, NULL);

  g_object_ref (tsampler);
  
  data = dexterplugin_isampler_data_deserialize (DEXTERPLUGIN_ISAMPLER (tsampler->plugin_sampler),
                                                 source_buffer, error);
  
  g_object_unref (tsampler);
  return data;
}


/**
 * dexter_timedsampler_input_new:
 * @tsampler: a #DexterTimedSampler.
 *
 * Allocates and returns a new timed sampler input structure.  The plugin implementing the timed sampler
 * defines this structure -- it is completely opaque to the library.
 *
 * The memory must be freed with dexter_timedsampler_input_free() when done.
 *
 * Returns: #gpointer to newly allocated input structure.
 **/
gpointer
dexter_timedsampler_input_new (DexterTimedSampler *tsampler)
{
  gpointer data;

  g_return_val_if_fail (DEXTER_IS_TIMEDSAMPLER (tsampler), NULL);
  g_return_val_if_fail (DEXTERPLUGIN_IS_ISAMPLER (tsampler->plugin_sampler), NULL);

  g_object_ref (tsampler);
  
  data = dexterplugin_isampler_input_new (DEXTERPLUGIN_ISAMPLER (tsampler->plugin_sampler));
  
  g_object_unref (tsampler);
  return data;
}


/**
 * dexter_timedsampler_input_free:
 * @tsampler: a #DexterTimedSampler.
 * @input: #gpointer to timed sampler input to free.
 *
 * Frees an input structure previously allocated with dexter_timedsampler_input_new().
 **/
void
dexter_timedsampler_input_free (DexterTimedSampler *tsampler,
                                gpointer            input)
{
  g_return_if_fail (DEXTER_IS_TIMEDSAMPLER (tsampler));
  g_return_if_fail (DEXTERPLUGIN_IS_ISAMPLER (tsampler->plugin_sampler));

  g_object_ref (tsampler);
  
  dexterplugin_isampler_input_free (DEXTERPLUGIN_ISAMPLER (tsampler->plugin_sampler), input);
  
  g_object_unref (tsampler);
}


gboolean
dexter_timedsampler_input_serialize (DexterTimedSampler  *tsampler,
                                     GString             *dest_buffer,
                                     gconstpointer        source_input,
                                     GError              **error)
{
  gboolean ret;

  g_return_val_if_fail (DEXTER_IS_TIMEDSAMPLER (tsampler), FALSE);
  g_return_val_if_fail (DEXTERPLUGIN_IS_ISAMPLER (tsampler->plugin_sampler), FALSE);
  g_return_val_if_fail (dest_buffer != NULL, FALSE);
  g_return_val_if_fail (source_input != NULL, FALSE);
  g_return_val_if_fail (error == NULL || *error == NULL, FALSE);

  g_object_ref (tsampler);
  
  ret = dexterplugin_isampler_input_serialize (DEXTERPLUGIN_ISAMPLER (tsampler->plugin_sampler),
                                               dest_buffer, source_input, error);
  
  g_object_unref (tsampler);
  return ret;
}


gpointer
dexter_timedsampler_input_deserialize (DexterTimedSampler  *tsampler,
                                       const GString       *source_buffer,
                                       GError             **error)
{
  gpointer data;

  g_return_val_if_fail (DEXTER_IS_TIMEDSAMPLER (tsampler), NULL);
  g_return_val_if_fail (DEXTERPLUGIN_IS_ISAMPLER (tsampler->plugin_sampler), NULL);
  g_return_val_if_fail (source_buffer != NULL, NULL);
  g_return_val_if_fail (error == NULL || *error == NULL, NULL);

  g_object_ref (tsampler);
  
  data = dexterplugin_isampler_input_deserialize (DEXTERPLUGIN_ISAMPLER (tsampler->plugin_sampler),
                                                  source_buffer, error);
  
  g_object_unref (tsampler);
  return data;
}

#define __DEXTER_TIMEDSAMPLER_C__
#include <dexter-aliasdef.c>
