/* $Id: dexter-tls-gnutls.c 393 2012-03-06 23:36:44Z pkovacs $ */
/*
 * libdexter - a plugin-based, distributed sampling library.
 * Copyright (C) 2006-2012 Philip Kovacs <pkovacs@users.sourceforge.net>.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 2.1
 * of the License, or (at your option) any later version.

 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.

 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
 * 02110-1301  USA
 *
*/

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <pthread.h>
#include <errno.h>
#ifdef HAVE_MSG_NOSIGNAL
#include <sys/types.h>
#include <sys/socket.h>
#else
#include <signal.h>
#endif
#include <time.h>
#include <gcrypt.h>
#include <gnutls/gnutls.h>
#include <gnutls/x509.h>
#include <dexter-tls.h>
#include <dexter-alias.h>

#define DH_BITS 1024

#ifdef HAVE_MSG_NOSIGNAL
               /* replacement for send() used by gnutls push, to avoid EPIPE signal */
static ssize_t _send_msg_nosignal    (gnutls_transport_ptr_t  ptr, 
                                      const void             *buf, 
                                      size_t                  buflen);
#endif

/* We will discard and regenerate the dh_params once a day */ 
static gnutls_dh_params_t _dh_params;
static gboolean           _dh_params_initialized = FALSE;
G_LOCK_DEFINE_STATIC (_dh_params_mutex);
G_LOCK_DEFINE_STATIC (_tls_lib_mutex);


/* GCRY_THREAD_OPTION_PTHREAD_IMPL uses malloc/free, so redirect them to glib versions */
#define malloc(x) g_malloc(x)
#define free(x) g_free(x)
GCRY_THREAD_OPTION_PTHREAD_IMPL;

struct _dexter_tls_t
{
  gboolean role;
  gnutls_session_t session;
  gnutls_certificate_credentials_t cred;
  gnutls_dh_params_t dh_params; /* server-side only */
  guint flags;
};


#ifdef HAVE_MSG_NOSIGNAL
static ssize_t
_send_msg_nosignal (gnutls_transport_ptr_t  ptr,
                    const void             *buf,
                    size_t                  buflen)
{
  ssize_t ret = send ( (int)ptr, buf, buflen, MSG_NOSIGNAL);
  if (ret < 0)
  {
    switch (errno)
    {
      case EPIPE:
      {
        errno=ENOTCONN;
        break;
      }
      default:
        break;
    }
    g_message ("%s: %s", G_STRFUNC, g_strerror (errno));
  }
  return ret;
}
#endif


gboolean
dexter_tls_library_init (GError **error)
{
  gcry_error_t gcry_err;
  int gnutls_err;

#ifdef G_ENABLE_DEBUG
  g_message (G_STRFUNC);
#endif

  G_LOCK (_tls_lib_mutex);

  /* Init gcrypt for multi-threading */
  if ((gcry_err = gcry_control (GCRYCTL_SET_THREAD_CBS, &gcry_threads_pthread)) != 0)
  {
    g_set_error (error,
                 DEXTER_TLS_ERROR,
                 DEXTER_TLS_ERROR_FAILED,
                 "%s: error inititalizing gcrypt for multi-threading: %s",
                 G_STRFUNC, gcry_strerror (gcry_err));
    G_UNLOCK (_tls_lib_mutex);
    return FALSE;
  }

  /* Check that gnutls headers are compatibile with actual library version */
  if (gnutls_check_version (LIBGNUTLS_VERSION) == NULL)
  {
    g_set_error (error,
                 DEXTER_TLS_ERROR,
                 DEXTER_TLS_ERROR_FAILED,
                 "%s: error: gnutls version check failed", G_STRFUNC);
    G_UNLOCK (_tls_lib_mutex);
    return FALSE;
  }

  if ((gnutls_err = gnutls_global_init ()) != 0)
  {
    g_set_error (error,
                 DEXTER_TLS_ERROR,
                 DEXTER_TLS_ERROR_FAILED,
                 "%s: error initializing gnutls: %s",
                 G_STRFUNC, gnutls_strerror (gnutls_err));
    G_UNLOCK (_tls_lib_mutex);
    return FALSE;
  }

  G_UNLOCK (_tls_lib_mutex);

#ifndef HAVE_MSG_NOSIGNAL
  /* if we cannot replace send() with _send_msg_nosignal() due to missing MSG_NOSIGNAL, \
   * go the old-fashioned route and simply ignore SIGPIPE. */
  signal (SIGPIPE, SIG_IGN);
#endif

  return (errno==0);
}


void
dexter_tls_server_seed (void)
{
#ifdef G_ENABLE_DEBUG
  g_message (G_STRFUNC);
#endif
  /* Generate dh params for certificate servers */
  G_LOCK (_dh_params_mutex);

  if (!_dh_params_initialized)
  {
    gnutls_dh_params_init (&_dh_params);
    _dh_params_initialized=TRUE;
  }

  gnutls_dh_params_generate2 (_dh_params, DH_BITS);

  G_UNLOCK (_dh_params_mutex);
}


dexter_tls_t *
dexter_tls_new (gboolean      role,
                const gchar  *cert_file,
                const gchar  *key_file,
                const gchar  *trust_file,
                const gchar  *crl_file,
                guint         flags,
                GError      **error)
{
#ifdef G_ENABLE_DEBUG
  g_message (G_STRFUNC);
#endif

  dexter_tls_t *tls;
  int           gnutls_error;

  g_return_val_if_fail (error == NULL || *error == NULL, FALSE);

  /* Require absolute paths on any files specified. */
  if (cert_file && *cert_file && !g_path_is_absolute (cert_file))
  {
    g_set_error (error,
                 DEXTER_TLS_ERROR,
                 DEXTER_TLS_ERROR_FILE,
                 "%s: %s is not an absolute path",
                 G_STRFUNC, cert_file);
    return NULL;
  }
  if (key_file && *key_file && !g_path_is_absolute (key_file))
  {
    g_set_error (error,
                 DEXTER_TLS_ERROR,
                 DEXTER_TLS_ERROR_FILE,
                 "%s: %s is not an absolute path",
                 G_STRFUNC, key_file);
    return NULL;
  }
  if (trust_file && *trust_file && !g_path_is_absolute (trust_file))
  {
    g_set_error (error,
                 DEXTER_TLS_ERROR,
                 DEXTER_TLS_ERROR_FILE,
                 "%s: %s is not an absolute path",
                 G_STRFUNC, trust_file);
    return NULL;
  }
  if (crl_file && *crl_file && !g_path_is_absolute (crl_file))
  {
    g_set_error (error,
                 DEXTER_TLS_ERROR,
                 DEXTER_TLS_ERROR_FILE,
                 "%s: %s is not an absolute path",
                 G_STRFUNC, crl_file);
    return NULL;
  }

  tls = g_slice_new0 (dexter_tls_t);
  tls->role = role;
  tls->flags = flags;

  /* Init gnutls_session_t */
  if ((gnutls_error = 
       gnutls_init (&tls->session, (role==DEXTER_TLS_ROLE_CLIENT) ? GNUTLS_CLIENT : GNUTLS_SERVER)) != 0)
  {
    g_set_error (error,
                 DEXTER_TLS_ERROR,
                 DEXTER_TLS_ERROR_FAILED,
                 "%s: error initializing TLS session: %s",
                 G_STRFUNC, gnutls_strerror (gnutls_error));
     gnutls_deinit (tls->session);
     g_slice_free (dexter_tls_t, tls);
    return NULL;
  }

  /* Set default priority on gnutls_session_t */
  if ((gnutls_error = gnutls_set_default_priority (tls->session)) != 0)
  {
    g_set_error (error,
                 DEXTER_TLS_ERROR,
                 DEXTER_TLS_ERROR_FAILED,
                 "%s: error setting TLS session priority: %s",
                 G_STRFUNC, gnutls_strerror (gnutls_error));
    gnutls_deinit (tls->session);
    g_slice_free (dexter_tls_t, tls);
    return NULL;
  }

  /* Init gnutls_certificate_credentials_t */
  if ((gnutls_error = 
       gnutls_certificate_allocate_credentials (&tls->cred)) < 0)
  {
    g_set_error (error,
                 DEXTER_TLS_ERROR,
                 DEXTER_TLS_ERROR_FAILED,
                 "%s: error initializing TLS credentials: %s",
                 G_STRFUNC, gnutls_strerror (gnutls_error));
    gnutls_deinit (tls->session);
    gnutls_certificate_free_credentials (tls->cred);
    g_slice_free (dexter_tls_t, tls);
    return NULL;
  }

  /* Set x509 trust file */
  if (trust_file && *trust_file)
  {
    /* On success, returns number of certs processed, so we want a non-negative back */
    if ((gnutls_error =
         gnutls_certificate_set_x509_trust_file (tls->cred, trust_file, GNUTLS_X509_FMT_PEM)) <= 0)
    {
      g_set_error (error,
                   DEXTER_TLS_ERROR,
                   DEXTER_TLS_ERROR_FILE,
                   "%s: error setting X.509 trust file: %s",
                   G_STRFUNC, gnutls_strerror (gnutls_error));
      gnutls_deinit (tls->session);
      gnutls_certificate_free_credentials (tls->cred);
      g_slice_free (dexter_tls_t, tls);
      return NULL;
    }
  }

  /* Set x509 cert revocation list file */
  if (crl_file && *crl_file)
  {
    /* On success, returns number of certs processed, so we want a non-negative back */
    if ((gnutls_error =
         gnutls_certificate_set_x509_crl_file (tls->cred, crl_file, GNUTLS_X509_FMT_PEM)) <= 0)
    {
      g_set_error (error,
                   DEXTER_TLS_ERROR,
                   DEXTER_TLS_ERROR_FILE,
                   "%s: error setting X.509 crl file: %s",
                   G_STRFUNC, gnutls_strerror (gnutls_error));
      gnutls_deinit (tls->session);
      gnutls_certificate_free_credentials (tls->cred);
      g_slice_free (dexter_tls_t, tls);
      return NULL;
    }
  }

  /* Set x509 key and cert files */
  if (key_file && *key_file && cert_file && *cert_file)
  {
    if ((gnutls_error =
         gnutls_certificate_set_x509_key_file (tls->cred, cert_file, key_file, GNUTLS_X509_FMT_PEM)) < 0)
    {
      g_set_error (error,
                   DEXTER_TLS_ERROR,
                   DEXTER_TLS_ERROR_FILE,
                   "%s: error setting X.509 key/cert file(s): %s",
                   G_STRFUNC, gnutls_strerror (gnutls_error));
      gnutls_deinit (tls->session);
      gnutls_certificate_free_credentials (tls->cred);
      g_slice_free (dexter_tls_t, tls);
      return NULL;
    }
  }

  /* Put x509 credentials on gnutls_session_t */
  if ((gnutls_error = gnutls_credentials_set (tls->session, GNUTLS_CRD_CERTIFICATE, tls->cred)) < 0)
  {
    g_set_error (error,
                 DEXTER_TLS_ERROR,
                 DEXTER_TLS_ERROR_FAILED,
                 "%s: error setting credentials for TLS session: %s",
                  G_STRFUNC, gnutls_strerror (gnutls_error));
    gnutls_deinit (tls->session);
    gnutls_certificate_free_credentials (tls->cred);
    g_slice_free (dexter_tls_t, tls);
    return NULL;
  }

  /* Server-side only */
  if (role==DEXTER_TLS_ROLE_SERVER)
  {
    /* Set up cert request from client */
    if (flags & DEXTER_TLS_CERT_REQUEST)
      gnutls_certificate_server_set_request (tls->session, GNUTLS_CERT_REQUEST);
    else if (flags & DEXTER_TLS_CERT_REQUIRE)
      gnutls_certificate_server_set_request (tls->session, GNUTLS_CERT_REQUIRE);

    /* Set number of prime bits */
    gnutls_dh_set_prime_bits (tls->session, DH_BITS);

    /* Init private copy of dh params */
    gnutls_dh_params_init (&tls->dh_params);
    G_LOCK (_dh_params_mutex);
    gnutls_dh_params_cpy (tls->dh_params, _dh_params);
    G_UNLOCK (_dh_params_mutex);
    gnutls_certificate_set_dh_params (tls->cred, tls->dh_params);
  }

  return tls;
}


gboolean
dexter_tls_handshake (dexter_tls_t *tls,
                      int           sockfd,
                      GError      **error)
{
  int gnutls_error;

#ifdef G_ENABLE_DEBUG
  g_message (G_STRFUNC);
#endif

  g_return_val_if_fail (tls != NULL, FALSE);
  g_return_val_if_fail (error == NULL || *error == NULL, FALSE);

  gnutls_transport_set_ptr (tls->session, (gnutls_transport_ptr_t) sockfd);
#ifdef HAVE_MSG_NOSIGNAL
  gnutls_transport_set_push_function (tls->session, _send_msg_nosignal);
#endif

  do
  {
    gnutls_error = gnutls_handshake (tls->session);
  }
  while ((gnutls_error == GNUTLS_E_AGAIN) || (gnutls_error == GNUTLS_E_INTERRUPTED));

  if (gnutls_error < 0)
  {
    g_set_error (error,
                 DEXTER_TLS_ERROR,
                 DEXTER_TLS_ERROR_FAILED,
                 "%s: error occurred during handshake: %s",
                 G_STRFUNC, gnutls_strerror (gnutls_error));
    return FALSE;
  }

  return TRUE;
}


gboolean
dexter_tls_verify_peer (dexter_tls_t *tls,
                        const gchar  *canonical_hostname,
                        GError      **error)
{
  const gnutls_datum_t *cert_list;
  unsigned int cert_list_size;
  unsigned int status;
  int gnutls_error;
  int iter;
  gnutls_x509_crt_t cert;
  time_t activation_time, expiration_time, current_time;

#ifdef G_ENABLE_DEBUG
  g_message (G_STRFUNC);
#endif

  g_return_val_if_fail (tls != NULL, FALSE);
  g_return_val_if_fail (error == NULL || *error == NULL, FALSE);

  /* Get peer's cert list */
  if ((cert_list = gnutls_certificate_get_peers (tls->session, &cert_list_size)) == NULL)
  {
    /* No cert was sent.  Check if that's ok */
    if ((tls->role == DEXTER_TLS_ROLE_CLIENT) || (tls->flags & DEXTER_TLS_CERT_REQUIRE))
    {
      g_set_error (error,
                   DEXTER_TLS_ERROR,
                   DEXTER_TLS_ERROR_CERTIFICATE,
                   "%s: error peer sent no certificate",
                   G_STRFUNC);
      return FALSE;
    }
    
    /* No cert is ok.  */
    return TRUE;
  }

  /* We have a cert list from the peer.  Verify it. */
  gnutls_certificate_set_verify_flags (tls->cred, GNUTLS_VERIFY_ALLOW_X509_V1_CA_CRT);
  if ((gnutls_error = gnutls_certificate_verify_peers2 (tls->session, &status)) != 0)
  {
    g_set_error (error,
                 DEXTER_TLS_ERROR,
                 DEXTER_TLS_ERROR_CERTIFICATE,
                 "%s: error verifying peer certificate: %s",
                 G_STRFUNC, gnutls_strerror (gnutls_error));
    return FALSE;
  }

  /* Check if invalid */
  if (status & GNUTLS_CERT_INVALID)
  {
    g_set_error (error,
                 DEXTER_TLS_ERROR,
                 DEXTER_TLS_ERROR_CERTIFICATE,
                 "%s: error invalid peer certificate",
                 G_STRFUNC);
    return FALSE;
  }

  /* Check if issuer found */
  if (status & GNUTLS_CERT_SIGNER_NOT_FOUND)
  {
    g_set_error (error,
                 DEXTER_TLS_ERROR,
                 DEXTER_TLS_ERROR_CERTIFICATE,
                 "%s: error peer certificate hasn't got a known issuer",
                 G_STRFUNC);
    return FALSE;
  }

  /* Check if revoked */
  if (status & GNUTLS_CERT_REVOKED)
  {
    g_set_error (error,
                 DEXTER_TLS_ERROR,
                 DEXTER_TLS_ERROR_CERTIFICATE,
                 "%s: error peer certificate has been revoked",
                 G_STRFUNC);
    return FALSE;
  }

  /* Check if X.509 cert */
  if (gnutls_certificate_type_get (tls->session) != GNUTLS_CRT_X509)
  {
    g_set_error (error,
                 DEXTER_TLS_ERROR,
                 DEXTER_TLS_ERROR_CERTIFICATE,
                 "%s: error peer certificate is not type X.509",
                 G_STRFUNC);
    return FALSE;
  }

  /* gnutls_certificate_verify_peers2 does not check activation/expiration times
   * on the cert chain, so we have to do that manually.  also, the first cert's
   * hostname must match our canonical name, so we know that the cert is coming 
   * from the peer we are connected to. */
  for (iter=0; iter < cert_list_size; iter++)
  {
    /* Init cert */
    if (gnutls_x509_crt_init (&cert) != 0)
    {
      g_set_error (error,
                   DEXTER_TLS_ERROR,
                   DEXTER_TLS_ERROR_CERTIFICATE,
                   "%s: error unable to init certificate",
                   G_STRFUNC);
      return FALSE;
    }

    /* Import cert */
    if (gnutls_x509_crt_import (cert, cert_list+iter, GNUTLS_X509_FMT_DER) != 0)
    {
      g_set_error (error,
                   DEXTER_TLS_ERROR,
                   DEXTER_TLS_ERROR_CERTIFICATE,
                   "%s: error unable to import certificate",
                   G_STRFUNC);

     gnutls_x509_crt_deinit (cert);
     return FALSE;

    }

    /* If canonical_hostname supplied, check hostname on first certificate. */
    if (canonical_hostname && (iter==0) && 
        (gnutls_x509_crt_check_hostname (cert, canonical_hostname) == 0))
    {
      g_set_error (error,
                   DEXTER_TLS_ERROR,
                   DEXTER_TLS_ERROR_CERTIFICATE,
                   "%s: error certificate hostname does not match",
                   G_STRFUNC);

      gnutls_x509_crt_deinit (cert);
      return FALSE;
    }

    /* Fetch cert activation time */
    if ((activation_time = gnutls_x509_crt_get_activation_time (cert)) < 0)
    {
      g_set_error (error,
                   DEXTER_TLS_ERROR,
                   DEXTER_TLS_ERROR_CERTIFICATE,
                   "%s: error unable to fetch certificate activation time",
                   G_STRFUNC);

      gnutls_x509_crt_deinit (cert);
      return FALSE;
    }

    /* Fetch cert expiration time */
    if ((expiration_time = gnutls_x509_crt_get_expiration_time (cert)) < 0)
    {
      g_set_error (error,
                   DEXTER_TLS_ERROR,
                   DEXTER_TLS_ERROR_CERTIFICATE,
                   "%s: error unable to fetch certificate expiration time",
                   G_STRFUNC);

      gnutls_x509_crt_deinit (cert);
      return FALSE;
    }  

    /* Get current time */
    if ((current_time = time (0)) < 0)
    {
      g_set_error (error,
                   DEXTER_TLS_ERROR,
                   DEXTER_TLS_ERROR_FAILED,
                   "%s: error unable to get system time",
                   G_STRFUNC);

      gnutls_x509_crt_deinit (cert);
      return FALSE;
    }

    /* Check activation time */
    if (activation_time > current_time)
    {
      g_set_error (error,
                   DEXTER_TLS_ERROR,
                   DEXTER_TLS_ERROR_CERTIFICATE,
                   "%s: error certificate is not yet activated",
                   G_STRFUNC);

      gnutls_x509_crt_deinit (cert);
      return FALSE;
    }

    /* Check expiration time */
    if (expiration_time < current_time)
    {
      g_set_error (error,
                   DEXTER_TLS_ERROR,
                   DEXTER_TLS_ERROR_CERTIFICATE,
                   "%s: error certificate has expired",
                   G_STRFUNC);

      gnutls_x509_crt_deinit (cert);
      return FALSE;
    }

    /* De-init cert */
    gnutls_x509_crt_deinit (cert);
  }

  /* Peer verified */
  return TRUE;
}


ssize_t
dexter_tls_recv (dexter_tls_t *tls,
                 void         *buffer,
                 size_t        bufsize,
                 GError      **error)
{
  ssize_t sz;

  g_return_val_if_fail (tls != NULL, -1);
  g_return_val_if_fail (buffer != NULL, -1);
  g_return_val_if_fail (bufsize >= 0, -1);
  g_return_val_if_fail (error == NULL || *error == NULL, -1);

  sz = gnutls_record_recv (tls->session, buffer, bufsize);
  if (sz == GNUTLS_E_INTERRUPTED)
  {
    g_set_error (error,
                 DEXTER_TLS_ERROR,
                 DEXTER_TLS_ERROR_EINTR,
                 "%s: a signal occurred before any data was delivered",
                 G_STRFUNC);
    return -1;
  }
  if (sz == GNUTLS_E_AGAIN)
  {
    g_set_error (error,
                 DEXTER_TLS_ERROR,
                 DEXTER_TLS_ERROR_EAGAIN,
                 "%s: operation would block",
                 G_STRFUNC);
    return -1;
  }
  if (sz < 0)
  {
    g_set_error (error,
                 DEXTER_TLS_ERROR,
                 DEXTER_TLS_ERROR_FAILED,
                 "%s: error occurred during recv: %s",
                 G_STRFUNC, gnutls_strerror (sz));
    return -1;
  }
  
  return sz;
}


ssize_t
dexter_tls_send (dexter_tls_t *tls,
                 const void   *buffer,
                 size_t        bufsize,
                 GError      **error)
{
  ssize_t sz;

  g_return_val_if_fail (tls != NULL, -1);
  g_return_val_if_fail (buffer != NULL, -1);
  g_return_val_if_fail (bufsize >= 0, -1);
  g_return_val_if_fail (error == NULL || *error == NULL, -1);

  sz = gnutls_record_send (tls->session, buffer, bufsize);
  if (sz == GNUTLS_E_INTERRUPTED)
  {
    g_set_error (error,
                 DEXTER_TLS_ERROR,
                 DEXTER_TLS_ERROR_EINTR,
                 "%s: a signal occurred before any data was transmitted",
                 G_STRFUNC);
    return -1;
  }
  if (sz == GNUTLS_E_AGAIN)
  {
    g_set_error (error,
                 DEXTER_TLS_ERROR,
                 DEXTER_TLS_ERROR_EAGAIN,
                 "%s: operation would block",
                 G_STRFUNC);
    return -1;
  }
  if (sz < 0)
  {
    g_set_error (error,
                 DEXTER_TLS_ERROR,
                 DEXTER_TLS_ERROR_FAILED,
                 "%s: error occurred during send: %s",
                 G_STRFUNC, gnutls_strerror (sz));
    return -1;
  }

  return sz;
}


void
dexter_tls_close (dexter_tls_t *tls)
{
  int gnutls_error = 0;

  g_return_if_fail (tls != NULL);

#ifdef G_ENABLE_DEBUG
  g_message (G_STRFUNC);
#endif

  do
  {
    /* GNUTLS_SHUT_WR instead of GNUTLS_SHUT_RDWR to avoid peer waits */
    gnutls_bye (tls->session, GNUTLS_SHUT_WR);
  }
  while ((gnutls_error == GNUTLS_E_AGAIN) || (gnutls_error == GNUTLS_E_INTERRUPTED));
}


void 
dexter_tls_free (dexter_tls_t *tls)
{
#ifdef G_ENABLE_DEBUG
  g_message (G_STRFUNC);
#endif
  
  if (G_UNLIKELY (!tls))
    return;

  gnutls_deinit (tls->session);
  gnutls_certificate_free_credentials (tls->cred);
  gnutls_dh_params_deinit (tls->dh_params);
  g_slice_free (dexter_tls_t, tls);
}


void
dexter_tls_library_exit (void)
{
#ifdef G_ENABLE_DEBUG
  g_message (G_STRFUNC);
#endif

  G_LOCK (_tls_lib_mutex);

  /* de-init dh params */
  G_LOCK (_dh_params_mutex);
  if (_dh_params_initialized)
    gnutls_dh_params_deinit (_dh_params);
  G_UNLOCK (_dh_params_mutex);

  /* de-init Gnutls lib */
  gnutls_global_deinit ();
  G_UNLOCK (_tls_lib_mutex);

}

#define __DEXTER_TLS_GNUTLS_C__
#include <dexter-aliasdef.c>
