/* $Id: dexter.c 393 2012-03-06 23:36:44Z pkovacs $ */
/*
 * libdexter - a plugin-based, distributed sampling library.
 * Copyright (C) 2006-2012 Philip Kovacs <pkovacs@users.sourceforge.net>.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 2.1
 * of the License, or (at your option) any later version.

 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.

 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
 * 02110-1301  USA
 *
*/

/**
 * SECTION:dexter
 * @title: Initialization
 * @short_description: Initialization and cleanup of Libdexter
 * @see_also: 
 *
 * Libdexter must be initialized with dexter_init() prior to use.
 *
 * This function causes Libdexter to read the configuration file; 
 * load the specified plugins and register authorized services; 
 * initialize the thread pool and set up the interfaces needed 
 * for plugins to interact with the library.
 *
 * When you are done using Libdexter, you should issue dexter_exit()
 * to free up the resources it uses; shutdown its thread pool, etc.
 *
 * Example of how to initialize and shutdown Libdexter:
 * <informalexample>
 * <programlisting>
 * #<!-- -->include &lt;dexter.h&gt;
 * <!-- -->
 * GError    *local_error = NULL;
 * <!-- -->
 * /<!-- -->* Init the library from a default configuration file. *<!-- -->/
 * dexter_init (NULL, &amp;local_error);
 * if (local_error) {
 *     /<!-- -->* deal with error *<!-- -->/
 *     g_error ("failed to init libdexter: %<!-- -->s", local_error->message);
 * }
 * <!-- -->
 * /<!-- -->* ----------------
 *    Use the library.
 *    ---------------- *<!-- -->/
 * ...
 * <!-- -->
 * /<!-- -->* Release library resources. *<!-- -->/
 * dexter_exit<!-- -->();
 * ...
 * </programlisting>
 * </informalexample>
 **/

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <glib/gstdio.h>
#include <glib-object.h>
#include <dexter.h>
#include <dexter-log.h>
#include <dexter-private.h>
#ifdef HAVE_TLS
#include <dexter-tls.h>
#endif
#include <dexter-alias.h>

#define DEFAULT_DEXTER_CONFIG1  ".dexterrc"
#define DEFAULT_DEXTER_CONFIG2  "/etc/libdexter/dexter.conf"

static gint dexter_set_state (gint state);
static volatile gint _dexter_state = DEXTER_STATE_DOWN;


GQuark
dexter_error_quark (void)
{
  static GQuark error_quark = 0;

  if (error_quark == 0)
    error_quark = g_quark_from_static_string ("dexter-error-quark");

  return error_quark;
}


/**
 * dexter_init:
 * @config:  full path to config file or %NULL for library defaults
 * @error:  return location for optional #GError
 *
 * Initialize libdexter using the indicated @config file.
 * If %NULL is specified for @config, the library will try to use:
 *
 * <filename>$HOME/.dexterrc</filename> or <filename>/etc/libdexter/dexter.conf</filename>, in that order.
 *
 * Returns: a #gboolean indicating success or failure.
 **/
gboolean
dexter_init (const gchar  *config,
             GError      **error)
{
  gchar  *tmp_filename = NULL;
  GError *local_error = NULL;
      
  g_return_val_if_fail (error == NULL || *error == NULL, FALSE);
  g_assert (dexter_get_state () == DEXTER_STATE_DOWN);

#ifdef G_ENABLE_DEBUG
  g_message (G_STRFUNC);
#endif

  dexter_set_state (DEXTER_STATE_STARTING);

  /* Init glib GType system */
  g_type_init ();
    
  /* if no config provided, try defaults */
  if (!config)
  {
      tmp_filename = g_build_filename ( g_get_home_dir (), DEFAULT_DEXTER_CONFIG1, NULL );
      if (!g_file_test (tmp_filename, G_FILE_TEST_EXISTS ))
      {
        g_free (tmp_filename);
        tmp_filename = g_build_filename ( DEFAULT_DEXTER_CONFIG2, NULL );
        if (!g_file_test (tmp_filename, G_FILE_TEST_EXISTS))
        {
          g_set_error (error,
                       DEXTER_ERROR,
                       DEXTER_ERROR_NO_CONFIG_FILE,
                       "%s: no config file found at $HOME/%s or %s",
                       G_STRFUNC,
                       DEFAULT_DEXTER_CONFIG1,
                       DEFAULT_DEXTER_CONFIG2);
          g_free (tmp_filename);
          dexter_set_state (DEXTER_STATE_DOWN);
          return FALSE;
        }
      }
  }
      
  /* Init logging system */
  dexter_log_init ();

  /* Init private objects  */
  dexter_private_init (config ? config : tmp_filename, &local_error);
  if (local_error)
  {
    g_free (tmp_filename);
    g_propagate_error (error, local_error);
    dexter_set_state (DEXTER_STATE_DOWN);
    return FALSE;
  }
  g_free (tmp_filename);
  
#ifdef HAVE_TLS
  /* Init TLS library */
  dexter_tls_library_init (&local_error);
  if (local_error)
  {
    local_error->domain = DEXTER_ERROR;
    local_error->code = DEXTER_ERROR_FAILED;
    g_propagate_error (error, local_error);
    return FALSE;
  }
#endif

  dexter_set_state (DEXTER_STATE_UP);

  return TRUE;
}


/**
 * dexter_exit:
 *
 * Shuts down libdexter and frees its resources.
 *
 * Note: Do not issue dexter_exit () from within any asynchronous callback generated by the library.
 **/
void  
dexter_exit (void)
{
#ifdef G_ENABLE_DEBUG
  g_message (G_STRFUNC);
#endif

  g_assert (dexter_get_state () == DEXTER_STATE_UP);

  dexter_set_state (DEXTER_STATE_STOPPING);

  dexter_private_exit();
 
#ifdef HAVE_TLS
  dexter_tls_library_exit ();
#endif

  dexter_log_exit ();

  dexter_set_state (DEXTER_STATE_DOWN);
}


static gint 
dexter_set_state (gint state)
{
  gint old_state;

  old_state = g_atomic_int_get (&_dexter_state);
  g_atomic_int_set (&_dexter_state, state);

  return old_state;
}

  
/**
 * dexter_get_state:
 *
 * Returns the current state of the Libdexter library.
 *
 * Returns: #gint from #DexterState indicating current state of the library.
 *
 * Since: 0.1.3
 **/
gint
dexter_get_state (void)
{
  return g_atomic_int_get (&_dexter_state);
}

#define __DEXTER_C__
#include <dexter-aliasdef.c>
