/* $Id: dexterplugin-iconfig.c 393 2012-03-06 23:36:44Z pkovacs $ */
/*
 * libdexter - a plugin-based, distributed sampling library.
 * Copyright (C) 2006-2012 Philip Kovacs <pkovacs@users.sourceforge.net>.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 2.1
 * of the License, or (at your option) any later version.

 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.

 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
 * 02110-1301  USA
 *
*/

/**
 * SECTION:dexterplugin-iconfig
 * @title: DexterPluginIConfig
 * @short_description: Interface to the Libdexter configuration file
 * @see_also:
 *
 * Plugins can access the configuration file from this interface.
 *
 * Example demonstrating how plugins can access the configuration file:
 * <informalexample>
 * <programlisting>
 * #<!-- -->include &lt;dexterplugin.h&gt;
 * <!-- -->
 * #define MY_SERVICE_UUID "86127315-6de6-43cd-a835-fbd25476390d"
 * <!-- -->
 * void some_plugin_function ()
 * {
 *   gchar *important_key;
 * <!-- -->
 *   /<!-- -->* Obtain key "Important" in config file section [86127315-6de6-43cd-a835-fbd25476390d]. *<!-- -->/
 *   if (!dexterplugin_iconfig_get_string (DEXTER_ICONFIG, 
 *                                         MY_SERVICE_UUID, "Important",
 *                                         &amp;important_key, NULL)) 
 *   {
 *       g_warning ("Important key not found");
 *       ...
 *   }
 *   /<!-- -->* use important_key and when done, free it. *<!-- -->/
 *   g_free (important_key);
 * }
 * </programlisting>
 * </informalexample>
 **/

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <dexterplugin-iconfig.h>
#include <dexterplugin-alias.h>

static void dexterplugin_iconfig_base_init (gpointer g_class);

static void
dexterplugin_iconfig_base_init (gpointer g_class)
{
  static gboolean initialized = FALSE;

  if (!initialized) 
  {
    initialized = TRUE;
  }
}

GType
dexterplugin_iconfig_get_type (void)
{
  static GType type_id = G_TYPE_INVALID;

  if (G_UNLIKELY (type_id == G_TYPE_INVALID))
  {
    static const GTypeInfo type_info =
    {
      sizeof (DexterPluginIConfigInterface),
      dexterplugin_iconfig_base_init,
      NULL,
      NULL,
      NULL,
      NULL,
      0,
      0,
      NULL,
      NULL,
    };

    type_id = g_type_register_static (G_TYPE_INTERFACE, "DexterPluginIConfig", &type_info, 0);
    g_type_interface_add_prerequisite (type_id, G_TYPE_OBJECT);
  }

  return type_id;
}


/**
 * dexterplugin_iconfig_has_group:
 * @iconfig: a #DexterPluginIConfig.
 * @group: #gchar group name within the config file.
 *
 * Returns whether or not the specified group exists in the config file.
 *
 * Returns: #gboolean indicating if group is in the config file or not.
 **/
gboolean 
dexterplugin_iconfig_has_group (DexterPluginIConfig *iconfig,
                                const gchar         *group)
{
  g_return_val_if_fail (DEXTERPLUGIN_IS_ICONFIG (iconfig), FALSE);
  return DEXTERPLUGIN_ICONFIG_GET_INTERFACE (iconfig) -> has_group (iconfig, group);
}


/**
 * dexterplugin_iconfig_get_groups:
 * @iconfig: a #DexterPluginIConfig.
 * @return_groups: address of a (#gchar **) return pointer to hold list of groups.
 * @length: address of optional #gsize to hold number of groups returned, or %NULL.
 *
 * Sets @return_groups to a list of all groups in the config file.
 *
 * Use g_strfreev() to free *return_groups.
 *
 * Returns: #gboolean indicating success or failure.
 **/
gboolean               
dexterplugin_iconfig_get_groups (DexterPluginIConfig   *iconfig,
                                 gchar               ***return_groups,
                                 gsize                 *length)
{
  g_return_val_if_fail (DEXTERPLUGIN_IS_ICONFIG (iconfig), FALSE);
  return DEXTERPLUGIN_ICONFIG_GET_INTERFACE (iconfig) -> get_groups (iconfig, return_groups, length);
}


/**
 * dexterplugin_iconfig_get_keys:
 * @iconfig: a #DexterPluginIConfig.
 * @group: #gchar group name within the config file.
 * @return_keys: address of a (#gchar **) return pointer to hold list of keys.
 * @length: address of optional #gsize to hold number of keys returned, or %NULL.
 * @error: return location for optional #GError.
 *
 * Sets @return_keys to a list of all keys in the config file for the specified group.
 *
 * Use g_strfreev() to free *return_keys.
 *
 * Returns: #gboolean indicating success or failure.
 **/
gboolean
dexterplugin_iconfig_get_keys (DexterPluginIConfig   *iconfig,
                               const gchar           *group,
                               gchar               ***return_keys,
                               gsize                 *length,
                               GError               **error)
{
  g_return_val_if_fail (DEXTERPLUGIN_IS_ICONFIG (iconfig), FALSE);
  return DEXTERPLUGIN_ICONFIG_GET_INTERFACE (iconfig) -> get_keys (iconfig, group, return_keys, length, error);
}


/**
 * dexterplugin_iconfig_get_string:
 * @iconfig: a #DexterPluginIConfig.
 * @group: #gchar group name within the config file.
 * @key: #gchar key name within the config file.
 * @return_string: address of a (#gchar *) return pointer to hold key value.
 * @error: return location for optional #GError.
 *
 * Sets @return_string to the key value in the config file for the specified group and key.
 *
 * Use g_free() to free *return_string.
 *
 * Returns: #gboolean indicating success or failure.
 **/
gboolean
dexterplugin_iconfig_get_string (DexterPluginIConfig  *iconfig,
                                 const gchar          *group,
                                 const gchar          *key,
                                 gchar               **return_string,
                                 GError              **error)
{
  g_return_val_if_fail (DEXTERPLUGIN_IS_ICONFIG (iconfig), FALSE);
  return DEXTERPLUGIN_ICONFIG_GET_INTERFACE (iconfig) -> get_string (iconfig, group, key, return_string, error);
}


/**
 * dexterplugin_iconfig_get_string_list:
 * @iconfig: a #DexterPluginIConfig.
 * @group: #gchar group name within the config file.
 * @key: #gchar key name within the config file.
 * @return_strings: address of a (#gchar **) return pointer to hold key value list.
 * @length: address of optional #gsize to hold number of strings returned, or %NULL.
 * @error: return location for optional #GError.
 *
 * Sets @return_strings to a key value list in the config file for the specified group and key.
 *
 * Use g_strfreev() to free *return_strings.
 *
 * Returns: #gboolean indicating success or failure.
 **/
gboolean
dexterplugin_iconfig_get_string_list (DexterPluginIConfig   *iconfig,
                                      const gchar           *group,
                                      const gchar           *key,
                                      gchar               ***return_strings,
                                      gsize                 *length,
                                      GError               **error)
{
  g_return_val_if_fail (DEXTERPLUGIN_IS_ICONFIG (iconfig), FALSE);
  return DEXTERPLUGIN_ICONFIG_GET_INTERFACE (iconfig) -> get_string_list (iconfig, 
                                                                          group, key, return_strings, length, 
                                                                          error);
}


/**
 * dexterplugin_iconfig_get_boolean:
 * @iconfig: a #DexterPluginIConfig.
 * @group: #gchar group name within the config file.
 * @key: #gchar key name within the config file.
 * @return_boolean: address of a #gboolean to hold key value.
 * @error: return location for optional #GError.
 *
 * Sets @return_boolean to the key value in the config file for the specified group and key.
 *
 * Returns: #gboolean indicating success or failure.
 **/
gboolean
dexterplugin_iconfig_get_boolean (DexterPluginIConfig  *iconfig,
                                  const gchar          *group,
                                  const gchar          *key,
                                  gboolean             *return_boolean,
                                  GError              **error)
{
  g_return_val_if_fail (DEXTERPLUGIN_IS_ICONFIG (iconfig), FALSE);
  return DEXTERPLUGIN_ICONFIG_GET_INTERFACE (iconfig) -> get_boolean (iconfig, 
                                                                      group, key, return_boolean, error);
}


/**
 * dexterplugin_iconfig_get_boolean_list:
 * @iconfig: a #DexterPluginIConfig.
 * @group: #gchar group name within the config file.
 * @key: #gchar key name within the config file.
 * @return_booleans: address of a (#gboolean *) return pointer to hold key value list.
 * @length: address of optional #gsize to hold number of booleans returned, or %NULL.
 * @error: return location for optional #GError.
 *
 * Sets @return_booleans to a key value list in the config file for the specified group and key.
 *
 * Use g_free() to free *return_booleans.
 *
 * Returns: #gboolean indicating success or failure.
 **/
gboolean
dexterplugin_iconfig_get_boolean_list (DexterPluginIConfig  *iconfig,
                                       const gchar          *group,
                                       const gchar          *key,
                                       gboolean            **return_booleans,
                                       gsize                *length,
                                       GError              **error)
{
  g_return_val_if_fail (DEXTERPLUGIN_IS_ICONFIG (iconfig), FALSE);
  return DEXTERPLUGIN_ICONFIG_GET_INTERFACE (iconfig) -> get_boolean_list (iconfig,
                                                                           group, key, return_booleans,
                                                                           length, error);
}


/**
 * dexterplugin_iconfig_get_integer:
 * @iconfig: a #DexterPluginIConfig.
 * @group: #gchar group name within the config file.
 * @key: #gchar key name within the config file.
 * @return_integer: address of a #gint to hold key value.
 * @error: return location for optional #GError.
 *
 * Sets @return_integer to the key value in the config file for the specified group and key.
 *
 * Returns: #gboolean indicating success or failure.
 **/
gboolean
dexterplugin_iconfig_get_integer (DexterPluginIConfig  *iconfig,
                                  const gchar           *group,
                                  const gchar           *key,
                                  gint                  *return_integer,
                                  GError               **error)
{
  g_return_val_if_fail (DEXTERPLUGIN_IS_ICONFIG (iconfig), FALSE);
  return DEXTERPLUGIN_ICONFIG_GET_INTERFACE (iconfig) -> get_integer (iconfig,
                                                                      group, key, return_integer, error);
}


/**
 * dexterplugin_iconfig_get_integer_list:
 * @iconfig: a #DexterPluginIConfig.
 * @group: #gchar group name within the config file.
 * @key: #gchar key name within the config file.
 * @return_integers: address of a (#gint *) return pointer to hold key value list.
 * @length: address of optional #gsize to hold number of integers returned, or %NULL.
 * @error: return location for optional #GError.
 *
 * Sets @return_integers to a key value list in the config file for the specified group and key.
 *
 * Use g_free() to free *return_integers.
 *
 * Returns: #gboolean indicating success or failure.
 **/
gboolean
dexterplugin_iconfig_get_integer_list (DexterPluginIConfig  *iconfig,
                                       const gchar          *group,
                                       const gchar          *key,
                                       gint                **return_integers,
                                       gsize                *length,
                                       GError              **error)
{
  g_return_val_if_fail (DEXTERPLUGIN_IS_ICONFIG (iconfig), FALSE);
  return DEXTERPLUGIN_ICONFIG_GET_INTERFACE (iconfig) -> get_integer_list (iconfig,
                                                                           group, key, return_integers,
                                                                           length, error);
}


/**
 * dexterplugin_iconfig_key_error:
 * @iconfig: a #DexterPluginIConfig.
 * @error: a #GError from a previous operation.
 *
 * Returns whether or not the specified GError is a key error indicating
 * that the group or key was not found.
 *
 * Returns: #gboolean indicating if the specified #GError is a key error.
 **/
gboolean
dexterplugin_iconfig_key_error (DexterPluginIConfig *iconfig,
                                GError              *error)
{
  g_return_val_if_fail (DEXTERPLUGIN_IS_ICONFIG (iconfig), FALSE);
  return DEXTERPLUGIN_ICONFIG_GET_INTERFACE (iconfig) -> key_error (iconfig, error);
}

#define __DEXTERPLUGIN_ICONFIG_C__
#include <dexterplugin-aliasdef.c>
