/* $Id: dexterplugin-isampler.c 393 2012-03-06 23:36:44Z pkovacs $ */
/*
 * libdexter - a plugin-based, distributed sampling library.
 * Copyright (C) 2006-2012 Philip Kovacs <pkovacs@users.sourceforge.net>.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 2.1
 * of the License, or (at your option) any later version.

 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.

 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
 * 02110-1301  USA
 *
*/

/**
 * SECTION:dexterplugin-isampler
 * @title: DexterPluginISampler
 * @short_description: Interface that Libdexter samplers must implement
 * @see_also:
 *
 * DexterPluginISampler is the GObject interface that Libdexter plugin services must implement
 * in order to operate as 
 * <ulink url="../libdexter/libdexter-Samplers.html">Samplers</ulink> or 
 * <ulink url="../libdexter/libdexter-Timed-Samplers.html">Timed Samplers</ulink>.  The interface
 * consists of a number of functions to:
 * <itemizedlist>
 * <listitem>Initialize the sampler</listitem>
 * <listitem>Perform data sampling</listitem>
 * <listitem>Exchange data</listitem>
 * <listitem>Serialize/Deserialize data</listitem>
 * </itemizedlist>
 **/

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <dexterplugin-isampler.h>
#include <dexterplugin-sampler.h>
#include <dexterplugin-alias.h>

static void  dexterplugin_isampler_base_init (gpointer g_class);
  
static void
dexterplugin_isampler_base_init (gpointer g_class)
{
  static gboolean initialized = FALSE;

  if (!initialized) 
  {
    initialized = TRUE;
  }
}


GType
dexterplugin_isampler_get_type (void)
{
  static GType type_id = G_TYPE_INVALID;

  if (G_UNLIKELY (type_id == G_TYPE_INVALID))
  {
    static const GTypeInfo type_info =
    {
      sizeof (DexterPluginISamplerInterface),
      dexterplugin_isampler_base_init,
      NULL,
      NULL,
      NULL,
      NULL,
      0,
      0,
      NULL,
      NULL,
    };

    type_id = g_type_register_static (G_TYPE_INTERFACE, "DexterPluginISampler", &type_info, 0);
    g_type_interface_add_prerequisite (type_id, DEXTERPLUGIN_TYPE_SAMPLER);
  }

  return type_id;
}


/**
 * dexterplugin_isampler_initialize:
 * @isampler: #DexterPluginISampler object.
 * @param: address of #gint parameter or %NULL.
 * @input: address of input structure or %NULL.
 * @error: return location of optional #GError.
 *
 * Perform one-time initialization of a sampler using supplied parameter and/or input.
 * The @param integer is a "free" parameter for plugin writers in the sense that no
 * serialization code need be written for it.  The @input pointer is meant for any
 * additional, structured or otherwise complex initialization data that the sampler
 * might need.  Plugin writers must write serialization code for this parameter.
 *
 * If the sampler requires no param or input for initialization, %NULL should be specified.
 *
 * After a sampler is created, it must be initialized once, and only once, using this function.  
 * 
 * The optional @error parameter can be used to return errors to the caller, e.g. a GLib error, 
 * or an error you describe using the  %DEXTERPLUGIN_SAMPLER_ERROR_FAILED error code from the 
 * %DEXTERPLUGIN_SAMPLER_ERROR error domain.
 *
 * Returns: #gboolean indicating success or failure.
 **/
gboolean
dexterplugin_isampler_initialize (DexterPluginISampler  *isampler,
                                  const gint            *param,
                                  gconstpointer          input,
                                  GError               **error)
{
  g_return_val_if_fail (DEXTERPLUGIN_IS_ISAMPLER (isampler), FALSE);
  g_return_val_if_fail (error == NULL || *error == NULL, FALSE);
  return DEXTERPLUGIN_ISAMPLER_GET_INTERFACE (isampler) -> initialize (isampler, param, input, error);
}


/**
 * dexterplugin_isampler_sample:
 * @isampler: #DexterPluginISampler object.
 * @error: return location of optional #GError.
 *
 * Perform the sampling operation implemented by this sampler service.  Samplers must store
 * the sample data internally so that the data may be exchanged with the user using the data 
 * api's described below.  
 *
 * The optional @error parameter can be used to return errors to the caller, e.g. a GLib error, 
 * or an error you describe using the %DEXTERPLUGIN_SAMPLER_ERROR_FAILED error code from the 
 * %DEXTERPLUGIN_SAMPLER_ERROR error domain.
 *
 * Returns: #gboolean indicating success or failure.
 **/
gboolean
dexterplugin_isampler_sample (DexterPluginISampler  *isampler,
                              GError               **error)
{
  g_return_val_if_fail (DEXTERPLUGIN_IS_ISAMPLER (isampler), FALSE);
  g_return_val_if_fail (error == NULL || *error == NULL, FALSE);
  return DEXTERPLUGIN_ISAMPLER_GET_INTERFACE (isampler) -> sample (isampler, error);
}


/**
 * dexterplugin_isampler_data_new:
 * @isampler: #DexterPluginISampler object.
 * @populate: #gboolean to request populated data structure.
 *
 * The sampler should return a newly allocated sample data structure, as defined by the
 * service implementation.  If @populate is %TRUE, the data structure should be populated 
 * with the sampler's current sample data.  If %FALSE, the structure should be zero-filled.
 *
 * The sample structure should be freed when done with dexterplugin_isampler_data_free().
 *
 * Returns: #gpointer to a newly allocated sample data structure.
 **/
gpointer
dexterplugin_isampler_data_new (DexterPluginISampler *isampler,
                                gboolean              populate)
{
  g_return_val_if_fail (DEXTERPLUGIN_IS_ISAMPLER (isampler), NULL);
  return DEXTERPLUGIN_ISAMPLER_GET_INTERFACE (isampler) -> data_new (isampler, populate);
}


/**
 * dexterplugin_isampler_data_dup:
 * @isampler: #DexterPluginISampler object.
 * @source_data: address of source data to duplicate.
 *
 * The sampler should duplicate the provided sample data structure, i.e. return a newly
 * allocated (deep) copy of the structure.
 *
 * The sample structure should be freed when done with dexterplugin_isampler_data_free().
 *
 * Returns: #gpointer to a newly allocated sample data structure.
 **/
gpointer
dexterplugin_isampler_data_dup (DexterPluginISampler *isampler,
                                gconstpointer         source_data)
{
  g_return_val_if_fail (DEXTERPLUGIN_IS_ISAMPLER (isampler), NULL);
  g_return_val_if_fail (source_data != NULL, NULL);
  return DEXTERPLUGIN_ISAMPLER_GET_INTERFACE (isampler) -> data_dup (isampler, source_data);
}


/**
 * dexterplugin_isampler_data_xfer:
 * @isampler: #DexterPluginISampler object.
 * @old_data: address of optional sampler data structure to receive existing data (before replacement).
 * @new_data: address of optional sampler data structure to replace existing data.
 *
 * The sampler should first (deep) copy its current, internal sample data to the @old_data structure,
 * if specified.  Then, the sampler should replace its internal sample data with a (deep) copy of
 * the @new_data structure, if specified.  Implementors must free any existing heap-allocated
 * data on the copy targets before performing the copies, so as not to leak memory.
 *
 * Note that it would be highly unusual for most users to change the internal data of a sampler
 * explicitly by specifying a @new_data parameter.  Libdexter uses that feature internally when
 * deserializing data transmitted over the network.  The preferred way to update the sampler's
 * internal sample data is simply to take a sample, i.e. to use dexterplugin_isampler_sample().
 **/
void
dexterplugin_isampler_data_xfer (DexterPluginISampler *isampler,
                                 gpointer              old_data,
                                 gconstpointer         new_data)
{
  g_return_if_fail (DEXTERPLUGIN_IS_ISAMPLER (isampler));
  DEXTERPLUGIN_ISAMPLER_GET_INTERFACE (isampler) -> data_xfer (isampler, old_data, new_data);
}


/**
 * dexterplugin_isampler_data_copy:
 * @isampler: #DexterPluginISampler object.
 * @dest_data: address of destination sample data structure.
 * @source_data: address of source sampler data structure.
 *
 * The sampler should (deep) copy the source sample data onto the destination structure,
 * first freeing any heap-allocated data on the destination structure so as not to leak
 * memory.
 **/
void
dexterplugin_isampler_data_copy (DexterPluginISampler *isampler,
                                 gpointer              dest_data,
                                 gconstpointer         source_data)
{
  g_return_if_fail (DEXTERPLUGIN_IS_ISAMPLER (isampler));
  g_return_if_fail (dest_data != NULL);
  g_return_if_fail (source_data != NULL);
  DEXTERPLUGIN_ISAMPLER_GET_INTERFACE (isampler) -> data_copy (isampler, dest_data, source_data);
}


/**
 * dexterplugin_isampler_data_free:
 * @isampler: #DexterPluginISampler object.
 * @data: address of sampler data structure.
 *
 * The sampler should free the indicated sampler data.
 **/
void
dexterplugin_isampler_data_free (DexterPluginISampler *isampler,
                                 gpointer              data)
{
  g_return_if_fail (DEXTERPLUGIN_IS_ISAMPLER (isampler));
  DEXTERPLUGIN_ISAMPLER_GET_INTERFACE (isampler) -> data_free (isampler, data);
}


/**
 * dexterplugin_isampler_data_serialize:
 * @isampler: #DexterPluginISampler object.
 * @dest_buffer: address of #GString destination buffer.
 * @source_data: address of sampler data structure.
 * @error: return location of optional #GError.
 *
 * The sampler should serialize the source sample data structure to the destination buffer.
 *
 * The optional @error parameter can be used to return errors to the caller, e.g. a GLib error,
 * or an error you describe using the %DEXTERPLUGIN_SAMPLER_ERROR_FAILED error code from the
 * %DEXTERPLUGIN_SAMPLER_ERROR error domain.
 *
 * Libdexter uses these serialization functions to transmit sampler data over the network
 * via <ulink url="../libdexter/libdexter-Channels.html">Channels</ulink>.
 * The format you use to serialize the sampler data is completely up to you.  You may perhaps
 * choose to implement several services that each do the same thing but that serialize data 
 * differently, e.g. plain text, binary, XML, etc.
 *
 * #GString is used as the destination buffer since it conveniently implements a
 * <emphasis>growable</emphasis> buffer and thus helps to avoid buffer overflow issues.
 * GString's can also contain binary data, embedded nulls, etc. See g_string_new_len().
 *
 * Returns: #gboolean indicating success or failure.
 **/
gboolean            
dexterplugin_isampler_data_serialize (DexterPluginISampler  *isampler,
                                      GString               *dest_buffer,
                                      gconstpointer          source_data,
                                      GError               **error)
{
  g_return_val_if_fail (DEXTERPLUGIN_IS_ISAMPLER (isampler), FALSE);
  g_return_val_if_fail (dest_buffer != NULL, FALSE);
  g_return_val_if_fail (source_data != NULL, FALSE);
  g_return_val_if_fail (error == NULL || *error == NULL, FALSE);
  return DEXTERPLUGIN_ISAMPLER_GET_INTERFACE (isampler) -> data_serialize (isampler, dest_buffer,
                                                                           source_data, error);
}


/**
 * dexterplugin_isampler_data_deserialize:
 * @isampler: #DexterPluginISampler object.
 * @source_buffer: address of #GString source buffer.
 * @error: return location of optional #GError.
 *
 * The sampler should deserialize the indicated #GString buffer of encoded sampler data,
 * allocate and return a data structure containing the deserialized data.  
 *
 * The optional @error parameter can be used to return errors to the caller, e.g. a GLib error, 
 * or an error you describe using the %DEXTERPLUGIN_SAMPLER_ERROR_FAILED error code from the
 * %DEXTERPLUGIN_SAMPLER_ERROR error domain.
 *
 * The sample structure should be freed when done with dexterplugin_isampler_data_free().
 *
 * Returns: #gpointer to a newly allocated sample data structure, or %NULL if failed.
 **/
gpointer
dexterplugin_isampler_data_deserialize (DexterPluginISampler  *isampler,
                                        const GString         *source_buffer,
                                        GError               **error)
{
  g_return_val_if_fail (DEXTERPLUGIN_IS_ISAMPLER (isampler), NULL);
  g_return_val_if_fail (source_buffer != NULL, NULL);
  g_return_val_if_fail (error == NULL || *error == NULL, NULL);
  return DEXTERPLUGIN_ISAMPLER_GET_INTERFACE (isampler) -> data_deserialize (isampler, source_buffer, error);
}


/**
 * dexterplugin_isampler_input_new:
 * @isampler: #DexterPluginISampler object.
 *
 * The sampler should allocate and return a zero-filled input data structure, or %NULL
 * if no such input structure is defined for the sampler service.
 *
 * The input structure should be freed when done with dexterplugin_isampler_input_free().
 *
 * Returns: #gpointer to a newly allocated input structure.
 **/
gpointer
dexterplugin_isampler_input_new (DexterPluginISampler *isampler)
{
  g_return_val_if_fail (DEXTERPLUGIN_IS_ISAMPLER (isampler), NULL);
  return DEXTERPLUGIN_ISAMPLER_GET_INTERFACE (isampler) -> input_new (isampler);
}


/**
 * dexterplugin_isampler_input_free:
 * @isampler: #DexterPluginISampler object.
 * @input: address of input structure.
 *
 * The sampler should free the indicated input structure.
 **/
void
dexterplugin_isampler_input_free (DexterPluginISampler *isampler,
                                  gpointer              input)
{
  DEXTERPLUGIN_ISAMPLER_GET_INTERFACE (isampler) -> input_free (isampler, input);
}


/**
 * dexterplugin_isampler_input_serialize:
 * @isampler: #DexterPluginISampler object.
 * @dest_buffer: address of #GString destination buffer.
 * @source_input: address of input structure.
 * @error: return location of optional #GError.
 *
 * If the sampler requires input data during dexterplugin_isampler_initialize(), the input data
 * must be serialized for transmission over the network when the sampler is operating over a
 * Libdexter <ulink url="../libdexter/libdexter-Channels.html">Channel</ulink>.
 *
 * The sampler should serialize the input data structure to the destination buffer.
 *
 * The optional @error parameter can be used to return errors to the caller, e.g. a GLib error,
 * or an error you describe using the %DEXTERPLUGIN_SAMPLER_ERROR_FAILED error code from the
 * %DEXTERPLUGIN_SAMPLER_ERROR error domain.
 *
 * Returns: #gboolean indicating success or failure.
 **/
gboolean
dexterplugin_isampler_input_serialize (DexterPluginISampler  *isampler,
                                       GString               *dest_buffer,
                                       gconstpointer          source_input,
                                       GError               **error)
{
  g_return_val_if_fail (DEXTERPLUGIN_IS_ISAMPLER (isampler), FALSE);
  g_return_val_if_fail (dest_buffer != NULL, FALSE);
  g_return_val_if_fail (source_input != NULL, FALSE);
  g_return_val_if_fail (error == NULL || *error == NULL, FALSE);
  return DEXTERPLUGIN_ISAMPLER_GET_INTERFACE (isampler) -> input_serialize (isampler, dest_buffer,
                                                                            source_input, error);
}


/**
 * dexterplugin_isampler_input_deserialize:
 * @isampler: #DexterPluginISampler object.
 * @source_buffer: address of a #GString source buffer.
 * @error: return location of optional #GError.
 *
 * The sampler should deserialize the indicated #GString buffer of encoded input data,
 * allocate and return an input structure containing the deserialized data.  
 *
 * The optional @error parameter can be used to return errors to the caller, e.g. a GLib error, 
 * or an error you describe using the %DEXTERPLUGIN_SAMPLER_ERROR_FAILED error code from the
 * %DEXTERPLUGIN_SAMPLER_ERROR error domain.
 *
 * The input structure should be freed when done with dexterplugin_isampler_input_free().
 *
 * Returns: #gpointer to a newly allocated input data structure, or %NULL if failed.
 **/
gpointer
dexterplugin_isampler_input_deserialize (DexterPluginISampler  *isampler,
                                         const GString         *source_buffer,
                                         GError               **error)
{
  g_return_val_if_fail (DEXTERPLUGIN_IS_ISAMPLER (isampler), NULL);
  g_return_val_if_fail (source_buffer != NULL, NULL);
  g_return_val_if_fail (error == NULL || *error == NULL, FALSE);
  return DEXTERPLUGIN_ISAMPLER_GET_INTERFACE (isampler) -> input_deserialize (isampler, source_buffer, error);
}

#define __DEXTERPLUGIN_ISAMPLER_C__
#include <dexterplugin-aliasdef.c>
