/* $Id: dexterplugin-iversion.c 393 2012-03-06 23:36:44Z pkovacs $ */
/*
 * libdexter - a plugin-based, distributed sampling library.
 * Copyright (C) 2006-2012 Philip Kovacs <pkovacs@users.sourceforge.net>.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 2.1
 * of the License, or (at your option) any later version.

 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.

 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
 * 02110-1301  USA
 *
*/

/**
 * SECTION:dexterplugin-iversion
 * @title: DexterPluginIVersion
 * @short_description: Interface to Libdexter version information
 * @see_also:
 *
 * Plugins can access version information using this interface.
 *
 * The plugin should use this interface during initialization to verify that 
 * it is being loaded by a version of Libdexter that is compatible with 
 * the version with which the plugin was compiled.
 *
 * <informalexample>
 * <programlisting>
 * #<!-- -->include &lt;dexterplugin.h&gt;
 * <!-- -->
 * void dexter_plugin_init (GTypeModule      *module)
 * {
 *   ...
 *   /<!-- -->* check for version compatibility with main library *<!-- -->/
 *   dexterplugin_iversion_check_version_compatible (dexterplugin_main_imain_get ()->iversion,
 *                                                   DEXTER_VERSION_MAJOR,
 *                                                   DEXTER_VERSION_MINOR,
 *                                                   DEXTER_VERSION_MICRO);
 *   <!-- -->
 *   /<!-- -->* or you can use the following shortcut macro *<!-- -->/
 *   DEXTER_CHECK_VERSION_COMPATIBLE;
 *   ...
 * }
 * </programlisting>
 * </informalexample>
 **/

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <dexterplugin-iversion.h>
#include <dexterplugin-alias.h>

static void  dexterplugin_iversion_base_init (gpointer g_class);

static void
dexterplugin_iversion_base_init (gpointer g_class)
{
  static gboolean initialized = FALSE;

  if (!initialized) 
  {
    initialized = TRUE;
  }
}

GType
dexterplugin_iversion_get_type (void)
{
  static GType type_id = G_TYPE_INVALID;

  if (G_UNLIKELY (type_id == G_TYPE_INVALID))
  {
    static const GTypeInfo type_info =
    {
      sizeof (DexterPluginIVersionInterface),
      dexterplugin_iversion_base_init,
      NULL,
      NULL,
      NULL,
      NULL,
      0,
      0,
      NULL,
      NULL
    };

    type_id = g_type_register_static (G_TYPE_INTERFACE, "DexterPluginIVersion", &type_info, 0);
    g_type_interface_add_prerequisite (type_id, G_TYPE_OBJECT);
  }

  return type_id;
}


/**
 * dexterplugin_iversion_get_version_major:
 * @iversion: a #DexterPluginIVersion.
 *
 * Returns the major version number of Libdexter.
 *
 * Returns: #gint major version number.
 **/
guint
dexterplugin_iversion_get_version_major (DexterPluginIVersion *iversion)
{
  g_return_val_if_fail (DEXTERPLUGIN_IS_IVERSION (iversion), -1);
  return DEXTERPLUGIN_IVERSION_GET_INTERFACE (iversion) -> get_version_major (iversion); 
}


/**
 * dexterplugin_iversion_get_version_minor:
 * @iversion: a #DexterPluginIVersion.
 *
 * Returns the minor version number of Libdexter.
 *
 * Returns: #gint minor version number.
 **/
guint
dexterplugin_iversion_get_version_minor (DexterPluginIVersion *iversion)
{
  g_return_val_if_fail (DEXTERPLUGIN_IS_IVERSION (iversion), -1);
  return DEXTERPLUGIN_IVERSION_GET_INTERFACE (iversion) -> get_version_minor (iversion); 
}


/**
 * dexterplugin_iversion_get_version_micro:
 * @iversion: a #DexterPluginIVersion.
 *
 * Returns the micro version number of Libdexter.
 *
 * Returns: #gint micro version number.
 **/
guint
dexterplugin_iversion_get_version_micro (DexterPluginIVersion *iversion)
{
  g_return_val_if_fail (DEXTERPLUGIN_IS_IVERSION (iversion), -1);
  return DEXTERPLUGIN_IVERSION_GET_INTERFACE (iversion) -> get_version_micro (iversion); 
}


/**
 * dexterplugin_iversion_check_version_exact:
 * @iversion: a #DexterPluginIVersion.
 * @major: #guint major version number.
 * @minor: #guint minor version number.
 * @micro: #guint micro version number.
 *
 * Verifies that the specified (@major, @minor, @micro) version number exactly matches
 * that of the Libdexter library.
 *
 * Used by plugins that require a specific library version.
 **/
void
dexterplugin_iversion_check_version_exact (DexterPluginIVersion *iversion,
                                           guint                 major, 
                                           guint                 minor, 
                                           guint                 micro)
{
  g_return_if_fail (DEXTERPLUGIN_IS_IVERSION (iversion));
  DEXTERPLUGIN_IVERSION_GET_INTERFACE (iversion) -> check_version_exact (iversion, major, minor, micro); 
}


/**
 * dexterplugin_iversion_check_version_compatible:
 * @iversion: a #DexterPluginIVersion.
 * @major: #guint major version number.
 * @minor: #guint minor version number.
 * @micro: #guint micro version number.
 *
 * Verifies that the specified (@major, @minor, @micro) version number is compatible
 * with the Libdexter library.
 *
 * It is generally sufficient to use this function to verify compatibility when loading plugins.
 **/
void
dexterplugin_iversion_check_version_compatible (DexterPluginIVersion *iversion,
                                                guint                 major, 
                                                guint                 minor, 
                                                guint                 micro)
{
  g_return_if_fail (DEXTERPLUGIN_IS_IVERSION (iversion));
  DEXTERPLUGIN_IVERSION_GET_INTERFACE (iversion) -> check_version_compatible (iversion, major, minor, micro); 
}

#define __DEXTERPLUGIN_IVERSION_C__
#include <dexterplugin-aliasdef.c>
