/* $Id: dexterplugin-macros.h 393 2012-03-06 23:36:44Z pkovacs $ */
/*
 * libdexter - a plugin-based, distributed sampling library.
 * Copyright (C) 2006-2012 Philip Kovacs <pkovacs@users.sourceforge.net>.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 2.1
 * of the License, or (at your option) any later version.

 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.

 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
 * 02110-1301  USA
 *
*/

/**
 * SECTION:dexterplugin-macros
 * @title: Libdexterplugin Macros
 * @short_description: Helper macros for defining Libdexter plugins
 * @see_also:
 *
 * Helper macros for defining plugin-based services.  They are similar to the standard GObject 
 * macros found in <filename>gtype.h</filename>. 
 *
 * Example of how to use the DEXTERPLUGIN_DEFINE_TYPE_WITH_CODE() macro to define a plugin 
 * <emphasis>sampler</emphasis> service:
 * <informalexample>
 * <programlisting>
 * #<!-- -->include &lt;dexterplugin.h&gt;
 * <!-- -->
 * ...
 * /<!-- -->* -------------------------------------------------------------------------------------------------
 *   Define a sampler service type from the required abstract base type DEXTERPLUGIN_TYPE_SAMPLER_BASE, 
 *   implementing the required interface DEXTERPLUGIN_TYPE_ISAMPLER and initializing that interface 
 *   in plugin function your_plugin_service_isampler_init().
 *   -------------------------------------------------------------------------------------------------- *<!-- -->/
 * DEXTERPLUGIN_DEFINE_TYPE_WITH_CODE (YourPluginService,
 *                                     your_plugin_service,
 *                                     DEXTERPLUGIN_TYPE_SAMPLER_BASE,
 *                                     DEXTERPLUGIN_IMPLEMENT_INTERFACE (DEXTERPLUGIN_TYPE_ISAMPLER,
 *                                                                       your_plugin_service_isampler_init));
 * ...
 * </programlisting>
 * </informalexample>
 **/

#ifndef __DEXTERPLUGIN_MACROS_H__
#define __DEXTERPLUGIN_MACROS_H__

/**
 * DEXTERPLUGIN_DEFINE_TYPE:
 * @TN: The name of the new type, in Camel case.
 * @t_n: The name of the new type, in lowercase, with words separated by '_'.
 * @T_P: The #GType of the parent type.
 *
 * Analog to G_DEFINE_TYPE() used to define object types for Libdexter plugins.
 *
 * This macro differs from the standard macro in that it registers a dynamic, module-based
 * type rather than a static type.
 **/
#define DEXTERPLUGIN_DEFINE_TYPE(TN, t_n, T_P) DEXTERPLUGIN_DEFINE_TYPE_EXTENDED (TN, t_n, T_P, 0, {})


/**
 * DEXTERPLUGIN_DEFINE_TYPE_WITH_CODE:
 * @TN: The name of the new type, in Camel case.
 * @t_n: The name of the new type, in lowercase, with words separated by '_'.
 * @T_P: The #GType of the parent type.
 * @_C_: Custom code that gets inserted in the *_register_type() function.
 *
 * Analog to G_DEFINE_TYPE_WITH_CODE() used to define object types for Libdexter plugins.
 *
 * This macro differs from the standard macro in that it registers a dynamic, module-based
 * type rather than a static type.
 **/
#define DEXTERPLUGIN_DEFINE_TYPE_WITH_CODE(TN, t_n, T_P, _C_) DEXTERPLUGIN_DEFINE_TYPE_EXTENDED (TN, t_n, T_P, 0, _C_)


/**
 * DEXTERPLUGIN_DEFINE_TYPE_EXTENDED:
 * @TypeName: The name of the new type, in Camel case.
 * @type_name: The name of the new type, in lowercase, with words separated by '_'.
 * @TYPE_PARENT: The #GType of the parent type.
 * @flags: flags for g_type_module_register_type().
 * @CODE: Custom code that gets inserted in the *_register_type() function.
 *
 * Analog to G_DEFINE_TYPE_EXTENDED() used to define object types for Libdexter plugins.
 *
 * This macro differs from the standard macro in that it registers a dynamic, module-based
 * type rather than a static type.
 *
 * <informalexample>
 * <programlisting>
 * DEXTERPLUGIN_DEFINE_TYPE_WITH_CODE (YourPluginService,
 *                                     your_plugin_service,
 *                                     DEXTERPLUGIN_TYPE_SAMPLER_BASE,
 *                                     DEXTERPLUGIN_IMPLEMENT_INTERFACE (DEXTERPLUGIN_TYPE_ISAMPLER,
 *                                                                       your_plugin_service_isampler_init));
 * </programlisting>
 * </informalexample>
 * expands to
 * <informalexample>
 * <programlisting>
 * static GType    your_plugin_service_type            = G_TYPE_INVALID;
 * static void     your_plugin_service_init              (YourPlugin      *self);
 * static void     your_plugin_service_class_init        (YourPluginClass *klass);
 * static gpointer your_plugin_service_parent_class    = NULL;
 * static void     your_plugin_service_class_intern_init (gpointer klass)
 * {
 *   your_plugin_service_parent_class = g_type_class_peek_parent (klass);
 *   your_plugin_service_class_init ((YourPluginServiceClass*) klass);
 * }
 * <!-- -->
 * GType
 * your_plugin_serice_get_type (void)
 * {
 *   return your_plugin_service_type;
 * }
 * <!-- -->
 * void
 * your_plugin_service_register_type (GTypeModule *module)
 * {
 *   GType dexterplugin_define_type_id;
 *   static const GTypeInfo dexterplugin_type_info = {
 *       sizeof (YourPluginServiceClass),
 *       NULL,
 *       NULL,
 *       (GClassInitFunc) your_plugin_service_class_intern_init,
 *       NULL,
 *       NULL,
 *       sizeof (YourPluginService),
 *       0,
 *       (GInstanceInitFunc) your_plugin_service_init,
 *       NULL
 *   };
 *   dexterplugin_define_type_id =
 *   g_type_module_register_type (module, DEXTERPLUGIN_TYPE_SAMPLER_BASE, "YourPluginService",
 *                                &amp;dexterplugin_type_info, (GTypeFlags) 0);
 *   your_plugin_service_type = dexterplugin_define_type_id;
 *   { 
 *     static const GInterfaceInfo dexterplugin_implement_interface_info = {
 *        (GInterfaceInitFunc) your_plugin_service_isampler_init
 *     };
 *     g_type_module_add_interface (module, dexterplugin_define_type_id,
 *                                  DEXTERPLUGIN_TYPE_ISAMPLER, &amp;dexterplugin_implement_interface_info);
 *   }
 * }
 * </programlisting>
 * </informalexample>
 **/
#define DEXTERPLUGIN_DEFINE_TYPE_EXTENDED(TypeName, type_name, TYPE_PARENT, flags, CODE) \
\
static GType    type_name##_type = G_TYPE_INVALID; \
static void     type_name##_init              (TypeName        *self); \
static void     type_name##_class_init        (TypeName##Class *klass); \
static gpointer type_name##_parent_class = NULL; \
static void     type_name##_class_intern_init (gpointer klass) \
{ \
  type_name##_parent_class = g_type_class_peek_parent (klass); \
  type_name##_class_init ((TypeName##Class*) klass); \
} \
\
GType \
type_name##_get_type (void) \
{ \
  return type_name##_type; \
} \
\
void \
type_name##_register_type (GTypeModule *module) \
{ \
  GType dexterplugin_define_type_id; \
  static const GTypeInfo dexterplugin_type_info = \
  { \
    sizeof (TypeName##Class), \
    NULL, \
    NULL, \
    (GClassInitFunc) type_name##_class_intern_init, \
    NULL, \
    NULL, \
    sizeof (TypeName), \
    0, \
    (GInstanceInitFunc) type_name##_init, \
    NULL \
  }; \
  dexterplugin_define_type_id = \
  g_type_module_register_type (module, TYPE_PARENT, #TypeName, &dexterplugin_type_info, (GTypeFlags) flags); \
                               type_name##_type = dexterplugin_define_type_id; \
  { CODE ; } \
}


/**
 * DEXTERPLUGIN_IMPLEMENT_INTERFACE:
 * @TYPE_IFACE: The #GType of the interface to add.
 * @iface_init: The interface init function.
 *
 * Analog to G_IMPLEMENT_INTERFACE() used to add an interface to Libdexter plugin types.
 *
 * This macro differs from the standard macro in that it adds the interface to a dynamic, module-based
 * type rather than a static type.
 **/
#define DEXTERPLUGIN_IMPLEMENT_INTERFACE(TYPE_IFACE, iface_init) \
{ \
  static const GInterfaceInfo dexterplugin_implement_interface_info = \
  { \
    (GInterfaceInitFunc) iface_init \
  }; \
  g_type_module_add_interface (module, dexterplugin_define_type_id, \
                               TYPE_IFACE, &dexterplugin_implement_interface_info); \
}

#endif /* __DEXTERPLUGIN_MACROS_H__ */
