/* $Id: dexterplugin-sampler.c 393 2012-03-06 23:36:44Z pkovacs $ */
/*
 * libdexter - a plugin-based, distributed sampling library.
 * Copyright (C) 2006-2012 Philip Kovacs <pkovacs@users.sourceforge.net>.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 2.1
 * of the License, or (at your option) any later version.

 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.

 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
 * 02110-1301  USA
 *
*/

/**
 * SECTION:dexterplugin-sampler
 * @title: DexterPluginSampler
 * @short_description: GObject base class for Libdexter samplers
 * @see_also:
 *
 * GObject base class from which all Libdexter samplers must be derived.
 *
 * This abstract base class is the foundation class of Libdexter samplers.  Plugin sampler
 * services <emphasis>must</emphasis> derive from this class to be operable within Libdexter.
 * <ulink url="libdexterplugin-DexterPluginMacros.html">Helper macros</ulink> are provided 
 * to simplify defining sampler services that derive from this class.
 **/

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <dexterplugin-sampler.h>
#include <dexterplugin-imain.h>
#include <dexterplugin-isampler.h>
#include <dexterplugin-ithreadpool.h>
#include <dexterplugin-alias.h>

/* Properties */
enum
{
  PROP_0,
  PROP_SAMPLER,
  PROP_CALLBACKS,
  PROP_ITHREADPOOL,
  PROP_CALLBACKS_SIZE
};

static void   dexterplugin_sampler_get_property     (GObject      *object,
                                                     guint         prop_id,
                                                     GValue       *value,
                                                     GParamSpec   *pspec);
static void   dexterplugin_sampler_set_property     (GObject      *object,
                                                     guint         prop_id,
                                                     const GValue *value,
                                                     GParamSpec   *pspec);

struct _DexterPluginSamplerPrivate
{
  GObject                         *sampler;         /* client's sampler which contains us */
  DexterPluginSamplerDataCallback *callbacks;       /* null-terminated array of data callbacks */
  DexterPluginIThreadPool         *ithreadpool;     /* threadpool for callbacks. may be NULL */
  gsize                            callbacks_size;  /* allocated size of callbacks array */
};

/* for threaded user callbacks */
typedef struct
{
  DexterPluginSampler             *plugin_sampler;
  GObject                         *sampler;
  DexterPluginSamplerDataCallback  callback;
  gpointer                         sampler_data;
} SamplerDataCallbackContainer;


static void dexterplugin_sampler_class_init (DexterPluginSamplerClass     *klass);
static void dexterplugin_sampler_init       (DexterPluginSampler          *plugin_sampler);
static void dexterplugin_sampler_finalize   (GObject                      *object);

            /* thread function for callbacks */
static void _sampler_callback_func          (SamplerDataCallbackContainer *cb);


GQuark
dexterplugin_sampler_error_quark (void)
{
  static GQuark error_quark = 0;

  if (error_quark == 0)
    error_quark = g_quark_from_static_string ("dexterplugin-sampler-error-quark");

  return error_quark;
}


G_DEFINE_ABSTRACT_TYPE (DexterPluginSampler, dexterplugin_sampler, G_TYPE_OBJECT);


static void
dexterplugin_sampler_class_init (DexterPluginSamplerClass *klass)
{
  GObjectClass  *g_object_class;

#ifdef G_ENABLE_DEBUG
  g_message (G_STRFUNC);
#endif
   
  g_object_class = G_OBJECT_CLASS (klass);
  g_object_class->finalize = dexterplugin_sampler_finalize;
  g_object_class->get_property = dexterplugin_sampler_get_property;
  g_object_class->set_property = dexterplugin_sampler_set_property;

  /**
   * DexterPluginSampler:private1:
   *
   * This property is private to the library.
   **/
  g_object_class_install_property (g_object_class,
                                   PROP_SAMPLER,
                                   g_param_spec_object ("private1", "private1", "private1",
                                                        G_TYPE_OBJECT,
                                                        G_PARAM_CONSTRUCT_ONLY | G_PARAM_WRITABLE));

  /**
   * DexterPluginSampler:private2:
   *
   * This property is private to the library.
   **/
  g_object_class_install_property (g_object_class,
                                   PROP_CALLBACKS,
                                   g_param_spec_pointer ("private2", "private2", "private2",
                                                         G_PARAM_CONSTRUCT_ONLY | G_PARAM_WRITABLE));

  /**
   * DexterPluginSampler:private3:
   *
   * This property is private to the library.
   **/
  g_object_class_install_property (g_object_class,
                                   PROP_ITHREADPOOL,
                                   g_param_spec_pointer ("private3", "private3", "private3",
                                                         G_PARAM_CONSTRUCT_ONLY | G_PARAM_WRITABLE));

  /**
   * DexterPluginSampler:private4:
   *
   * This property is private to the library.
   **/
  g_object_class_install_property (g_object_class,
                                   PROP_CALLBACKS_SIZE,
                                   g_param_spec_uint ("private4", "private4", "private4",
                                                      0, G_MAXSIZE, 0,
                                                      G_PARAM_CONSTRUCT_ONLY | G_PARAM_WRITABLE));
}


static void
dexterplugin_sampler_init (DexterPluginSampler *plugin_sampler)
{
#ifdef G_ENABLE_DEBUG
  g_message (G_STRFUNC);
#endif
      
  plugin_sampler->priv = g_slice_new0 (DexterPluginSamplerPrivate);
  plugin_sampler->priv->sampler = NULL;
  plugin_sampler->priv->callbacks = NULL;
  plugin_sampler->priv->ithreadpool = NULL;
  plugin_sampler->priv->callbacks_size = 0; 
}


static void
dexterplugin_sampler_finalize (GObject *object)
{
  DexterPluginSampler *plugin_sampler = DEXTERPLUGIN_SAMPLER (object);
    
#ifdef G_ENABLE_DEBUG
  g_message (G_STRFUNC);
#endif

  if (plugin_sampler->priv->callbacks)
    g_slice_free1 (plugin_sampler->priv->callbacks_size, plugin_sampler->priv->callbacks);

  g_slice_free (DexterPluginSamplerPrivate, plugin_sampler->priv);

  /* chain up */
  G_OBJECT_CLASS (dexterplugin_sampler_parent_class)->finalize (object);
}


static void
dexterplugin_sampler_get_property (GObject      *object,
                                   guint         prop_id,
                                   GValue       *value,
                                   GParamSpec   *pspec)
{
#ifdef G_ENABLE_DEBUG
  g_message (G_STRFUNC);
#endif

  g_warning ("%s: object properties may not be read", G_STRFUNC);
}


static void
dexterplugin_sampler_set_property (GObject      *object,
                                   guint         prop_id,
                                   const GValue *value,
                                   GParamSpec   *pspec)
{
  DexterPluginSampler *plugin_sampler = DEXTERPLUGIN_SAMPLER (object);

#ifdef G_ENABLE_DEBUG
  g_message (G_STRFUNC);
#endif

  switch (prop_id)
  {
    case PROP_SAMPLER:
      plugin_sampler->priv->sampler = g_value_get_object (value);
      break;
    case PROP_CALLBACKS:
      plugin_sampler->priv->callbacks = g_value_get_pointer (value);
    case PROP_ITHREADPOOL:
      plugin_sampler->priv->ithreadpool = g_value_get_pointer (value);
      break;
    case PROP_CALLBACKS_SIZE:
      plugin_sampler->priv->callbacks_size = g_value_get_uint (value);
      break;
    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
      break;
  }
}

void
dexterplugin_sampler_callbacks (DexterPluginSampler *plugin_sampler)
{
#ifdef G_ENABLE_DEBUG
  g_message (G_STRFUNC);
#endif

  g_return_if_fail (DEXTERPLUGIN_IS_SAMPLER (plugin_sampler));

  g_object_ref (plugin_sampler);

  if (plugin_sampler->priv->callbacks && plugin_sampler->priv->ithreadpool)
  {
    /* submit callbacks onto threadpool */
    SamplerDataCallbackContainer    *cb;
    DexterPluginSamplerDataCallback *func;

    for (func=plugin_sampler->priv->callbacks; func && *func; func++)
    {
      /* Create a callback sampler and load it with everything needed for the callback. */
      cb = g_slice_new0 (SamplerDataCallbackContainer);
      cb->plugin_sampler = plugin_sampler;         /* ref'd in thread function */
      cb->sampler = plugin_sampler->priv->sampler; /* ref'd in thread function */
      cb->callback = *func;
      cb->sampler_data = dexterplugin_isampler_data_new (DEXTERPLUGIN_ISAMPLER (cb->plugin_sampler), TRUE);
      /* Submit job onto threadpool. */
      dexterplugin_ithreadpool_push_task (plugin_sampler->priv->ithreadpool, 
                                          (GThreadFunc) _sampler_callback_func,
                                          (gpointer) cb, NULL);
    }
  }
  else if (plugin_sampler->priv->callbacks)
  {
    /* perform callbacks synchronously */
    DexterPluginSamplerDataCallback *func;
    DexterPluginISampler            *isampler = DEXTERPLUGIN_ISAMPLER (plugin_sampler);

    for (func=plugin_sampler->priv->callbacks; func && *func; func++)
    {
      gpointer sampler_data = dexterplugin_isampler_data_new (isampler, TRUE);

      (*func) ((gpointer)plugin_sampler->priv->sampler, sampler_data);

      dexterplugin_isampler_data_free (isampler, sampler_data);
    }
  }

  g_object_unref (plugin_sampler);
}


/*  thread function for callback */
static void
_sampler_callback_func (SamplerDataCallbackContainer *cb)
{
  g_return_if_fail (cb != NULL);

  /* keep objects alive by ref'ing them */
  g_object_ref (cb->sampler);
  g_object_ref (cb->plugin_sampler);

  /* Perform the user's callback */
  (*cb->callback) ((gpointer)cb->sampler, cb->sampler_data);
    
  /* Free the sample_data for the user */
  dexterplugin_isampler_data_free (DEXTERPLUGIN_ISAMPLER (cb->plugin_sampler), cb->sampler_data);
    
  /* Release our references */
  g_object_unref (cb->plugin_sampler);
  g_object_unref (cb->sampler);
    
  /* Free the callback container */
  g_slice_free (SamplerDataCallbackContainer, cb);
}

#define __DEXTERPLUGIN_SAMPLER_C__
#include <dexterplugin-aliasdef.c>
