/* $Id: dexterplugin-serialize.c 393 2012-03-06 23:36:44Z pkovacs $ */
/*
 * libdexter - a plugin-based, distributed sampling library.
 * Copyright (C) 2006-2012 Philip Kovacs <pkovacs@users.sourceforge.net>.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 2.1
 * of the License, or (at your option) any later version.

 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.

 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
 * 02110-1301  USA
 *
*/

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <errno.h>
#include <stdlib.h>
#include <string.h>
#include <dexterplugin-serialize.h>
#include <dexterplugin-alias.h>

typedef enum
{
  PARSER_START,
  PARSER_SERVICELIST,
  PARSER_SERVICE,
  PARSER_GTYPE,
  PARSER_NAME,
  PARSER_BLURB,
  PARSER_UUID
} ParserState;

typedef struct 
{
  GQueue               *state;
  GList               *services;
  DexterPluginService *service;
  /* text accumulators */
  GString             *str_gtype;
  GString             *str_name;
  GString             *str_blurb;
  GString             *str_uuid;
} ParserData;

static void  parser_start_element_handler (GMarkupParseContext  *context,
                                           const gchar          *element_name,
                                           const gchar         **attribute_names,
                                           const gchar         **attribute_values,
                                           gpointer              user_data,
                                           GError              **error);
static void  parser_end_element_handler   (GMarkupParseContext  *context,
                                           const gchar          *element_name,
                                           gpointer              user_data,
                                           GError              **error);
static void  parser_text_handler          (GMarkupParseContext  *context,
                                           const gchar          *text,
                                           gsize                 text_len,
                                           gpointer              user_data,
                                           GError              **error);
static void  parser_data_destroy          (ParserData           *service_parser_data);

static void  _free_text_accumulators      (ParserData           *service_parser_data);

static GMarkupParser markup_parser =
{
  parser_start_element_handler,
  parser_end_element_handler,
  parser_text_handler,
  NULL,
  NULL
};


static void         
parser_start_element_handler (GMarkupParseContext  *context,
                              const gchar          *element_name,
                              const gchar         **attribute_names,
                              const gchar         **attribute_values,
                              gpointer              user_data,
                              GError              **error)
{
  ParserData *service_parser_data = (ParserData *) user_data;

  if (strcmp (element_name, "servicelist") == 0)
  {
    gint *state = g_slice_new0 (gint);
    *state = PARSER_SERVICELIST;
    g_queue_push_head (service_parser_data->state, state);
  }
  else if (strcmp (element_name, "service") == 0)
  {
    gint *state = g_slice_new0 (gint);
    *state = PARSER_SERVICE;
    g_queue_push_head (service_parser_data->state, state);

    /* create service structure to hold service data */
    service_parser_data->service = dexterplugin_service_new (G_TYPE_INVALID,NULL,NULL,NULL);
  }
  else if (strcmp (element_name, "gtype") == 0)
  {
    gint *state = g_slice_new0 (gint);
    *state = PARSER_GTYPE;
    g_queue_push_head (service_parser_data->state, state);
    service_parser_data->str_gtype = g_string_new (NULL);
  }
  else if (strcmp (element_name, "name") == 0)
  {
    gint *state = g_slice_new0 (gint);
    *state = PARSER_NAME;
    g_queue_push_head (service_parser_data->state, state);
    service_parser_data->str_name = g_string_new (NULL);
  }
  else if (strcmp (element_name, "blurb") == 0)
  {
    gint *state = g_slice_new0 (gint);
    *state = PARSER_BLURB;
    g_queue_push_head (service_parser_data->state, state);
    service_parser_data->str_blurb = g_string_new (NULL);
  }
  else if (strcmp (element_name, "uuid") == 0)
  {
    gint *state = g_slice_new0 (gint);
    *state = PARSER_UUID;
    g_queue_push_head (service_parser_data->state, state);
    service_parser_data->str_uuid = g_string_new (NULL);
  }
  else
  {
    g_set_error (error,
                 G_MARKUP_ERROR,
                 G_MARKUP_ERROR_UNKNOWN_ELEMENT,
                 "%s: unknown element <%s>",
                 G_STRFUNC, element_name);
  }
}


static void
parser_end_element_handler (GMarkupParseContext  *context,
                            const gchar          *element_name,
                            gpointer              user_data,
                            GError              **error)
{
  ParserData *service_parser_data = (ParserData *) user_data;
  gint *state = g_queue_pop_head (service_parser_data->state);
    
  if (strcmp (element_name, "servicelist") == 0)
  {
    if (*state != PARSER_SERVICELIST)
      goto wrong_state;
  }
  else if (strcmp (element_name, "service") == 0)
  {
    if (*state != PARSER_SERVICE)
      goto wrong_state;

    /* add service to list */
    service_parser_data->services = 
      g_list_prepend (service_parser_data->services, service_parser_data->service);
    /* service struct now on list, so clear this (now unneeded) pointer to it */
    service_parser_data->service=NULL;
  }
  else if (strcmp (element_name, "gtype") == 0)
  {
    if (*state != PARSER_GTYPE)
      goto wrong_state;

    errno=0;
    service_parser_data->service->type = 
      strtoll ((const char *)service_parser_data->str_gtype->str, (char **)NULL, 10);
    if (errno)
    {
      g_set_error (error,
                   G_MARKUP_ERROR,
                   G_MARKUP_ERROR_INVALID_CONTENT,
                   "%s: error converting text for state (%d)",
                   G_STRFUNC, *state);
    }

    g_string_free (service_parser_data->str_gtype, TRUE);
    service_parser_data->str_gtype=NULL;
  }
  else if (strcmp (element_name, "name") == 0)
  {
    if (*state != PARSER_NAME)
      goto wrong_state;

    service_parser_data->service->name = service_parser_data->str_name->str;
    g_string_free (service_parser_data->str_name, FALSE);
    service_parser_data->str_name=NULL;
  }
  else if (strcmp (element_name, "blurb") == 0)
  {
    if (*state != PARSER_BLURB)
      goto wrong_state;

    service_parser_data->service->blurb = service_parser_data->str_blurb->str;
    g_string_free (service_parser_data->str_blurb, FALSE);
    service_parser_data->str_blurb=NULL;
  }
  else if (strcmp (element_name, "uuid") == 0)
  {
    if (*state != PARSER_UUID)
      goto wrong_state;

    service_parser_data->service->uuid = service_parser_data->str_uuid->str;
    g_string_free (service_parser_data->str_uuid, FALSE);
    service_parser_data->str_uuid=NULL;
  }
  else 
  {
    g_set_error (error,
                 G_MARKUP_ERROR,
                 G_MARKUP_ERROR_UNKNOWN_ELEMENT,
                 "%s: unknown closing element </%s>",
                 G_STRFUNC, element_name);
  }

  g_slice_free (gint, state);
  return;

wrong_state:
  g_set_error (error,
               G_MARKUP_ERROR,
               G_MARKUP_ERROR_PARSE,
               "%s: unexpected closing element </%s>",
               G_STRFUNC, element_name);

  g_slice_free (gint, state);
}


static void         
parser_text_handler (GMarkupParseContext  *context,
                     const gchar          *text,
                     gsize                 text_len,
                     gpointer              user_data,
                     GError             **error)
{
  ParserData *service_parser_data = (ParserData *) user_data;
  gint *state = g_queue_peek_head (service_parser_data->state);

  /* GLib docs state that text may be delivered in multiple chunks,
   * so we accumulate text as it is seen. */
  switch (*state)
  {
    case PARSER_GTYPE:
    {
      g_string_append_len (service_parser_data->str_gtype, text, text_len);
      break;
    }
    case PARSER_NAME:
    {
      g_string_append_len (service_parser_data->str_name, text, text_len);
      break;
    }
    case PARSER_BLURB:
    {
      g_string_append_len (service_parser_data->str_blurb, text, text_len);
      break;
    }
    case PARSER_UUID:
    {
      g_string_append_len (service_parser_data->str_uuid, text, text_len);
      break;
    }
    default:
      break;
    }
}


static void
parser_data_destroy             (ParserData    *service_parser_data)
{
  if (!service_parser_data)
    return;

  if (service_parser_data->state && !g_queue_is_empty (service_parser_data->state))
  {
    gint *state=NULL;

    while ((state = g_queue_pop_head (service_parser_data->state)))
      g_slice_free (gint, state);
  }
  if (service_parser_data->state)
  {
    g_queue_free (service_parser_data->state);
    service_parser_data->state=NULL;
  }

  if (service_parser_data->services && g_list_length (service_parser_data->services) > 0 )
  {
    GList *element=NULL;

    for (element=g_list_first (service_parser_data->services);
         element;
         element=g_list_next (element))
    {
      dexterplugin_service_free ( (DexterPluginService*) element->data);
    }
  }
  if (service_parser_data->services)
  {
    g_list_free (service_parser_data->services);
    service_parser_data->services=NULL;
  }

  _free_text_accumulators (service_parser_data);

  if (service_parser_data->service) /* only if something went wrong */
  {
    dexterplugin_service_free (service_parser_data->service);
    service_parser_data->service=NULL;
  }

  /* the service_parser_data itself is on the stack and need not be freed */
}


void
dexterplugin_serialize_service_group_serialize (DexterPluginServiceGroup  *service_group,
                                                GString                   *buffer,
                                                GError                   **error)
{
  DexterPluginService *service;
  gchar               *name_escaped;
  gchar               *blurb_escaped;
  gchar               *uuid_escaped;
  gint                 iter;

#ifdef G_ENABLE_DEBUG
  g_message (G_STRFUNC);
#endif

  g_return_if_fail (service_group != NULL);
  g_return_if_fail (buffer != NULL);
  g_return_if_fail (error == NULL || *error == NULL);

  /* output the servicelist as a small XML doc */
  g_string_append_printf (buffer, "<servicelist>\n");
  for (iter=0; iter<service_group->n_services; iter++)
  {
    service = service_group->services + iter;

    /* escape text for XML transmission, e.g. "A & P Service" -> "A &amp; P Service" */
    name_escaped = service->name ? g_markup_escape_text (service->name, -1) : NULL;
    blurb_escaped = service->blurb ? g_markup_escape_text (service->blurb, -1) : NULL;
    uuid_escaped = service->uuid ? g_markup_escape_text (service->uuid, -1) : NULL;

    g_string_append_printf (buffer, " <service>\n");
    g_string_append_printf (buffer, "  <gtype>%lu</gtype>\n", (gulong) service->type);
    g_string_append_printf (buffer, "  <name>%s</name>\n", name_escaped);
    g_string_append_printf (buffer, "  <blurb>%s</blurb>\n", blurb_escaped);
    g_string_append_printf (buffer, "  <uuid>%s</uuid>\n", uuid_escaped);
    g_string_append_printf (buffer, " </service>\n");

    g_free (name_escaped);
    g_free (blurb_escaped);
    g_free (uuid_escaped);
  }
  g_string_append_printf (buffer, "</servicelist>");
}


DexterPluginServiceGroup *
dexterplugin_serialize_service_group_deserialize (const GString  *buffer,
                                                  GError        **error)
{
  DexterPluginServiceGroup *service_group = NULL;
  GMarkupParseContext      *context = NULL;
  ParserData                service_parser_data;
  gint                     *start_state = NULL;
  GError                   *local_error = NULL;

#ifdef G_ENABLE_DEBUG
  g_message (G_STRFUNC);
#endif

  g_return_val_if_fail (buffer != NULL, NULL);
  g_return_val_if_fail (error == NULL || *error == NULL, NULL);

  service_parser_data.state = g_queue_new ();
  service_parser_data.services = NULL;
  service_parser_data.service = NULL;
  service_parser_data.str_name = NULL;
  service_parser_data.str_blurb = NULL;
  service_parser_data.str_uuid = NULL;
  start_state = g_slice_new0 (gint);
  *start_state = PARSER_START;
  g_queue_push_head (service_parser_data.state, start_state);

  context = g_markup_parse_context_new (&markup_parser, 0, 
                                        &service_parser_data, 
                                        (GDestroyNotify) parser_data_destroy);

  g_markup_parse_context_parse (context, buffer->str, buffer->len, &local_error);
  if (local_error)
  {
    g_propagate_error (error, local_error);
    goto group_list_deserialize_done;
  }

  g_markup_parse_context_end_parse (context, &local_error);
  if (local_error)
  {
    g_propagate_error (error, local_error);
    goto group_list_deserialize_done;
  }

  /* parse succeeded.  create a service group. */
  {
    GList *element = NULL;
    gint   iter;

    service_group = dexterplugin_service_group_new (g_list_length (service_parser_data.services));

    for (element = g_list_first (service_parser_data.services), iter=0;
         element;
         element=g_list_next (element), iter++)
    {
      dexterplugin_service_copy (service_group->services + iter, (DexterPluginService*) element->data);
    }
  }

group_list_deserialize_done:
  g_markup_parse_context_free (context);
  return service_group;
}

static void
_free_text_accumulators (ParserData *service_parser_data)
{
  g_return_if_fail (service_parser_data != NULL);
 
  if (service_parser_data->str_gtype)
  {
    g_string_free (service_parser_data->str_gtype, TRUE);
    service_parser_data->str_gtype=NULL;
  }
  if (service_parser_data->str_name)
  {
    g_string_free (service_parser_data->str_name, TRUE);
    service_parser_data->str_name=NULL;
  }
  if (service_parser_data->str_blurb)
  {
    g_string_free (service_parser_data->str_blurb, TRUE);
    service_parser_data->str_blurb=NULL;
  }
  if (service_parser_data->str_uuid)
  {
    g_string_free (service_parser_data->str_uuid, TRUE);
    service_parser_data->str_uuid=NULL;
  }
}

#define __DEXTERPLUGIN_SERIALIZE_C__
#include <dexterplugin-aliasdef.c>
