/* $Id: dexterplugin-service.c 393 2012-03-06 23:36:44Z pkovacs $ */
/*
 * libdexter - a plugin-based, distributed sampling library.
 * Copyright (C) 2006-2012 Philip Kovacs <pkovacs@users.sourceforge.net>.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 2.1
 * of the License, or (at your option) any later version.

 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.

 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
 * 02110-1301  USA
 *
*/

/**
 * SECTION:dexterplugin-service
 * @title: DexterPluginService
 * @short_description: Service descriptors for identifying and using Libdexter services
 * @see_also:
 *
 * Service descriptors provide identification information for Libdexter services.  
 * Collections of service descriptors are delivered to the user by  
 * <ulink url="../libdexter/libdexter-Service-Brokers.html">Service Brokers</ulink> as 
 * <emphasis>service groups</emphasis>.  Each descriptor contains a 
 * <emphasis>universally unique identifier</emphasis> (uuid) to distinquish it from all other 
 * services. During service registration, each service is assigned a #GType by the #GTypeModule 
 * system. That type is used by GLib/GObject to instantiate objects implementing the service.
 *
 * Note that the service #GType is dynamic in nature and <emphasis>will vary</emphasis>.
 * Do not use it to identify services. Use the uuid instead.
 **/

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <dexterplugin-serialize.h>
#include <dexterplugin-service.h>
#include <dexterplugin-alias.h>


GQuark
dexterplugin_service_error_quark (void)
{
  static GQuark error_quark = 0;

  if (error_quark == 0)
    error_quark = g_quark_from_static_string ("dexterplugin-service-error-quark");

  return error_quark;
}


/**
 * dexterplugin_service_new:
 * @type: service type.
 * @name: service name.
 * @blurb: service description.
 * @uuid: service uuid (univerally unique identifier). 
 * 
 * Creates a new service descriptor.  It must be freed with dexterplugin_service_free().
 *
 * Returns: #DexterPluginService service descriptor.
 **/
DexterPluginService *
dexterplugin_service_new (GType        type,
                          const gchar *name,
                          const gchar *blurb,
                          const gchar *uuid)
{
  DexterPluginService *service;
    
#ifdef G_ENABLE_DEBUG
  g_message (G_STRFUNC);
#endif

  service = g_slice_new0 (DexterPluginService);
  service->type = type;
  service->name = g_strdup (name);
  service->blurb = g_strdup (blurb);
  service->uuid = g_strdup (uuid);

  return service;
}


/**
 * dexterplugin_service_dup:
 * @service: #DexterPluginService to duplicate.
 *
 * Creates a duplicate of a service descriptor.  It must be freed with dexterplugin_service_free().
 *
 * Returns: #DexterPluginService service descriptor.
 **/
DexterPluginService *
dexterplugin_service_dup (const DexterPluginService *service)
{
  g_return_val_if_fail (service != NULL, NULL);
    
#ifdef G_ENABLE_DEBUG
  g_message (G_STRFUNC);
#endif
    
  return dexterplugin_service_new (service->type, service->name, service->blurb, service->uuid);
}


/**
 * dexterplugin_service_copy:
 * @dest: destination address of a #DexterPluginService.
 * @source: source address of a #DexterPluginService.
 *
 * Copies a service descriptor.  Existing data on the destination descriptor will be freed
 * prior to the copy.
 **/
void
dexterplugin_service_copy (DexterPluginService       *dest,
                           const DexterPluginService *source)
{
  g_return_if_fail (dest);
  g_return_if_fail (source);

#ifdef G_ENABLE_DEBUG
  g_message (G_STRFUNC);
#endif
    
  g_free (dest->name);
  g_free (dest->blurb);
  g_free (dest->uuid);

  dest->type = source->type;
  dest->name = g_strdup (source->name);
  dest->blurb = g_strdup (source->blurb);
  dest->uuid = g_strdup (source->uuid);
}


/**
 * dexterplugin_service_free:
 * @service: address of a #DexterPluginService service descriptor.
 *
 * Frees a service descriptor.
 **/
void
dexterplugin_service_free (DexterPluginService *service)
{
#ifdef G_ENABLE_DEBUG
  g_message (G_STRFUNC);
#endif

  if (G_UNLIKELY (!service))
    return;

  g_free (service->name);
  g_free (service->blurb);
  g_free (service->uuid);
  g_slice_free (DexterPluginService, service);
}


/**
 * dexterplugin_service_group_new:
 * @n_services: size of contained #DexterPluginService array.
 *
 * Allocates a #DexterPluginServiceGroup with a #DexterPluginService array of size @n_services.
 * The elements of the contained #DexterPluginService array must be populated, e.g. with 
 * dexterplugin_service_copy(). The service group should be freed with dexterplugin_service_group_free().  
 * Plugin writers should always use this api for allocation of service groups.
 *
 * Returns: #DexterPluginServiceGroup service group.
 *
 * Since: 0.1.4
 **/
DexterPluginServiceGroup *
dexterplugin_service_group_new (guint n_services)
{
  DexterPluginServiceGroup *service_group;
#ifdef G_ENABLE_DEBUG
  g_message (G_STRFUNC);
#endif

  service_group = g_slice_new0 (DexterPluginServiceGroup);
  service_group->n_services = n_services;
  service_group->services = (DexterPluginService *) g_slice_alloc0 (n_services * sizeof (DexterPluginService));

  return service_group;
}


/**
 * dexterplugin_service_group_free:
 * @service_group: #DexterPluginServiceGroup collection of service descriptors.
 *
 * Frees a service group created with dexterplugin_service_group_new().  
 **/
void
dexterplugin_service_group_free (DexterPluginServiceGroup *service_group)
{
  gint iter;
#ifdef G_ENABLE_DEBUG
  g_message (G_STRFUNC);
#endif

  if (G_UNLIKELY (!service_group))
    return;
  
  for (iter=0; iter<service_group->n_services; iter++)
  {
    g_free (service_group->services[iter].name);
    g_free (service_group->services[iter].blurb);
    g_free (service_group->services[iter].uuid);
  }

  g_slice_free1 (service_group->n_services * sizeof (DexterPluginService), service_group->services);
  g_slice_free (DexterPluginServiceGroup, service_group);
}


/**
 * dexterplugin_service_group_find_uuid:
 * @service: address of a pointer to a #DexterPluginService service descriptor.
 * @service_group: #DexterPluginServiceGroup collection of service descriptors.
 * @uuid: uuid of service to find.
 *
 * Finds a service by uuid in the @service_group. If found, the @service pointer
 * will be set to point into the <userinput>service_group->services</userinput> array.
 *
 * Returns: #gboolean indicating if the service was found or not.
 **/
gboolean
dexterplugin_service_group_find_uuid (DexterPluginService      **service,
                                      DexterPluginServiceGroup  *service_group,
                                      const gchar               *uuid)
{
  DexterPluginService *tmp_service;
  gint                 iter;        

  g_return_val_if_fail (service != NULL && *service == NULL, FALSE);
  g_return_val_if_fail (service_group != NULL, FALSE);
  g_return_val_if_fail (uuid != NULL, FALSE);

  for (iter=0; iter<service_group->n_services; iter++)
  {
    tmp_service = service_group->services + iter;
    if (g_ascii_strncasecmp (tmp_service->uuid, uuid, 36)==0)
    {
      *service = tmp_service;
      return TRUE;
    }
  }

  return FALSE;
}

#define __DEXTERPLUGIN_SERVICE_C__
#include <dexterplugin-aliasdef.c>
