/* $Id: dxt-null.c 393 2012-03-06 23:36:44Z pkovacs $ */
/*
 * dxt-null - libdexter sampling plugin that does nothing at all.
 * Copyright (C) 2006-2012 Philip Kovacs <pkovacs@users.sourceforge.net>.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 2.1
 * of the License, or (at your option) any later version.

 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.

 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
 * 02110-1301  USA
 *
*/

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <dexterplugin.h>
#include <dxt-null.h>
#include <dxt-null-public.h>


static void           dxt_null_class_init        (DxtNullClass          *klass);
static void           dxt_null_init              (DxtNull               *self);
static void           dxt_null_finalize          (GObject               *object);
static void           dxt_null_isampler_init     (DexterPluginISamplerInterface  *iface);

                      /* helpers */
static DxtNullData  *_data_new0                  (void);
static void          _data_free                  (DxtNullData           *data);
static void          _data_copy                  (DxtNullData           *dest,
                                                  const DxtNullData     *source);

                      /* DexterPluginISamplerInterface implementation */
static gboolean       dxt_null_initialize        (DexterPluginISampler  *isampler,
                                                  const gint            *param,
                                                  gconstpointer          input,
                                                  GError               **error);
static gboolean       dxt_null_sample            (DexterPluginISampler  *isampler,
                                                  GError               **error);
static gpointer       dxt_null_data_new          (DexterPluginISampler  *isampler,
                                                  gboolean               populate);
static gpointer       dxt_null_data_dup          (DexterPluginISampler  *isampler,
                                                  gconstpointer          source_data);
static void           dxt_null_data_xfer         (DexterPluginISampler  *isampler,
                                                  gpointer               old_data,
                                                  gconstpointer          new_data);
static void           dxt_null_data_copy         (DexterPluginISampler  *isampler,
                                                  gpointer               dest_data,
                                                  gconstpointer          source_data);
static void           dxt_null_data_free         (DexterPluginISampler  *isampler,
                                                  gpointer               data);
static gboolean       dxt_null_data_serialize    (DexterPluginISampler  *isampler,
                                                  GString               *dest_buffer,
                                                  gconstpointer          source_data,
                                                  GError               **error);
static gpointer       dxt_null_data_deserialize  (DexterPluginISampler  *isampler,
                                                  const GString         *source_buffer,
                                                  GError               **error);
static gpointer       dxt_null_input_new         (DexterPluginISampler  *isampler);
static void           dxt_null_input_free        (DexterPluginISampler  *isampler,
                                                  gpointer               input);
static gboolean       dxt_null_input_serialize   (DexterPluginISampler  *isampler,
                                                  GString               *dest_buffer,
                                                  gconstpointer          source_input,
                                                  GError               **error);
static gpointer       dxt_null_input_deserialize (DexterPluginISampler  *isampler,
                                                  const GString         *source_buffer,
                                                  GError               **error);

struct _DxtNull
{
  DexterPluginSampler  parent;
  gboolean             initialized;
  DxtNullData         *null_data;
};


struct _DxtNullClass 
{
  DexterPluginSamplerClass parent_class;
};


DEXTERPLUGIN_DEFINE_TYPE_WITH_CODE (DxtNull,
                                    dxt_null,
                                    DEXTERPLUGIN_TYPE_SAMPLER,
                                    DEXTERPLUGIN_IMPLEMENT_INTERFACE (DEXTERPLUGIN_TYPE_ISAMPLER,
                                                                      dxt_null_isampler_init));

static void
dxt_null_class_init (DxtNullClass *klass)
{
  GObjectClass *gobject_class;

#ifdef G_ENABLE_DEBUG
  g_message (G_STRFUNC);
#endif

  gobject_class = G_OBJECT_CLASS (klass);
  gobject_class->finalize = dxt_null_finalize;
}


static void
dxt_null_init (DxtNull *dxt_null)
{
#ifdef G_ENABLE_DEBUG
  g_message (G_STRFUNC);
#endif

  dxt_null->initialized = FALSE;
  dxt_null->null_data = _data_new0 ();
}


static void
dxt_null_finalize (GObject *object)
{
  DxtNull *dxt_null = DXT_NULL (object);

#ifdef G_ENABLE_DEBUG
  g_message (G_STRFUNC);
#endif

  _data_free (dxt_null->null_data);
    
  /* chain up */
  G_OBJECT_CLASS (dxt_null_parent_class)->finalize (object);
}


static void
dxt_null_isampler_init (DexterPluginISamplerInterface *iface)
{
#ifdef G_ENABLE_DEBUG
  g_message (G_STRFUNC);
#endif
   
  iface->initialize = dxt_null_initialize;
  iface->sample = dxt_null_sample;
  iface->data_new = dxt_null_data_new;
  iface->data_dup = dxt_null_data_dup;
  iface->data_xfer = dxt_null_data_xfer;
  iface->data_copy = dxt_null_data_copy;
  iface->data_free = dxt_null_data_free;
  iface->data_serialize = dxt_null_data_serialize;
  iface->data_deserialize = dxt_null_data_deserialize;
  iface->input_new = dxt_null_input_new;
  iface->input_free = dxt_null_input_free;
  iface->input_serialize = dxt_null_input_serialize;
  iface->input_deserialize = dxt_null_input_deserialize;
}


static DxtNullData *
_data_new0 (void)
{
  return g_slice_new0 (DxtNullData);
}


static void
_data_free (DxtNullData *data)
{    
  if (G_LIKELY (data))
    g_slice_free (DxtNullData, data);
}


static void
_data_copy (DxtNullData       *dest,
            const DxtNullData *source)
{
  g_return_if_fail (dest != NULL);
  g_return_if_fail (source != NULL);
}


static gboolean
dxt_null_initialize (DexterPluginISampler  *isampler,
                     const gint            *param,
                     gconstpointer          input,
                     GError               **error)
{
  DxtNull *dxt_null;
    
#ifdef G_ENABLE_DEBUG
  g_message (G_STRFUNC);
#endif

  g_return_val_if_fail (DEXTERPLUGIN_IS_ISAMPLER (isampler), FALSE);
  g_return_val_if_fail (error == NULL || *error == NULL, FALSE);

  dxt_null = DXT_NULL (isampler);

  if (dxt_null->initialized)
  { 
    g_set_error (error,
                 DEXTERPLUGIN_SAMPLER_ERROR,
                 DEXTERPLUGIN_SAMPLER_ERROR_FAILED,
                 "Null sampler already initialized");
    return FALSE;
  }
    
  dxt_null->initialized = TRUE;

  return TRUE;
}


static gboolean
dxt_null_sample (DexterPluginISampler  *isampler,
                 GError               **error)
{
  DxtNull *dxt_null; 

#ifdef G_ENABLE_DEBUG
  g_message (G_STRFUNC);
#endif

  g_return_val_if_fail (DEXTERPLUGIN_IS_ISAMPLER (isampler), FALSE);
  g_return_val_if_fail (error == NULL || *error == NULL, FALSE);

  dxt_null = DXT_NULL (isampler);

  if (!dxt_null->initialized)
  {
    g_set_error (error,
                 DEXTERPLUGIN_SAMPLER_ERROR,
                 DEXTERPLUGIN_SAMPLER_ERROR_FAILED,
                 "Null sampler not initialized");
    return FALSE;
  }

  return TRUE;
}


static gpointer
dxt_null_data_new (DexterPluginISampler *isampler,
                   gboolean              populate)
{
  DxtNullData *data;
  DxtNull     *dxt_null; 

#ifdef G_ENABLE_DEBUG
  g_message (G_STRFUNC);
#endif
    
  g_return_val_if_fail (DEXTERPLUGIN_IS_ISAMPLER (isampler), NULL);
  dxt_null = DXT_NULL (isampler);

  data = _data_new0 ();

  if (populate)
    _data_copy (data, dxt_null->null_data);

  return data;
}


static gpointer
dxt_null_data_dup (DexterPluginISampler *isampler,
                   gconstpointer         source_data)
{
  DxtNullData *data;

#ifdef G_ENABLE_DEBUG
  g_message (G_STRFUNC);
#endif
    
  g_return_val_if_fail (DEXTERPLUGIN_IS_ISAMPLER (isampler), NULL);
  g_return_val_if_fail (source_data != NULL, NULL);
  
  data = _data_new0 ();

  _data_copy (data, source_data);
            
  return data;
}


static void
dxt_null_data_xfer (DexterPluginISampler *isampler,
                    gpointer              old_data,
                    gconstpointer         new_data)
{
  DxtNull *dxt_null; 

#ifdef G_ENABLE_DEBUG
  g_message (G_STRFUNC);
#endif

  g_return_if_fail (DEXTERPLUGIN_IS_ISAMPLER (isampler));
  dxt_null = DXT_NULL (isampler);
  
  if (old_data)
    _data_copy ((DxtNullData *)old_data, dxt_null->null_data);

  if (new_data)
    _data_copy (dxt_null->null_data, (DxtNullData *)new_data);
}


static void 
dxt_null_data_copy (DexterPluginISampler *isampler,
                    gpointer              dest_data,
                    gconstpointer         source_data)
{
  DxtNullData       *null_dest;
  const DxtNullData *null_source;

#ifdef G_ENABLE_DEBUG
  g_message (G_STRFUNC);
#endif

  g_return_if_fail (DEXTERPLUGIN_IS_ISAMPLER (isampler));
  g_return_if_fail (dest_data != NULL);
  g_return_if_fail (source_data != NULL);

  null_dest = (DxtNullData *)dest_data;
  null_source = (const DxtNullData *)source_data;
  
  _data_copy (null_dest, null_source);
} 


static void
dxt_null_data_free (DexterPluginISampler *isampler,
                    gpointer              data)
{
#ifdef G_ENABLE_DEBUG
  g_message (G_STRFUNC);
#endif
  
  g_return_if_fail (DEXTERPLUGIN_IS_ISAMPLER (isampler));

  if (G_LIKELY (data))
    _data_free ((DxtNullData *)data);
}


static gboolean
dxt_null_data_serialize (DexterPluginISampler  *isampler,
                         GString               *dest_buffer,
                         gconstpointer          source_data,
                         GError               **error)
{
#ifdef G_ENABLE_DEBUG
  g_message (G_STRFUNC);
#endif

  g_return_val_if_fail (DEXTERPLUGIN_IS_ISAMPLER (isampler), FALSE);
  g_return_val_if_fail (dest_buffer != NULL, FALSE);
  g_return_val_if_fail (source_data != NULL, FALSE);

  return TRUE;
}


static gpointer
dxt_null_data_deserialize (DexterPluginISampler  *isampler,
                           const GString         *source_buffer,
                           GError               **error)
{
#ifdef G_ENABLE_DEBUG
  g_message (G_STRFUNC);
#endif

  g_return_val_if_fail (DEXTERPLUGIN_IS_ISAMPLER (isampler), NULL);
  g_return_val_if_fail (error == NULL || *error == NULL, NULL);

  return NULL;
}


static gpointer
dxt_null_input_new (DexterPluginISampler *isampler)
{
#ifdef G_ENABLE_DEBUG
  g_message (G_STRFUNC);
#endif
    
  g_return_val_if_fail (DEXTERPLUGIN_IS_ISAMPLER (isampler), NULL);

  return NULL;
}


static void
dxt_null_input_free (DexterPluginISampler *isampler,
                     gpointer              input)
{
#ifdef G_ENABLE_DEBUG
  g_message (G_STRFUNC);
#endif
    
  g_return_if_fail (DEXTERPLUGIN_IS_ISAMPLER (isampler));
}


static gboolean
dxt_null_input_serialize (DexterPluginISampler  *isampler,
                          GString               *dest_buffer,
                          gconstpointer          source_input,
                          GError               **error)
{
#ifdef G_ENABLE_DEBUG
  g_message (G_STRFUNC);
#endif

  g_return_val_if_fail (DEXTERPLUGIN_IS_ISAMPLER (isampler), FALSE);
  g_return_val_if_fail (dest_buffer != NULL, FALSE);
  g_return_val_if_fail (source_input != NULL, FALSE);
  g_return_val_if_fail (error == NULL || *error == NULL, FALSE);

  return TRUE;
}


static gpointer
dxt_null_input_deserialize (DexterPluginISampler  *isampler,
                            const GString         *source_buffer,
                            GError               **error)
{
#ifdef G_ENABLE_DEBUG
  g_message (G_STRFUNC);
#endif

  g_return_val_if_fail (DEXTERPLUGIN_IS_ISAMPLER (isampler), NULL);
  g_return_val_if_fail (error == NULL || *error == NULL, NULL);

  return NULL;      
}
