/* $Id: dexterd.c 393 2012-03-06 23:36:44Z pkovacs $ */
/*
 * dexterd - a generic daemon server for libdexter. 
 * Copyright (C) 2006-2012 Philip Kovacs <pkovacs@users.sourceforge.net>.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 2.1
 * of the License, or (at your option) any later version.

 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.

 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
 * 02110-1301  USA
 *
*/

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif
#include <fcntl.h>
#ifdef DEXTER_UID_CHANGE
#include <grp.h>
#include <pwd.h>
#endif
#include <signal.h>
#include <stdio.h>
#include <sys/types.h>
#include <unistd.h>
#include <dexter.h>

#ifdef HAVE_DAEMON
#include "daemon.h" 
#endif

static gboolean done = FALSE;
static gchar *config = NULL;
#ifdef HAVE_DAEMON
static gboolean daemonize = FALSE;
#endif
static gchar *pidfile = NULL;
#ifdef DEXTER_UID_CHANGE
static gchar *user = NULL;
#endif
static gboolean version = FALSE;

static GOptionEntry entries[] =
{
  { "config", 'c', 0, G_OPTION_ARG_STRING, &config, "config file to read", NULL },
#ifdef HAVE_DAEMON
  { "daemon", 'd', 0, G_OPTION_ARG_NONE, &daemonize, "become a daemon process", NULL },
#endif
  { "pidfile",'p', 0,  G_OPTION_ARG_STRING, &pidfile, "pid file to create", NULL },
#ifdef DEXTER_UID_CHANGE
  { "user", 'u', 0, G_OPTION_ARG_STRING, &user, "run as indicated user", NULL},
#endif
  { "version", 'v', 0, G_OPTION_ARG_NONE, &version, "display version", NULL },
  { NULL }
};

void server_shutdown (int sig_num)
{
  done = TRUE;
}

int main (int argc, char *argv[])
{
  GOptionContext          *context;
  DexterServer            *server;
  GError                  *err = NULL;

  /* Abort on any warning within this program's domain */
  g_log_set_fatal_mask (G_LOG_DOMAIN, G_LOG_FATAL_MASK | G_LOG_LEVEL_CRITICAL | G_LOG_LEVEL_WARNING);

  /* parse command line options */
  context = g_option_context_new ("- a general-purpose server for Libdexter applications");
  g_option_context_add_main_entries (context, entries, NULL);
  g_option_context_parse (context, &argc, &argv, &err);
  if (err)
    g_warning ("%s", err->message);
    
  /* possibly display version and exit */
  if (version)
  {
#ifdef HAVE_CONFIG_H
    printf ("%s\n", PACKAGE_VERSION);
#else
    printf ("version unknown\n");
#endif
    g_option_context_free (context);
    return 0;
  }

#ifdef HAVE_DAEMON
  /* possibly daemonize */
  if (daemonize) 
  {
    if (daemon (0,0) < 0)
      g_warning ("unable to daemonize");
  }
#endif

  /* possibly create a pid file */
  if (pidfile)
  {
    int fd;
    FILE *f;

    if ( ((fd = open (pidfile, O_WRONLY|O_CREAT|O_EXCL, 0644)) < 0) ||
         ((f = fdopen (fd, "w")) == NULL) )
      g_warning ("cannot create pidfile %s", pidfile);

    if (fprintf (f, "%d\n", getpid()) < 0)
      g_warning ("cannot write to pidfile");

    if (fclose (f) == EOF)
      g_warning ("error closing pidfile");
  }

#ifdef DEXTER_UID_CHANGE
  /* possibly change user id */
  if (user)
  {
    struct passwd *pwent;

    if (geteuid () != 0)
      g_warning ("-user flag requires root privilege");

    if ((pwent = getpwnam (user)) == NULL)
      g_warning ("error getting passwd entry for user `%s'", user);

    if (setgid (pwent->pw_gid))
      g_warning ("error setting group id");

    if (setgroups (0, NULL))
      g_warning ("error dropping supplemental groups");

    if (initgroups (user, pwent->pw_gid))
      g_warning ("error initing supplemental groups");

    if (setuid (pwent->pw_uid))
      g_warning ("error setting user id to `%s'", user);
  }
#endif

  /* catch SIGINT/SIGQUIT/SIGTERM to server_shutdown */
  signal (SIGINT, server_shutdown);
  signal (SIGQUIT, server_shutdown);
  signal (SIGTERM, server_shutdown);
#ifndef HAVE_MSG_NOSIGNAL
  signal (SIGPIPE, SIG_IGN);
#endif
    
  /* init libdexter */
  dexter_init (config, &err);
  if (err)
    g_warning ("%s", err->message);
    
  /* create and start server */
  if (!(server = dexter_server_new (NULL)))
    g_warning ("unable to create server");

  dexter_server_start (server, &err);
  if (err)
    g_warning ("%s", err->message);

  /* main thread does nothing but wait around for a signal */
  while (!done)
    g_usleep (0.20*G_USEC_PER_SEC); 

  /* stop the server and clean up libdexter */
  g_message ("shutting down server ...");
  dexter_server_stop (server, &err);
  if (err)
    g_warning ("%s", err->message);
  dexter_server_free (server);
  dexter_exit ();

  /* free command option argument context */
  g_option_context_free (context);

  return 0;
}
